package Lento::ACT;
use Tie::RefHash;
use strict;
use Carp;

use POE;
use Lento::Debuggable;

@Lento::ACT::ISA = qw(Lento::Debuggable);
$Lento::ACT::level = -1000;

=pod

=head1  NAME

Lento::ACT - Asynchronous Completion Tokens

=head1 SYNOPSIS

$act_table = new Lento::ACT();

my $session = $act_table->token_to_session($token);

my $sessions = $act_table->sessions(); # returns 

my $newact = $act_table->newact($session);

$act_table->get($session);             # give a reference to the table

$act_table->remove_token($token);      # this token is no longer used

$act_table->put($session);    # no more session use of this table

=head1 Description 

=over 4

=item b<new> Create a new lookup table. 

=cut

sub new($) {
    my $class = shift;
    my $self = {};

    $self->{sessions} = {};
    tie %{$self->{sessions}}, 'Tie::RefHash';
    $self->{acts} = {};
    $self->{last} = 0;

    return bless $self, $class;
}

=pod

=item B<newact>

=head1 DESCRIPTION

An ACT table maps integers to sessions.  When a new ACT is installed
for a session an alias for the session is set, which prevents garbage
collection of the session (XXX IS THIS NEEDED, POSSIBLY NOT - this has
a nasty side effect namely that sessions registering ACTs must have
an alias_set).

Attach integers to sessions for dispatch to sessions.

=cut

# arg0 is class, arg1 is session, arg2 is connection
sub newact ($$) {
    my $self = shift;
    my $session = shift;

    $self->{last}++;
    my $token = $self->{last};
    $self->get($session);

    $self->{sessions}->{$session}->{$token} = undef;
    $self->{acts}->{$token} = $session;

    DEBUG("new ACT $token for session ".$session->get_heap()->{session_name}."\n");

    return $token;
}

sub get {
    my $self = shift;
    my $session = shift || confess;

    if (!defined $self->{sessions}->{$session} ) {
        DEBUG("new ACT for session ".$session->get_heap()->{session_name}." $session\n");
        $poe_kernel->post($session, 'alias_set', "$self-$session");
        $self->{sessions}->{$session} = {};
    }
}

sub sessions { 
    my $self = shift;

    my @sessions = keys %{$self->{sessions}};
    return \@sessions
}

sub remove_token ($$) {
    my $self = shift;
    my $token = shift || die;

    my $session = $self->{acts}->{$token};
    die "No Session! Bug" unless $session;

    delete $self->{acts}->{token};
    delete $self->{sessions}->{$session}->{$token};

    my @keys = keys %{$self->{acts}};
    DEBUG("Removing ACT %d (%d ACTs remain in table)\n", $token,
          scalar(@keys));
}

sub put ($$) {
    my $self = shift;
    my $session = shift || confess;

    foreach my $act ( keys %{$self->{sessions}->{$session}} ) { 
        delete $self->{acts}->{$act};
        my @keys = keys %{$self->{acts}};
        DEBUG("Removing ACT %d (%d ACTs remain in table)\n", $act,
              scalar(@keys));
    }
    $poe_kernel->post($session, 'alias_remove', "$self-$session");
    delete $self->{sessions}->{$session};
}

sub destroy ($) {
    my $self = shift;

    foreach my $session (@{$self->sessions()}) {
        $self->put($session);
    }

    delete $self->{sessions};
    delete $self->{acts};
    delete $self->{last};
}

# arg0 is class, arg1 is ACT
sub token_to_session ($$) {
    return $_[0]->{acts}->{$_[1]};
}

sub printme {
    my $self = shift;
    my $buf = "";

    foreach my $session (keys %{$self->{sessions}}) {
        $buf .=  "$session->get_heap()->{session_name} : [";
        $buf .= join(" ", keys %{$self->{sessions}->{$session}});
        $buf .= "]\n";
    }
    $buf;
}

1;

__END__
