package Lento::Acceptor;
use Socket;
use POSIX;
use POE qw(Wheel::SocketFactory Wheel::ReadWrite Driver::SysRW Filter::Line);
use POE::Session;
use strict;

use Lento::Connection;
use Lento::EventSupport;
use Lento::Debuggable;
@Lento::Acceptor::ISA = qw(Lento::Debuggable);
$Lento::Acceptor::level = 1;


=head1 NAME

Acceptor - Accept incoming TCP connections.  

=head1 SYNOPSIS

Lento::Acceptor->new($port, $bindaddr);

=head1 DESCRIPTION

For each accepted socket a connection object is instantiated, and it
is attached to the accepted socket.

=head1 Methods & Events

=over 4

=item B<start> (method)

Start an Acceptor session. The port is a mandatory argument, the
bindaddress allows to only listen to incoming connection request on
that address (optional, useful for multihoming lento''s).

B<Example:> Acceptor->start(port, bindaddr);


=item B<_start> (event)

Create the listening socket, this event is delivered by the _start method.

=item B<got_connection> (event)

Delivered by the SocketFactory Wheel.  This event instantiates a
connection object and attaches a session to the object.

=item B<got_error> (event)

If something goes bad with the listening socket, the Wheel notifies
the Acceptor here.

=back

=cut

# instantiate an Acceptor class controlled session

# method
sub start {
    my $class = shift;
    my $port = shift;
    my $bindaddr = shift;

    new POE::Session([ $port, $bindaddr ], 
		     "Lento::EventSupport", ['_signal', '_child', '_stop', 
									 'alias_set', 'alias_remove'],
		     "Lento::Acceptor", ['_start', 'got_client', 'got_error']
		    );
    if( defined($bindaddr) ) {
       LOG "----- Acceptor started for port: $port, bindaddr: $bindaddr\n";
    }
    else {
       LOG "----- Acceptor started for port: $port (no bindaddr)\n";
    }
    return undef;
}

# event
sub _start {
    my ($heap, $port, $bindaddr) = @_[HEAP, ARG0, ARG1];

    $bindaddr = INADDR_ANY unless $bindaddr;

    $_[HEAP]->{session_name} = $_[SESSION]->ID().':Acceptor';
    Lento::Debuggable::set_option($_[SESSION], trace => 1, default => 1);
    # start a socket factory
    $heap->{'wheel'} = new POE::Wheel::SocketFactory
      (SocketDomain   => AF_INET,        # in the AF_INET address family
       SocketType     => SOCK_STREAM,    # create stream sockets
       SocketProtocol => 'tcp',          # using the tcp protocol
       BindAddress    => $bindaddr,      # can run multiple servers
       BindPort       => $port,          # port
       ListenQueue    => 150,              # listen, with a 5-connection queue
       Reuse          => 'yes',          # reusing the address and port
       SuccessState   => 'got_client',   # sending this message when connected
       FailureState   => 'got_error',    # sending this message upon failure
      );
}

# Process an incoming connection.

# event
sub got_client {
    my ($socket, $peer_addr, $peer_port) = @_[ARG0, ARG1, ARG2];
    my $peer = gethostbyaddr($peer_addr, AF_INET);
    my $conn = new Lento::Connection();
    LOG ">>    Incoming connection $peer (port $peer_port)\n";
    # spawn the server session, no instantiator on Server
    $conn->attach($socket, undef, $peer, "from", $peer_addr);
    $conn->level(LVL_FULL);
}

# event
sub got_error {
    my ($operation, $errnum, $errstr) = @_[ARG0, ARG1, ARG2];
    DEBUG "Acceptor detected $operation error $errnum: $errstr\n";
    die;
}

1;

__END__
