package Lento::Bootstrap;

use strict;

use Socket;
use XML::Simple;

use Lento::Psdev;
use Lento::CacheTable;
use Lento::Fsetdb;
use Lento::Acceptor;
use Lento::ClientDB;

use Lento::Debuggable;
use Getopt::Long;
use POE::Session;
use Lento::ReqDispatcher;

@Lento::Bootstrap::ISA = qw(Lento::Debuggable);


=head1 NAME

Lento::Bootstrap - argument parsing and startup of InterMezzo

=head1 SYNOPSIS

Lento::Bootstrap->start();

=head1  DESCRIPTION

The Bootstrap package is responsible for parsing lento's command line
arguments and starting the services.

=cut
#' # stupid emacs

=pod

=over 4

=item B<Set up a server database> $::serverdb which servers do we know about

=item B<Set up a fileset database> $::fsetdb on which server can we find a fileset

=item B<Initialize kernel communication driver> If we are a client or
server, which needs presto initialize a handle for this device.  The
$::psdev object of type Psdev is structured as a POE::Driver object.
The device is opened.

=item B<Request dispatcher> A session is instantiated.

=item B<CallBackTable> Always instantiate an acceptor on port 369.
Set up a ServerSemantics object $::CallbackTable if we are running a
server. (not currently used).

=back

=cut


sub print_usage {
    print(
"Usage : $0 [--idfile=my_idfile][--configdir=my_config_dir]\n" .
"           [--debuglevel=db_level][--deletedb][--dbdir=my_dbir]\n" .
"           [--printinterval=delay]\n" .
"\n" .
"--idfile=path\n" .
"\tGives the filename of an alternate file containing the system id of\n" .
"\tthis machine (default /etc/intermezzo/sysid).\n" .
"\n" .
"--configdir=path\n" .
"\tGives the pathname of the configuration directory of this machine\n" .
"\t(default /etc/intermezzo).\n" .
"\n" .
"--debuglevel=level\n" .
"\tOverride the global debuglevel.\n" .
"\n" .
"--deletedb\n" .
"\tCauses the persistent databases to be removed before lento starts.\n" .
"\tThis is only used to recover an inconsistent system.\n" .
"\n" .
"--dbdir=path\n" .
"\tGives the pathname of the database directory of this machine\n" .
"\t(default /var/intermezzo).\n" .
"\n" . 
"--printinterval=delay\n" .
"\tCauses the session manager to print out the session state every\n" .
"\t\"delay\" seconds.  This can be useful for debugging.\n" .
"\n"
);
}

my $debugging; # is debugging on or off
my $sysid;     # id of this lento
my $idfile;    # file containing system id of this lento
my $fsetdbfile;    # file containing system id of this lento
my $prestodev;   # kernel pseudo device for this lento
my $dbbase;    # /var/intermezzo - unless specified
my $configdir; # /etc/intermezzo - unless specified
my $deletedb= 0;  # flag to indicate if databases should be cleared
                  # on startup/
my $print_session_interval= -1.0;

sub args {
    # startup options
    my ($debugging, $fork);

    $::compress=0;
    $::checksum=0;
    $::nofsuid=0;
    if (! GetOptions("debuglevel=s" => \$debugging,
                     "dbdir=s" => \$dbbase,
                     "configdir=s" => \$configdir,
                     "idfile=s" => \$idfile,
                     "fsetdbfile=s" => \$fsetdbfile,
                     "deletedb!" => \$deletedb,
                     "printinterval=f" => \$print_session_interval,
                     "compress!" => \$::compress,
                     "checksum!" => \$::checksum,
                     "noserver!" => \$::noserver,
                     "nofsuid!" => \$::nofsuid,
                     "fork!" => \$fork) ) {
            print_usage();
            exit(1);
    }

    $::global_debug = $debugging ? $debugging : -100;
    LOG ">>>>> Debug level set to $::global_debug\n";
    $dbbase = "/var/intermezzo" unless $dbbase;
    $::configdir = $configdir ? $configdir : "/etc/intermezzo";
    if ( $idfile ) {
            if ( ! -f $idfile ) {
                $idfile = $::configdir . "/" . $idfile;
            }
    } else {
            $idfile = $::configdir . "/sysid";
    }
    LOG(">>>>> IdFile is $idfile\n");

    if ( $fsetdbfile ) {
            if ( ! -f $fsetdbfile ) {
                $fsetdbfile = $::configdir . "/" . $fsetdbfile;
            }
    } else {
            $fsetdbfile = $::configdir . "/fsetdb";
    }
    $::fsetdbfile=$fsetdbfile;
    LOG(">>>>> FsetdbFile is $fsetdbfile\n");

    if ($::compress) {
        eval { require 'Compress::Zlib';
               import Compress::Zlib;
        };   

        if ($@) {
            warn "Compress::Zlib load failed; error $@\n";
            warn "ignoring compression option\n";
            $::compress=0;
        }
        $::compress=1;
    } else {
        $::compress=0;
    }

    if ($::nofsuid) {
        LOG(" lento will not set fsuid while reintegrating\n");
    }

    if ($fork) {
            if (!defined ($fork = fork())) {
                die "Fork failed";
            } elsif ($fork != 0) {
                exit(0);
            }
    }
}

sub set_sysid { 
    my $sysip = undef;
    my $hname= undef;

    LOG ">>>>> Initializing SYSID of this Lento\n";
    die "No --idfile or /etc/intermezzo/sysid" 
            unless -f $idfile;
    
    my $buf = eval { XMLin($idfile) };
    die "Cannot parse idfile $idfile\n" if ($@);

    $::mysysid = $buf->{name};
    $prestodev = $buf->{psdev}; 
    $::mysysip = $buf->{bindaddr};

    $prestodev = "/dev/intermezzo0" unless $prestodev;
    $::mysysip = "" unless $::mysysip;

    if (! -c $prestodev ) {
            print("InterMezzo character device ($prestodev) doesn't exist\n");
            die( "Check that IdFile \"$idfile\" is correct.\n" );
    }

    # set no_filter back to 0 on this device
    # this might be 1 to allow booting
    my $filterfile = $prestodev;
    $filterfile =~ s:/dev/:/proc/sys/intermezzo/:;
    $filterfile .= "/no_filter";
    `/bin/echo 0 > $filterfile`;

    # Check the device number.
    my $rdev = (stat $prestodev)[6];
    my ($major, $minor) = ($rdev >> 8, $rdev & 0xff);
    if ($major != 185 && $major != 186) {
        die "InterMezzo character device ($prestodev) has major $major, not 185\n";
    }

    $::prestodev=$prestodev;

    LOG ">>    mysysid=$::mysysid, prestodev=$prestodev (IP $::mysysip)\n";

    LOG "<<<<< Initializing SYSID\n";
}
    
sub setup_dbdir {
    $::dbdir = $dbbase . "/" . "$::mysysid";

    if ($deletedb) {
        $_= $dbbase;
        if ( !/\/var\/intermezzo/ ) {
            print("The --deletedb flag can only be used if the default " .
                  "database directory\nis used.  Don't use the " .
                  "--deletedb flag with the --dbdir flag.\n");
            exit(1);
        }
        LOG(">>> DELETING OLD DATABASE RECORDS\n");
        system("rm -rf $::dbdir");
    }

    system ("mkdir -p $::dbdir") unless -d $::dbdir;
    chmod 0700, $::dbdir;
}

sub start {
    my $class = shift;
    my $self = bless {}, $class;

    new POE::Session([], 'Lento::Bootstrap', ['_start' , '_stop']);
    return undef;
}


sub _start {

    $_[HEAP]->{session_name} = $_[SESSION]->ID().": Bootstrapper";
    # deal with the command line
    &args();

    # set my sysid, fix filter files
    &set_sysid();

    # make directory for KML databases; clean old databases if required.
    &setup_dbdir();

    # server database
    Lento::Server::setup_server_db();

    # see if we are a server
    $::thisserver = Lento::Server::find($::mysysid);
    if ( $::thisserver ) {
        LOG ">>>>> This an InterMezzo server, id $::thisserver->{name}\n";
    }

    # need a client table
    if ($::thisserver) {
      use Lento::ServerSemantics;
      $::CallbackTable = Lento::ServerSemantics->new();
      $::clientdb = new Lento::ClientDB;
      die unless $::clientdb;
    }

    # initialize presto dev
    unless ($::noclient) {
        LOG ">>>>> Opening $prestodev...\n";
        $::psdev = Lento::Psdev->new($prestodev);
        $::psdev->open();
        $::psdev->setpid();
        $::psdev->psdev_no_filter(0);
        LOG "<<<<< done Opening $prestodev.\n";
    }

    # initialize caches
    $::cachetable = Lento::CacheTable->new();
    $::cachetable->read_mounts();

    # fset location database
    $::fsetdb = Lento::Fsetdb->new();
    $::fsetdb->init();

    $::dispatcher = Lento::ReqDispatcher->new($::psdev);

    # umask to 0
    umask 00000; 

    # if we are a server fire up an acceptor
    if ( $::thisserver ) {
        Lento::Acceptor->start($::thisserver->{port}, 
                               $::thisserver->{bindaddr});
    }

    # print an overview of all the sessions
    if ( $print_session_interval > 0.0 ) {
            Lento::Debuggable::debug_sessions( $print_session_interval );
    }

}

sub _stop {
    my $object = $_[OBJECT];

    #
    # Release the cache if we are the server ... do this by sending
    # an upcall to presto saying it should clear the PID associated
    # with the given cache.  Fsetdb handles this.
    #
    unless ($::noclient) {
        LOG ">>>>> Clearing presto fileset information ...\n";
#       $::fsetdb->clear_mounttable();
        LOG "<<<<< done Clearning presto fileset information\n";
    }

    LOG "Lento is exiting, bye world...\n";
    exit;
}

1;

__END__
