######################################################################
#
# Methods for manipulating a bulk descriptor.  A bulk descriptor
# comes in two flavors, Source and Sink.  They are used to manage
# the state and to carry out the work involved in transferring data
# from one Lento to another.  The data can originate or be stored into
# a number of different storage mechanisms (buffers, files, inodes).
# Desc is sort of like a base class for Source and Sink.
#
######################################################################

package Lento::Bulk::Desc;
use Lento::Bulk;
use Lento::Bulk::Sink;
use Lento::Bulk::Source;
use File::stat;
use POSIX qw(:sys_stat_h :errno_h :unistd_h :fcntl_h);
use IO::File;
use Lento::Debuggable;
use POE;
use strict;
use Carp;

@Lento::Bulk::Desc::ISA = qw(Lento::Debuggable);
$Lento::Bulk::Desc::level = -1;


######################################################################
#
# static method Lento::Bulk::Desc::new
#
# Initialize a new descriptor for a bulk transfer.  The descriptor
# describes how the transfer occurs: direction, on each size whether
# they are a source or sink, where the data goes or comes from, and
# other characteristics.
#
# To use, invoke this method and assign the relevant state variables
# the proper values.  For example:
#
#       my $bulkdesc = Lento::Bulk::Desc->new(
#         { source_type      => $Lento::Bulk::Source::SOURCE_FILE,
# 	    source_object    => $path,
# 	    source_offset    => $start_offset,
# 	    bytesize         => $size,
# 	    role             => $Lento::Bulk::SOURCE,
# 	    direction        => $Lento::Bulk::CLIENT_TO_SERVER,
# 	    connection       => $_[HEAP]->{connection_obj},
# 	    debug            => 1,
# 	    callback_event   => 'finished_sending_kml',
# 	    callback_session => $_[SESSION] });
#
# Member variables that are common to all Desc object (Source and Sink):
#   - role ............. SOURCE or SINK (basically, output or input).
#   - direction ........ CLIENT_TO_SERVER or SERVER_TO_CLIENT.  Really
#                        this means how do the events transfer between
#                        source and sink when the connection is set up.
#   - connection ....... Connection to use to talk to other Lento.
#   - callback_session . Session that this transfer works within.
#   - callback_event ... Event sent to the callback_session at completion.
#   - debug ............ If set to 1, use debugging output.
#
# There are Sink- and Source-specific items as well, as described
# in Sink.pm and Source.pm.
#
######################################################################

sub new {
    my $class = shift;
    my $self = shift;
    bless $self, $class;

    #
    # We must have a session and connection from the start, and something
    # to work with.
    #
    confess "No callback_session" unless defined $self->{callback_session};
	
 	confess "no connection" if (!defined $self->{connection});

    if ($self->iam_sink()) {
	    confess "No sink_object" unless defined $self->{sink_object};
    } else {
	    confess "No source_object" unless defined $self->{source_object};
    }

    #
    # Make sure the use has selected how the communication should be
    # handled.
    #

    DEBUG "direction $self->{direction}, $Lento::Bulk::CLIENT_TO_SERVER\n";
    if ( ($self->{direction} != $Lento::Bulk::CLIENT_TO_SERVER) &&
	 ($self->{direction} != $Lento::Bulk::SERVER_TO_CLIENT)) {
	    confess "Direction ill-defined";
    }

    #
    # If we are a server we always need a client ACT: this client ACT
    # might either have been explicitly sent by the client in the REQ or
    # be the ctoken from the packet.
    # If we are a client we get a server ACT only with the first DAT or
    # START.
    #
    if ($self->iam_server()) {
	    confess "No source ACT" unless defined $self->{source_act};
    } 

    confess "No callback session" unless defined $self->{callback_session};

    return $self;
}


###########################################################################
#
# method Lento::Bulk::Desc::setup
#
# Based on whether this bulk descriptor is for a sink (receiving data)
# or a source (sending data), initialize it so that it can start doing
# its job.
#
# Inputs: none
# Returns: none
#
###########################################################################

sub setup {
    my $desc = shift;

    if ($desc->iam_sink()) {
	    return Lento::Bulk::Sink::setup($desc);
    } else {
	    return Lento::Bulk::Source::setup($desc);
    }
}


###########################################################################
#
# method Lento::Bulk::Desc::iam_server
#
# Return 1 if this descriptor is for a server, 0 if not.  A server does not
# necessarily send out data, but it does have different behavior than a
# client.
#
###########################################################################

sub iam_server {
    my $desc = shift;

    DEBUG("Direction %d  role: %d \n", $desc->{direction}, $desc->{role});

    if ( ($desc->{direction} == $Lento::Bulk::CLIENT_TO_SERVER) &&
	 ($desc->{role} == $Lento::Bulk::SINK)) {
	return 1;
    }

    if ( ($desc->{direction} == $Lento::Bulk::SERVER_TO_CLIENT) &&
	 ($desc->{role} == $Lento::Bulk::SOURCE)) {
	    return 1;
    }

    return 0;
}


###########################################################################
#
# method Lento::Bulk::Desc::iam_sink
#
# Return 1 if this descriptor is for a sink, 0 if not.  A sink receives
# data from a source and puts it in the destination storage (buffer or
# file).  Servers or clients can be sinks.
#
###########################################################################

sub iam_sink {
    my $desc = shift;
    DEBUG("Direction %d  role: %d \n", $desc->{direction}, $desc->{role});
    return  ($desc->{role} == $Lento::Bulk::SINK);
}


###########################################################################
#
# method Lento::Bulk::Desc::prepare_file(filepath, openmode)
#
# Open a file, and prepare it for reading or writing.  This may require
# seeking to a certain point.
#
# Inputs: filepath = pathname of file to prepare for reading or writing.
#         openmode = numeric mode for opening.
# Return: 0 if OK, or an error code if not OK.
#
###########################################################################

sub prepare_file {
    my $desc = shift;
    my $path = shift;
    my $mode = shift;
    my $error = 0;
    my $st;
    my $count;
	
    $desc->{handle} = new IO::File || die "NO handle";
    if ($desc->iam_sink()) {
        $count=0;
        $st=0;
        for ($count=0;$count < $Lento::Bulk::RETRY;$count++) {
	        # this is actually a sysopen
	        $error = $desc->{handle}->open($path, O_RDWR|O_CREAT, 0700);
            DEBUG " Error $! opening $path\n" if $!;
	        $st = stat($path) && last;
        }; 
        if (!($st=stat($path))) {
            # presto generates (create/close) upcalls even when files have 
            # been deleted. These upcalls show up with a path with prefix 
            # (deleted). Unfortunately, we cannot trap this earlier since an 
            # application might be using the (deleted) prefix for its own files.
            # Instead, we wait for the open to fail before checking for this
            # condition.
            # XXX: This is ugly. Presto should be changed.
            if ($path =~ / [(]deleted[)]$/) {
                DEBUG "ignoring deleted file ($path)\n";
                return ENOENT;
            }
    
            confess "No sink file ($path) was created: $!" ;
        }

	    if (defined $desc->{sink_offset}) {
	        sysseek($desc->{handle}, $desc->{sink_offset}, SEEK_SET);
	    }
	    $desc->{handle}->autoflush(1);
    } else {
	    $st = stat($path);
        # OS might return an EINTR or EAGAIN
        for ($count=0;!$st && $count < $Lento::Bulk::RETRY;$count++) {
	        $st = stat($path) && last;
        }; 
	    if (!$st) {
	        # confess "No such source file\n";
	        # file might be deleted while backfetch, just ignore them.
	        return ENOENT;
	    }
	    if (S_ISDIR ($st->mode)) {
	        # this file is a directory
	        confess "Source not a file\n";
	        return EISDIR;
	    }
	    $error = $desc->{handle}->open($path, O_RDONLY);
    }

    $desc->{stat} = $st;
    if (defined $desc->{source_offset}) {
	    DEBUG "**** offset: $desc->{source_offset} \n";
	    sysseek($desc->{handle}, $desc->{source_offset}, SEEK_SET);
    }

    if ($error < 0) {
	    # access denied, or similar
	    $desc->{handle} = undef;
	    confess "ERROR opening $error\n";
	    return EPERM;
    }
    
    $desc->{source_offset} = 0;
    return 0;
}

# Error hook into the descriptor...needed for compatibility with the
# Rsync descriptor.
sub error {
    my $desc=shift;
    my $session=$poe_kernel->get_active_session();

    # presumable caller does not know about the local got_error
    # event handlers, we will help it along by installing the
    # right handlers -SHP
    $poe_kernel->state("sink_got_error","Lento::Bulk::Sink")
            and die "no session in &Lento::Bulk::Desc::error\n";
    $poe_kernel->state("source_got_error","Lento::Bulk::Source")
            and die "no session in &Lento::Bulk::Desc::error\n";

    $desc->{state}="quitting";

    if ($desc->iam_sink()) {
	    $poe_kernel->call($session,'sink_got_error',@_);
    } else {
	    $poe_kernel->call($session,'source_got_error',@_);
    }
}

###########################################################################
#
# method Lento::Bulk::Desc::printme
#
# Print out this descriptor.
#
###########################################################################

sub printme {
    my $desc = shift;
    my $str;

    $str = "Bulk Descriptor:\n";

    if ($desc->iam_sink()) {
	$str .= " - I am sink\n";
	$str .= " - Sink type: File\n"
	  if $desc->{sink_type} eq $Lento::Bulk::SINK_FILE;
	$str .= " - Sink type: Buffer\n"
	  if $desc->{sink_type} eq $Lento::Bulk::SINK_BUFFER;
	$str .= " - Sink object: $desc->{sink_object}\n";
    } else {
	$str .= " - I am source\n";
	$str .= " - Source type: File\n"
	  if $desc->{source_type} eq $Lento::Bulk::SOURCE_FILE;
	$str .= " - Source type: Buffer\n"
	  if $desc->{source_type} eq $Lento::Bulk::SOURCE_BUFFER;
	$str .= " - Source type: Inode\n"
	  if $desc->{source_type} eq $Lento::Bulk::SOURCE_INODE;
	$str .= " - Source object: $desc->{source_object}\n";
    }

    $str .= " - Offset: $desc->{source_offset}\n" 
	if defined $desc->{source_offset};

    $str .= " - Bytes: " .
      (defined $desc->{bytesize} ? $desc->{bytesize} : "unknown") . "\n";

    $str .= " - Processed bytes: $desc->{processed}\n"
      if defined $desc->{processed};

    $str .= " - Quota: $desc->{quota}\n"
      if defined $desc->{quota};

    $str .= " - Source_act: $desc->{source_act}\n" if defined $desc->{source_act};
    $str .= " - Sink_act: $desc->{sink_act}\n" if defined $desc->{sink_act};

    my $peer = $desc->{connection}->{sysid};
    my $me = $::mysysid;
    my $msg;

    if ($desc->iam_sink()) {
	$msg = " from $peer to $me";
    } else {
	$msg = " from $me to $peer";
    }

    $str .= " - Client to server: $msg\n" if 
	$desc->{direction} == $Lento::Bulk::CLIENT_TO_SERVER;

    $str .= " - Server to Client: $msg\n" if 
	$desc->{direction} == $Lento::Bulk::SERVER_TO_CLIENT;

    $str .= " - Running as uid $desc->{uid}\n" if defined $desc->{uid};
    return $str;
}
