package Lento::Bulk::Sink;

use strict;
use Carp;
use POE;

use Lento::Connection;
use Lento::Packet;
use Lento::Debuggable;
use Lento::Bulk;
use POSIX qw(:sys_stat_h :errno_h :unistd_h :fcntl_h);

@Lento::Bulk::Sink::ISA = qw(Lento::Debuggable);
$Lento::Bulk::Sink::level = -1;

# A hash for all active sinks...
%Lento::Bulk::Sink::Active_Sinks = ();

# METHOD - setup()
#
# setup takes the bulk transfer descriptor for this sink, and prepares
# the destination for the data.  This can be:
#   - a file, if $desc->{sink_object} is a string with a filename
#   - a buffer, if $desc->{sink_object} is a *reference* to a string.  In
#     that case it is treated like a buffer that is used to accumulate the
#     data sent to this sink.
#   - if this completes the target file or buffer has been instantiated.
sub setup  {
    my $desc = shift;
    my $rc;

    # Hook in our event handlers
    $poe_kernel->state('DAT', 'Lento::Bulk::Sink') and die "no session";
    $poe_kernel->state('EOD', 'Lento::Bulk::Sink') and die "no session";
    $poe_kernel->state('done_sinking', 'Lento::Bulk::Sink')
        and die "no session";
    
    if ($desc->{debug}) {
        Lento::Debuggable::set_option($poe_kernel->get_active_session(),
                                      trace => 1, default => 1);
    }
    
    # Descriptor setup might happen right after a disconnect.
    if (!Lento::Connection::connected($desc->{connection})) {
        DEBUG "No connection\n";
        $poe_kernel->yield('got_error','disconnected',ENOTCONN,'no connection');
        return (ENOTCONN, undef);
    }
    
    DEBUG "Creating new sink:\n";
    DEBUG "Expecting $desc->{bytesize} bytes\n" if defined $desc->{bytesize};
    
    # for now handle the sink inside the calling session
    $desc->{session} = $desc->{callback_session};
    $desc->{session}->get_heap()->{desc} = $desc;
    
    # note that we haven't processed any bytes on this yet
    $desc->{processed} = 0;
    
    # get a handle for I/O
    if ($desc->{sink_type} eq $Lento::Bulk::SINK_FILE) {
        # the data will go into a file
        DEBUG "Sink: file=$desc->{sink_object}\n";
        
        # If we were already sinking to this file, stop the other sink
        # session with a got_error. This way we do not leak fds
        if (exists ($Lento::Bulk::Sink::Active_Sinks{$desc->{sink_object}})) {
            DEBUG "Found an active sink for this file...closing\n";
            # Do a call to make sure that the file is closed
            $poe_kernel->call(
                              $Lento::Bulk::Sink::Active_Sinks{$desc->{sink_object}},
                              'got_error','handle', ESTALE, 'new sink opened');
            #$desc->{connection}->set_error('handle', ESTALE, 'sink conflict');
        }
        $Lento::Bulk::Sink::Active_Sinks{$desc->{sink_object}}=
            $poe_kernel->get_active_session();
        
        $rc = $desc->prepare_file($desc->{sink_object});
        confess "NO HANDLE" unless $desc->{handle};
    } elsif ($desc->{sink_type} eq $Lento::Bulk::SINK_BUFFER) {
        # the data will go into a buffer
        DEBUG "Sink: buffer=$desc->{sink_object}\n";
    } else {
        confess "Unknown sink type $desc->{sink_type}!\n";
    }
    
    # If there is an error return immediately
    if ($rc) {
        return ($rc, undef);
    }
    
    # register for session timeouts
    $desc->{connection}->register_in_monitor_list
        ($poe_kernel->get_active_session());
    
    # if this is a server and sink, we must tell the client to start
    # sending us the data.  We do this by replying with a START event.
    if ($desc->iam_server() && $desc->{role} == $Lento::Bulk::SINK) {
        confess "No source token"  unless $desc->{source_act};
        confess "No sink token" unless $desc->{sink_act};
        sendstart($desc);
    } 
    
    return (0,$desc->{session});
    
}


# METHOD - sendstart(), sent from sink-server to source-client. 
sub sendstart {
    my $desc = shift;
    if ($desc->iam_server()) {
        DEBUG "**** source act: $desc->{source_act},".
            " sink act $desc->{sink_act}\n";
        $desc->{connection}->sendpacket
            ('START', [ $desc->{source_act}, $desc->{sink_act} ]);
    }
}

# DAT packets are supposed to arrive in order.
sub DAT  {
    my $desc = $_[HEAP]->{desc};
    my $packet = $_[ARG0];
    my $count = length($packet->[PKT_BODY]);
    my $newprocessed = $count + $desc->{processed};
    
    if (!defined($desc)) {
        $packet->[PKT_CONNECTION]->set_error("DAT",EBADF,"No Descriptor");
        return;
    }
    
    DEBUG "Sink - processing DAT event, length = $count\n";
    
    # check the sizes, if necessary.  If the descriptor was not set up
    # with a known size, we just keep receiving data until EOD.  But if
    # we know about a size, we make sure we don't receive too much.
    # This is where quota-checking could also go.
    if (defined $desc->{bytesize}) {
        DEBUG "Checking bytesize = $desc->{bytesize} ...\n";
        if ($newprocessed > $desc->{bytesize}) {
            DEBUG "Size is too large: $newprocessed > $desc->{bytesize}\n";
            $count = $desc->{bytesize} - $desc->{processed};
            DEBUG "New count = $count\n";
        }
    }
    
    # output the data, either to a buffer or a file.  We only use the amount
    # of data indicated by $count, which might be smaller than sent.
    if ($count > 0) {
        if ($desc->{sink_type} eq $Lento::Bulk::SINK_BUFFER) {
            # append the data to the end of the buffer
            DEBUG "Sink - appending data to buffer.\n";
            #    print "Got:\n" . $packet->[PKT_BODY];
            $ {$desc->{sink_object}} .= substr($packet->[PKT_BODY], 0, $count);
        } elsif ($desc->{sink_type} eq $Lento::Bulk::SINK_FILE) {
            # write the data to the end of the output file
            DEBUG "Sink - writing data to file.\n";
            
            if (!$desc->{handle}) {
                DEBUG "Handle closed...check for got_errors\n";
                $poe_kernel->yield('got_error','DAT',EBADF,"fd closed");
                return;
            }
            
            my $rc = $desc->{handle}->syswrite($packet->[PKT_BODY], $count);
            if ($! == ENOSPC ||
               (defined $rc and $rc != $count) ) {
                LOG("NO space left on device for %s\n", $desc->{sink_object}); 
                LOG("Lento will exit.  To proceed first remove this file" . 
                    " on the system where is being modified, then " .
                    " restart this lento, free up space and re-create the ".
                    " file.\n"); 
                exit 0; 
            }
            if (!defined($rc) or $rc < 0) { 
                confess "Write to file $desc->{sink_object} failed: $!\n";
                $packet->[PKT_CONNECTION]->set_error
                    ('writing', $!, 'file write');
            }
        } else {
            confess "Unknown sink type $desc->{sink_type}!\n";
        }
        
        # indicate we've processed more bytes.
        $desc->{processed} += $count;
    }
}


# EVENT - EOD, done, cleanup
sub EOD  {
    my $desc = $_[HEAP]->{desc};
    my $packet = $_[ARG0];
    my $sizeleft = 0;
    my $rc = shift @{$packet->[PKT_PARMS]};
    my $sent = shift @{$packet->[PKT_PARMS]};
    
    if (!defined($desc)) {
        $packet->[PKT_CONNECTION]->set_error("EOD",EBADF,"No Descriptor");
        return;
    }
    if (!$packet->[PKT_CONNECTION]->connected()) {
        return; 
    }
    # make sure we received everything we were expecting; if we didn't,
    # log a message but still continue.
    if (defined $desc->{bytesize}) {
        $sizeleft = $desc->{bytesize} - $desc->{processed};
        if ($sizeleft > 0) {
            LOG "Source did not send full byte count for $desc->{sink_object}:\n".
                "       still expecting $sizeleft bytes.\n";
        }
    }
    
    if ($rc) {
        LOG "Error $rc at source for object $desc->{sink_object}.\n";
    }
    
    DEBUG "Sink - processing EOD event.  Left to write = $sizeleft\n";
    DEBUG "We processed $desc->{processed} bytes\n";
    DEBUG "Source sent result code $rc\n" if defined $rc;
    DEBUG "Source sent $sent bytes\n" if defined $sent;
    
    # generate an event to ourselves indicating we're done sinking; when
    # that event arrives, everything gets cleaned up and the data is
    # processed.
    
    if ( defined $sent and $desc->{processed} != $sent) {
        # We have a problem, not everything that was sent out was received
        LOG (" Source sent $sent, we only received $desc->{processed}\n");
        # This could become a problem so disconnect
        $packet->[PKT_CONNECTION]->set_error('sinking', EINTR,'incomplete file');
    } else {
        $_[KERNEL]->yield('done_sinking', 0);
    }
}


###########################################################################
# EVENT - got_error events
###########################################################################
sub got_error {
    $_[KERNEL]->call($_[SESSION],'sink_got_error',$_[ARG0],$_[ARG1],$_[ARG2]);
}

sub sink_got_error  {
    my $desc = $_[HEAP]->{desc};
    
    if (!defined $desc) {
        DEBUG "Sink - Descriptor is not defined\n";
        $_[KERNEL]->yield('finish');  
        return;
    }
    
    LOG "Sink - Fetchfile error for $desc->{sink_object}!\n";
    
    if ($desc->{handle}) {
        DEBUG "Closing output file handle.\n";
        
        if (defined $_[HEAP]->{closeinfo}) {
            my $closeinfo=$_[HEAP]->{closeinfo};
            $closeinfo->{fd}=$desc->{handle}->fileno()
                if (!defined $closeinfo->{fd});
            
            DEBUG "Using closeinfo to release fd\n";
            DEBUG("Closing: fd $closeinfo->{fd}: recno $closeinfo->{recno},
                   offset: $closeinfo->{kml_offset} + $closeinfo->{len}, flags
                   $closeinfo->{flags}); \n");
            
            # now close without journalling...note the 0x0 for flags
            $closeinfo->{psdev}->lento_close
                ($closeinfo->{fd}, $closeinfo->{slot_offset},
                 $closeinfo->{recno},
                 $closeinfo->{kml_offset} + $closeinfo->{len},
                 0x0) if defined $closeinfo->{fd};
            
            delete   $_[HEAP]->{closeinfo};
        } else {
            LOG("Error closing file handle\n");
        }
    }
    
    # We already received an error so deregister
    if (defined $desc->{connection}) {
        $desc->{connection}->remove_from_monitor_list
            ($poe_kernel->get_active_session());
    }
    
    delete $Lento::Bulk::Sink::Active_Sinks{$desc->{sink_object}};
    
    # There is an error state so set the error variable and inform the
    # caller.
    $_[HEAP]->{error}=1;
    $_[KERNEL]->post($desc->{callback_session}, $desc->{callback_event}, 1);
    
    delete $desc->{callback_session};
    delete $desc->{session};
    delete $desc->{handle} if $desc->{handle};
    delete $desc->{wheel} if $desc->{wheel};
    undef $desc;
    delete $_[HEAP]->{desc};
}


###########################################################################
# EVENT - done_sinking
#
# We send this event to ourselves when we're finished receiving all
# the data.  This cleans up the sink data in the heap, and deals with
# what to do with the results.
###########################################################################

sub done_sinking  {
    my $desc = $_[HEAP]->{desc};
    
    # This might happen if the session receives a got_error before the 
    # descriptor is set up.
    # XXX: This is ugly...fix it by putting descriptors into their own 
    # sessions?
    if (!defined $desc) {
        DEBUG "Sink - done_sinking Descriptor is not defined\n";
        $_[KERNEL]->yield('finish');  
        return;
    }
    
    # We are no longer interested in being monitored so unregister
    if (defined $desc->{connection}) {
        $desc->{connection}->remove_from_monitor_list
            ($poe_kernel->get_active_session());
    }
    
    my $st = $desc->{stat};
    
    if ($desc->{sink_type} eq $Lento::Bulk::SINK_BUFFER) {
        # destination is buffer
        DEBUG "Sink - done_sinking buffer\n";
        1;
    } elsif ($desc->{sink_type} eq $Lento::Bulk::SINK_FILE) {
        if (!defined $st) {
            confess " done_sinking: error in stat? file = %s.\n" , 
            $desc->{sink_object}; 
        }
        
        # close callback -  this will set the attributes of the file.
        if ($desc->{handle} && defined $desc->{close_callback}) {
            DEBUG "Closing output file handle using $desc->{close_callback}.\n";
            no strict 'refs';
            &{"$desc->{close_callback}"};
            use strict 'refs';
        } elsif ($desc->{handle} ) {
            # ordinary close
            DEBUG "Closing output file handle.\n";
            $_[HEAP]->{desc}->{handle}->close();
        }
        
        DEBUG "Sink - done_sinking, file = %s\n", $desc->{sink_object};
    } else {
        confess "Unknown sink type $desc->{sink_type}!\n";
    }
    
    delete $Lento::Bulk::Sink::Active_Sinks{$desc->{sink_object}};
    
    # OK, we're completely done with the receive, so notify the interested
    # party by posting the callback event in the descriptor.
    $_[KERNEL]->post($desc->{callback_session}, $desc->{callback_event}, 0);
    
    delete $desc->{callback_session};
    delete $desc->{session};
    delete $desc->{handle} if $desc->{handle};
    delete $desc->{wheel} if $desc->{wheel};
    undef $desc;
    delete $_[HEAP]->{desc};
}

1;
