package Lento::Bulk::Source;
use Carp;
use File::stat;
use POSIX qw(:sys_stat_h :errno_h :unistd_h :fcntl_h);
use Lento::Debuggable;
use Lento::Packet;
use Lento::Connection;
use Lento::Bulk;
use strict;
use POE qw(Filter::Stream);

@Lento::Bulk::Source::ISA = qw(Lento::Debuggable);
$Lento::Bulk::Source::level = -1;

# Block size used for reading if stream is to be compressed
$Lento::Bulk::Source::RECORD_SIZE  = 1048576; # 1M

# Smallest file/buffer that we will compress
$Lento::Bulk::Source::MIN_SIZE  = 131072; # 128k

###########################################################################
# METHOD - setup_wheel()
#
# Create the POE ReadWrite wheel to read the data from the file.
# This will read in the file, and send us input_ready events when
# data is read.  This is done in chunks, so we'll get multiple
# input_ready's.  Reading will start right after this routine is
# called, so don't call this until it is safe to start reading and
# send out the data.
###########################################################################

sub setup_wheel {
    my $desc = shift;
    my $blocksize= $desc->{blocksize} || &Lento::Connection::BLOCK_SIZE;
    
    DEBUG "Using a block size of $blocksize bytes\n";
    
    # create the read-write wheel
    $desc->{wheel} = new POE::Wheel::ReadWrite
        (Handle => $desc->{handle},
         #Offset => $desc->{source_offset},
         Driver => new POE::Driver::SysRW(BlockSize => $blocksize),
         Filter => new POE::Filter::Stream,
         InputState => 'input_ready',
         FlushedState => undef,
         ErrorState => 'file_error'
        );
    
    # Tell the Connection that we're interested in receiving flow
    # control events
    $desc->{connection}->register_for_flow();
    
}


###########################################################################
# METHOD - setup()
#
# setup a source session - return (error_code, bulk_session)
###########################################################################

sub setup {
    my $desc = shift;
    my $rc;

    # Hook in the event handlers we need to use
    $poe_kernel->state('START', 'Lento::Bulk::Source') 
        and die "no session";
    $poe_kernel->state('low_water', 'Lento::Bulk::Source') 
        and die "no session";
    $poe_kernel->state('high_water', 'Lento::Bulk::Source') 
        and die "no session";
    $poe_kernel->state('complete', 'Lento::Bulk::Source') 
        and die "no session";
    $poe_kernel->state('input_ready', 'Lento::Bulk::Source') 
        and die "no session";
    $poe_kernel->state('file_error', 'Lento::Bulk::Source') 
        and die "no session";
    $poe_kernel->state('complete', 'Lento::Bulk::Source') 
        and die "no session";
    
    $desc->{error}=0;
    $desc->{eof}=0;
    
    # This variable determines if files are to be compressed
    $desc->{use_compression}=0; # No compression by default
    
    # This variable determines if checksumming is to be used on this file
    # Note that setting this forces an EOR record with the checksum after every 
    # record size bytes
    $desc->{use_checksum}=0;    # No checksum by default
    
    # This buffer holds the records if any
    # records are only generated if checksumming
    $desc->{record_buffer}="";
    
    # The first DAT that is sent out sets this variable to false
    $desc->{first_dat}=1;
    
    # This should never happen!
    if (defined($desc->{bytesize}) && $desc->{bytesize} <= 0) {
        LOG(" Unexpected bytesize ".$desc->{bytesize}." <= 0?\n");
        $rc=EINVAL;
    } else {
        # make sure we know where we are getting data, and set that source up
        if ($desc->{source_type} eq $Lento::Bulk::SOURCE_FILE) {
            DEBUG("Setting up Source desc: infile = $desc->{source_object}\n");
            $rc = $desc->prepare_file($desc->{source_object}, &O_RDONLY);
            
            $desc->{use_compression}=1 
                if $desc->{stat} and 
                    $desc->{stat}->size > $Lento::Bulk::Source::MIN_SIZE;
            
        } elsif ($desc->{source_type} eq $Lento::Bulk::SOURCE_BUFFER) {
            DEBUG("Setting up Source desc: buffer=${$desc->{source_object}}\n");
            if (!defined $desc->{bytesize}) {
                $desc->{bytesize} = length($ {$desc->{source_object}});
                DEBUG("Initializing source buffer bytesize = $desc->{bytesize}\n");
            }
            $desc->{use_compression}=1 
                if $desc->{bytesize} > $Lento::Bulk::Source::MIN_SIZE;
            
        } elsif ($desc->{source_type} eq $Lento::Bulk::SOURCE_INODE) {
            DEBUG("Setting up Source desc: inode = $desc->{source_object}\n");
            my $inodeinfo = $desc->{source_object};
            $desc->{source_object} = sprintf("/...ino:%lu:%u", 
                                             $inodeinfo->{ino}, $inodeinfo->{generation});
            
            $rc = $desc->prepare_inode($desc->{source_object}, &O_RDONLY);
        } else {
            confess "Unknown source type $desc->{source_type}!\n";
        }
    }
    
    DEBUG "  File/Buffer is large enough\n" if $desc->{use_compression};
    
    $desc->{use_compression}=0 unless $::compress;
    DEBUG "  Compression will be used\n" if $desc->{use_compression};
    
    # blocksize to be used for file reads
    $desc->{blocksize}=$desc->{use_compression} ? 
        $Lento::Bulk::Source::RECORD_SIZE : &Lento::Connection::BLOCK_SIZE;
    
    $desc->{packetsize}=&Lento::Connection::BLOCK_SIZE;
    
    # if there was an error setting up the source, just return now
    # If I am the server signal client that there is no more data forthcoming
    # else setup an error state.
    if ($rc) {
        DEBUG("Error %d ocurred. \n", $rc);
        if ( $desc->iam_server() ) {
            DEBUG("Sending EOD to client.\n");
            my $packet=$poe_kernel->get_active_session()->get_heap()->{req_packet};
            $packet->[PKT_CONNECTION]->sendpacket('EOD', $packet,[$rc,0]);
            # $rc is non zero and session is undef if I am server
            return ($rc, undef);
        } else {
            DEBUG("Deferring Error handling till START packet is received.\n");
            $desc->{error}=$rc;
        }
    }
    
    # do we want to debug stuff here? 
    if ($desc->{debug}) {
        Lento::Debuggable::set_option($poe_kernel->get_active_session(),
                                      trace => 1, default => 1);
    }
    
    # for now we run the bulk transport inside the calling session
    $desc->{session} = $desc->{callback_session};
    $desc->{session}->get_heap()->{desc} = $desc;
    $desc->{state} = "";
    
    # note that we haven't processed any bytes on this yet
    $desc->{processed} = 0;
    
    # if we are the server, set up the wheel to get
    # I/O going, else wait for START and do it there
    if ( $desc->iam_server() ) {
        $desc->{dataclone_packet} = 
            $poe_kernel->get_active_session()->get_heap()->{req_packet};
        setup_wheel($desc) unless $rc;
    }
    
    # In case of error $rc is non zero and session is set if I am the client
    return ($rc, $desc->{session});
}


###########################################################################
# EVENT - START
#
# sent from server, only when client is source.  If we receive this,
# we are a client and should now start up the read/write wheel to get
# data out to the server.
###########################################################################

sub START {
    my $packet = $_[ARG0];
    my $desc = $_[HEAP]->{desc};
    
    DEBUG("START packet received.\n");
    
    # reverse the tokens in this packet -- see Connection.pm for
    # the protocol description
    my $tmp = $packet->[PKT_STOKEN];
    $packet->[PKT_STOKEN] =  $packet->[PKT_CTOKEN];
    $packet->[PKT_CTOKEN] =  $tmp;
    
    if ($desc->{error}) {
        DEBUG("Error during setup, sending EOD with rc=%d to server.\n", 
              $desc->{error});
        $desc->{state} = 'quitting';
        $packet->[PKT_CONNECTION]->sendpacket('EOD', $packet, [$desc->{error},0]);
        $_[KERNEL]->yield('complete');
        return;
    }
    
    $desc->{dataclone_packet} = $packet;
    $desc->{sink_act} = $packet->[PKT_STOKEN];
    DEBUG("SERVER ACT %d\n", $desc->{sink_act});
    confess "No server ACT" unless defined $desc->{sink_act};
    
    setup_wheel($_[HEAP]->{desc});
}


###########################################################################
# EVENT - low_water
#
# the network connection has notified us there are buffers...
# reactivate the file wheel.
###########################################################################

sub low_water {
    DEBUG("low_water received.\n");
    
    # If descriptor already reached end of file send an EOD
    if ($_[HEAP]->{desc}->{eof}) {
        $_[KERNEL]->yield('input_ready',"");
    }
    
    # re-enable the read select and event delivery
    if (defined $_[HEAP]->{desc} and defined $_[HEAP]->{desc}->{wheel}) {
        $poe_kernel->select_resume_read($_[HEAP]->{desc}->{handle});
    }
}


###########################################################################
# EVENT - high_water
#
# buffers are full, stop sending
###########################################################################

sub high_water {
    DEBUG("high_water received.\n");
    
    # disable the read select
    if (defined $_[HEAP]->{desc}and defined $_[HEAP]->{desc}->{wheel}) {
        $poe_kernel->select_pause_read($_[HEAP]->{desc}->{handle});
    }
}


###########################################################################
# EVENT - input_ready
#
# our source wheel is doing this when it is allowed to read.  When it
# has read some data (perhaps not all of it, but the next available
# amount), this event is generated.  We take the data and send it on
# to the destination.
###########################################################################

sub input_ready  {
    my $desc = $_[HEAP]->{desc};
    return if (!defined $desc or 
               ! defined $desc->{state} or 
               $desc->{state} eq 'quitting');
    
    my $packet = $desc->{dataclone_packet};
    my $buffer = $_[ARG0];
    my $buflen = length($buffer);
    
    # check how many bytes we have sent so far, and only send out the
    # requested amount.
    if (defined $desc->{bytesize}) {
        if ($desc->{processed} + $buflen > $desc->{bytesize}) {
            DEBUG("Extra bytes in input_ready buffer\n");
            $buflen = $desc->{bytesize} - $desc->{processed};
            $buffer = substr($buffer, 0, $buflen);
            DEBUG("Changed buflen from %d to %d\n", length($buffer), $buflen);
        }
    }
    
    # update the count of bytes that have been sent out
    $desc->{processed} += $buflen;
    DEBUG("New processed = $desc->{processed} (buflen $buflen)\n");
    
    # if there is anything to send, do so now.  If we send more than
    # necessary, then the receiving side will ignore the extra.
    if ($buflen > 0) {
        DEBUG("Sending DAT to $packet->[PKT_STOKEN], length: %d\n", $buflen);
        $packet->[PKT_CONNECTION]->sendpacket('DAT', $packet, \$buffer);
    }
    
    if (defined $desc->{bytesize}) {
        # check if a fixed-size send is complete
        if ($desc->{processed} == $desc->{bytesize}) {
            DEBUG("Reached final send amount from Source.\n");
            $desc->{eof}=1;
        }
    } else {
        # check if we've reached the end of our input buffer or file
        if (defined $desc->{stat} and 
            $desc->{processed} >= $desc->{stat}->size) {
            DEBUG("Reached end-of-file from Source.\n");
            $desc->{eof}=1;
        }
    }
    
    # if we've reached the end of the input send out EOD
    if ($desc->{eof}) {
        DEBUG "sending EOD\n";
        $desc->{state} = 'quitting';
        $packet->[PKT_CONNECTION]->sendpacket('EOD', $packet, 
                                              [0,$desc->{processed}]);
        $_[KERNEL]->yield('complete');
    }
}


###########################################################################
# EVENT - file_error
#
# arrives if there was an error reading the input file.  If so, signal
# EOD.
###########################################################################

sub file_error {
    my $desc = $_[HEAP]->{desc};
    my $packet = $desc->{dataclone_packet};
    my $error = $_[ARG1];
    return if (!defined $desc or $desc->{state} eq 'quitting');
    
    if ( $_[HEAP]->{desc}->{stat}->size == 0 or $desc->{eof}) {
        $desc->{state} = 'quitting';
        $packet->[PKT_CONNECTION]->sendpacket('EOD', $packet, 
                                              [0, $desc->{processed}]);
        $_[KERNEL]->yield('complete');
    } else {
        if ($error != 0) {
            $_[KERNEL]->yield('got_error',"file error",$error,"file read");
            DEBUG "!!!! Error reading size %d, $_[ARG0], $_[ARG1], $_[ARG2]\n",
            $_[HEAP]->{desc}->{stat}->size;
            $packet->[PKT_CONNECTION]->sendpacket('EOD', $packet, 
                                                  [$error, $_[HEAP]->{desc}->{processed}]);
            confess;
        } else {
            DEBUG "Received EOF\n";
            $desc->{eof}=1;
            # If we receive an EOF make sure that we force an extra input_ready
            # event
            $_[KERNEL]->yield('input_ready',"");
        }
    }
}


###########################################################################
# EVENT - got_error event handlers
###########################################################################
sub got_error {
    $_[KERNEL]->call($_[SESSION],'source_got_error',$_[ARG0],$_[ARG1],$_[ARG2]);
}

sub source_got_error  {
    # since we Fedex'd this got_error, it's possible to get another event
    # afterwards (say, send_data) that will want to do scary things.
    # Don't let it do those scary things.
    $_[HEAP]->{desc}->{state} = 'quitting' if defined $_[HEAP]->{desc};
    
    # I think this is all of the necessary cleanup.
    $_[KERNEL]->yield('complete');
}


###########################################################################
# EVENT - complete
#
# We send this event to ourselves when we're done sending out the data.
# This is after we've sent EOD.  At this point, invoke the user's
# callback event stored in $desc->{callback_event}.
###########################################################################

sub complete {
    # This might happen if the session receives a got_error before the 
    # descriptor gets setup.
    # XXX: This is ugly...fix it by putting descriptors into their own 
    # sessions?
    if (!defined $_[HEAP]->{desc}) {
        $_[KERNEL]->yield('finish');
        return;
    }
    #confess "No descriptor in heap" unless defined $desc;
    
    # invoke the callback event, if there is one
    $_[KERNEL]->post($_[HEAP]->{desc}->{callback_session}, 
                     $_[HEAP]->{desc}->{callback_event}, 0);
    
    # potential circular reference: callback_session might be the 
    # bulk session
    delete $_[HEAP]->{desc}->{callback_session};
    
    # Another potential circular reference
    delete $_[HEAP]->{desc}->{session};
    
    # shut down our connection and wheel
    $_[HEAP]->{desc}->{connection}->remove_for_flow()
        if defined $_[HEAP]->{desc}->{connection};
    delete $_[HEAP]->{desc}->{connection};
    delete $_[HEAP]->{desc}->{wheel} if defined $_[HEAP]->{desc}->{wheel};
    
    # close the input file if we have to
    delete $_[HEAP]->{desc}->{handle};
    
    # Remove the cloned packet reference
    delete $_[HEAP]->{desc}->{dataclone_packet};
    
    # set the state to quitting for further low_water/input_ready evts
    $_[HEAP]->{desc}->{state} = 'quitting'; 
}

1;

__END__

=head1 NAME

Bulk::Source - InterMezzo cache manager bulk transfer sender

=head1 SYNOPSIS

=head1 DESCRIPTION

This file describes methods and event handlers for a simple bulk
transport method, simply implemented using our Packets.
