package Lento::Connection;
use strict;
use Socket;
use POE qw(Wheel::SocketFactory Wheel::ReadWrite Driver::SysRW Filter::Line);
use Lento::ACT;
use Lento::Bulk::Source;
use Lento::Debuggable;
use Lento::EventSupport;
use Lento::Packet;
use Lento::Filter::Packet;
use Lento::List;
use Carp qw(cluck confess);
use POSIX qw(:errno_h);
use Exporter;

sub LVL_NOTCONN { 0 };
sub LVL_PING { 1 };
sub LVL_BIND  { 2 } ;
sub LVL_RECOVER { 3 };
sub LVL_FULL  { 4 };
# XXX: do an actual dump on a DAT packet and check this
sub METADATA_SIZE { 1024; };
sub BLOCK_SIZE  { 8192; };
sub PACKET_SIZE { METADATA_SIZE + BLOCK_SIZE; };

@Lento::Connection::ISA = qw(Lento::Debuggable Exporter);
@Lento::Connection::EXPORT = qw(LVL_NOTCONN LVL_PING LVL_BIND LVL_RECOVER LVL_FULL BLOCK_SIZE);
$Lento::Connection::level = -1;

# Maximum time in secs we allow a connection to be silent (no received packets)
# before we disconnect. 
$Lento::Connection::MAX_SILENCE = 30;

# Packet watermarks
$Lento::Connection::packet_low_water_mark = 2;
$Lento::Connection::packet_high_water_mark = 8;

# But POE requires these in octets
$Lento::Connection::octet_low_water_mark = 
    $Lento::Connection::packet_low_water_mark * PACKET_SIZE + 1;
$Lento::Connection::octet_high_water_mark = 
    $Lento::Connection::packet_high_water_mark * PACKET_SIZE - 1;

=head1 NAME

Connections - InterMezzo Connection class

=head1 SYNOPSIS

Lento::Connection->new($sysid);

=head1 DESCRIPTION

Connections manage an accepted and connected TCP sockets.  When in a
connected state, each object has POE session with a Packet Wheel to
receive and send data.

This dispatching function is performed by the got_data event handler,
invoked by the Wheel and Filter.  B<Incoming> packets need a
destination session. This is determined with the routine
packet_to_session, which follows the following rules:

=over 4

=item * the ReqDispatcher [type == REQ, session == none]

=item * an already active ReqSession expecting data.  

This session is resolved (with the help of the Asynchronous Completion
Token (ACT) package) through the B<tokens> in the packet.  

After the initial setup there are two types of packets needed: further
packets from the client to the server, and packets from the server to
the client.  The client will not have the server''s stoken until at
least a first packet has been received from the server.  The REP
packet type is always dispatched to the client, i.e. using the
B<ctoken>.  Similarly for START packets the B<ctoken> is used, and for
MSG packets the B<stoken> is used.  DAT and EOD packets are dispatched
to the {ctoken} found in the packet, but this packet may be either a
REQ or a START packet, and is known in the source as a
clonedata_packet.

=back

If the PacketWheel tells us that the connection was closed, this class will:

=over 4

=item * destroy the connection session, and wheel set up by the attach
method of the Connection, and close the socket.

=item * notify the Connector with a 'got_error' event.

=back

Connections objects are attached only instantiated by Connectors and
the Acceptor.

=head1 Events propagated to other sessions

=over 4

=item I<got_wheel>: posted to the connector when the socket has been
instantiated, and the Connection session is ready for use.

=item I<got_error>: propagate operation and errnum to all users and
Connector.

=item I<REQ,REP,DAT,EOD,START,MSG> Propagate these events to other
sessions when packets come in.

=back

=head1 Methods and Events

=cut

my %CONNDB;

sub new ($) {
    my $class = shift;
    my $sysid = shift;

    $sysid = "" unless defined $sysid;
    my $self = {};
    $self->{act_table} = new Lento::ACT;
    $self->{sysid} = $sysid;
    DEBUG "Setting Connection Level to 0\n";
    $self->{connection_level} = LVL_NOTCONN;

    $CONNDB{$sysid} = $self if $sysid;
    return bless $self, $class;
}


sub level {
    my $self = shift;
    my $level = shift;


    if ( defined $level ) {
        DEBUG "Setting Connection Level to $level\n";
        $self->{connection_level} = $level;
    }

    return $self->{connection_level};
}

# has the connection joined the barrier of a session's level? 
# if not, the session should generally experience a disconnection
sub joined_barrier {
    my $connection = shift;
    my $level = shift;

    if (!defined $connection or !defined $connection->{session} or
        $connection->level() < $level) {
        return 0;
    }
    return 1;
}

# Similar to joined barrier except we don't test for the connection level
sub connected {
    my $connection=shift;

    if (!defined $connection or !defined $connection->{session}) {
        return 0;
    }
    return 1;
}

sub find {
    my $sysid = shift;
    my $result = $CONNDB{$sysid}; 
    return $result;
}

sub setsysid {
    my $conn = shift;
    my $sysid = shift;
    $conn->{sysid} = $sysid;
    $CONNDB{$sysid} = $conn;
}

=pod

=over 4

=item B<attach> (Method)

Set set up a session owning the SysRW wheel for an accepted or
connected socket.

B<Example:>  $conn->attach($socket, $connector, $sysidid);

=back

=cut

sub attach {
    my $self = shift;

    $self->{socket} = shift;
    $self->{connector_obj} = shift;
    $self->{peer} = shift;
    $self->{direction} = shift;
    $self->{peer_addr} = shift;

    $self->{max_silence} = $Lento::Connection::MAX_SILENCE;
    $self->{monitor_delay} = 5;
    $self->{last_recv} = time();
    $self->{flow_sessions} = new Lento::List;
    $self->{monitor_sessions} = new Lento::List;

    DEBUG "Setting Connection Level to 1\n";
    $self->{connection_level} = LVL_PING;

    # $self->{session} should NOT be assigned here but in _start!!
    # the reason is that an error call may be executed right after
    # _start, but before this assignment takes place.
    new POE::Session
        ([$self], # one argument to _start
         $self, ['_start', 'got_error', 'got_packet', 'do_monitor'],
         'Lento::Connection', ['REQ', 'REP', 'DAT', 'EOD', 'EOR', 'START'],
         'Lento::EventSupport', ['_signal', '_child', '_stop',
                                 'alias_set', 'alias_remove'],
         
         # called when the packet buffer has emptied out a fair bit
         low_water => sub {
             $self->under_low_water();
         },
        );
}

sub under_low_water {
    my $self = $_[0];

    # XXX We may wish to do something different here
    $self->{flow_sessions}->iterate
      (sub {
           DEBUG "Posting low_water to ".$_[0]->get_heap()->{session_name}." ($_[0])\n";
           $poe_kernel->post(@_);
       }, 'low_water'
      );
}

sub over_high_water {
    my $self = $_[0];

    # XXX We may wish to do something different here
    $self->{flow_sessions}->iterate
      (sub {
           DEBUG "Posting high_water to ".$_[0]->get_heap()->{session_name}." ($_[0])\n";
           $poe_kernel->post(@_);
       }, 'high_water'
      );
}

=pod

=over 4

=item B<_start> (Event)

Event is delivered by the attach method, and attached a SysRW Wheel to the accepted or connected socket.

=back

=cut

sub _start {
    my $self = $_[ARG0];
    my $packet;

    $self->{session} = $_[SESSION];
    $_[HEAP]->{self} = $self;

    $_[HEAP]->{session_name} = $_[SESSION]->ID().":Connection ($self->{direction} $self->{peer})";
    LOG ">>>>> Connection $self->{sysid} starting\n";

    Lento::Debuggable::set_option($_[SESSION], trace => 1, default => 1);
    $_[KERNEL]->alias_set($_[SESSION]);

    # start the read/write wheel
    $_[HEAP]->{wheel} = new POE::Wheel::ReadWrite
      (Handle => $self->{socket},          # on this socket handle
       Driver => POE::Driver::SysRW->new(), # use sysread and syswrite
       Filter => Lento::Filter::Packet->new(),  # parse streams as packets
       InputState   => 'got_packet', # generate this event upon receipt
                                     # of input
       ErrorState   => 'got_error',  # and this event if an error occurs
       FlushedState => 'low_water',  # this event when all output is sent

       HighMark => $Lento::Connection::octet_high_water_mark,
       LowMark => $Lento::Connection::octet_low_water_mark,
       LowState => 'low_water',
       HighState => 'high_water'
      );

    $_[KERNEL]->delay('do_monitor', $self->{monitor_delay});
}

=pod

=over 4

=item  B<do_monitor> (Event)

Check if the socket has not received a packet for max_silence
or more.  If so, raise and error.

=back

=cut

sub do_monitor {
    my $self=$_[OBJECT];
    my $posted=0;

    if (!$_[HEAP]->{wheel}) {
        # The Connection has already disconnected.
        # XXX This can be removed if we ever make that POE patch to forcefully
        # destroy sessions by removing pending events!
        return;
    }

    DEBUG $_[HEAP]->{wheel}->get_driver_out_octets()." Bytes in buffer\n";
    DEBUG $_[HEAP]->{wheel}->get_driver_out_messages()." messages buffered\n";

    $self->{monitor_sessions}->iterate
      (sub {
        return if $posted;

        DEBUG "monitoring ".$_[0]->get_heap()->{session_name}." time ".time().
                " last_recv=".$_[0]->get_heap()->{last_recv}. 
                " max silence= $self->{max_silence}\n";

        if (time() - $_[0]->get_heap()->{last_recv} > $self->{max_silence}) {
            DEBUG "Monitor error in session ".$_[0]->get_heap()->{session_name}.
                " ($_[0]) posting got_error\n";
            # registering in the monitor list does an implicit get()
            #$poe_kernel->post($_[0], 'got_error', 'monitor', ETIMEDOUT, 
            #            'session timeout');

            # XXX: are we doing too much here? Should we just send the
            # session a got_error? On the other hand most sessions believe
            # there is disconnection when they get got_errors, so disconnect
            # for now.
            $poe_kernel->call($self->{session}, 'got_error', 'monitor', 
                ETIMEDOUT, 
                $_[0]->get_heap()->{session_name}.' session timeout');
            $posted=1;

            # XXX: got_error should take care of this.
            # $self->remove_from_monitor_list($_[0]);
        }
       }
      ); 

    return if $posted;

    if (time() - $self->{last_recv} > $self->{max_silence}) {
        # Be greedy about signalling errors, so call rather than post
        $_[KERNEL]->call($self->{session}, 'got_error', 'monitor', ETIMEDOUT,
                            'timeout');
        LOG(">>>> $_[HEAP]->{session_name} dead, disconnecting.\n");
        LOG("Stats: time:".time()." time last_rcvd: $self->{last_recv}"
            ." max silence: self->{max_silence}\n");
    } else {
        $_[KERNEL]->delay('do_monitor', $_[OBJECT]->{monitor_delay});
    }
}

=over 4

=item B<got_error> (event)

Arguments: operation, errnum, errstr.  Invoked by the Wheel when an
error or disconnect occurs.  The action taken is to destroy the Wheel,
close the socket, and to remove the Connection session from the object.

If the connection was instantiated by a Connector, the Connector is
notified with a 'got_error' event.  All users of the connection are
notified.

=back

=cut

sub got_error {
    my ($operation, $errnum, $errstr) = @_[ARG0, ARG1, ARG2];
    my $self = $_[OBJECT];

    DEBUG "got_error sent by ".$_[SENDER]->get_heap()->{session_name}.
          " ($_[SENDER])\n";

    if ($errnum) { # non-zero errnum is an error
        LOG( ">     Connection to $self->{sysid} ($_[HEAP]->{session_name}) detected $operation error $errnum: $errstr\n");
    } else { # zero errnum is a plain disconnect
        LOG( ">     Connection to $self->{sysid} ($_[HEAP]->{session_name}) detected a remote disconnect.\n");
        $errnum=ENOTCONN;
    }

    DEBUG "Setting Connection Level to 0\n";
    $self->{connection_level} = LVL_NOTCONN;

    # either way, disconnect
    delete $_[HEAP]->{wheel};
    shutdown($self->{socket}, 2);
    close $self->{socket};
    $_[KERNEL]->alias_remove($_[SESSION]);
    delete $self->{session};
    # must be a fedex/call to avoid race condition
    if ($_[OBJECT]->{connector_obj}) {
        $_[KERNEL]->call($self->{connector_obj}->{session},
                         'got_error', $operation, $errnum, $errstr);
    }

    foreach my $session (@{$self->{act_table}->sessions()}) {
        DEBUG "Posting (op $operation) to ".
            $session->get_heap()->{session_name}."\n";
        $poe_kernel->post($session, 'got_error', $operation, $errnum, 
                          $errstr);
        $self->{act_table}->put($session);
    }

    # Clear out the list of sessions to receive flow control events
    $self->{flow_sessions} = new Lento::List;

    # Clear out session monitor list...these sessions should take care of
    # themselves if they want to re-register!
    $self->{monitor_sessions} = new Lento::List;
}

=pod

=over 4

=item B<got_packet> (event)

Invoked by the Wheel/Filter when a packet arrives.

=back

=cut

sub got_packet {
    my $packet = $_[ARG0];

    # return if ($_[HEAP]->{'protocol state'} eq 'quitting');

    $_[OBJECT]->{last_recv} = time();

    $packet->[PKT_CONNECTION] = $_[OBJECT];
    my $session = packet_to_session($packet);

    if (! defined $session) {
       # This can happen if packets are delayed in WAN pipes.
       warn "!!! No session in got_packet...disconecting!!!\n";

       # XXX: the following line needs to be commented to  stop a
       # disconnect. I think that this might be overly aggressive...but the code
       # needs to be tested in WAN environments. In particular, a missing 
       # session does need some kind of error recovery.
       $_[KERNEL]->call($_[SESSION],'got_error', 'got_packet', ENOENT,
              'no such session');
       return;
    }

    DEBUG "dispatching packet: type $packet->[PKT_TYPE] stoken".  
          " $packet->[PKT_STOKEN], ctoken $packet->[PKT_CTOKEN] to session ".  
          $session->ID()." [".$session->get_heap()->{session_name}."]\n";
    $session->get_heap()->{last_recv}=time();
    $_[KERNEL]->post($session, $packet->event_type(), $packet);

    undef $packet;
    undef $_[ARG0];
}

=pod

=over 4

=item B<packet_to_session> (method)

Used by the got_packet method to determine the destination of the packet.

=back

=cut

sub packet_to_session {
    my $packet = $_[0];
    my $conn = $packet->[PKT_CONNECTION] || die;
    my $session;

    if ($packet->[PKT_TYPE] == $Packet::REQ) {
            $session =  $::dispatcher->{session}; # 'ReqDispatcher';
    } elsif ($packet->[PKT_TYPE] == $Packet::MSG) {
            $session = $conn->{act_table}->token_to_session($packet->[PKT_STOKEN]);
    } else {
            $session = $conn->{act_table}->token_to_session($packet->[PKT_CTOKEN]);
    }
    DEBUG "No session: type $packet->[PKT_TYPE] stoken $packet->[PKT_STOKEN], ctoken $packet->[PKT_CTOKEN]" unless $session;

    return $session;
}

sub register_for_flow {
    my $self = shift;
    my $session = $poe_kernel->get_active_session() || die;

    $self->{act_table}->get($session);

    die "no flow list?\n" if !defined $self->{flow_sessions};

    DEBUG "***** adding ".$session->get_heap()->{session_name}.
          " ($session) to flow control list!\n";
    $self->{flow_sessions}->add($session, $session->get_heap()->{session_name});
}

sub remove_for_flow {
    my $self = shift;
    my $session = $poe_kernel->get_active_session() || die;

    DEBUG "***** removing ".$session->get_heap()->{session_name}.
          " from flow control list!\n";
    $self->{flow_sessions}->del($session)
        if defined $self->{flow_sessions};
}

sub register_in_monitor_list {
    my $self = shift;
    my $session = $poe_kernel->get_active_session() || die;

    die "no monitor list?\n" if !defined $self->{monitor_sessions};

    # setup an initial value for last_recv
    $session->get_heap()->{last_recv}=time();
    $self->{act_table}->get($session);

    DEBUG "***** adding ".$session->get_heap()->{session_name}.
          " to monitor list!\n";
    $self->{monitor_sessions}->add($session, 
                $session->get_heap()->{session_name});
}

sub remove_from_monitor_list {
    my $self = shift;
    my $session = $poe_kernel->get_active_session() || die;

    DEBUG "***** removing ".$session->get_heap()->{session_name}.
          " from monitor list!\n";
    $self->{monitor_sessions}->del($session) 
            if defined $self->{monitor_sessions};
}

# Report hard errors to the connection. This function will tear down
# an existing connection if any.
sub set_error {
    my $connection_obj = shift;
    my $operation = shift;
    my $errno = shift;
    my $reason = shift;

    if (defined $connection_obj->{session}) {
        # XXX: should this be a call or a post? We make it a call since
        # an unrecoverable error has happened implying that we should
        # disconnect ASAP without waiting for the event to be scheduled.
        $poe_kernel->call($connection_obj->{session}, 'got_error', 
            $operation, $errno, $reason);
    }
}

#
#  SENDING stuff
#

=pod

=over 4

=item B<sendpacket> (method)

This function is a wrapper for the individual packet sending events
delivered to the session attached to the connection.  This wrapper is
present to handle failed connections gracefully; here the connection
object will persist, but the session may disappear, and the wrapper is
the entity that knows about this.

Any session calling this is added to the list of users.  These users
are notified if an error occurs.

B<Examples:>

$conn->sendpacket('REQ', ["method", "arg1", ...]);

$conn->sendpacket('REP', $packet, ["result", "arg1", ...]);

$conn->sendpacket('DAT', $packet, \$buffer);

$conn->sendpacket('EOD', $packet, ["result", "numbytes", ...]);

$conn->sendpacket('EOR', $packet);

$conn->sendpacket('START', $packet);

=back

=cut

sub sendpacket {
    my $self = shift;
    my $type = $_[0];

    my $heap = $poe_kernel->get_active_session()->get_heap();

    DEBUG("LEVEL: Packet sent from session  level $heap->{session_level} conn level $self->{connection_level}  \n");

    if ( ! defined $heap->{session_level} ) {
        $poe_kernel->yield('got_error', ENOTCONN, 'Not Connected');
        return;
    }

    if ($heap->{session_level} > $self->{connection_level}) {
        $poe_kernel->yield('got_error', ENOTCONN, 'Not Connected');
        LOG("*** Session level too high for connection state!\n");
        return;
    }

    if ($heap->{session_level} < $self->{connection_level}) {
        # XXX Ping sessions do this all the time -- perhaps 
        # we can adjust the level of the Pinger sessions
        # LOG "**&&** Packet sent from low level session\n";
    }

    # REQ does this itself with a newact!
    if ( $type ne 'REQ' ) {
            $self->get();
    }

    if ($self->{session}) {
        DEBUG "using session ". $self->{session}."\n";
        $poe_kernel->post($self->{session}, @_);
    } else {
        DEBUG "no session while sending packet\n";
        $poe_kernel->yield('got_error', 'sendpacket', ENOTCONN, 'Not Connected');
    }
}

=pod

=over 4

=item B<put> (method)

Called from within a session when the connection is no longer used by
the session.  Error notifications will not be delivered after this is
called.

=back

=cut

sub put {
    my $self = shift;
    my $session = shift || $poe_kernel->get_active_session();

    $self->{act_table}->put($session);
    delete $self->{flow_sessions}->{list}->{$session};
    delete $self->{monitor_sessions}->{list}->{$session};
}

sub get {
    my $self = shift;
    my $session = shift || $poe_kernel->get_active_session();

    $self->{act_table}->get($session);
}

=pod

=head1 Sending Packet events

One sends packets by posting events to a Connection session.  This is
only done from sendpacket and not directly, to avoid sending messages
to evaporated sessions.  The various packet types are detailed below,
with the appropriate event name in brackets.

=over 4

=item * Request [REQ]

Request requires at least one argument, the method to be executed on the remote
system.  Additional optional arguments are passed to that method.

B<Example:>

$_[KERNEL]->post($conn->{session}, 'REQ', ["method", "arg1", ...]);

=cut

sub REQ {
    my $args = $_[ARG0];
    my $conn_obj = $_[HEAP]->{self};
    my $packet = new Lento::Packet;

    $packet->[PKT_TYPE] = $Packet::REQ;
    if (scalar(@{$args}) == 0) {
        confess "No method in req packet\n";
    }

    my $act = $conn_obj->{act_table}->newact($_[SENDER]);
    $packet->[PKT_CTOKEN] = $act;

    # freeze method with args
    $packet->[PKT_BODY] = Storable::nfreeze($args);
    $packet->[PKT_STOKEN] = 0;
    $packet->[PKT_SIZE] = length($packet->[PKT_BODY]);
    if (!$_[HEAP]->{wheel}) {
        $_[KERNEL]->yield('got_error', 'sendpacket', EBADF,
                            'dead connection');
        DEBUG "not sending REQ on dead connection\n";
        return;
    }

    DEBUG "$_[SENDER]->get_heap()->{session_name} using $_[HEAP]->{session_name} to send REQ\n";
    if ($_[HEAP]->{wheel}->put($packet)) {
        # put returns a true value if we're over the high water mark
        DEBUG "put returned HighWater\n";
        $conn_obj->over_high_water();
    }
}

=pod

=item * Reply [REP]

Reply requires two arguments: the packet which is being replied to, and the
result code.  Additional optional arguments are passed to the reply handler.

B<Example:>

$_[KERNEL]->post($conn->{session}, 'REP', $packet, ["result", "arg1", ...]);

=cut

# pass this method: packet, result code, other return vals
sub REP {
    my $oldpkt = $_[ARG0];
    my $conn_obj = $_[HEAP]->{self};
    my $repargs = $_[ARG1];

    my $packet = $oldpkt->clone();
    my $result = $repargs->[0];
    confess "no return code in rep packet\n" unless defined ($result);

    $packet->[PKT_TYPE] = $Packet::REP;
    $packet->[PKT_BODY] = Storable::nfreeze($repargs);
    $packet->[PKT_SIZE] = length($packet->[PKT_BODY]);

    DEBUG "$_[SENDER]->get_heap()->{session_name} using $_[HEAP]->{session_name} to send REP\n";
    if (!$_[HEAP]->{wheel}) {
            $_[KERNEL]->yield('got_error', 'sendpacket', EBADF,
                          'dead connection');
        DEBUG "not sending REP on dead connection\n";
        return;
    }

    if ($_[HEAP]->{wheel}->put($packet)) {
            # put returns a true value if we're over the high water mark
            $conn_obj->over_high_water();
    }

}

=pod

=item * Data [DAT]

Data requires one and only one argument, the packet which is being replied to.
packet->{body} B<must> be set before one tries to send a data packet.

B<Example:>

$kernel->post($conn->{session}, 'DAT', $packet, \$buffer);

=cut

# arguments: packet.  packet must already have a valid body!
sub DAT {
    my $oldpkt = $_[ARG0];
    my $conn_obj = $_[HEAP]->{self};
    my $refbuf = $_[ARG1];

    my $packet = $oldpkt->clone();

    $packet->[PKT_TYPE] = $Packet::DAT;
    $packet->[PKT_BODY] = $$refbuf;
    $packet->[PKT_SIZE] = length($packet->[PKT_BODY]);

    DEBUG $_[SENDER]->get_heap()->{session_name}." using $_[HEAP]->{session_name} to send DAT\n";
    if (!$_[HEAP]->{wheel}) {
            $_[KERNEL]->yield('got_error', 'sendpacket', EBADF,
                          'dead connection');
        DEBUG "not sending DAT on dead connection\n";
        return;
    }


    if ($_[HEAP]->{wheel}->put($packet)) {
            # put returns a true value if we're over the high water mark
            $conn_obj->over_high_water();
    }
}

=pod

=item * End of Record [EOR]

EOR requires one and only one argument, the packet which is being replied to.
packet->{body} B<must> be set before one tries to send an EOR  packet.

B<Example:>

$kernel->post($conn->{session}, 'EOR', $packet, \$buffer);

=cut

# arguments: packet.  packet must already have a valid body!
sub EOR {
    my $oldpkt = $_[ARG0];
    my $conn_obj = $_[HEAP]->{self};
    my $refbuf = $_[ARG1];

    my $packet = $oldpkt->clone();

    $packet->[PKT_TYPE] = $Packet::EOR;
    $packet->[PKT_BODY] = $$refbuf;
    $packet->[PKT_SIZE] = length($packet->[PKT_BODY]);

    DEBUG $_[SENDER]->get_heap()->{session_name}." using $_[HEAP]->{session_name} to send DAT\n";
    if (!$_[HEAP]->{wheel}) {
            $_[KERNEL]->yield('got_error', 'sendpacket', EBADF,
                          'dead connection');
        DEBUG "not sending EOR on dead connection\n";
        return;
    }

    if ($_[HEAP]->{wheel}->put($packet)) {
            # put returns a true value if we're over the high water mark
            $conn_obj->over_high_water();
    }
}

=pod

=item * End of Data [EOD]

End of Data takes as its first argument the packet which is being
replied to and two optional arguments: number of bytes sent 
and takes an argument for the return code.  Other arguments
may be added if the protocol requires them. If the 
return code is not specified a value of 0 is assumed. 
The packet's body, size, etc are automatically cleared.

B<Example:>

$kernel->post($conn->{session}, 'EOD', $packet, [$result, $numbytes,...]);

=cut

# arguments: packet
sub EOD {
    my $oldpkt = $_[ARG0];
    my $conn_obj = $_[HEAP]->{self};
    my $packet = $oldpkt->clone();
    my $repargs = $_[ARG1];

    $repargs=[0] unless defined $repargs;

    my $result = $repargs->[0];
    $repargs->[0]=0 unless defined ($result);

    $packet->[PKT_TYPE] = $Packet::EOD;
    $packet->[PKT_BODY] = Storable::nfreeze($repargs);
    $packet->[PKT_SIZE] = length($packet->[PKT_BODY]);

    DEBUG $_[SENDER]->get_heap()->{session_name}." using $_[HEAP]->{session_name} to send EOD with result code $repargs->[0]\n" if defined $_[SENDER];
    DEBUG "$repargs->[1] bytes were sent\n";
    if (!$_[HEAP]->{wheel}) {
            $_[KERNEL]->yield('got_error', 'sendpacket', EBADF,
                          'dead connection');
        DEBUG "not sending EOD on dead connection\n";
        return;
    }


    if ($_[HEAP]->{wheel}->put($packet)) {
            # put returns a true value if we're over the high water mark
            $conn_obj->over_high_water();
    }
}

=pod

=item * Begin bulk transfer START

START indicates to a client (sender of REQ) that the servers bulk
transfer handler is ready to receive data.  The START packet contains
the correct server ACT field which the client should use to send data. 

B<Example:>

$kernel->post($conn->{session}, 'START', $packet);

=back

=cut

# arguments: packet
# note: this is sent from server to client, when the server is a sink
sub START {
    my $source_act = $_[ARG0]->[0];
    my $sink_act = $_[ARG0]->[1];
    my $conn_obj = $_[HEAP]->{self};
    my $packet = Lento::Packet->new();

    $packet->[PKT_TYPE] = $Packet::START;
    $packet->[PKT_STOKEN] = $sink_act;
    $packet->[PKT_CTOKEN] = $source_act;
    $packet->[PKT_SIZE] = 0;
    $packet->[PKT_BODY] = "";

    DEBUG $_[SENDER]->get_heap()->{session_name}." using $_[HEAP]->{session_name} to send START, sink_act $sink_act, source_act $source_act\n";
    if (!$_[HEAP]->{wheel}) {
            $_[KERNEL]->yield('got_error', 'sendpacket', EBADF,
                          'dead connection');
        DEBUG "not sending START on dead connection\n";
        return;
    }


    if ($_[HEAP]->{wheel}->put($packet)) {
            # put returns a true value if we're over the high water mark
            $conn_obj->over_high_water();
    }
}

1;


package TestClientConnection;
use POE;

sub _start {
  my ($object, $heap, $socket) = @_[OBJECT, HEAP, ARG0];
  my $packet;
                                        # start the read/write wheel
  $heap->{'wheel'} = new POE::Wheel::ReadWrite
    ( Handle     => $socket,                # on this socket handle
      Driver     => new POE::Driver::SysRW, # using sysread and syswrite
      Filter     => new Lento::Filter::Packet,   # and parsing streams as lines
      InputState => 'got_response',         # generate this event upon input
      ErrorState => 'got_error'             # and this event in an error occurs
    );

  $packet = Lento::Packet->new();
  $packet->init(1, 2, 3, "Moose\n");
  $heap->{wheel}->put($packet);

  $packet = Lento::Packet->new();
  $packet->init(1, 2, 3, "Hello\n");
  $heap->{wheel}->put($packet);

  $packet = Lento::Packet->new();
  $packet->init(1, 2, 3, "");
  $heap->{wheel}->put($packet);

}

1;

__END__
