package Lento::Connector;
use strict;
use Lento::ACT;
use Lento::Connection;
use Lento::Debuggable;
use Lento::EventSupport;
use Lento::List;
use Lento::Packet;
use Socket;
use POSIX qw(:errno_h);
use POE;
use Carp qw(cluck confess);

@Lento::Connector::ISA = qw(Lento::Debuggable);
$Lento::Connector::level = 1;

=head1 NAME

Connectors - a class to establish a connected TCP socket with a peer.

=head1 SYNOPSIS

$connector = Lento::Connector->new($connection_obj, $peer, $peerport, $bindaddr);


=head1 DESCRIPTION

A connector object is a daemon session which manages the Connection
session to a peer/port pair, from a client perspective.  It is used on
the systems which establish connections. It is not for accepting
connections.

It knows of all the sessions that requested such a connection in a
List of clients.

The connection session is not explicitly held by the connector, to
avoid race conditions in its destruction upon errors.  The Connection
object is held.

=head1 Responsibilities

The responsibilities of this class and session are:

=over 4

=item * to do an asynchronous connect 

=item * to initiate a connection to a peer on behalf of a client session.

=item * to notify connecting client sessions of error conditions which
led the socket to be closed before the connection was fully established.

=back

=head1 Access methods

=over 4

=item I<new>: instantiate a new connector

=item I<connectme>: try to connect

Yield with a I<got_connection> message to the caller if an existing
Connection session exists.  Else set up a Connection session to the
peer/port.  When a new Connection has been created, all client
sessions will be notified with a I<got_connection> event in the
I<got_wheel> event handler.


=back

=head1 Events propagated to other sessions

=over 4

=item I<got_connection>: posted to the caller(s) of B<connectme> when the
Connection session is ready for use.

=item I<got_error>: propagate errors to clients that have places a
got_connection request.  Such error messages are no longer delivered
after the clients have been sent a got_connection event, at which
point they need to register with the connection object to get error
messages.   It sends, operation and errnum to all clients.

=back 

=head1 Other events handled internally

=over 4

=item I<get_connection> (internal Connector event) Performs a connect
to a peer, using a connecting SocketFactory wheel.  We will be
notified of success at the internal got_connection event.  Sent by the
connectme method.

=item I<connect_done>: the connect has succeeded.  The socket is now
connected. Proceed by attaching the Connection to a SysRW wheel using
this socket.

=item I<got_wheel>: The socket with the SysRW Wheel is now ready for
use.  Give all clients to the Connection object, using \"get\".
Notify all clients that they can start using the connection, by
sending the got_connection event to client sessions.

=item I<got_error>: an error arose during connection setup. This event
can be delivered by the connecting wheel or by the Connection during
connecting or socket establishment operations. It notifies all clients
of the error.

=item I<check_init>: set up a check with 5 secs delay to verify
connection was established.

=item I<check>: carry out the check.  Close and shutdown any socket that was instantiated. Also send a got_error to the connector session.


=back

=head1 Problematic issues

=over 4

=item * POE can deliver incorrect got_connections right after
got_errors.  We mmanage this by maintaining a state variable.
Got_connection is only accepted as success if the state was
connecting.

=item * We want the following: a client using the connection should be
either using the Connector or the Connection, never both. This is
achieved by handing over all clients to the connection upon attach.

=item * Connectors and Connections should post different error
messages

=back

=cut
#' stupid emacs.


# make a new connector object
sub new ($$$) {
    my $class = shift;
    my $self = bless { }, $class;

    $self->{connection} = shift || confess "No connection passed";
    $self->{peer} = shift  || die "No peer";
    $self->{peerport} = shift || die "No peer port";
    $self->{bindaddr} = shift;

    $self->{clients} = new Lento::List;
    $self->{state} = "idle";

    LOG "Connector to $self->{peer}, $self->{peerport} being created.\n";

    # wrap this object in a POE session
    new POE::Session
        ([$self], # one argument to _start
         "Lento::EventSupport", ['_signal', '_child', '_stop', 'alias_remove',
                                'alias_set'],
         $self, ['get_connection', 'connect_done', 
                    'got_error', 'check', 'check_init'],
         _start => sub {
                my $connector = $_[ARG0];
                $connector->{session} = $_[SESSION];
                $_[HEAP]->{session_name} = $_[SESSION]->ID().
                                       ": Connector (" . $connector->{peer} . ")";
                Lento::Debuggable::set_option($_[SESSION],
                                                trace => 1, default => 1);
                $_[KERNEL]->alias_set($_[SESSION]);
         }
        );

    return $self;
}

# event handler: internal asynchronous connect
sub get_connection {
    my $bindaddr = undef;
    die "Connected already!" 
                if Lento::Connection::connected($_[OBJECT]->{connection});
    die "Connecting already!" 
                if ($_[HEAP]->{wheel});

    $_[OBJECT]->{state} = 'connecting';

    if ($_[OBJECT]->{bindaddr}) {
        $bindaddr = $_[OBJECT]->{bindaddr}; 
    }

    # the wheel doing a connect to the peer
    $_[HEAP]->{wheel} = POE::Wheel::SocketFactory->new
      (SocketDomain    => AF_INET,    # Internet address family
       SocketType      => SOCK_STREAM, # create stream sockets
       SocketProtocol  => 'tcp',          # using the tcp protocol
       RemoteAddress   => $_[OBJECT]->{peer}, 
       RemotePort      => $_[OBJECT]->{peerport},
       BindAddress     => $bindaddr,     # be sure to use our outgoing IP
       Reuse           => 'yes',            # reusing address and port
       SuccessState    => 'connect_done', # send when connected
       FailureState    => 'got_error',  # send this upon failure
      );
}

# method: external access point to get a connection
sub connectme {
    my $self = shift;

    # XXX INJECT connection got_error here
    # ASSERT state is not 'connected'

    if ($self->{state} eq 'connected') {
        DEBUG("CONNECTOR: got_connection -- connected\n");
        $self->{connection}->get(); 
        $poe_kernel->yield('got_connection');
    } else {
        my $session = $poe_kernel->get_active_session();
        my $heap = $session->get_heap();
        
        DEBUG "Adding ".$heap->{session_name}.
            "(".$session.") to clients\n";
        $self->{clients}->add($session, $heap->{session_name});
        if ( $self->{state} ne 'connecting') {
            $self->{state} = 'connecting';
            $poe_kernel->post($self->{session}, 'get_connection');
            $poe_kernel->post($self->{session}, 'check_init');
        }
    }
}

# connect went through, now set up the wheel
sub connect_done {
    my $socket = $_[ARG0];
    my $connector = $_[OBJECT];
    my $self = $_[OBJECT];

    delete $_[HEAP]->{wheel};

    if ($_[OBJECT]->{state} eq "connecting") {

        DEBUG "State is connecting, attaching socket\n";

        $_[OBJECT]->{connection}->attach
            ($socket, $_[OBJECT], "$_[OBJECT]->{peer}", "to");
    }

    # recheck the state since a disconnection got_error right after _start
    # in Connection would call into our error routines
    if ($_[OBJECT]->{state} eq "connecting") {
        setsockopt($socket, SOL_SOCKET, 0x01, 1)
                or die "Couldn't disable Nagle's algorithm: $!\a\n";

        $self->{state} = 'connected';
        $self->{clients}->iterate
            (sub {    my $session = shift;
                      my $connection = shift;
                      $connection->get($session);
                      $poe_kernel->post($session, 'got_connection');
                      LOG ">>>>> Connector ".$session->get_heap()->{session_name}.
                          ": got_connection.\n";
                  }, $self->{connection}
            );
        
        # empty out our client list - we've given them to the connection
        $self->{clients} = new Lento::List;
    } else {
        # Our state was reset due to an error.
        DEBUG "State is $_[OBJECT]->{state} - post error\n";
        $_[KERNEL]->yield('got_error', 'connector got connection',
                          ENOTCONN,'not connected');
    }
}

# needed since there is no cross-session delay command
sub check_init {
    $_[KERNEL]->delay('check', 5);
}

# check if the socket is now "up".  If not, kill it.
sub check {
    my $connector = $_[OBJECT];

    if ( $_[OBJECT]->{state} eq "connecting") {
        $_[OBJECT]->{state} = "error pending";
        if ($_[HEAP]->{wheel} && 
            defined $_[HEAP]->{wheel}->{handle}) {
            $_[KERNEL]->select($_[HEAP]->{wheel}->{handle});
            shutdown($_[HEAP]->{wheel}->{handle}, 2);
            close ($_[HEAP]->{wheel}->{handle});
        }

        delete $_[HEAP]->{wheel};
        $_[KERNEL]->yield('got_error', 'Connector check', 
                          ETIMEDOUT, 'Time out');
    }
}

# send all clients a got_error event
# state -> idle
sub got_error {
    my ($operation, $errnum, $errstr) = @_[ARG0, ARG1, ARG2];
    my $self = $_[OBJECT];
    my $client;

    if( !defined( $errstr ) ) { $errstr= "NONE"; }
    if( !defined( $errnum ) ) { $errnum= 0; }

    LOG("ERROR: connector got_error $errstr ($errnum) in state $self->{state} sent by" . 
        $_[SENDER]->get_heap()->{session_name} . " ($_[SENDER]) operation ($operation)\n");

    delete $_[HEAP]->{wheel};
    $self->{state} = 'idle';
    $self->{clients}->iterate
        (sub {    $poe_kernel->post(@_);
                  DEBUG "Sending got_error to $_[0]->get_heap()->{session_name} ($_[0])\n";
              }, 'got_error', $self->{peer}, $operation, $errnum
        );
    
    $self->{clients} = new Lento::List;
}

1;

__END__
