package Lento::EventSupport;
use strict;
use Carp;
use POE;
use Lento::Debuggable;

@Lento::EventSupport::ISA = qw(Debuggable);
@Lento::EventSupport::level = -1000;


sub  _signal {
    if (!defined $_[HEAP]->{session_name}) {
        confess "nameless";
    }

    if ( defined($_[ARG2]) ) {
        DEBUG "<<    $_[HEAP]->{session_name} got signal $_[ARG2].\n";
    } else {
        DEBUG "<<    $_[HEAP]->{session_name} got signal.\n";
    }
}
 
sub  _child {
    DEBUG "<<    $_[HEAP]->{session_name} got _child: $_[ARG0], " .
      " (child is $_[ARG1]->get_heap()->{session_name}.)\n";
 1;
}

sub _stop  {
    delete $_[HEAP]->{self}; # facilitate garbage collection
    DEBUG "<<    $_[HEAP]->{session_name} going away.\n";
}

=pod

The wait requirements in Lento are manifold, and we should combine
this into a single session waiting mechanism.

=item * Often a session wants to wait for multiple other sessions to
complete.  XXX currently we do not manage the counter associated with
the number of sessions to wait for correctly.  However, such a counter
is managed for upcall waiting.

=item * Some sessions may wait for a specific event to happen in
another session and then want to be woken up.  For example, an upcall,
received in the ReqDispatchter, with a given number may need to be
waited for.  Or an upcall with a cookie (waitsequence number) may need
to be waited for.

Here the session that wakes up the other sessions needs to match the
cookie or upcall number to locate the sessions that need to be woken.

=cut

###############
# GROUP 1
###############

#  Send Events from Waker to Waiters.
sub WakeUpWaitList { 
  my @post_args= @_;
  my $waitlist = $poe_kernel->get_active_session()->get_heap()->{waitlist};
  if (defined $waitlist) {
  DEBUG("  >>  IN WakeUpWaitList waitlist= $waitlist\n");
  }

  if ( defined $waitlist  ) {
      $waitlist->iterate
                (
                 sub {
                        my $arg = shift;
                        my $session = $arg->[0];
                        my $callbackevt = $arg->[1];
                        my $waker =$poe_kernel->get_active_session();
                        DEBUG(">>>> Waking Up: %s $callbackevt\n", 
                                        $session->get_heap()->{session_name});  
                        $poe_kernel->post($session, $callbackevt, @post_args);
                        $poe_kernel->call($session, 'alias_remove', 
                                        "WAITER: ". $session->ID(). 
                                        "for $callbackevt from " . $waker->ID()); 
                        $waitlist->del($arg);
         }
        );
    }
}

#  Send event from Waiter to Waker
sub WaitForSession {
        my $waker = shift;
        my $callbackevt = shift;
        my $name = $poe_kernel->get_active_session()->get_heap()->{session_name};
        $name = $poe_kernel->get_active_session() unless defined $name;
        confess unless defined $waker;
        confess unless defined $waker->get_heap()->{session_name};

        my $waiter = $poe_kernel->get_active_session();
        
        DEBUG("  >>  $name WaitForSession: $callbackevt from ".$waker->get_heap()->{session_name}."\n");
        $poe_kernel->alias_set("WAITER: " . $waiter->ID() . "for $callbackevt from " . $waker->ID());
        $poe_kernel->call($waker, 'AddWaiter', $poe_kernel->get_active_session(), 
                          $callbackevt);
}

#  Event handled by Waker
#  Only used in the previous method as event handler
sub AddWaiter {
        my $session = $_[ARG0];
        my $evt = $_[ARG1];

        DEBUG(" AddWaiter to %s\n", $poe_kernel->get_active_session() );
        if ( ! defined $_[HEAP]->{waitlist} ) {
          $_[HEAP]->{waitlist} = new Lento::List;
        }
        $_[HEAP]->{waitlist}->add([$session, $evt]);
}

#########################
# GROUP 2 - UpcallWaiting
#########################

# Called by the Waker - send events to matching Waiters.
# XXX changes the waitcount in the Waiter with RemoveUpcallWaiter
sub WakeUpcallList { 
    my $seq_num= shift;
    my $waitlist = $poe_kernel->get_heap()->{uc_waitlist};
    DEBUG(" IN WakeUpcallList ($seq_num) waitlist= $waitlist\n");
    if ( defined $waitlist  ) {
        $waitlist->iterate
            (
             sub {
                 my $arg = shift;
                 my $actual_seq_num= shift;
                 my $session = $arg->[0];
                 my $evt = $arg->[1];
                 my $snum= $arg->[2];
                 if ( $actual_seq_num == $snum ) {
                     
                     DEBUG("WakeUpcallList :: calling $session $evt\n");
                     $poe_kernel->call($session, 'RemoveUpcallWaiter');
                     $poe_kernel->post($session, $evt); # , $post_args);
                     $waitlist->del($arg);
                 }
             }, $seq_num
            );
    }
}

# Send event to Waker, including the $upc_num to mathc on. 
sub WaitForUpcall {
    my $upc_num = shift;
    my $callbackevt = shift;
 
    DEBUG("WaitForUpcall to %s for upcall $upc_num ($callbackevt)\n", 
        $poe_kernel->get_active_session() );
    my $res= $poe_kernel->call($::dispatcher->{session}, 'AddUpcallWaiter', 
                               $poe_kernel->get_active_session(), $callbackevt, $upc_num);
};

# Event handler in the Waker
# XXX changes the waitcount in the Waker with RemoveUpcallWaiter
sub AddUpcallWaiter {
    my $session = $_[ARG0];
    my $evt = $_[ARG1]; 
    my $upc_num= $_[ARG2];

    DEBUG("AddUpcallWaiter to %s for upcall $upc_num\n",
       $session );
    if ( ! defined $_[HEAP]->{uc_waitlist} ) {
            $_[HEAP]->{uc_waitlist} = new Lento::List;
    }
    $_[HEAP]->{uc_waitlist}->add([$session, $evt, $upc_num]);
    if ( ! defined $_[HEAP]->{num_uc_waiting_sessions} ) {
            $_[HEAP]->{num_uc_waiting_sessions}= 1;     
    } else {
            $_[HEAP]->{num_uc_waiting_sessions}++;
    }
};


# Event handled by Waiter
# XXX Changes wait count in the Waiter
sub RemoveUpcallWaiter {
   $_[HEAP]->{num_uc_waiting_sessions}--; 
};

# Called in the Waker
sub init_upcall_waiting {
    $_[HEAP]->{num_uc_waiting_sessions}= 0;
};

###########################
# GROUP 3 -- Cookie waiting
###########################

my $next_response_cookie= 1;

# this method wakes a session that is waiting for a particular cookie.
sub RespondWithCookie {
    my $upcall = shift;
    my $cookie = $upcall->{cookie};
        
    Lento::EventSupport::WakeCookieList( $cookie );
}

sub get_cookie {
    if ( ! defined $_[HEAP]->{my_cookie} ) {
            $_[HEAP]->{my_cookie}= $next_response_cookie++;
    }
    return $_[HEAP]->{my_cookie};
};


sub WaitForCookie {
    my $cookie = shift;
    my $callbackevt = shift;
    my $res= $poe_kernel->call($::dispatcher->{session}, 'AddCookieWaiter', 
                               $poe_kernel->get_active_session(), $callbackevt, $cookie);
};


sub WakeCookieList {
    my $cookie = shift;
    my $waitlist =  $poe_kernel->get_active_session()->get_heap()->
                            {cookie_waitlist};

    if ( defined $waitlist  ) {
        $waitlist->iterate
            (
             sub {
                    my $arg = shift;
                    my $session = $arg->[0];
                    my $evt = $arg->[1];
                    my $ecookie= $arg->[2];
                 
                    if ( $ecookie == $cookie ) {
                        DEBUG("WakeCookieList :: calling $session $evt\n");
                        $poe_kernel->post($session, $evt);
                        $waitlist->del($arg);
                    }
             }
            );
    }
};

# Event handler in the Waker
sub AddCookieWaiter {
    my $session = $_[ARG0];
    my $evt = $_[ARG1]; 
    my $cookie= $_[ARG2];

    my $waiter_session=$poe_kernel->get_active_session();

    DEBUG("AddCookieWaiter for ".$session->get_heap()->{session_name}.
            " with cookie $cookie\n");
    if ( !defined $waiter_session->get_heap()->{cookie_waitlist} ) {
            $waiter_session->get_heap()->{cookie_waitlist} = new Lento::List;
    }
    $waiter_session->get_heap()->{cookie_waitlist}->add(
            [$session, $evt, $cookie]);
};

# Support for persistent sessions
sub alias_set {
    my $alias = $_[ARG0];
    $_[KERNEL]->alias_set($alias);
}

sub alias_remove {
    my $alias = $_[ARG0];
    $_[KERNEL]->alias_remove($alias);
}

# manages state bitmaps in a class
sub state {
    my $self = shift;
    my $var = shift;
    my $val = shift;

    if (defined $val) {
        if ( $val != 1 && $val != 0 ) {
            confess "Bad value!";
        }
        if ( $val ) {
            $self->{state} |= $var;
        } else {
            $self->{state} &= ~$var;
        }
    }
    $self->{state} & $var;
}


1;

__END__
