package Lento::Fileset;
use POSIX qw(:errno_h);
use strict;
use File::stat;
use Carp;
use Lento::Packet;
use Lento::Client;
use Lento::KML;
use Lento::LML;
use Lento::Connection;
use Lento::Debuggable;
use Lento::Replicator;
use Lento::List;
use Lento::Server;
use Lento::Fsetdb;
use POE;

@Lento::Fileset::ISA = qw(Lento::Debuggable);
$Lento::Fileset::level = 1;


=pod

=head1 NAME

Fileset - InterMezzo file set abstractions

=head1 SYNOPSIS

No synopsis yet.

=head1 DESCRIPTION

Fileset have state




=over 4
=item  B<name>: name of the fileset
=item  B<serversysid>: sysid of the server holding this fileset
=item  B<mountpoint>: where is this fileset mounted?
=item  B<status>: (mounted, online)
=item  B<callbacks>: a list of session that needs notification when
                     reintegration completes
=item  B<kml>: the modification log for this fileset
=back

On servers of filesets some server specific state is used:
=over 4

=item B<fshl>: file system holding location of this fileset: can either
be an InterMezzo mount point, or in case of a server not running a
client, another file system location.

=item  B<permitholder>: the sysid of the system holding the permit
=item  B<replist>: a list head for replicators of this fileset, used mainly 
       for update forwarding
=back

On clients 
=over 4

=item B<state>: disconnected, synced, sending, receiving
=back 

=head1 METHODS

=over 4
=item * I<cache_path>: concatenation of mount point and relative path.
=item * I<fshl_path>: concatenation of fshl and relative path
=item * I<fs_path>: relative path wrt to the fileset mountpoint
=item * I<register_session_callback>: add a session to the list
=item * I<reint_done>: deliver reint_done event to sessions on list
=item * 
=back

=cut


sub new ($$$$$$) {
    my $class = shift || die "No class";
    my $self = {};
    bless ($self, $class);
    my $ro_flag;
    my $stat;

    $self->{name} = shift;
    $self->{serversysid} = shift;
    $self->{mtpt} = shift;
    $ro_flag = shift;
    $self->{fetch_on_demand}= shift;
    $self->{steal_permit}= shift;
    $self->{jclose_on_write}= shift;
    $self->{fetchtype}= shift;
    my $replicators = shift;

    # need the mount point of the cache
    $stat = stat($self->{mtpt}) || 
		die "Cannot stat fset's $self->{name} mount point $self->{mtpt}";
    $self->{cache_mtpt} = $::cachetable->{dev2mtpt}->{$stat->dev}; 

    # Initialize the list of session to be notified when reintegration finishes
    $self->{callback} = new Lento::List;

    # we set these once we know if we are a server for this fileset.
    $self->{lento_read_only}= 0;
    $self->{presto_read_only}= 0;
    
    # tell the kernel that the mountpoint is a file set root.
    # If this call fails the fset is already open and incomplete closes
    # have already been recovered...
    $self->{recovered_localcloses}= 
            !$::psdev->setfsetroot($self->{mtpt}, $self->{name}, 1, 0x0);
    $self->{recovered_backfetches}=0;

    DEBUG(" recovered localcloses: $self->{recovered_localcloses}\n");

    # get the KML - this requires the cache mount point
    $self->{kml} = Lento::KML->new
		("$self->{cache_mtpt}/.intermezzo/$self->{name}/kml", $self);

    # get the LML - this requires the cache mount point...
    $self->{lml} = Lento::LML->new
		("$self->{cache_mtpt}/.intermezzo/$self->{name}/lml", $self);

	# ...and create a hash of backfetch lml records
	# This could be done along with Lento::LML::new, but then lml-print would
	# suffer from an unecessary overhead.
    $self->{lml}->populate_backfetch_hash();

    if ($self->{steal_permit}) {
        $::psdev->enable_permit_stealing($self->{mtpt});
    } else {
        $::psdev->disable_permit_stealing($self->{mtpt});
    }

    if ($self->{jclose_on_write}) {
        $::psdev->enable_jclose_on_write($self->{mtpt});
    } else {
        $::psdev->disable_jclose_on_write($self->{mtpt});
    }

    my $server = Lento::Server::find($self->{serversysid});
    if (!defined($server)) {
	    die "Server $self->{serversysid} not known (fset $self->{name})";
    }

    LOG ">>   - Seting fset state to disconnected for fset $self->{name}\n";
    $self->{state} = "disconnected"; 

    # is this a syncing fileset? 
    if ( $self->synced () || $self->ishere() ) {
	    DEBUG("$self->{name} is IN SYNC\n");
	    $::psdev->setfsetinsync($self->{mtpt});
    }

    # list of the server or the client replicators
    $self->{replist} = new Lento::List;

    # the beginning of life
    $self->{permitholder} = $self->{serversysid};
    # are we the holder of this fileset?
    if ( $self->{serversysid} eq $::mysysid ) {
	    LOG ">>    Fset $self->{name} is served by this system!\n";

	    #LOG ">>    Fileset state reset to connected\n";
	    #$self->{state} = "connected";

	    # if client_read_only was set we don't let lento write back
	    # to this fileset.
	    if ( $ro_flag ) {
	        $self->{lento_read_only}= 1;
	        $self->{presto_read_only}= 0;
	    }
	
	    # add replicators with KML offset 0
	    foreach my $repid (@{$replicators}) {
            next if ($repid eq $::mysysid);

	        LOG ">>    Added replicator $repid for fileset $self->{name}\n";

	        # possibly create a new client structure for this replicator
	        my $client = $::clientdb->find_by_sysid($repid);
	        if (!defined $client) {
		        LOG(">>     Creating client object ($repid) for " .
		            "replication of fileset $self->{name}\n");
		        $client = new Lento::Client($repid, undef);
		        $::clientdb->add($client);
	        }
	    
	        #instantiate the replicator
	        my $replicator = new Lento::Replicator($self, $repid);

	        # put it on the lists
#	        if ( $self->{fetch_on_demand} == 0 ) {
		    $self->{replist}->add($replicator);
#	        }
	        $client->{replist}->add($replicator);
	    }
	} else {
	    # We are a client for this file set
	    my $replicator = Lento::Replicator->new($self, $self->{serversysid});
	    $self->{replist}->add($replicator);
	    $server->{replist}->add($replicator);
  
	    # if client_read_only is set don't let presto write
	    if ( $ro_flag ) {
	        $self->{lento_read_only}= 0;
	        $self->{presto_read_only}= 1;
            $::psdev->disable_permit_stealing($self->{mtpt});
 	        $::psdev->set_presto_read_only($self->{mtpt});
	    }
	
    }

    $self->{sync} = 'no';

    return $self;
}

sub logical ($) {
    my $self = shift;
    return defined $self->{physical_fileset};
}

sub kml ($) {
    my $self = shift;
    return $self->{kml};
}

sub name ($) {
    my $self = shift;

    return $self->{name};
}

sub mtpt ($$) {
    my $self = shift;
    my $arg = shift;

    if ( defined $arg ) {
	$self->{mtpt} = $arg;
	$::fsetdb->setmtpt($self);
    }

    return $self->{mtpt};
}

sub kml_size($) {
    my $self = shift;

    $::psdev->get_kml_size($self->mtpt());
}

sub serversysid ($) {
    my $self = shift;

    return $self->{serversysid};
}

sub server ($) {
    my $self = shift;

    return Lento::Server::find($self->{serversysid});
}

sub ishere($) {
    my $self = shift;

    if ( $self->{serversysid} eq $::mysysid ) {
	return 1;
    } else {
	return 0;
    }
}

sub iswget($) {
    my $self = shift;

    if ( $self->{fetchtype} eq "Wget" ) {
	    return 1;
    } else {
	    return 0;
    }
}

sub isrsync($) {
    my $self = shift;

    if ( $self->{fetchtype} eq "Rsync" ) {
	    return 1;
    } else {
	    return 0;
    }
}
sub is_kernelreint($) {
    my $self = shift;

    return $self->{kernel_reint};
}

# XXX deal with this
sub fshl ($) {
    my $self = shift;

    return $self->{mtpt};
}

#
# returns everything after fileset root.  example:
# fs_path("/izo/foo") = "/foo"
#
sub fs_path ($$) {
    my $self = shift;
    my $path = shift;

    return substr($path, length($self->mtpt()));
}

#
# returns proper path for cache files, given relative path (fs_path).  example:
# cache_path("foo") = "/izo/foo"
#
sub cache_path ($$) {
    my $self = shift;
    my $path = shift;

    return $self->mtpt() . $path;
}

#
# returns proper path for storing files on server, given fs_path.  example:
# fshl_path("foo") = "/tmp/dvto/foo"
#

sub fshl_path ($$) {
    my $self = shift;
    my $path = shift;
    
    if ( $self->{mtpt} ) {
	    return $self->{mtpt} . $path;
    } else {
	    return $self->{fshl} . $path;
    }
}


sub reconnect ($$) {
    my $fileset = shift;
    my $replicator = shift;
    my $sync_session;

    if ($fileset->serversysid() eq $::mysysid) {
	    confess "Not syncing a local fileset\n";
    }

    if ( $fileset->{fetch_on_demand} ) {
	    # XXX need to validate cache here.
	    # For now reintegrate.
	    $sync_session= $fileset->reintegrate();
    } else {
	    $sync_session= $fileset->sync($replicator);
    }

    return $sync_session;
}

sub sync () {
    my $fileset = shift;
    my $replicator = shift;
    
    LOG(">>>>  fset->sync session.\n");
    my $session= new POE::Session
	([$fileset, $replicator], # one argument to _start
	 $::dispatcher, ['connectme'],
	 'Lento::EventSupport' , ['alias_set', 'alias_remove', '_stop', 
				  '_child', '_signal', 'AddWaiter', 'get_cookie'], 
	 _start => sub {
	     $_[HEAP]->{session_name} = $_[SESSION]->ID().":fset->sync ($_[ARG0]->{name})";
	     $_[HEAP]->{session_level} = LVL_RECOVER;
	     $_[HEAP]->{fileset} = $_[ARG0];
	     $_[HEAP]->{replicator} = $_[ARG1];

	     #XXX: do we want to do debugging stuff here?
	     Lento::Debuggable::set_option($_[SESSION],trace => 1,default => 1);
	     
	     $_[KERNEL]->alias_set($_[SESSION]);

	     $_[KERNEL]->yield('get_lock', $_[HEAP]->{fileset});
	 }, 
	 get_lock => sub { 
	     my $fset = $_[HEAP]->{fileset};
	     # denest syncs (should not be needed)...note that subsequent 
	     # permits will go  through only after the syncs are done.
	     if (defined  $fset->{sync_session}) {
		 DEBUG "!!!!!This should not be happening!!!!!!\n";
		 DEBUG "fileset $fset->{name} already syncing...waiting\n";
         	 # I can think of some circumstances where this might happen.
         	 # die would be an overreaction! -shp
		 #die ""; 
		 Lento::EventSupport::WaitForSession
		     ($fset->{sync_session}, '_get_lock');
		 return;
	     } 

	     # lock the sync table...
	     $fset->{sync_session} = $_[SESSION];
	     $_[KERNEL]->call
			($_[SESSION],'connectme', $_[HEAP]->{fileset}, LVL_RECOVER);
	 },
	 got_connection => sub {
	     my $fset = $_[HEAP]->{fileset};
	     $_[HEAP]->{connection} = $fset->server()->{connection_obj};
	     
		 # this will now be taken care of by sendpacket
         #if (!$_[HEAP]->{connection}->connected()) {
         #   $_[KERNEL]->yield('got_error',"sync",ENOTCONN,"no connection");
         #   return;
         #}

         if (!defined $_[HEAP]->{replicator}) {
            $_[KERNEL]->yield('got_error',"sync",ENOTCONN,"no replicator");
            return;
         }

	     DEBUG "sending Initial ReplicatorStatus for ".$fset->name()."\n";
	     my $fsetstatus = $_[HEAP]->{replicator}->repstatus();
	     $_[KERNEL]->alias_set("$_[SESSION] waiting for permit/sync");
	     $_[HEAP]->{connection}->register_in_monitor_list($_[SESSION]);
	     $_[HEAP]->{connection}->sendpacket
		 		('REQ', ['ReplicatorStatus',$fset->name(),$fsetstatus]);
	 },
	 REP => sub {
	     my $packet= $_[ARG0];
	     my $rc=shift @{$packet->[PKT_PARMS]};
	     my $fsetinfo=shift @{$packet->[PKT_PARMS]};
	     my $fset = $_[HEAP]->{fileset};
	     
	     $_[HEAP]->{connection}->remove_from_monitor_list($_[SESSION]);
	     
	     if ($rc != 0) {
		 die "Server does not like us and returns error code $rc!\n";
	     }
	     
	     $replicator->fix_offsets($fsetinfo);

	     # Now get the permit. This should get any pending KML.
	     DEBUG "ReplicatorStatus done...now fetching permit\n"; 
	     Lento::RPCS::GetPermit::GetPermit($fset, 'got_permit', LVL_RECOVER); 
	 },
	 got_permit => sub {
	     my $rc=$_[ARG0];
	     my $fset = $_[HEAP]->{fileset};
	     
	     $_[KERNEL]->alias_remove("$_[SESSION] waiting for permit/sync");
	     
	     if ($rc) {
		 	LOG("got_permit returned error $rc\n");
            # We will accept EROFS only if this is a read only client.
            if (($rc == EROFS) and $fset->{presto_read_only}) {
                # XXX: Actually we should go through the entire process of
                # sending KML to the peers since this lento might have become
                # read only on restart. However, for now, we shall assume that
                # this fset has always been purely readonly.
	            $_[KERNEL]->yield('done', 0); 
            } else {
		 	    $_[KERNEL]->yield('got_error', 
                                  "get permit", $rc, "permit error");
            }

		 	return;
	     }
	     
         # XXX: Where does this really go?
	     if ($fset->{state} eq "conflict") {
		    print "Conflicts during initial GetPermit - reint\n";
		    exit 101;
	     }
	     $fset->synced('yes');

		 # Before sending kml, generate kml for incomplete closes only if
		 # this is the beginning of life.
		 if (!$fset->{recovered_localcloses}) {
			# This is the beginning of life. We first revoke the permit
			# temporarily to complete closes 
			# XXX: we should really also be locking the permit table since
			# a process might in theory access the file system during this
			# operation, however since we just sent in a sync, permit
			# calls will not go through.
		 	my $cookie= $_[KERNEL]->call($_[SESSION],'get_cookie');
		 	my $kernel_cookie=$::psdev->revoke_permit($fset->mtpt(), $cookie);

		 	if ($kernel_cookie) {
				DEBUG "Need to wait for $kernel_cookie\n";
				Lento::EventSupport::WaitForCookie
						($kernel_cookie,'complete_closes');
		 	} else {
				$_[KERNEL]->yield('complete_closes');
		 	}
		 } else {
			$_[KERNEL]->yield('reintegrate_kml');
		 }
	 },
     complete_closes => sub {
	     my $fset = $_[HEAP]->{fileset};

		 LOG("Fileset $fset->{name}: recovering lost close records\n");
		 $::psdev->complete_closes($fset->{mtpt});
		 $fset->{recovered_localcloses}=1;

		 $_[KERNEL]->yield('reintegrate_kml');
	 },
	 reintegrate_kml => sub {
	     my $fset = $_[HEAP]->{fileset};

	     my $rep_count = $fset->{kml}->ReintegrateKML
		                  (undef, 'complete_reconnect', LVL_RECOVER);
	     
	     if (!$rep_count) {
		    $_[KERNEL]->yield('complete_reconnect', 0);
	     }
	 },
	 complete_reconnect => sub {
	     delete $_[HEAP]->{fileset}->{sync_session};
	     LOG("<<<<  $_[HEAP]->{session_name} going away\n");
	     $_[KERNEL]->yield('done', 0); # or maybe pick up $rc from above?
	 },
	 got_error => sub {
		 # This causes the code to loop, but it may still be the best thing 
		 # to do
	     #$_[KERNEL]->yield('connectme', $_[HEAP]->{fileset}, LVL_RECOVER);

	     # XXX: fix in Server.pm being tested. Basically we simply
	     # refuse to bring the fileset online if the server state
	     # is connecting.
	     LOG "**** CHECK ME************************************\n";
	     delete $_[HEAP]->{fileset}->{sync_session};
	     $_[KERNEL]->yield('done', 1);
	 },
	 send_done => sub { 
	     $_[KERNEL]->yield('done', 0);
	 },   
	 done => sub {
	     $_[HEAP]->{connection}->put() 
            if ($_[HEAP]->{connection} and !$_[ARG0]);
	     Lento::EventSupport::WakeUpWaitList();
	     $_[KERNEL]->alias_remove($_[SESSION]);
	 }
	);
    return $session;
}


sub synced {
    my $fileset = shift;
    my $val = shift;
    my $syncedfile = $::dbdir . "/" . $fileset->{name} . "-synced";
    
    if ($fileset->{fetch_on_demand}) {
	    LOG(">>> Called synced for $fileset->{name} : fetching on demand.\n");
	    return 0;
    }

    LOG(">>    Setting $fileset->{name} to synced\n");

    if ( ! defined $val ) {
	    if ( -f $syncedfile ) {
	        return 1;
	    } else {
	        return 0;
	    }
    }
    
    if ( $val ) {
	    open FH , ">" . $syncedfile or die;
	    close FH;
	    $::psdev->setfsetinsync($fileset->{mtpt});
    } else {
	    unlink ($syncedfile) or die ;
	    # XXX clear kernel flag!
    }
}

sub printme ($) {
    my $self = shift;

    my $msg = ("[Fileset \"$self->{name}\", server \"$self->{serversysid}\", " .
	       "mtpt \"$self->{mtpt}\", state \"$self->{state}\", replicators: [ ");
    
    if ( !defined $self->{replist} ) {
	return $msg . "]]\n";
    }

    $self->{replist}->iterate
      ( sub {
	    my $replicator = shift;
	    my $rmsg = shift;
	    $$rmsg .= $replicator->{sysid} . " ";
	}, \$msg
      );
    return $msg . "]]\n";
}


1;
