# unpacking kernel arguments
#use strict;
package Lento::Upcall;
use POSIX qw(:errno_h);
use Lento::Psdev;
use Lento::Debuggable;
use POE;

@Lento::Upcall::ISA = qw(Lento::Debuggable);
$Lento::Upcall::level = 1;

# opcodes
my %op;
my $PERMIT=  1;   $op{$PERMIT}="Permit";
my $JOURNAL= 2;   $op{$JOURNAL}="Journal";
my $OPENDIR= 3;   $op{$OPENDIR}="OpenDir";
my $OPEN=    4;   $op{$OPEN}="OpenFile";
my $SIGNAL=  5;   $op{$SIGNAL}="Signal";
my $KML=  6;      $op{$KML}="KML";
my $Cookie=  7;   $op{$Cookie}="Cookie";

# Class Method
sub new ($$) {
    shift; # class
    my $buf = shift;
    my $self;

    $self->{opc} =""; 
    $self->{uniq} =-1;
    $self->{pid} = -1;
    $self->{uid} = -1; 
    $self->{path} = "";
    $self->{fileset} = "";


    ($self->{opc}, $self->{uniq}, $self->{pid}, $self->{uid}) =
      unpack "IIss", $buf;

    die "Upcall::new, opcode $self->{opc} not known" unless
      defined $op{$self->{opc}};

    &{"Lento::Upcall::$op{$self->{opc}}::init"}($self, $buf, 12);

    DEBUG $self->printme;
    return $self;
}

# Method
sub op {
    my $self = shift;
    $op{$self->{opc}};
}


# Method
sub printme {
    my $self = shift;

    return "$op{$self->{opc}} [uniq $self->{uniq} pid " .
      "$self->{pid} uid $self->{uid}] path: $self->{path}, " .
      "fileset: $self->{fileset}\n";
}

# method
# replies to the kernel won't block: do them synchronously
sub complete {
    my $self = shift;
    my $result = shift;

    my $answer = pack "iii", $self->{opc}, $self->{uniq}, $result;

    # mount does opendir on / and is async, so is Journal: don't send a reply
    return if $self->{async}; 

    DEBUG "Upcall %s result: $result\n", $self->op();

    my $retval = $::psdev->write($answer, length($answer));
    if ($retval < 0 && $retval != (-1 * ESRCH)) {
            die "Psdev->write (error $retval)";
    }
}

#
# Subclasses for each upcall TYPE
#
# Upcall::Permit
package Lento::Upcall::Permit;
use strict;
use Lento::Debuggable;

@Lento::Upcall::Permit::ISA = qw(Lento::Upcall Lento::Debuggable);
$Lento::Upcall::Permit::level = 1;

sub init($$$) {
    my ($self, $buf, $offset) = @_;
    my $off;
    $self->{pathlen} = unpack "x$offset i", $buf;
    $self->{fsetnamelen} = unpack "x$offset x4 i", $buf;
    $self->{path} = substr($buf, $offset + 8, $self->{pathlen});
    $self->{fileset} = substr($buf, $offset + 8 + 1 + $self->{pathlen}, $self->{fsetnamelen});

    DEBUG "Permit path: $self->{path}, fsetname: $self->{fileset}\n";

    return bless $self;
}

# Upcall::Opendir
package Lento::Upcall::OpenDir;
use strict;
use Lento::Debuggable;

@Lento::Upcall::OpenDir::ISA = qw(Lento::Upcall Lento::Debuggable);
$Lento::Upcall::OpenDir::level = -1;

sub init ($$$) {
    my ($self, $buf, $offset) = @_;
    ($self->{async}, $self->{pathlen}) = unpack "x$offset ii", $buf;
    $self->{path} = substr($buf, $offset + 8, $self->{pathlen});

    DEBUG "OpenDir path: $self->{path}\n";

    return bless $self;
}

=pod

=over 4

=item B<Upcall::KML> This informs Lento there is new stuff in the KML

=back

=cut

package Lento::Upcall::KML;
use strict;
use Lento::Debuggable;

@Lento::Upcall::KML::ISA = qw(Lento::Upcall Lento::Debuggable);
$Lento::Upcall::KML::level = 1;

sub init ($$$) {
    my ($self, $buf, $offset) = @_;
    ($self->{first_rec_offset}, $self->{first_recno}, $self->{length}, 
     $self->{last_recno}, $self->{fsetlen}) = unpack "x$offset IIIIi", $buf;
    $self->{fileset} = substr($buf, $offset + 20 , $self->{fsetlen});

    DEBUG("KML: fileset $self->{fileset}, first $self->{first_recno} length $self->{length} last_recno $self->{last_recno}\n");
    die unless defined $self->{length};
    die unless defined $self->{first_recno};
    die unless defined $self->{last_recno};
    bless $self;
}

=pod

=over 4

=item B<Upcall::Cookie> This informs Lento when the kernel sends a cookie

=back

=cut

package Lento::Upcall::Cookie;
use strict;
use Lento::Debuggable;

@Lento::Upcall::Cookie::ISA = qw(Lento::Upcall Lento::Debuggable);
$Lento::Upcall::Cookie::level = 1;

sub init ($$$) {
    my ($self, $buf, $offset) = @_;
    $self->{cookie}=unpack "x$offset i", $buf;
    die "no cookie in cookie upcall?" unless defined $self->{cookie};

    DEBUG("Cookie: $self->{cookie}\n");
    bless $self;
}

# Upcall::OpenFile
package Lento::Upcall::OpenFile;
use strict;
use Lento::Debuggable;
use POE;
use POSIX qw(:fcntl_h);

@Lento::Upcall::OpenFile::ISA = qw(Lento::Debuggable Lento::Upcall);
$Lento::Upcall::OpenFile::level = 1;

sub init ($$$) {
    my ($self, $buf, $offset) = @_;
    ($self->{pathlen}) = unpack "x$offset i", $buf;
    $self->{path} = substr($buf, $offset + 4, $self->{pathlen});

    DEBUG "Openfile path: $self->{path}\n";
    return bless $self;
}

#
# Upcall::Journal
# This call does not always need completion to the kernel
#
package Lento::Upcall::Journal;
use strict;
use Lento::Debuggable;

@Lento::Upcall::Journal::ISA = qw(Lento::Upcall Lento::Debuggable);
$Lento::Upcall::Journal::level = 1;

sub init ($$$) {
    my ($self, $buf, $offset) = @_;
    my $fset;
    $self->{async} = 1;
    $self->{buf} = substr($buf, $offset);

    return bless $self;
}

# Upcall::Signal
package Lento::Upcall::Signal;
use strict;
use Lento::Debuggable;
@Lento::Upcall::Signal::ISA = qw(Lento::Upcall Lento::Debuggable);
$Lento::Upcall::Signal::level = 0;

sub init ($$$) {
    my ($self, $buf, $offset) = @_;

    return bless $self;
}
1;

#
#  The filter for the Upcall Wheel
#
package Lento::Filter::Upcall;
#use Upcall;

sub new {
  my $type = shift;
  my $self = bless { }, $type;
  $self->{basetype} = $type;

  $self;
}

sub put {
    my ($self, $packets) = @_;
    my @raw;
    foreach my $packet (@$packets) {
        DEBUG($packet->printme());
        push(@raw, $packet->packme());
    }

    \@raw;
}

sub get {
    my ($self, $stream) = @_;
    my @result;

    foreach my $buf (@$stream) {
        my $upcall = Lento::Upcall->new($buf);
        push @result, $upcall;
    }
    \@result;
}

1;
