package Lento::Fsetdb;
use XML::Simple;
use IO::File;
use strict;
use Lento::Debuggable;
use Lento::Fileset;
use Lento::Server;
use POE;

@Lento::Fsetdb::ISA = qw(Lento::Debuggable);
$Lento::Fsetdb::level = 1;


sub new ($) { 
    my $class = shift;
    my $self = {};

    $self->{fset_by_mtpt} = {};
    $self->{fset_by_name} = {};

    bless $self;
}

sub init ($) {
    my $self = shift;
    my $hash;
    my $buffer;    

    $::fsetdbfile=$::configdir . "/fsetdb" if !defined $::fsetdbfile;
    LOG( ">>>>> Initializing FileSetDB : $::fsetdbfile\n" );

    $hash = eval { XMLin($::fsetdbfile, forcearray => 1) };
    die "Can't open fileset database $::fsetdbfile\n" if ($@);
    my $filesets = $hash->{fileset}; 

    foreach my $fsetname (keys %{$filesets} ) {

        my $entry = $filesets->{$fsetname};

        # A. Servers
        LOG ">>>  Fsetname: $fsetname servername: $entry->{servername}\n";
        # referential integrity with another table: belongs in Fsetdb. 
        # XXX need to actually look up in the serverdb.
        die "\"$fsetname\" has an invalid servername!" unless 
            $entry->{servername};
        
        if ($entry->{client_read_only}) {
            $entry->{client_read_only} = 1;
            LOG(">>>   - $fsetname is client read-only\n");
        } else {
            $entry->{client_read_only} = 0;
            LOG(">>>   - $fsetname is client read-write\n");
        }

        if (! defined $entry->{fetchtype} ) {
            #$entry->{fetchtype} = "Desc"
            $entry->{fetchtype} = "Rsync"
        }
        elsif ($entry->{fetchtype} eq "Wget" ) {
            die "\"$fsetname\" requires docroot for wget!" 
               unless $entry->{docroot};
            die "\"$fsetname\" requires httpmnt for wget!"
                unless $entry->{httpmnt};
        }

        if ($entry->{kernel_reint}) {
            $entry->{kernel_reint} = 1;
        } else {
            $entry->{kernel_reint} = 0;
        }

        if (! defined $entry->{steal_permit}) {
            if (defined ($entry->{no_steal_permit})) {
                # Use the opposite of no_steal_permit.
                $entry->{steal_permit} = ! $entry->{no_steal_permit};
            } else {
                # Steal permits by default
                $entry->{steal_permit} = 1;
            }
        }

        if ($entry->{steal_permit}) {
            LOG (">>>   - Permit stealing is enabled\n");
        } else {
            LOG (">>>   - Permit stealing is disabled\n");
        }

        if ($entry->{jclose_on_write}) {
            LOG (">>>   - Will journal closes on writes\n");
        } 

        if (! $entry->{mount_point}) {
            my $mountpoint = $::cachetable->fset2cache($fsetname);
            die "File set $fsetname has no mountpoint" unless $mountpoint;
            $entry->{mount_point} = $mountpoint;
        }

        LOG ">>>   - Fset mount point: $entry->{mount_point}\n";
        # not implemented yet, activate after 1.0
        #       if (defined $entry->{fetch_on_demand}) {
        #           LOG(">>> $fsetname set to fetch on demand.\n");
        #       } else {
        $entry->{fetch_on_demand}= 0;
        #       }
        my $fset = Lento::Fileset->new($fsetname, 
                                     $entry->{servername},
                                     $entry->{mount_point},  
                                     $entry->{client_read_only},
                                     $entry->{fetch_on_demand},
                                     $entry->{steal_permit},
                                     $entry->{jclose_on_write},
                                     $entry->{fetchtype},
                                     $entry->{replicator});
        die "Error in initializing fileset $fsetname\n" unless $fset;
        $self->add($fset);
    }
    LOG "<<<<< Initializing FileSetDB\n";
    
    return $self;
}

# give it a path, return the fset object
sub find_by_path ($$) {
    my $self = shift;
    my $path = shift;

    while (!defined($self->{fset_by_mtpt}->{$path}) && $path) {
            my $foo = rindex($path, "/");
        # "/" paths are a headache for this loop. 
        # We'll count on peter to make this look nice. 
        # we need this initial silly test so we don't loop forever
        # if '/' is not a valid intermezzo mountpoint. 
        if ($path eq "/") { 
            # been here, done that, no dice
            $path = "";
        } elsif ($foo < 0) {
                $path = "";
            } elsif ($foo == 0) {
            # no much left to do here ... It's just /
                $path = "/";
        } else {
                $path = substr($path, 0, $foo);
            }
    }

    if ( defined $path) {
            return $self->{fset_by_mtpt}->{$path};
    } else {
            return undef;
    }
}

sub find_by_name ($$) {
    my $self = shift;
    my $name = shift;

    return $self->{fset_by_name}->{$name};
}

sub add ($$) {
    my $self = shift; 
    my $fset = shift;

    LOG("<<    Added file set: %s:%s with mtpt %s\n", 
           $fset->serversysid(),
           $fset->name(), 
           $fset->mtpt());

    $fset->server()->pushfset($fset);
    $self->{fset_by_name}->{$fset->name()} = $fset;
    $self->{fset_by_mtpt}->{$fset->mtpt()} = $fset;
}

sub setmtpt ($$) {
    my $self = shift;
    my $fset = shift;

    $self->{fset_by_mtpt}->{$fset->mtpt()} = $fset;
}

sub printdb ($) {
    my $self = shift;
    my $result;
    my $name;

    foreach $name (keys %{$self->{fset_by_name}}) {
            $result .= $self->{fset_by_name}->{$name}->printme();
    }

    return $result;
}


1;

__END__
