package Lento::InterMezzo::ReqHandler;
use strict;

use POE;
use Carp;
use Lento::ACT;
use Lento::Client;
use Lento::ClientDB;
use Lento::CreateML;
use Lento::Connection;
use Lento::Debuggable;
use Lento::Packet;
use Lento::ReqDispatcher;
use Lento::Replicator;
use Lento::Server;
use Lento::EventSupport;
use Lento::Bulk;
use Lento::Bulk::Desc;
use Lento::Bulk::Rsync;
use Lento::KML::Rec;
use Lento::KML::Reint;
use Lento::RPCS::GetPermit;

use POSIX qw(:fcntl_h :errno_h tmpnam);

@Lento::InterMezzo::ReqHandler::ISA = qw(Lento::Debuggable);
$Lento::InterMezzo::ReqHandler::level = 100;

# These events are common to all ReqHandl'ing sessions
# ARG0 is the request packet, ARG1 is the name
sub _start {
    $_[HEAP]->{session_name} = $_[SESSION]->ID().": ".$_[ARG1]. " $_[SESSION]";
    Lento::Debuggable::set_option($_[SESSION], trace => 1, default => 1);
    $_[HEAP]->{req_packet} = $_[ARG0];
    $_[HEAP]->{connection} = $_[HEAP]->{req_packet}->[PKT_CONNECTION];
    $_[HEAP]->{state} = 'idle';
    
    # All session need an alias, otherwise they get garbage collected.
    $_[KERNEL]->alias_set($_[SESSION]);
    
    $_[KERNEL]->yield('REQ');
}

sub finish {

    # This should be enough to break all connection related circular references
    $_[HEAP]->{connection}->put() if $_[HEAP]->{connection};

    # See comment before alias_set in _start
    $_[KERNEL]->alias_remove($_[SESSION]);
}

sub got_error {
    # Hookup to descriptor error method
    $_[HEAP]->{desc}->error($_[ARG0],$_[ARG1],$_[ARG2])
        if defined $_[HEAP]->{desc};

    $poe_kernel->yield('finish');
}

sub Ping {
    my $packet = shift;
    
    new POE::Session
        ([$packet, "Ping"],
         'Lento::InterMezzo::ReqHandler', ['_start', 'finish', 'got_error'],
         'Lento::EventSupport', ['alias_set', 'alias_remove', '_stop',
                                 '_child', '_signal'],
         REQ => sub {
             my $time = @{$_[HEAP]->{req_packet}->[PKT_PARMS]}[0];
             $_[HEAP]->{session_level} = LVL_PING;
             $_[HEAP]->{connection} = $_[HEAP]->{req_packet}->[PKT_CONNECTION];
             $_[HEAP]->{connection}->sendpacket('REP', $_[HEAP]->{req_packet}, 
                                                [$time]);
             DEBUG ("REP: for Ping to %s\n", 
                    $_[HEAP]->{req_packet}->[PKT_CONNECTION]->{sysid});
             $poe_kernel->yield('finish', [$_[HEAP]->{req_packet}, 0]);
         }
        );

}

sub SysId {
    my $packet = shift;
    
    new POE::Session
        ([$packet, "SysId"],
         'Lento::EventSupport', ['alias_set', 'alias_remove', '_stop',
                                 '_child', '_signal'],
         'Lento::InterMezzo::ReqHandler', ['_start', 'finish', 'got_error'],
         REQ => sub {
             my $sysid = @{$_[HEAP]->{req_packet}->[PKT_PARMS]}[0];
             # XXX
             # check security here!!
             $_[HEAP]->{session_level} = LVL_BIND;
             
             my $client = $::clientdb->find_by_sysid($sysid);
             if ( !defined $client ) {
                 LOG "New client instantiated, sysid $sysid\n";
                 $client = new Lento::Client($sysid);
                 $::clientdb->add($client);
             } 
             
             # Bind the connection
             $_[HEAP]->{req_packet}->[PKT_CONNECTION]->setsysid($sysid);
             
             $_[HEAP]->{req_packet}->[PKT_CONNECTION]->sendpacket('REP', 
                                            $_[HEAP]->{req_packet}, [0]);
             $poe_kernel->yield('finish', [$_[HEAP]->{req_packet}, 0]);
         }
        );

}

# Synchronizing fsetstatus

sub ReplicatorStatus {
    my $packet = shift;
    
    new POE::Session
        ([$packet, "ReplicatorStatus"], # two arguments to _start
         'Lento::EventSupport', ['alias_set', 'alias_remove', 
                                 'get_cookie', '_stop',
                                 '_child', '_signal', 'AddWaiter',
                                 'RemoveUpcallWaiter'], 
         'Lento::InterMezzo::ReqHandler', ['_start', 'finish'],
         REQ => sub {
             my $packet = $_[HEAP]->{req_packet};
             my $connection = $packet->[PKT_CONNECTION];
             my $fsetname = @{$packet->[PKT_PARMS]}[0];
             my $fsetinfo = @{$packet->[PKT_PARMS]}[1];

             $_[HEAP]->{fsetname} = $fsetname;
             $_[HEAP]->{connection}= $connection;
             $_[HEAP]->{session}= $_[SESSION];
             $_[HEAP]->{fsetinfo} = $fsetinfo;
             $_[HEAP]->{session_name} .= " ($fsetname)";
             
             # All errors go through at minimum connection level
             $_[HEAP]->{session_level}=LVL_BIND;
            
                         if (!defined($fsetinfo) || !defined($fsetname)) {
                                DEBUG "No fsetinfo or fsetname...invalid fsetstatus request\n";
                                $_[KERNEL]->yield('sendrep', EINVAL); 
                                return;
                         }
                         
                         if (!$::fsetdb->find_by_name($fsetname)) {
                                DEBUG " This fset is not known here! \n";
                                $_[KERNEL]->yield('sendrep', ENOENT); 
                                return;
                         }

             my $replicator = $_[HEAP]->{replicator} =
                 Lento::Replicator::find($connection->{sysid}, $fsetname);
             
             if (!defined $replicator) {
                 DEBUG "no replicator\n";
                                 $_[KERNEL]->yield('sendrep', EPERM);
                 return;
             }
             
             my $fset = $_[HEAP]->{fset} = $replicator->{fileset};

             LOG(">>>> ReplicatorStatus request by %s for fset %s\n", 
                    $connection->{sysid}, $fset->{name});
             
             # Take the permit from the client, since we know that the client
                         # must have relinquished its permit.
             if ($fset->{steal_permit} and 
                 $fset->{permitholder} eq $connection->{sysid}) {
                 DEBUG "taking permit from $connection->{sysid}";
                 $fset->{permitholder} = $::mysysid;
             }
             
             Lento::Debuggable::set_option
                 ($_[SESSION], trace => 1, default => 1);
             
             # our offsets need to synced to restore correct last_rcvd state.
             $replicator->fix_offsets($_[HEAP]->{fsetinfo});
             $_[HEAP]->{session_level} = LVL_RECOVER;
             $_[KERNEL]->yield('get_lock');
         },
         get_lock => sub {
             
             my $fset=$_[HEAP]->{fset}; 
             my $fsetname=$_[HEAP]->{fsetname}; 
             my $connection=$_[HEAP]->{connection}; 
             
             # Just in case we get a got_error while in a get_lock loop.
             if ($_[HEAP]->{state} eq 'finished') {
                 LOG ("!!! session $_[SESSION] finished while in get_lock\n");
                 # XXX: this should not be neccessary
                 $_[KERNEL]->yield('finish');
                 return;
             }
             
             # test sync lock
             if (defined $fset->{sync_session}) {
                 DEBUG "fileset $fsetname is syncing...waiting\n";
                 Lento::EventSupport::WaitForSession
                     ($fset->{sync_session}, 'get_lock');
                 return;
             }
             
             DEBUG "Acquired sync lock for fileset $fsetname\n";
             $fset->{sync_session}=$_[SESSION];
             
             $_[KERNEL]->yield('sendrep', 0);
         },
         sendrep => sub {  
             # get here when $fset->{permit} has reintegrated
             my $status = $_[ARG0];
             my $packet = $_[HEAP]->{req_packet};
             my $connection = $_[HEAP]->{connection};
             my $replicator = $_[HEAP]->{replicator}; 
             my $fsetstatus=undef;
             
                         confess "Heap changed" if !defined $connection; 
             
                         if (!Lento::Connection::connected($connection)) {
                                DEBUG "No connection\n";
                                $_[KERNEL]->yield('got_error',"sendrep",ENOTCONN,
                                                                        "no connection");
                                return;
                         }
             
                         if (!defined $replicator) {
                                $status=ENOENT unless defined $status;
                $connection->sendpacket('REP', $packet, [$status,0]);
                $_[KERNEL]->yield
                                        ('got_error',"sendrep", ENOTCONN,"no replicator");
                $connection->set_error
                                        ("repstatus", ENOTCONN, "no replicator");
                                return;
                         }

             $status = 0 unless defined $status;

             DEBUG " setting fsetstatus for ".$_[HEAP]->{fset}->name()."\n";
             $fsetstatus = $replicator->repstatus();
             
             $_[HEAP]->{state}="finished";
             $connection->sendpacket('REP', $packet, [$status,$fsetstatus]);
             $_[KERNEL]->yield('release_lock');
             
         },
         got_error => sub {
             LOG "Error in ReplicatorStatus handler\n";
             
             $_[HEAP]->{state}='error';
             $_[KERNEL]->call($_[SESSION],'release_lock');
         }, 
         release_lock => sub {
             $_[HEAP]->{state}='finished';
             
             # release sync lock
             if (defined $_[HEAP]->{fset}->{sync_session} and 
                 $_[HEAP]->{fset}->{sync_session} == $_[SESSION]) {
                 DEBUG "releasing sync lock\n";
                 delete $_[HEAP]->{fset}->{sync_session};
             } else {
                 DEBUG "In release_lock, no sync lock, got a got_error!\n";
                 DEBUG "Lock holder is $_[HEAP]->{fset}->{sync_session}\n" 
                     if defined $_[HEAP]->{fset}->{sync_session};
             }
             
             # Either way wake up lock waiters...this list will be empty if we 
             # did not hold the sync lock
             Lento::EventSupport::WakeUpWaitList();
             $_[KERNEL]->yield('finish');
         }
        );

}

#
# Get the KML from current permit holder
# Send updates to requester
# Grant the permit
#

sub Permit {
    my $packet = shift;
    
    new POE::Session
        ([$packet, "Permit"],   # two arguments to _start
         'Lento::EventSupport', ['alias_set', 'alias_remove', 
                                 'get_cookie', '_stop',
                                 '_child', '_signal', 'AddWaiter',
                                 'RemoveUpcallWaiter'], 
         'Lento::InterMezzo::ReqHandler', ['_start', 'finish'],
         REQ => sub {
             my $packet = $_[HEAP]->{req_packet};
             my $connection = $packet->[PKT_CONNECTION];
             my $fsetname = @{$packet->[PKT_PARMS]}[0];
             my $fset = $::fsetdb->find_by_name($fsetname);
             my $replicator =
                 Lento::Replicator::find($connection->{sysid}, $fsetname);
             
             $_[HEAP]->{fset} = $fset;
             $_[HEAP]->{fsetname} = $fsetname;
             $_[HEAP]->{connection}= $connection;
             $_[HEAP]->{session}= $_[SESSION];
             $_[HEAP]->{replicator}=$replicator;
             $_[HEAP]->{session_name} .= " ($fsetname)";
             
             $_[HEAP]->{state}='acquiring_permit';
             
             # All errors go through at minimum connection level
             $_[HEAP]->{session_level}=LVL_BIND;
             
             if ( !defined $fset ) {
                 # Invalid fileset name sent by client
                 # XXX 
                 # XXX return ENOENT - client should remove file set
                 # XXX from replicator list
                 LOG "ERROR: Permit request for fileset $fsetname:not known\n";
                 $_[KERNEL]->yield('REP', EINVAL);
                 return;
             }
             
             DEBUG ">>>> Permit request by %s for fset %s (current holder %s)\n", 
             $connection->{sysid}, $fset->{name}, $fset->{permitholder};
             
             Lento::Debuggable::set_option
                 ($_[SESSION], trace => 1, default => 1);
             
             
             # if we are a client, only respond if we have the permit
             if ( ! $fset->ishere()  && $fset->{permitholder} ne $::mysysid ){
                 LOG("ERROR: This client does not hold permit\n");
                 $_[KERNEL]->yield('REP', EINVAL);
                 return;
             }
             
             $_[HEAP]->{session_level} = LVL_FULL;
             # if we are the server and the fileset is read-only, then
             # we will send in the KML and then reply with an EROFS
             if ( $fset->ishere() && $fset->{lento_read_only} ) {
                 LOG(">>>> ERROR: Client \"%s\" attempted to gain permit " .
                     "for read-only fileset \"%s\"\n",
                     $connection->{sysid}, $_[HEAP]->{fsetname} );
                 DEBUG(" Read only client asked for permit...sending KML\n");
                 my $count=$fset->{kml}->ReintegrateRep($replicator, 
                                'ro_client');
                 $_[KERNEL]->yield('ro_client') unless $count;
                 return;
             }
             
             # This might happen if a peer recovers before the connection 
             # times out. If this host is the client, it indicates that
             # the server believes the permit is here. If this host is the server
             # the client restarted while it had the permit. In either, case
             # behave as if we have the permit here and finished reintegration.
             # otherwise nasty permit deadlocks might happen!
             # This might also happen when a client is performing initial sync 
             # after a disconnection.
             # XXX: check these conditions for proxies 
             # (drop test for server sysid?)
             if ( ($fset->{permitholder} eq "" &&  
                   $connection->{sysid} eq $fset->{serversysid}) ||  
                  $fset->{permitholder} eq $connection->{sysid} ) {
                 
                 # Behave as if we have the permit...even though we really 
                 # don't. 
                 DEBUG("WARNING :: permitholder asking for permit\n");
                 
                 #This allows us to trap got_errors correctly
                 #Note that remote believes we have permit, so might as well
                 #pretend that we do!
                 $_[HEAP]->{state}='have_permit';
                 
                 # otherwise the reply will never get through and the peer might
                 # deadlock
                 $_[HEAP]->{session_level} = LVL_BIND;
                 $_[KERNEL]->yield('reint_done');
                 return;
             }
             
             $_[KERNEL]->yield('get_lock');
         },
         ro_client => sub {
             DEBUG(" Read only client...sending EROFS\n");

             # We send in an EROFS reply
             $_[KERNEL]->yield('REP',EROFS);
         },
         get_lock => sub {
                         # NOTE: we first test if we can get both locks. 
                         # It is important that WaitForSession does not block
                         # in the POE kernel (it doesn't) !!
                         # when we can get both locks we take them. 
             
             my $fset=$_[HEAP]->{fset}; 
             my $fsetname=$_[HEAP]->{fsetname}; 
             my $connection=$_[HEAP]->{connection}; 
             
             # Just in case we get a got_error while in a get_lock loop.
             if ($_[HEAP]->{state} eq 'finished') {
                 LOG ("!!! session $_[SESSION] finished while in get_lock\n");
                 # XXX: this should not be neccessary
                 $_[KERNEL]->yield('finish');
                 return;
             }
             
             # join sync barrier
             if (defined $fset->{sync_session}) {
                 DEBUG "fileset $fsetname is syncing...waiting\n";
                 Lento::EventSupport::WaitForSession
                     ($fset->{sync_session}, 'get_lock');
                 return;
             }
             
             # test permit lock
             if (defined $fset->{permit_session}) { 
                 DEBUG "permit table locked for $fsetname...waiting\n";
                 Lento::EventSupport::WaitForSession
                     ($fset->{permit_session}, 'get_lock');
                 return;
             } 
             
             # Add this session to hash.
             DEBUG "got lock for $fsetname \n";
             $fset->{permit_session} = $_[SESSION];
             
             # A retry counter for revoking permits
             #$_[HEAP]->{retries}=0;
             
             # if we are a server now do a Permit call to the permit holder
             # this will break the permit there and send us the updates
             if ( $fset->ishere() && $fset->{permitholder} ne $::mysysid ) {
                 # start session to get the permit.
                 $_[HEAP]->{wait_session}= Lento::RPCS::GetPermit::GetPermit
                     ($fset,'permit_acquired',$_[HEAP]->{session_level},
                                                $_[SESSION]);
                 die "no getpermit session"
                                        unless defined $_[HEAP]->{wait_session};
             } else {
                 # We set this greedily here in case we get a got_error right after
                 # this subroutine finishes.
                 $_[HEAP]->{state}='have_permit';
                 $_[KERNEL]->yield('permit_acquired',0);
             }
         }, 
         permit_acquired => sub {
             my $rc=$_[ARG0];
                         delete $_[HEAP]->{wait_session};
             
                         # If a prior error occured, we still wait for the
                         # underlying getpermit session to finish, so we need to explicitly
                         # release the lock
             if ($_[HEAP]->{state} eq 'error' or 
                        $_[HEAP]->{state} eq 'finished') {
                                $_[KERNEL]->yield('release_lock');
                                return;
                         }
             
             if ($rc) {
                 DEBUG "GetPermit returns an error $rc!\n";
                 $_[KERNEL]->yield('got_error','permit_acquired',
                                   EAGAIN,'can not get permit');
                 return;
             }
             
             $_[HEAP]->{state}='have_permit';
             
             $_[KERNEL]->yield('revoke_permit');
         },
         revoke_permit => sub {
             my $packet = $_[HEAP]->{req_packet};
             my $fset = $_[HEAP]->{fset};
             my $cookie= $_[KERNEL]->call($_[SESSION],'get_cookie');
             my $kernel_cookie= 
                 $::psdev->revoke_permit($fset->mtpt(), $cookie);
             
             if ($kernel_cookie) {
                 DEBUG "Need to wait for $kernel_cookie\n";
                 Lento::EventSupport::WaitForCookie($kernel_cookie,  'reintegrate');
             } else {
                 $_[KERNEL]->yield('reintegrate');
             }
         },
         reintegrate => sub {
             #$_[HEAP]->{retries}++;
             #if ($_[HEAP]->{retries} == 1) {
             #    # XXX: We try this twice to work around an ordering problem
             #    # with presto upcalls and downcalls
             #    $_[KERNEL]->yield('revoke_permit');
             #    return;
             #}
             
             DEBUG("  >>  Permit::reintegrate to $_[HEAP]->{connection}->{sysid}".
                 " from $::mysysid\n");
             
             my $fset = $_[HEAP]->{fset};
             
             if ( $fset->{fetch_on_demand} ) {
                 $_[KERNEL]->call($_[SESSION], 'reint_done');
             } else {   
                 my $packet = $_[HEAP]->{req_packet};
                 my $connection = $packet->[PKT_CONNECTION];
                 my $replicator = $_[HEAP]->{replicator}; 
                 
                 # replicator might have been deleted while we were reintegrating
                 if (!defined $replicator) {
                     $_[KERNEL]->yield('got_error','reintegrate',ENOTCONN,
                                       'no replicator');
                     return;
                 }
                 
                 # setup a local alias to prevent garbage collection.
                 $_[KERNEL]->alias_set("$_[SESSION] waiting for send_KML");
                 
                 my $count = $fset->{kml}->ReintegrateRep($replicator, 
                                                          'reint_done');
                 $_[KERNEL]->yield('reint_done') unless $count; 
             }
         },
         reint_done => sub {
             
             # Remove the local alias, send_KML was completed.
             $_[KERNEL]->alias_remove("$_[SESSION] waiting for send_KML");
             
             # might happen since send_KML always sends back a reint_done even 
             # after this permit session has seen an error.
             return if ($_[HEAP]->{state} eq 'error' or 
                        $_[HEAP]->{state} eq 'finished');
             
             # Reintegration complete; issue the permit.
             my $connection = $_[HEAP]->{req_packet}->[PKT_CONNECTION];
             
             DEBUG "giving permit to $connection->{sysid}\n";
             $_[HEAP]->{fset}->{permitholder} = $connection->{sysid};
             DEBUG("<<<<<  Permitholder of fileset %s is now %s\n",  
                 $_[HEAP]->{fset}->{name}, $connection->{sysid});
             
             $_[KERNEL]->yield('REP', 0);
         },
         REP => sub {  
             return if ($_[HEAP]->{state} eq 'error' or 
                        $_[HEAP]->{state} eq 'finished');
             
             # get here when $fset->{permit} has reintegrated
             DEBUG("  PERMIT REP : holder= $_[HEAP]->{fset}->{permitholder}\n");
             my $status = $_[ARG0];
             my $packet = $_[HEAP]->{req_packet};
             my $connection = $packet->[PKT_CONNECTION];
             my $replicator = $_[HEAP]->{replicator}; 
             my $fsetstatus=undef;
             
             $status = 0 unless defined $status;
             
             # replicator might have been deleted while we were reintegrating
             if (!defined $replicator) {
                 $_[KERNEL]->yield('got_error', 'reply to permit request', 
                                    ENOTCONN, 'no connection');
                 return;
             }
             
             # Connection might have disconnected...force an
             # early got_error
             if (!Lento::Connection::joined_barrier
                        ($connection, $_[HEAP]->{session_level})) {
                                 DEBUG "Session level $_[HEAP]->{session_level} < ".
                                           " connection level $connection->{connection_level}\n";
                                 DEBUG " connection session $connection->{session}\n";
                 $_[KERNEL]->yield('got_error','REP', ENOTCONN, 'no connection');
             } else {
                 # Once we send a REP all bets are off, so signal finished state
                 # Note that even if a got_error comes at this stage, we can not
                 # be certain that the recipient did not receive our reply.
                 $_[HEAP]->{state}="finished";
                 $connection->sendpacket('REP', $packet, [$status]);
                 $_[KERNEL]->yield('release_lock');
             }
         },
         got_error => sub {
             return if ($_[HEAP]->{state} eq 'error' or 
                        $_[HEAP]->{state} eq 'finished');

             my $errno = $_[ARG1];
             my $packet = $_[HEAP]->{req_packet};
             my $connection = $packet->[PKT_CONNECTION];
             
             $errno=EAGAIN unless (defined $errno and $errno != 0);
             
             LOG "Error in Permit handler\n";
             
             # Did we already acquire the permit?
             if ($_[HEAP]->{state} eq 'have_permit') {
                 # Got an error. We are more aggressive here in reaquiring our permit
                 # because the new steal permit flag can throw the permit into limbo
                 # if that flag is not enabled, we can wait till a GetPermit instead.
                 # XXX: also do a grant permit instead of waiting for an upcall?
                 DEBUG "got error trying to release permit...reaquiring\n";
                 DEBUG "taking permit\n";
                 $_[HEAP]->{fset}->{permitholder} = $::mysysid;
             } else {
                 # could not acquire permit from permit holder...tell remote
                 # This might also cause a get error which will be trapped 
                 $_[HEAP]->{state}='error';
                 
                 $connection->sendpacket('REP', $packet, [$errno]);
             }
             
             $_[HEAP]->{state}='error';

                         # If there is a getpermit in progress the callback will take
                         # care of releasing the lock. 
             $_[KERNEL]->yield('release_lock') 
                                unless defined $_[HEAP]->{wait_session};
         }, 
         release_lock => sub {
             $_[HEAP]->{state}='finished';
             
             # unlock permit table if we locked it...
             if (defined $_[HEAP]->{fset}->{permit_session} and 
                 $_[HEAP]->{fset}->{permit_session} == $_[SESSION]) {
                 DEBUG "unlocking permit table\n";
                 delete $_[HEAP]->{fset}->{permit_session};
             } else {
                 DEBUG "In release_lock without a lock, check for got_errors!\n";
                 DEBUG "Lock holder is $_[HEAP]->{fset}->{permit_session}\n" 
                     if defined $_[HEAP]->{fset}->{permit_session};
             }
             
             # Either way wake up lock waiters...this list will be empty if we 
             # did not hold the permit lock
             Lento::EventSupport::WakeUpWaitList();
             $_[KERNEL]->yield('finish');
         }
        );

}

# 
sub BackFetch {
    my $packet = shift;
    &MarkLentoBusy;
    new POE::Session
        ([$packet, "BackFetch"], # two arguments to _start
         'Lento::EventSupport' , ['alias_set', 'alias_remove', '_stop', 
                                  '_child', '_signal'], 
         'Lento::InterMezzo::ReqHandler', ['_start', 'finish', 'got_error'],
         REQ => sub {
             my $packet = $_[HEAP]->{req_packet};
             my $fsetname = @{$packet->[PKT_PARMS]}[0];
             my $inodeinfo = @{$packet->[PKT_PARMS]}[1];
             # one of Rsync, Desc or Wget for now -SHP
             my $fetchtype  = @{$packet->[PKT_PARMS]}[2];

             # set default to traditional Bulk Descriptor for backward
             # compatibility
             $fetchtype="Desc" unless defined $fetchtype;
             
             $_[HEAP]->{session_level} = LVL_RECOVER;
             # does this fileset actually exist here?
             if (!defined($_[HEAP]->{fset} = $::fsetdb->find_by_name($fsetname))) {
                 # Invalid fileset name sent by client
                 #$_[HEAP]->{connection}->sendpacket('REP', $packet, [EINVAL]);
                 $_[KERNEL]->yield('finish');
             }
             # set up source desc
             my $path = $_[HEAP]->{fset}->mtpt();
             DEBUG "*** fsetname $fsetname, inode $inodeinfo->{ino} mtpt $path\n";
             
             $path .= sprintf("/...ino:%lu:%u", $inodeinfo->{ino}, $inodeinfo->{generation});
             DEBUG "**** path $path, fetchtype $fetchtype\n";
             DEBUG "****direction $Lento::Bulk::SERVER_TO_CLIENT,$Lento::Bulk::SOURCE\n";

             # instantiate a descriptor of the correct type $fetchtype
             # Note that all descriptors must share the Lento::Bulk namespace
             my $desc = "Lento::Bulk::$fetchtype"->new
                 ({ source_type      => $Lento::Bulk::SOURCE_FILE,
                    source_object    => $path,
                    direction        => $Lento::Bulk::SERVER_TO_CLIENT,
                    role             => $Lento::Bulk::SOURCE,
                    connection       => $packet->[PKT_CONNECTION],
                    debug            => 1,
                    callback_event   => 'finished_sending',
                    callback_session => $_[SESSION],
                    source_act => $_[HEAP]->{req_packet}->[PKT_CTOKEN],
                    sink_act => $_[HEAP]->{req_packet}->[PKT_STOKEN] } 
                 );

             if (!$desc) {
                $packet->[PKT_CONNECTION]->set_error
                    ("descriptor instantiation",ENOENT, "No descriptor");
                return;
             }
             
             DEBUG $desc->printme();
             
             my ($rc, $bulk_session) = $desc->setup();
             if ($rc) {
                 if ($rc == ENOENT) {
                     #$_[HEAP]->{connection}->sendpacket('REP', $packet, [ENOENT]);
                     $_[KERNEL]->yield('finish');
                     return;
                 } elsif ($rc == ENOTCONN) {
                     $_[KERNEL]->yield('got_error',"backfetch",ENOTCONN,
                                       "No connection");
                     return;
                 } else {
                     $!=$rc;
                     confess "Error setting up bulk transfer ($!)\n";
                 }
             }

             $desc->start();
         },
         finished_sending => sub {
             DEBUG "Got finished_sending in BackFetch\n";
             #     $_[HEAP]->{connection}->sendpacket('REP', $packet, [0]);
             #   DEBUG "Ping delta (server): " . (time() - $time) . "\n";
             $poe_kernel->yield('finish', [ 0]);
         }
        );

}

# Arguments: (fsetname, offset, gap, length)
# Receiver of KML
sub Reintegrate_KML {
    my $packet = shift;
    &MarkLentoBusy;
    new POE::Session
        ([$packet, "Reintegrate_KML"], # two arguments to _start
         'Lento::EventSupport' , ['alias_set', 'alias_remove', '_stop', 
                                  '_child', '_signal'], 
         'Lento::InterMezzo::ReqHandler', ['_start', 'finish', 'got_error'],
         REQ => sub {
             my $fsetname = shift @{$_[HEAP]->{req_packet}->[PKT_PARMS]};
             # use the naming used by the sender of the KML
             $_[HEAP]->{next_to_send} = 
                 shift @{$_[HEAP]->{req_packet}->[PKT_PARMS]};
             my $fsetinfo= $_[HEAP]->{fsetinfo} = 
                 shift @{$_[HEAP]->{req_packet}->[PKT_PARMS]};
             my $segment_length = shift @{$_[HEAP]->{req_packet}->[PKT_PARMS]};
             $_[HEAP]->{segment_length} = $segment_length;
             $_[HEAP]->{connection} = $_[HEAP]->{req_packet}->[PKT_CONNECTION];
             
             # offsets/recnos in remote's nomenclature, so remote's remote is us!
             ($_[HEAP]->{remote_kmlsize}, $_[HEAP]->{last_remote_recno},
              $_[HEAP]->{last_remote_offset}, $_[HEAP]->{last_local_recno},
              $_[HEAP]->{last_local_offset}) =  unpack("NNNNN", $fsetinfo);
             
             $_[HEAP]->{kmlbuf} = "";
             $_[HEAP]->{session_level} = LVL_RECOVER;
             
             # does this fileset actually exist here?
             if (!defined($_[HEAP]->{fset} = $::fsetdb->find_by_name($fsetname))) {
                 # Invalid fileset name sent by client
                 $_[HEAP]->{connection}->sendpacket('REP', 
                                    $_[HEAP]->{req_packet}, [EINVAL]);
                 $_[KERNEL]->yield('done');
                 return;
             }
             
             # do we have a replication relationship with the peer? 
             if ( ! defined ($_[HEAP]->{replicator} 
                             = Lento::Replicator::find 
                             ($_[HEAP]->{connection}->{sysid}, $fsetname))) {
                 LOG ("No replicator for client $_[HEAP]->{connection}->{sysid}".
                      "fileset $fsetname\n");
                 $_[HEAP]->{connection}->sendpacket('REP', 
                                            $_[HEAP]->{req_packet}, [EINVAL]);
                 $_[KERNEL]->yield('done');
                 return;
             }
             
             # are we forwarding or reintegrating?
             # is the source of the records OK? 
             # XXX: Fix for proxies
             if ( $_[HEAP]->{fset}->{serversysid} eq  $::mysysid ) {
                 $_[HEAP]->{session_level} = LVL_FULL;
                 if ($_[HEAP]->{fset}->{permitholder} ne 
                     $_[HEAP]->{connection}->{sysid} ) {
                     warn "-=- REINTEGRATE WITHOUT PERMIT -----------\n";
                     warn "-=- permitholder: $_[HEAP]->{fset}->{permitholder}\n";
                     warn "-=- connectsysid: $_[HEAP]->{connection}->{sysid}\n";
                     $_[HEAP]->{connection}->sendpacket('REP', 
                                            $_[HEAP]->{req_packet},[EPERM]);
                     $_[KERNEL]->yield('complete');
                     return 0;
                 }
             } else {
                 my $sender =  $_[HEAP]->{connection}->{sysid};
                 die "$fsetname KML not coming from server but from $sender" 
                     unless $_[HEAP]->{fset}->{serversysid} eq $sender;
             }
             
             DEBUG("last_remote_offset %d, last_local_offset %d,".
                   " last_remote_recno %d,  last_local_recno %d\n".
                   " replicator %s\n, length %d, S-ACT %d\n",
                   $_[HEAP]->{last_remote_offset}, 
                   $_[HEAP]->{last_local_offset}, 
                   $_[HEAP]->{last_remote_recno},  
                   $_[HEAP]->{last_local_recno},  
                   $_[HEAP]->{replicator}->printme(), 
                   $segment_length, 
                   $_[HEAP]->{req_packet}->[PKT_STOKEN]);
             
             my $fset = $_[HEAP]->{fset};
             my $replicator= $_[HEAP]->{replicator};
             
             # check that we are getting the record we expect
             my $local_last_state=$replicator->test_direction
                 ($_[HEAP]->{last_remote_recno},$_[HEAP]->{last_remote_offset},
                  $_[HEAP]->{last_local_recno},$_[HEAP]->{last_local_offset});
             
             # Note that remote's remote is us!
             if ($local_last_state eq "sending" ) {
                 if ($_[HEAP]->{replicator}->next_to_send() != 
                     $_[HEAP]->{last_remote_offset}) {
                     LOG("fset: %s, peer %s, last sent (%d), last remote (%d)\n %s \n %s",
                          $fsetname, 
                          $_[HEAP]->{connection}->{sysid},
                          $_[HEAP]->{next_to_send}, 
                          $_[HEAP]->{last_remote_offset}, 
                          $replicator->printme(), $fset->printme());
                     LOG("Permit switched but previous reintegration incomplete?\n");
                     LOG("Attempting to adjust next_to_send offsets from "
                         .$_[HEAP]->{replicator}->next_to_send()." to ".
                         "$_[HEAP]->{last_remote_offset} bytes\n");
                     # XXX: make sure that the if clause is not needed!
                     # If records were lost by the remote, 
                     # our conflict detection should kick in as if this were a local
                     # conflict situation.
                     #if ( $_[HEAP]->{last_remote_offset} 
                     #     < $_[HEAP]->{replicator}->next_to_send()) {
                     #  die "records were lost...unable to adjust\n";
                     #} else {
                     DEBUG("Resetting next_to_send\n");
                     $_[HEAP]->{replicator}->next_to_send(
                                                          $_[HEAP]->{last_remote_offset});
                     #}
                 }
             } else {
                 if ($_[HEAP]->{replicator}->last_remote_offset() != 
                     $_[HEAP]->{next_to_send}) {
                     LOG ("fset: %s, peer %s last sent (%d)!= last received\n %s\n %s",
                          $fsetname, 
                          $_[HEAP]->{connection}->{sysid}, $_[HEAP]->{next_to_send}, 
                          $replicator->printme(), $fset->printme());
                     LOG("Record sequence error in reintegration\n");
                     LOG("attempting to adjust offsets from "
                         .$_[HEAP]->{replicator}->last_remote_offset()." to ".
                         "$_[HEAP]->{next_to_send} bytes\n");
                     
                     # if the remote thinks it sent us fewer records than we have
                     # all we need to do is to shift our offs_incr pointer. In
                     # The other case, we need to disconnect and redo a ReplicatorStatus
                     if (($_[HEAP]->{offs_incr} 
                          = $_[HEAP]->{replicator}->last_remote_offset() 
                          - $_[HEAP]->{next_to_send}) < 0) {
                         # This might happen if a connection reestablishes before 
                         # it is officially terminated by lento...the peer will
                         # run ahead of us and we must reset.
                                # die "records were lost... adjust, resync manually\n";
                         LOG(" RECORDS WERE LOST...MANUAL RESYNC MAY BE NEEDED\n");
                         DEBUG "******lost records...disconnecting*****\n"; 
                         $_[KERNEL]->yield('got_error',"reintegrate kml",ESTALE,
                                           "bad record sequence");
                         
                         # and disconnect
                         $_[HEAP]->{connection}->set_error
                                ("reintegrate kml",ESTALE,"bad record sequence");
                         return;
                     }
                 }
             }
             
             my ($remote_recno, $remote_offset, $local_recno, $local_offset) 
                 = $_[HEAP]->{replicator}->last_rcvd();
             
             my $max_local_offset= 
                 ($local_offset > $_[HEAP]->{last_remote_offset})?
                     $local_offset : $_[HEAP]->{last_remote_offset};
             
             # check for possible conflicts
             # do we have non-reintegrated data? 
             if ($::psdev->get_kml_size($fset->mtpt()) - $max_local_offset) {
                 DEBUG($replicator->printme()." kml size=".$::psdev->get_kml_size($fset->mtpt())." local offs= $max_local_offset\n");
                 print "Possible conflict; resync manually\n";
		 exit 100;
             }
             
             my $desc = Lento::Bulk::Desc->new
                 ({ sink_type => $Lento::Bulk::SINK_BUFFER, 
                    sink_object => \$_[HEAP]->{kmlbuf},
                    bytesize => $segment_length,
                    role =>  $Lento::Bulk::SINK,
                    direction => $Lento::Bulk::CLIENT_TO_SERVER,
                    connection => $_[HEAP]->{req_packet}->[PKT_CONNECTION],
                    callback_session => $_[SESSION],
                    callback_event => 'got_kml',
                    source_act => $_[HEAP]->{req_packet}->[PKT_CTOKEN],
                    sink_act => $_[HEAP]->{req_packet}->[PKT_STOKEN],
                    debug => 1 } );
             
             DEBUG("Created sink descriptor for receiving KML:\n");
             DEBUG $desc->printme();
             
             
             my ($rc, $bulk_session) = $desc->setup();
             if ($rc) {
                 if ($rc == ENOTCONN) {
                    $_[KERNEL]->yield('got_error',"reintegrate kml", ENOTCONN,
                                      "No connection");
                    return;
                 }
                 $!=$rc;
                 confess "Error setting up bulk transfer ($!)\n";
             }

         },
         got_kml => sub { 
             # check for errors
             my $replicator = $_[HEAP]->{replicator};
             my $error = $_[ARG0];

             if ($error || $_[HEAP]->{error}) { 
                 $_[KERNEL]->yield('finish') ;
                 return;
             }
             
             # offs_incr might have been fudged earlier.
             $_[HEAP]->{offs_incr} = 0 if (!$_[HEAP]->{offs_incr});
             $_[HEAP]->{error}=0;
             $_[KERNEL]->yield('do_reint');
             
         },
         do_reint => sub {
             my $error = $_[ARG0];

             $_[HEAP]->{error}= $error if $error;
             
             if ($_[HEAP]->{error}) { 
                 $_[KERNEL]->yield('finish') ;
                 return;
             }
             
             if (!defined($_[HEAP]->{replicator})) {
                 LOG ("Could not find replicator\n");
                 $_[KERNEL]->yield('got_error','do reint',ENOTCONN,"no replicator");
                 return;
             }
             
             # XXX: hack, if do_reint is called after finish fset gets undefined
             # for some reason
             if (!defined($_[HEAP]->{fset})) {
                 if (!defined($_[HEAP]->{fset} 
                              = $::fsetdb->find_by_name($_[HEAP]->{replicator}->{fsetname}))) {
                     # Invalid fileset name sent by client
                     LOG ("Could not find fileset %s\n",
                          $_[HEAP]->{replicator}->{fsetname});
                     $_[KERNEL]->yield('got_error','do reint',ENOTCONN,"no fileset");
                     return;
                 }
             }
             
             my ($remote_recno, $remote_offset, $local_recno, $local_offset) 
                 = $_[HEAP]->{replicator}->last_rcvd();
             
             # offs_incr is to deal with backfetches
             # if a backfetch is needed, the reint_kmlbuf routine 
             # returns and needs to restart in the middle of the buffer
             # after the backfetch has reported completion. 
             my ($offs_incr, $recno_incr, $rc) = 
                 Lento::KML::Reint->reint_kmlbuf
                     ($_[SESSION],
                      $::psdev,
                      $_[HEAP]->{kmlbuf}, 
                      $_[HEAP]->{fset}->mtpt(), 
                      $_[HEAP]->{fset}->{name},
                      $_[HEAP]->{replicator}->{slot_offset},
                      $_[HEAP]->{next_to_send} + $_[HEAP]->{offs_incr},
                      $_[HEAP]->{offs_incr}, $remote_recno,
                      $_[HEAP]->{replicator}
                     );
             
             $_[HEAP]->{offs_incr} = $offs_incr;
             $_[HEAP]->{recno_incr} = $recno_incr;
             
             DEBUG "did reint: last_local: $_[HEAP]->{last_local_offset}";
             DEBUG " offs_incr: $_[HEAP]->{offs_incr}";
             DEBUG " kmllen: $_[HEAP]->{segment_length}\n";
             DEBUG "last recno from presto is ".
                 $::psdev->get_last_recno($_[HEAP]->{fset}->mtpt())."\n";
             
             if ((!$rc) && 
                  ($_[HEAP]->{offs_incr} ==  $_[HEAP]->{segment_length})) {
                 #If there are backfetches pending then got_backfetch_file will
                 #take care of sending the confirmation.
                 $_[KERNEL]->yield('send_confirmation', 'done');
             } else {
                 # We are here because the reint loop was broken
                 # Only continue if a backfetch was not seen
                 $_[KERNEL]->yield('do_reint') if !($rc);
             }
         },
         got_backfetch_file => sub {
             my $error = $_[ARG0];
             
             $_[HEAP]->{error}= $error if $error;
             
             if ($_[HEAP]->{error}) {
                 # do not reintegrate anything else if there was an error!
                 $_[KERNEL]->yield('finish');
                 return;
             }
             
             if ($_[HEAP]->{offs_incr} == $_[HEAP]->{segment_length}) {
                 # All done! Send in the confirmation.
                 $_[KERNEL]->yield('send_confirmation', 'done') 
             } else {
                 # We broke reintegration to accomodate this backfetch
                 # continue with the reintegration 
                 $_[KERNEL]->yield('do_reint');
             }
             
             DEBUG("got_backfetch: offs_incr: $_[HEAP]->{offs_incr}, kmllen: $_[HEAP]->{segment_length}\n");
         },
         send_confirmation => sub {
             my $status = $_[ARG0];
             my $packet = $_[HEAP]->{req_packet};
             
             DEBUG("CONFIRMATION: offs_incr: $_[HEAP]->{offs_incr}\n");
             if ($status eq 'done') { 
                 if (defined $_[HEAP]->{connection}) {
                     $_[HEAP]->{connection}->sendpacket
                       ('REP', $packet, 
                        [0, $_[HEAP]->{offs_incr}, $_[HEAP]->{recno_incr}]);
                 } else {
                    $_[KERNEL]->yield('got_error',"sending confirmation",
                                    ENOTCONN, "Connection deleted");
                    return;
                 }
             }
             $_[KERNEL]->yield('complete');
         },
         complete => sub {
             $_[KERNEL]->yield('finish');
             
             # XXX - if another replicator has requested the permit,
             # then we really should send it the KML and the permit before
             # sending the KMLs to the other replicators; this should
             # speed up the movement of permits Not doing this yet ...
             # Currently send_KMLs are asynchronously done when a KML upcall
             # is seen by lento. We need to carefully check interactions between 
             # the upcalls, permit handlers and this function.
             
             # This should be taken care of asynchronously by the KML upcalls
             #$_[HEAP]->{fset}->{kml}->ReintegrateKML($_[HEAP]->{replicator}->sysid());
         }
        );
}

1;

__END__
