
=pod

=head1 NAME

KML - Methods for manipulating the Kernel Modification Log file

=head1 SYNOPSIS

=over 4

=item B<new>: Build a new KML object.

=item B<start_offset>: Set the bytewise start offset is an arg is supplied.  
                Return the start offset in any case.

=item B<last_offset>: Set the bytewise start offset is an arg is supplied.  
                 Return the start offset in any case.
=item B<readrec>: Read a single record from a KML file into a buffer.

=item B<ReintegrateKML>: Forward KML to the replicators for reintegration.
                   Why is this called ReintegrateKML?

=item B<send_KML>: Send KML records to a replicator.

=back

=head1 DESCRIPTION

FIXME: Write a description.

=head1 BUGS

=back

=head1 USAGE

=cut

package Lento::KML;
use Lento::KML::Rec;
use Lento::Bulk;
use Lento::Bulk::Source;
use Lento::Connection;
use Lento::Packet;
use POE;
use Lento::Debuggable;

use POSIX qw(:sys_stat_h :errno_h :unistd_h :fcntl_h);
use IO::File;
use strict;
use Carp;

@Lento::KML::ISA=qw(Lento::Debuggable);
$Lento::KML::level=1;

#Retry counter
$Lento::KML::Retry=3;

sub CHUNK_BITS  { return 16; }

######################################################################
#
# static method Lento::KML::new
#
# Create a new KML object with the inputs:
#   class name
#   path for the KML file
#   associated fileset (optional)
#
# Initialize the member data:
#   {path}         To the KML file
#   {handle}       File handle for the KML file
#   {fset}          The KML fileset, if it was input.
#   {start_offset} The initial offset in the KML file, set to zero.
#
# We keep track of two pointers into the KML file:
#
# start_offset: Our best guess as to the next byte to send to the server.
#
# last_offset: One past the last byte we've been authorized 
#              by the kernel to send.
#
#
######################################################################

sub new { 
    my $class = shift;
    my $path = shift;        # path of KML file
    my $fset = shift;        # associated fileset (when defined)
    my $self = {};
    
    $self->{path} = $path;
    $self->{handle} = new IO::File;
    $self->{handle}->open($path, &O_RDWR) or confess "Cannot open";
    $self->{start_offset} = 0;
    $self->{last_offset} = 0;
        $self->{fset} = $fset if (defined $fset);

        bless $self;
}

sub printme { 
    my $kml = shift;
    my $kml_offset = shift;
    my $briefly = shift;
    my $result = "";
    my $count = 0;
    
    $briefly = "" unless defined $briefly;
    $kml_offset = 0 unless defined $kml_offset;
    
    while ($kml_offset < $kml->last_offset()) {
        my ($size, $offset, $buf) = $kml->readrec($kml_offset);
        last if $size == 0;

        DEBUG "skipped chunk boundary at $kml_offset to $offset\n" 
            if ($offset > $kml_offset);
    
        my $rec = Lento::KML::Rec->Unpack($buf, 0, $offset);
    
        $kml_offset = $offset + $size;
        $count++;
    
        $result .= sprintf "Record $count size: $size, offset $kml_offset,  (total kml length %d)\n", $kml->last_offset();
    
        if (defined $rec) {
                $result .= $rec->printme($briefly);
                TACITUS($rec->printme("verybrief",$kml->{fset}));
        } else {
                DEBUG "Warning! KML unpack failed, look out for errors!\n";
                return "";
        }
        
        if ($briefly ne "brief") {
                $result .= "----------------------------------------------------------------\n";
        }
        }
    $result;
}

# read last record suffix (see presto/linux/intermezzo_kml.h)
my $SUFFIX_SIZE = 16;

sub get_last_suffix { 
    my $self = shift;
    my $buf;
    
    
    my $st = stat($self->path) || die "cannot stat KML $self->{path}";
    my $filesize = $st->size;
    my $offset = $st->size - 4;
    DEBUG "File size %d path %s\n", $filesize, $self->{path};
    my $word = 1;
    while ( 1 ) {
        $self->{fh}->seek($offset, SEEK_SET); 
        $self->read($self->{fh}, $buf, 4);
        $word = unpack "I", $buf;
        last if $word;
        $offset -= 4;  # go to the previous integer
    }


    DEBUG "Found last record at offset %d\n", $offset + 4;
    if ($offset <= 0) {
        return (0, 0, 0); 
    }

    $offset -= ($SUFFIX_SIZE - 4); 
    $self->{fh}->seek($offset, SEEK_SET); 
    $self->{fh}->read($buf, $SUFFIX_SIZE);

    my ($prevrec, $recno, $time, $len) = unpack $buf, "IIII"; 

    return ($recno, $len, $filesize); 
}



# With no args, return the start offset.
# With an integer arg, set the start offset.
sub start_offset {
    my $self = shift;
    my $arg = shift;
    
    if ($arg) {
        $self->{start_offset} = $arg;
    }
    
    return $self->{start_offset};
}


# With no args return the last offset.
# With an integer arg, set the last offset from argument (deprecated)
sub last_offset { 
    my $self = shift;
    my $arg = shift;
   
    $self->{last_offset}= $arg if defined $arg;
    
    return $self->{last_offset};
}

# Read a single record from the KML into a buffer.  If called with
# optional position, first seek to that position then read the record.
# This routine also skips chunk boundaries if any are detected.
sub readrec () {
    #
    # Get the KML object and the optional position to seek to.
    #
    my $self = shift;
    my $pos = shift;
    my $buf = "";
    my $size=0; 
    my $count;
    my $res;
    
    
    while (!$size) {

        # did we skip past the last offset...this can happen as a kml 
        # record is being written at the chunk boundary, before presto
        # actually updates its last_offset
        if ($pos >= $self->last_offset()) {
            warn "KML at $pos beyond last offset: $self->{last_offset}\n";
            return (0, $pos, undef) 
        }

        sysseek($self->{handle}, $pos, SEEK_SET) if ( defined $pos );

        $res = $self->{handle}->sysread($buf, 4);
        if (!$res) {
            LOG("kml read failed: $!\n");
            return (0, $pos, undef) 
        }

        # We should be able to read all four bytes in the prefix
        confess "Corrupt KML record only $res bytes read: error $!\n"
            if $res != 4;

        $size = unpack "I", $buf;
        if (!$size) {
            # Possible chunk boundary...
            my $chunksize = 1<< &CHUNK_BITS;
            my $chunkmask = ~($chunksize - 1);
            my $chunkboundary = ($pos + $chunksize) & $chunkmask;

            #...move to start of next chunk
            $pos=$chunkboundary;
        }

    }
    
    sysseek($self->{handle}, $pos, SEEK_SET) if ( defined $pos );
    
    $res = $self->{handle}->sysread($buf, $size);
    DEBUG("Res=$res, size=$size, pos=$pos, length=%d\n", length($buf));
    # if the first 4 bytes could be read, the entire kml record should be
    # readable...
    confess "Could not read record in KML::readrec: os error $!\n"  
        if ( $res != $size );
    
    return ($res, $pos, $buf);
}


######################################################################
#
# method Lento::KML::ReintegrateKML
#
# Forward a KML from client to all replicators
# Skip the "not_send" replicator (commonly, one replicator is skipped)
# Do start_host first (not yet implemented).
#
# Inputs:
#    the KML object
#    a sysid for a host to skip. (optional)
#    A closure to call when we're done. (optional)
#
# Returns:
#    The number of callbacks we expect, which is the number of 
#    successful send_KML's we got.
#
######################################################################

sub ReintegrateKML ($$$$) {
    my $kml = shift;
    my $skip_host = shift;
    my $callback_count = 0;
    my $closure = shift;
    my $level = shift;
    # my $start_host = shift; 
    
    $level = LVL_FULL unless defined $level; 
    $skip_host = "" unless defined $skip_host;
    
    # Loop over all of the replicators, sending the KML to each.
    $kml->{fset}->{replist}->iterate
    ( sub {
          my $replicator = shift;
          
          if ($skip_host eq $replicator->sysid()) {
            DEBUG("Skipping $skip_host\n");
            return;
          }
          
          my ($err, $session) = $kml->send_KML($replicator, $level);
          
          # If the send was successful, register a closure for that
          # sending session.
          if (!$err && defined $session) {
          $callback_count++;
          if (defined $closure) {
              Lento::EventSupport::WaitForSession($session, $closure);
          }
          }
      }
    );
    
    # Return the number of callbacks we expect.
    return $callback_count;
}


sub ReintegrateRep ($$$$) {
    my $kml = shift;
    my $replicator = shift;
    my $callback_count = 0;
    my $closure = shift;
    
    my ($err, $session) = $kml->send_KML($replicator, undef);
    
    if (!$err && defined $session) {
    $callback_count++;
    if (defined $closure) {
        Lento::EventSupport::WaitForSession($session, $closure);
    }
    }
    return $callback_count;
}


=pod

=over 4

=item B<method Lento::KML::send_KML>

Set up a session that will send a block of KML to a particular
    replicator.  The session will create the necessary bulk descriptors
    used to transfer the KML data.

=back

# use the WaitForSession semantics
# return (errorcode, session)

=cut
    
sub send_KML ($) {
    my $kml = shift;
    my $replicator = shift;
    my $level = shift;
    
    #XXX: We may have huge problems here if the first send_KML does not go
    # through...how do you check for that case?
    $level = LVL_FULL unless defined $level;
    
    #LOG(">>    send_KML: $kml->{path} to $replicator->{sysid}, lvl $level\n");
    
    if (!defined $replicator) {
        LOG("No replicator in send_KML - die for debugging!\n");
        die " No replicator in send_KML";
        return (1,undef);
    }
    
    # get the connection object.
    my $connection = Lento::Connection::find($replicator->sysid());
    
    # Has the connection recovered to our level? 
    if (!Lento::Connection::joined_barrier($connection, $level)) { 
        # LOG("No connection in send_KML\n");
        return (1,undef);
    }

        # ask the kernel where the last record boundary lies.
    $kml->last_offset($::psdev->get_kml_size($kml->{fset}->mtpt()));
    
    DEBUG("send_KML: last_offset %d\n%s\n", 
            $kml->last_offset(), $replicator->printme());
    
    # Is the replicator already sending stuff?
    # Return no error and the existing session.
    if ( $replicator->send_session() ) {
        DEBUG("More pending for %s; not reinitiating send_KML\n", 
              $replicator->{sysid});
        $replicator->morepending(1);
        return (0, $replicator->send_session());
    }
    
    # Check for the case of next_to_send==last_offset,
    # in which case we don't have anything to do.
    # 
    # State of replicator:
    #   no more pending
    #   not sending 
    # Return no error.
    if ($replicator->next_to_send() == $kml->last_offset()) {
        DEBUG "Nothing to ship; returning; last_offset: $kml->{last_offset}\n";
        $replicator->morepending(0);
        return (0, undef);
    }
    
    # Check to see if the replicator has gone off line.
    # If it has: don't send, return an error.
    if ( !$replicator->online() ) {
        DEBUG("Fileset off-line: no send_KML for %s to %s\n", 
              $kml->{path}, $replicator->{sysid});
        $replicator->morepending(0);
        #$replicator->clear_send_session();
        return (1, undef);
    }
   
        # temporary placeholder for the send_session
        $replicator->{send_session}=1;

    &MarkLentoBusy;

    # start a session 
    $replicator->{send_session}= new POE::Session
        (  
         # three arguments to _start:
         [$kml, $replicator, $level],
         'Lento::EventSupport' , ['alias_set', 'alias_remove',
                                  '_child', '_signal', '_stop', 'AddWaiter'],
         _start => sub {
             # Get the kml, replicator and session level from the constructor.
             my $kml = $_[HEAP]->{kml} = $_[ARG0];
             my $replicator = $_[HEAP]->{replicator} = $_[ARG1];
             $_[HEAP]->{session_level} = $_[ARG2];
         
         
             # Get the sysid from the replicator so we can get a good name.
             my $id = $replicator->sysid();
             $_[HEAP]->{session_name} = $_[SESSION]->ID().
                 ": send_KML from $kml->{fset}->{name} to $id $_[SESSION]";
         
             Lento::Debuggable::set_option($_[SESSION], trace => 1, 
                                           default => 1);
         
             DEBUG "setting alias to send_KML-$_[HEAP]->{session_name}\n";
             $_[KERNEL]->alias_set("$_[SESSION]") && 
                 die "Unable to set alias";
         
             $_[KERNEL]->yield('sendreq');
         },
         sendreq => sub {
             # Build the bulk transfer descriptor and initiate the
             # reintegration by sending a REQ packet with the
             # Reintegrate method.     
             my $kml = $_[HEAP]->{kml};
             
             $_[HEAP]->{connection} = Lento::Connection::find
                 ($_[HEAP]->{replicator}->sysid());
             
             DEBUG "Using connection ". $_[HEAP]->{connection} ."\n";
             
             if ( !Lento::Connection::connected($_[HEAP]->{connection})) {
                 DEBUG("No connection to %s\n",$_[HEAP]->{replicator}->sysid());
                 $_[KERNEL]->yield('got_error', "kml send request", 
                                   ENOTCONN, "Not connected");
                 return;
             }
             
             # We'll start sending at the 'next_to_send' offset....
             DEBUG("^^%%^^ Sending segment:  up to $kml->{last_offset}".
                   " fileset %s to %s:\n %s\n", 
                   $_[HEAP]->{replicator}->{fsetname}, 
                   $_[HEAP]->{replicator}->sysid(), 
                   $_[HEAP]->{replicator}->printme());
             
             my $fsetstatus =  $_[HEAP]->{replicator}->repstatus();
             
             # How many bytes should we send? We will send till we hit a 
             # chunk boundary. Beyond this we create a new descriptor.
             my $chunksize = 1 << &CHUNK_BITS;
             my $chunkmask = ~($chunksize - 1);
             my $chunkboundary = ($_[HEAP]->{replicator}->next_to_send() + 
                                  $chunksize) & $chunkmask; 
             
             my $last_to_send = $kml->last_offset();

             DEBUG("Chunk Boundary at %d/%d=%x, last_to_send %d\n",
                 $chunkboundary,$chunksize,$chunkmask,$last_to_send);
             if ($last_to_send > $chunkboundary) {
                 # KML is too large to send, send upto chunk boundary instead
                 $last_to_send = $chunkboundary;
                 
                 # Inform replicator that there is more kml to send
                 $_[HEAP]->{replicator}->morepending(1);
                 
                 DEBUG "Adjusted last offset to send to $last_to_send\n";
             }

             # We're sending a file from the KML file, starting at start_offset
             # and the number of bytes is up to last_to_send.  
             # When it finishes, call finished_sending_kml in this session.
             #
             # Constructing this descriptor does not cause any communication.
             my $bytecount = $last_to_send - 
                 $_[HEAP]->{replicator}->next_to_send();

             DEBUG ("Sending $bytecount bytes of KML\n");
             my $bulkdesc = Lento::Bulk::Desc->new
                 ( {
                    source_type      => $Lento::Bulk::SOURCE_FILE,
                    source_object    => $kml->{path},
                    source_offset   => $_[HEAP]->{replicator}->next_to_send(), 
                    bytesize         => $bytecount,
                    role             => $Lento::Bulk::SOURCE,
                    direction        => $Lento::Bulk::CLIENT_TO_SERVER,
                    connection       => $_[HEAP]->{connection},
                    debug            => 1,
                    callback_event   => 'finished_sending_kml',
                    callback_session => $_[SESSION]});
             
             DEBUG $bulkdesc->printme();
             
             
             # Handle some setup for the descriptor.
             my ($rc, $bulk_session) = $bulkdesc->setup();
             $_[HEAP]->{bulk_session}=$bulk_session;
             if ($rc) {
                 $_[KERNEL]->yield('got_error', "send_KML bulk setup", 
                                   EIO, "IO Error");
                 return;
             }
             
             # Send the REQ packet to the peer.
             # The peer will send START back to us.
             $_[HEAP]->{connection}->sendpacket
                 ('REQ', ['Reintegrate_KML', $kml->{fset}->name(), 
                          $_[HEAP]->{replicator}->next_to_send(), $fsetstatus,
                          $bytecount ]);
         },
         finished_sending_kml => sub { 
                         # we might get stuck here waiting for a rep
             DEBUG "$_[HEAP]->{connection}:$_[HEAP]->{connection}->{session}\n";
             if (!Lento::Connection::connected($_[HEAP]->{connection})) {
                 $_[KERNEL]->yield('got_error', 'finished_sending_kml', 
                                   ENOTCONN, 'no connection');
             }
             # now wait for rep 
             1;
         },
         REP => sub {
             my $packet = $_[ARG0];
             my $kml = $_[HEAP]->{kml};
             my $replicator = $_[HEAP]->{replicator};
         
             my $rc = shift @{$packet->[PKT_PARMS]};
             # report the server error
             if ($rc != 0) {
                 confess "ERROR from Reintegrate_KML: server" . 
                     $kml->{fset}->server()->{name} .
                     "returns error code $rc for fileset $kml->{fset}->{name}";
             }
             
             # The peer sends us a packet every time it completes the
             # reintegration of some records.  Inside the packet is the
             # extentsize of the KML integrated up to that point
             my $confirmed_len = shift @{$packet->[PKT_PARMS]};
             $replicator->incr_next_to_send($confirmed_len);
             DEBUG("  >>  Last confirmed extent: $confirmed_len, nts: %d\n", 
                 $replicator->next_to_send());
             
             $_[KERNEL]->yield('finish', 0);
         },
         got_error => sub {
             my $operation = $_[ARG0];
             my $errno = $_[ARG1];
             my $msg = $_[ARG2];
             LOG "  >>  error in send_KML: $msg ($errno)\n";
             $_[HEAP]->{replicator}->morepending(0);
         
             $_[HEAP]->{replicator}->clear_send_session($_[SESSION]);

             # signal error to the source to prevent fd leaks.
             $_[HEAP]->{desc}->error() if $_[HEAP]->{desc};
             
             # We now signal completion...use a call to make this "atomic"
             $_[KERNEL]->call($_[SESSION],'finish', 1);
         },
         finish => sub {
             my $status = $_[ARG0];
             my $kml = $_[HEAP]->{kml};

             $_[HEAP]->{connection}->put();

             # If bulk session is this session we have a circular reference
             delete $_[HEAP]->{bulk_session};
             
             if ( $_[HEAP]->{replicator}->morepending() && !$status &&
                  $_[HEAP]->{replicator}->next_to_send()<$kml->last_offset() ) {
                 DEBUG("  >>  RESTART sending KML\n");
                 $_[KERNEL]->yield('sendreq');
             } else {
                 DEBUG("send_KML done; last_offset: %d, next_to_send \n", 
                        $kml->{last_offset}, 
                                                $_[HEAP]->{replicator}->next_to_send());
                 $_[HEAP]->{replicator}->clear_send_session($_[SESSION]);
                 $_[HEAP]->{replicator}->morepending(0);
                 
                 Lento::EventSupport::WakeUpWaitList($status);
                 
                 $_[KERNEL]->alias_remove("$_[SESSION]");
                 DEBUG "<<<<  send_KML: done\n";
             }
         }
        );
    
    #
    # return that we got no error and the session we created.
    #
    return (0,$replicator->send_session());
}
          
1;
