# performs reintegration of KML entries

package Lento::KML::Reint;

use strict;
use Carp;
use Lento::Debuggable;
use File::stat;
use File::Copy;
use POSIX qw(:unistd_h :fcntl_h);
use Lento::KML;
use Lento::KML::Rec;
use Lento::Psdev;
#use Lento::KML::ReintFS; --- fix the warnings please
use Lento::KML::ReintLento;
use SetFS;
    
@Lento::KML::Reint::ISA = qw(Lento::Debuggable);
$Lento::KML::Reint::level = 1;

#Maximum number of records to reintegrate at a time
#Required to solve livelock problems caused by reint_kmlbuf
$Lento::KML::Reint::MAX_REINT = 10;

my %opcode = (1 => "reint_create",
          2 => "reint_mkdir",
          3 => "reint_unlink",
          4 => "reint_rmdir",
          5 => "reint_close",
          6 => "reint_symlink",
          7 => "reint_rename",
          8 => "reint_setattr",
          9 => "reint_link",
          10 => "reint_open",
          11 => "reint_mknod",
          15 => "reint_setextattr",
          16 => "reint_delextattr");

my $parentclass = "Lento::KML";

sub LENTO_FL_KML   {         0x0001; };
sub LENTO_FL_EXPECT {        0x0002; };
sub LENTO_FL_VFSCHECK {       0x0004; };
sub LENTO_FL_JUSTLOG {       0x0008; };
sub LENTO_FL_WRITE_KML {      0x0010; };
sub LENTO_FL_CANCEL_LML {     0x0020; };
sub LENTO_FL_WRITE_EXPECT {   0x0040; };
sub LENTO_FL_IGNORE_TIME {   0x0080; };


# given a buffer containing multiple KML records from the Sender
# loop through all KML records and reintegrate into Receiver
sub reint_kmlbuf {
    my $class = shift;
    my $session = shift;
    my $psdev = shift;
    my $kmlbuf = shift; # sender's KML buffer -  unknown number of KML records
    my $fsetmtpt = shift; # mountpoint of local fileset 
    my $fsetname = shift; # mountpoint of local fileset 
    my $slot_offset = shift; # slot offset for the sender of the KML buffer
    my $kml_offset = shift;  # kml offset for the sender of the KML buffer
    my $buf_offset = shift;  # offset to start at in the buffer
    my $last_remote_recno = shift; # Last record number we received from remote
    my $replicator = shift; # replicator to use to adjust offsets for chunks

    my $kmlbuf_size = length($kmlbuf);
    my $flags = LENTO_FL_KML | LENTO_FL_EXPECT; 

    # loop through the entire kmlbuf stripping off recs and
    # reintegrating them
    my $count = 0;
    my $rc = 0;
    my $msg = "";

        DEBUG "last_remote_recno $last_remote_recno\n";

    while ($buf_offset < $kmlbuf_size && 
           $count < $Lento::KML::Reint::MAX_REINT) {
        # strip off a KML record
        DEBUG "XXXXXXXXXXXXXX kmloffset: $kml_offset, buf_offset $buf_offset\n";
        # if the length is 0 we need to jump to the next 64k boundary
        my $len = unpack("x$buf_offset I", $kmlbuf);

        if (!$len) {
            my $chunksize = 1<< &Lento::KML::CHUNK_BITS;
            my $chunkmask = ~($chunksize - 1); 
            my $chunkboundary = ($kml_offset + $chunksize - 1) &
                $chunkmask; 
       
                        DEBUG " hit chunk boundary at $kml_offset\n"; 
            # move to start of next record
            $buf_offset = $buf_offset + $chunkboundary - $kml_offset;
            $kml_offset  = $chunkboundary;
                        DEBUG " restarting at offset $kml_offset\n"; 

            # We skipped a bunch of bytes so reset the last remote offset
            $replicator->set_remote_offset($kml_offset);
           
            # we may be at end of the kml chunk so recheck.
            next;
        }

        my $rec = Lento::KML::Rec->Unpack($kmlbuf, $buf_offset, $kml_offset);

        die "Unable to unpack KML record\n" unless defined $rec;

        DEBUG $rec->printme();

        # XXX: add checks for duplicate records here.
        # reintegrate it locally - if we return 1 it was a close and 
        # we break the loop to wait for the file
        if ($rec->{recno} > $last_remote_recno) {
                        DEBUG "reint record\n";
            ($rc, $msg) = reint_rec("ReintLento", $rec, $fsetmtpt, $psdev, 
                               $slot_offset, $flags, $fsetname, $replicator);
            $last_remote_recno=$rec->{recno} if (!$rc);
        }
        $count++;   # after $Lento::KML::Reint::MAX_REINT records yield 

        # move to start of next record
        $buf_offset += $rec->{len};
        $kml_offset += $rec->{len};
    
        DEBUG( "Record $rec->{recno}, buf_offset: $buf_offset, kml_offset: $kml_offset, size: $rec->{len}, offset $buf_offset, (log length %d)\n", $kmlbuf_size);
        DEBUG "----------------------------------------------------------------\n";

        last if $rc;
    }

    return ($buf_offset, $count, $rc); 
    
}

sub reint_rec {
    my $reint_class = shift;
    my $rec = shift;

    my $method = $parentclass . "::" . $reint_class . 
        "::" . $opcode{$rec->{opcode}};
    DEBUG "calling $method for recno: $rec->{recno}\n";

    no strict 'refs';
    return &{"$method"}($rec, @_);
}

1;

__END__


=head1 NAME

KML - Kernel Modification Log Reintegration

=head1 SYNOPSIS

FIXME: Write synopsis

=head1 DESCRIPTION

=item 


=cut
