package Lento::KML::ReintFS;

use strict;
use Carp;
use Lento::Debuggable;
use File::stat;
use File::Copy;
use POSIX qw(:unistd_h :fcntl_h);
use SetFS;
    
@Lento::KML::ReintFS::ISA = qw(Lento::Debuggable);
$Lento::KML::ReintFS::level = -100;

# a bunch of flags, not all needed at this time
sub ATTR_MODE      {1;}
sub ATTR_UID       {2;}
sub ATTR_GID       {4;}
sub ATTR_SIZE      {8;}
sub ATTR_ATIME     {16;}
sub ATTR_MTIME     {32;}
sub ATTR_CTIME     {64;}
sub ATTR_ATIME_SET {128;}
sub ATTR_MTIME_SET {256;}
sub S_IFMT   {00170000;}
sub S_IFREG   {0100000;}
sub S_IFLNK   {0120000;}
sub S_IFSOCK  {0140000;}
sub S_IFDIR   {0040000;}
sub S_IFCHR   {0020000;}
sub S_IFBLK   {0060000;}
sub S_IFIFO   {0010000;}
sub S_ISUID   {0004000;}
sub S_ISGID   {0002000;}
sub S_ISVTX   {0001000;}
sub S_IRWXU     {00700;}
sub S_IRWXG     {00070;}
sub S_IRWXO     {00007;}
sub S_IRWXUGO {(S_IRWXU|S_IRWXG|S_IRWXO);}
sub S_IALLUGO {(S_ISUID|S_ISGID|S_ISVTX|S_IRWXUGO);}

sub S_ISLNK {
    my $mode = shift;
    return (($mode & S_IFMT) == S_IFLNK) ;
}

sub S_ISREG {
    my $mode = shift;
    return (($mode & S_IFMT) == S_IFREG) ;
}

sub S_ISDIR {
    my $mode = shift;
    return (($mode & S_IFMT) == S_IFDIR) ;
}

sub S_ISCHR {
    my $mode = shift;
    return (($mode & S_IFMT) == S_IFCHR) ;
}
sub S_ISBLK {
    my $mode = shift;
    return (($mode & S_IFMT) == S_IFBLK) ;
}
sub S_ISFIFO {
    my $mode = shift;
    return (($mode & S_IFMT) == S_IFIFO) ;
}
sub S_ISSOCK {
    my $mode = shift;
    return (($mode & S_IFMT) == S_IFSOCK) ;
}

sub DEBUG {
    my ($class, $file, $line) = caller();

    # this next line -really- confuses xemacs, for what it's worth
    my @args =  @_;
    $args[0] = "[$file: $line] $_[0]";
    
#    printf @args;
    
    return;
}

my $parentclass = "Lento::KML::ReintFS";


# STATIC METHODS

sub reint_rec {
    my $rec = shift;

    my $method = $parentclass . "::reint_" . lc $rec->{op};

    no strict 'refs';
    return &{"$method"}($rec, @_);

}


sub reint_pre_secure {
    my $rec = shift;
    my $hmsg = "reint_pre_secure: ";

    # prepare to set the group array to the correct groups
    # the first number must be the effective gid
    my @groups = ($rec->{fsgid});
    #  figure out groups for correct owner and push onto the groups array
    setgrent;
    my ($name,$passwd,$gid,$members);
    my $login = (getpwuid($rec->{fsuid}))[0];
    while (($name, $passwd, $gid, $members) = getgrent) {
	push(@groups, $gid) if ($members =~ m/$login/);
    }

    # sefstuid to correct owner/group
    return(-1, $hmsg . "cannot setfsuid: $^E\n")
	if (setfsuid($rec->{fsuid}) != $<);
    return(-1, $hmsg . "cannot setfsgid: $^E\n") 
	if (setfsgid($rec->{fsgid}) != $));

    # setgroups, in perl setting $) sets the effective group array
    $) = join (" ",@groups);

    return 0;
}

sub reint_post_secure {
    my $rec = shift;
    my $hmsg = "reint_post_secure: ";

    # restore root owner/group
    return(-1, $hmsg . "cannot restore fsuid: $^E\n")
	if (setfsuid($<) != $rec->{fsuid});
    return(-1, $hmsg . "cannot restore fsgid $^E\n") 
	if (setfsgid($<) != $rec->{fsgid});

    # restore group array to the real group array by
    # setting the effective group array $) to the real group array $)
    $) = $(;

    return 0;
}

sub reint_setattr () {
    my $rec = shift;
    my $mtpt = shift;
    my $path = $mtpt . $rec->{path}; 
    my $hmsg = "reint_setattr: ";
    
    my $filestat = lstat($path);
    if (not defined $filestat) {
	my $type = "mode" if ($rec->{valid} & ATTR_MODE);
	$type = "chown" if (($rec->{valid} & ATTR_UID) ||
			    ($rec->{valid} & ATTR_GID));
	$type = "mtime" if ($rec->{valid} & ATTR_MTIME);
	my $msg = sprintf(" %s on %s failed because file does not exist\n", 
		       $type, $path);
	return(-1, $hmsg . $msg);
    }
    my ($uid, $gid);
    
    my ($rc, $msg) = reint_pre_secure($rec);
    if ($rc != 0) {
	reint_post_secure($rec);
	return($rc, $hmsg . $msg);
    }
    if ($rec->{valid} & ATTR_MODE) {
	DEBUG($hmsg . "orig mode: %o\tnew: %o\n", 
	      $filestat->mode, $rec->{mode});
 	# how to check if we are allowed to do this?
	my $mode = $rec->{mode};
	DEBUG($hmsg . "pre-and mode is currently: %o\n", $mode);
	$mode = $mode & 0xfff;
	($rc, $msg) = setattr($path, "mode", $mode);
	if ($rc != 0) {
	    reint_post_secure($rec);
	    return($rc, $hmsg . $msg);
	}
	#$mode = ($mode & S_IALLUGO); #???
    }
    
    # yes this is weird logic, leave in for now
    my $chownit = 0;
    if ($rec->{valid} & ATTR_UID) {
	DEBUG($hmsg . "orig uid: %d\tnew: %d\n", 
	      $filestat->uid, $rec->{uid});
	$uid = $rec->{uid};
	$chownit = 1;
    }
    if ($rec->{valid} & ATTR_GID) {
	DEBUG($hmsg . "orig gid: %d\tnew: %d\n", 
	      $filestat->gid, $rec->{gid});
	$gid = $rec->{gid};
	$chownit = 1;
    }
    if ( $chownit ) {
	if ( defined $uid && defined $gid ) {
	    ($rc, $msg) = setattr($path, "chown", $uid, $gid);
	    if ($rc != 0) {
		reint_post_secure($rec);
		return($rc, $hmsg . $msg);
	    }
	} else {
	    LOG($hmsg . "recno %d chown: uid or gid not defined, can't execute\n",
		 $rec->{recno});
	}
    }

    # only muck with the mtime
    # if there was a deliberate change of the mtime
    # i.e. setattr for mtime
    if ($rec->{valid} & ATTR_MTIME) {
	DEBUG($hmsg . "orig mtime: %d\tnew: %d\n", $filestat->mtime,
	      $rec->{mtime});
	#	    $atime = $filestat->atime;
	($rc, $msg) = setattr($path, "mtime", $rec->{mtime});
	if ($rc != 0) {
	    reint_post_secure($rec);
	    return($rc, $hmsg . $msg);
	}
    } 

    return reint_post_secure($rec);
}

sub reint_create () {
    my $rec = shift;
    my $mtpt = shift;
    my $path = $mtpt . $rec->{path}; 
    my $hmsg = "reint_create: ";
    my $filestat = lstat($path);
    if (defined $filestat) {
	my $msg = "";
	if (S_ISREG($filestat->mode)) { 
	    $msg = sprintf("failed for %s - already exists.\n", $path);
	    return(-2, $hmsg . $msg);
	} 
	$msg = sprintf("failed for %s - exists but not a file.\n", 
		       $path);
	return(-1, $hmsg . $msg);
    }

    DEBUG("ReintFS::" . $hmsg . "$path %o \n",$rec->{mode});

    my ($rc, $msg) = reint_pre_secure($rec);
    if ($rc != 0) {
	reint_post_secure($rec);
	return($rc, $hmsg . $msg);
    }
    my $localfh = new IO::File;
    if (! $localfh->open($path, "w")) {
	    $msg = "cannot open $path for writing: $^E\nVerify groups are the same on server and client\n";
	    reint_post_secure($rec);
	    return(-1, $hmsg . $msg);
    }  
    $rc = $localfh->close();
    if ($rc < 1) {
	    $msg = "failed close on $path: $^E\n";
	    reint_post_secure($rec);
	    return(-1, $hmsg . $msg);
    }
    ($rc, $msg) = reint_post_secure($rec);
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }
    
    # correct mode/mtime for the object
    my $mode = $rec->{mode};
    DEBUG($hmsg . "pre-and mode is currently: %o\n",$mode);
    $mode = $mode & 0xfff;
    ($rc, $msg) = setattr($path, "mtime mode", $rec->{new_obj_mtime}, $mode);
    if ($rc != 0) {
	    reint_post_secure($rec);
	    return($rc, $hmsg . $msg);
    }

    # update the mtime for the parent directory to the correct 
    #    version mtime
    my $parentdir = $path;
    $parentdir =~ s/(.+)\/.+$/$1/;
    ($rc, $msg) = setattr($parentdir, "mtime", $rec->{new_pdir_mtime});
    if ($rc != 0) {
	    return($rc, $hmsg . $msg);
    }

    0;

}

sub reint_link () {
    my $rec = shift;
    my $mtpt = shift;
    my $oldname = $mtpt . $rec->{path}; 
    my $newname = $mtpt . $rec->{target};
    my $hmsg = "reint_link: ";
    my $filestat = lstat($newname);
    if (defined $filestat) {
	my $msg = "";
	if (S_ISLNK($filestat->mode) or S_ISDIR($filestat->mode)) {
	    $msg = sprintf("failed for %s - exists as an incorrect type.\n", 
			   $newname);
	    return(-1, $hmsg . $msg);
	} 
	$msg = sprintf("failed for %s - already exists.\n", $newname);
	return(-2, $hmsg . $msg);
    }

    my ($rc, $msg) = reint_pre_secure($rec);
    if ($rc != 0) {
	reint_post_secure($rec);
	return($rc, $hmsg . $msg);
    }
    $rc = link($oldname, $newname);
    if ($rc < 1) {
	$msg = "failed link($oldname, $newname): $^E\n";
	reint_post_secure($rec);
	return(-1, $hmsg . $msg);
    }
    
    ($rc, $msg) = reint_post_secure($rec);
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }

    # update the mtime for the parent directory to the correct 
    #    version mtime
    my $parentdir = $newname;
    $parentdir =~ s/(.+)\/.+$/$1/;
    ($rc, $msg) = setattr($parentdir, "mtime", $rec->{new_pdir_mtime});
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }

    0;
}

sub reint_unlink () {
    my $rec = shift;
    my $mtpt = shift;
    my $path = $mtpt . $rec->{path} . '/' . $rec->{target}; 
    my $hmsg = "reint_unlink: ";

    my $filestat = lstat($path);
    if (not defined $filestat) {
	my $msg = "";
	$msg = sprintf("failed for %s - already unlinked.\n", $path);
	return(-2, $hmsg . $msg);
    }

    DEBUG("ReintFS::" . $hmsg . "$path \n");

    my ($rc, $msg) = reint_pre_secure($rec);
    if ($rc != 0) {
	reint_post_secure($rec);
	return($rc, $hmsg . $msg);
    }
    $rc = unlink($path); 
    if ($rc < 1) {
	$msg = "failed unlink($path): $^E\n";
	reint_post_secure($rec);
	return(-1, $hmsg . $msg);
    }

    ($rc, $msg) = reint_post_secure($rec);
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }

    # update the mtime for the parent directory to the correct 
    #    version mtime
    ($rc, $msg) = setattr($mtpt . $rec->{path}, "mtime", 
			  $rec->{new_pdir_mtime});
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }

    0;
}

sub reint_symlink () {
    my $rec = shift;
    my $mtpt = shift;
    my $oldname = $rec->{target};
    my $newname = $mtpt . $rec->{path}; 
    my $hmsg = "reint_symlink: ";
    my $filestat = lstat($newname);
    if (defined $filestat) {
	my $msg = "";
	if (S_ISLNK($filestat->mode)) {
	    if (readlink($newname) eq $oldname) {
		$msg = sprintf("failed for %s - already exists.\n", 
			       $newname);
		return(-2, $hmsg . $msg);
	    } 
	    
	} else {
	    $msg = sprintf("failed for %s - exists but not as a symlink.\n", 
			   $newname);
	    return(-1, $hmsg . $msg);
	}
    }

    DEBUG("ReintFS::" . $hmsg . "$oldname $newname\n");

    my ($rc, $msg) = reint_pre_secure($rec);
    if ($rc != 0) {
	reint_post_secure($rec);
	return($rc, $hmsg . $msg);
    }
    $rc = symlink($oldname, $newname);
    if ($rc < 1) {
	$msg = "failed symlink($oldname, $newname): $^E\n";
	reint_post_secure($rec);
	return(-1, $hmsg . $msg);
    }

    ($rc, $msg) = reint_post_secure($rec);
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }

    # update the mtime for the parent directory to the correct 
    #    version mtime
    my $parentdir = $newname;
    $parentdir =~ s/(.+)\/.+$/$1/;
    ($rc, $msg) = setattr($parentdir, "mtime", $rec->{new_pdir_mtime});
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }

    0;
}


sub reint_mkdir {
    my $rec = shift;
    my $mtpt = shift;
    my $path = $mtpt . $rec->{path}; 
    my $hmsg = "reint_mkdir: ";
    my $filestat = lstat($path);
    if (defined $filestat) {
	my $msg = "";
	if (S_ISDIR($filestat->mode)) { 
	    $msg = sprintf("failed for %s - already exists.\n", $path);
	    return(-2, $hmsg . $msg);
	} 
	$msg = sprintf("failed for %s - exists but not a dir.\n", 
		       $path);
	return(-1, $hmsg . $msg);
    }

    DEBUG("ReintFS::" . $hmsg . "$path %o \n",$rec->{mode});

    my ($rc, $msg) = reint_pre_secure($rec);
    if ($rc != 0) {
	reint_post_secure($rec);
	return($rc, $hmsg . $msg);
    }
    $rc = mkdir($path, $rec->{mode});
    if ($rc < 1) {
	$msg = sprintf("mkdir(%s, %o) failed\n", $path, $rec->{mode});
	reint_post_secure($rec);
	return(-1, $hmsg . $msg);
    }
    ($rc, $msg) = reint_post_secure($rec);
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }

    # correct mode/mtime for the object
    my $mode = $rec->{mode};
    DEBUG ($hmsg . "pre-and mode is currently: %o\n",$mode);
    $mode = $mode & 0xfff;
    ($rc, $msg) = setattr($path, "mtime mode", $rec->{new_obj_mtime}, $mode);
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }

    # update the mtime for the parent directory to the correct 
    #    version mtime
    my $parentdir = $path;
    $parentdir =~ s/(.+)\/.+$/$1/;
    ($rc, $msg) = setattr($parentdir, "mtime", $rec->{new_pdir_mtime});
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }

    0;
}

sub reint_rmdir () {
    my $rec = shift;
    my $mtpt = shift;
    my $path = $mtpt . $rec->{path} . '/' . $rec->{target}; 
    my $hmsg = "reint_rmdir: ";
    my $filestat = lstat($path);
    if (not defined $filestat) {
	my $msg = sprintf("failed for %s - already rm'd.\n", $path);
	return(-2, $hmsg . $msg);
    }

    DEBUG("ReintFS::" . $hmsg . "$path \n");

    my ($rc, $msg) = reint_pre_secure($rec);
    if ($rc != 0) {
	reint_post_secure($rec);
	return($rc, $hmsg . $msg);
    }
    $rc = rmdir($path);
    if ($rc < 1) {
	$msg = sprintf("rmdir(%s) failed\n", $path);
	reint_post_secure($rec);
	return(-1, $hmsg . $msg);
    }
    ($rc, $msg) = reint_post_secure($rec);
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }

    # update the mtime for the parent directory to the correct 
    #    version mtime
    ($rc, $msg) = setattr($mtpt . $rec->{path}, "mtime", 
			  $rec->{new_pdir_mtime});
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }

    0;
}

sub reint_mknod () {
    my $rec = shift;
    my $mtpt = shift;
    my $path = $mtpt . $rec->{path}; 
    my $hmsg = "reint_mknod: ";
    my $type;
    my $filestat = lstat($path);
    if (defined $filestat) {
	my $msg = "";
	if (S_ISCHR($filestat->mode) || S_ISBLK($filestat->mode) ||
	    S_ISFIFO($filestat->mode))  {
	    $msg = sprintf("failed for %s - already exists.\n", $path);
	    return(-2, $hmsg . $msg);
	} 
	$msg = sprintf("failed for %s - exists but not a special file.\n", 
		       $path);
	return(-1, $hmsg . $msg);
    }

    my $mode = $rec->{mode};

    # need to figure out type - that must come from mode
    if (S_ISCHR($mode)) {
	$type = 'c';
    } elsif (S_ISBLK($mode)){
	$type = 'b';
    } elsif (S_ISFIFO($mode)) {
	$type = 'p';
    } else {
	return (-1, $hmsg .  "invalid special device type\n");
    }
    DEBUG ($hmsg . "pre-and mode is currently: %o\n",$mode);
    $mode = $mode & 0xfff;
    
    DEBUG ("ReintFS::" . $hmsg . "$path $type %o %d %d\n",
	   $mode, $rec->{major}, $rec->{minor});
    my ($rc, $msg) = reint_pre_secure($rec);
    if ($rc != 0) {
	reint_post_secure($rec);
	return($rc, $hmsg . $msg);
    }
    ($rc, $msg) = mknod($path, "$mode", $type, $rec->{major}, $rec->{minor});
    if ($rc != 0) {
	reint_post_secure($rec);
	return($rc, $hmsg . $msg);
    }

    ($rc, $msg) = reint_post_secure($rec);
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }

    # update the mtime for the object to the correct time
    ($rc, $msg) = setattr($path, "mtime\n", $rec->{new_obj_mtime});
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }

    # update the mtime for the parent directory to the correct 
    #    version mtime
    my $parentdir = $path;
    $parentdir =~ s/(.+)\/.+$/$1/;
    ($rc, $msg) = setattr($parentdir, "mtime", $rec->{new_pdir_mtime});
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }

    0;
}

sub reint_rename () {
    my $rec = shift;
    my $mtpt = shift;
    my $oldname = $mtpt . $rec->{path}; 
    my $newname = $mtpt . $rec->{target};
    my $hmsg = "reint_rename: ";
    my $oldfilestat = lstat($oldname);
    my $newfilestat = lstat($newname);
    if ((not defined $oldfilestat) and (defined $newfilestat)) {
	my $msg = sprintf("failed for %s - already exists.\n", 
		       $newname);
	return(-2, $hmsg . $msg);
    }


    DEBUG("ReintFS::" . $hmsg . "$oldname $newname\n");

    my ($rc, $msg) = reint_pre_secure($rec);
    if ($rc != 0) {
	reint_post_secure($rec);
	return($rc, $hmsg . $msg);
    }
    $rc = rename($oldname, $newname);
    if ($rc < 1) {
	$msg = sprintf("rename(%s, %s) failed\n", $oldname, $newname);
	reint_post_secure($rec);
	return(-1, $hmsg . $msg);
    }
    ($rc, $msg) = reint_post_secure($rec);
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }

    # set the mtime for the parent directories to the correct 
    #    version mtime
    #  this is the target's parent directory
    my $parentdir = $newname;
    $parentdir =~ s/(.+)\/.+$/$1/;
    ($rc, $msg) = setattr($parentdir, "mtime", $rec->{new_tgt_mtime});
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }
    #  this is really the object's old parent directory
    my $parentdir = $oldname;
    $parentdir =~ s/(.+)\/.+$/$1/;
    ($rc, $msg) = setattr($parentdir, "mtime", $rec->{new_obj_mtime});
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }

    0;
}

sub reint_close () {
    my $rec = shift;
    my $localmtpt = shift;
    my $backfetch_func = shift;
    my $syncdata = shift;
    my $localfile = $localmtpt . $rec->{path};
    my $hmsg = "reint_close: ";
    my ($rc, $msg);

    # fetch file by inode and gen#
    return (-1, $hmsg . "attempt to fetch file $rec->{path} with no inode #\n")
       if not defined $rec->{ino};
    return (-1, $hmsg . "attempt to fetch file $rec->{path} with no generation #\n")
       if not defined $rec->{generation};
    my $remotefile = sprintf("/...ino:%lu:%u", $rec->{ino}, $rec->{generation});

    DEBUG("ReintFS::" . $hmsg . "$rec->{path} $localfile\n");

    # save data for setting attributes later
    my $filestat = lstat($localfile);
    if (not defined $filestat) {
	return (-1, $hmsg . "attempt to close non-existant file $localfile\n");
    }
    my $uid = $filestat->uid;
    my $gid = $filestat->gid;
    my $mode = $filestat->mode;
    my $parentdir = $localfile;
    $parentdir =~ s/(.+)\/.+$/$1/;
    my $pdir_stat = stat($parentdir);
    my $pdir_mtime = $pdir_stat->mtime;


    # when do you do this?
    #reint_pre_secure($rec);
    #reint_post_secure($rec);

    # request backfetch of remotefile
#    if (! &{$backfetch_func}($syncdata, $remotefile, $localfile) ) {
    ($rc, $msg) = &{$backfetch_func}($syncdata, $remotefile, $localfile);
    if ($rc < 0 ) {
	return (-1, "backfetch_func failed for $remotefile to $localfile\n");
    } 

    # now set attributes, etc...
    ($rc, $msg) = setattr($localfile, "chown mtime mode\n", 
			  $uid, $gid, $rec->{new_obj_mtime}, $mode);
    if ($rc != 0) {
	return($rc, $hmsg . $msg);
    }

    # update the mtime for the parent directory to the correct 
    #    version mtime
    ($rc, $msg) = setattr($parentdir, "mtime", $pdir_mtime);
    if ($rc != 0) {
       return($rc, $hmsg . $msg);
    }

    0;
}

sub setattr () {
    my $path = shift;
    my $type = shift;
    my ($rc, $msg);
    my $hmsg = "setattr: ";
    my $filestat = lstat($path);
    if (not defined $filestat) {
	$msg = sprintf("%s on %s failed because file does not exist\n", 
		       $type, $path);
	return(-1, $hmsg . $msg);
    }
    my ($uid, $gid);

    DEBUG($hmsg . "$path $type\n");
    if ($type =~ m/chown/) {
	my $uid = shift;
	my $gid = shift;
        # if the file is really a symlink we need to use 
        #   lchown, which unfortunately doesn't exist in 
        #   the standard perl distribution.
        # in linux (after 2.1.80) the chown command does a lchown
        #   underneath unless a dereference is explicitly 
        #   requested.  
	if (S_ISLNK($filestat->mode)) {
	    `chown $uid.$gid $path`;
	    if ($? != 0) {  # $? = 256 is a permission problem.
		$rc = 0;
	    } else {
		$rc = 1;
	    }
	} else {
	    $rc = chown($uid, $gid, $path);
	}
	if ($rc < 1) {
	    $msg = sprintf("chown( %d, %d, %s) failed\n", $uid, $gid, $path);
	    return(-1, $hmsg . $msg);
	}
    } 
    if ($type =~ m/mtime/) {
	my $mtime = shift;
	DEBUG($hmsg . "orig mtime: %d\tnew: %d\n", $filestat->mtime, $mtime);
	my $atime = $filestat->atime;
        # if the file is a symlink we can't change the mtime
        #   but let's continue on
	if (S_ISLNK($filestat->mode)) {
	    $msg = sprintf("cannot change utimes on symlink (%d, %d, %s)\n", 
			   $atime, $mtime, $path);
	    return(0, $hmsg . $msg);
	} 
	$rc = utime($atime, $mtime, $path);
	if ($rc < 1) {
	    $msg = sprintf("utime(%d, %d, %s) failed: $^E\n", $atime,
			  $mtime, $path);
	    return(-1, $hmsg . $msg);
	}
    }
    if ($type =~ m/mode/) {
	my $mode = shift;
	DEBUG($hmsg . "orig mode: %o\tnew: %o\n", $filestat->mode, $mode);
        # if the file is a symlink we can't change the mode
        #   but let's continue on
	if (S_ISLNK($filestat->mode)) {
	    $msg = sprintf("cannot change mode on symlink (%o, %s)\n", 
			   $mode, $path);
	    return(0, $hmsg . $msg);
	} 
	$rc = chmod($mode, $path);
	if ($rc < 1) {
	    $msg = sprintf("chmod( %o, %s) failed\n", $mode, $path);
	    return(-1, $hmsg . $msg);
	}
    }

    0;
}

sub mknod () {
 my ($path, $mode, $type, $major, $minor) = @_; 
 my $hmsg = "mknod: ";
 my $str;

 if ($type ne 'p') {
     DEBUG($hmsg . "nonpipe: mknod -m %o $path $type %d %d\n", 
	   $mode, $major, $minor);
     $str .= sprintf ("mknod -m %o $path $type %d %d", $mode, $major, $minor);
     `$str`;
 } else {
     DEBUG($hmsg . "pipe: mknod -m %o $path $type\n", $mode);
     $str .= sprintf ("mknod -m %o $path $type", $mode);
     `$str`;
 }

 0;
}

1;

__END__


=head1 NAME

ReintFS - Kernel Modification Log (KML) Reintegration for a local Filesystem

=head1 SYNOPSIS

  use ReintFS;

  # reintegrate a record for a specific operation

  # all but the reintegration of a close record have the same 
  # call and parameters:
  # $rec is a pointer to a hash containing an unpacked KML record, 
  #    different fields of which are used for different operations
  # $mtpt is the root mountpoint for the filesystem that will 
  #    receive the reintegration.
  ReintFS::reint_op($rec, $mtpt);

  # the current operations are:
  ReintFS::reint_setattr
  ReintFS::reint_create 
  ReintFS::reint_link 
  ReintFS::reint_unlink 
  ReintFS::reint_symlink
  ReintFS::reint_mkdir
  ReintFS::reint_rmdir
  ReintFS::reint_mknod
  ReintFS::reint_rename 
  ReintFS::reint_close 

  # the close record reintegration has extra arguments which 
  #    are required in order to do the backfetch of the file
  #    that was closed.
  # $rec is a pointer to a hash containing an unpacked KML record, 
  # $localmtpt is the root mountpoint for the filesystem that will 
  #    receive the reintegration.
  # $backfetch_func is a pointer to the function to use for the 
  #    backfetch operation.
  # $syncdata is a ref to a hash containing all the data to 
  #    pass on to the backfetch function.
  ReintFS::reint_close($rec, $localmtpt, $backfetch_func, $syncdata);

=head1 DESCRIPTION

=item 

  ReintFS reintegrations KML records into a local file system. 

  ReintFS is subordinate to Reint - only Reint should call ReintFS
  methods.

=head1 Uses

  use with the InterSync package.
  use for the testing framework.
  

=head1 Bugs

  * Currently doing the backfetch by *name* and not by inode. Under 
    certain circumstances (involving rename reintegrations) the 
    correct file is not fetched from the server, leaving a zero
    size file. This should be resolved with the new iopen (open by inode).

=cut

