# performs reintegration of KML entries

# August 1, 2001: Will now do Rsync if appropriate fset option is set

package Lento::KML::ReintLento;

use strict;
use Carp;
use Lento::Debuggable;
use File::stat;
use File::Copy;
use POE;
use POSIX qw(:unistd_h :fcntl_h :errno_h);
use Lento::KML::Rec;
use Lento::Psdev;
use Lento::Bulk;
use Lento::Bulk::Desc;
use Lento::Bulk::Rsync;
use SetFS;

@Lento::KML::ReintLento::ISA = qw(Lento::Debuggable);
$Lento::KML::ReintLento::level = 1;

sub ATTR_MODE      {1;}
sub ATTR_UID       {2;}
sub ATTR_GID       {4;}
sub ATTR_SIZE      {8;}
sub ATTR_ATIME     {16;}
sub ATTR_MTIME     {32;}
sub ATTR_CTIME     {64;}
sub ATTR_ATIME_SET {128;}
sub ATTR_MTIME_SET {256;}

my %opcode_reint = (1 => \&reint_create,
                    2 => \&reint_mkdir,
                    3 => \&reint_unlink,
                    4 => \&reint_rmdir,
                    5 => \&reint_close,
                    6 => \&reint_symlink,
                    7 => \&reint_rename,
                    8 => \&reint_setattr,
                    9 => \&reint_link,
                    10 => \&reint_open,
                    11 => \&reint_mknod);

sub reint_pre_secure {
    
    return if $::nofsuid;
    
    my $rec = shift;
    
    if ($rec->{ngroups}) {
        # Record has a groups array...just use it
        $) = $rec->{fsgid}." ".join(" ",@{$rec->{groups}});
        DEBUG "groups $) should be ".join(" ",@{$rec->{groups}})."\n"
    } else {
        
        # XXX not like this - we need a file set flag or so
        # fall back to default reintfs like behaviour 
        my @groups = ($rec->{fsgid});
        setgrent;
        my ($name, $passwd, $gid, $members);
        my $login= (getpwuid($rec->{fsuid}))[0];
        while (($name, $passwd, $gid, $members) = getgrent) {
            push(@groups, $gid) if ($members =~ m/$login/);
        }
        
        # setgroups, in perl setting $) sets the effective group array
        $) = join (" ",@groups);
    }
    
    my $fsgid;
    # sefstuid to correct owner/group
    confess "can not setfsuid\n" if (setfsuid($rec->{fsuid}) != $<);
    confess "can not setfsgid $fsgid != $)\n" if (($fsgid=setfsgid($rec->{fsgid})) != $));
    
    DEBUG "groups $) should be ".join(" ",@{$rec->{groups}})."\n"
}

sub reint_post_secure {
    
    return if $::nofsuid;
    
    my $rec = shift;
    # restore root owner/group
    confess "cannot restore fsuid\n" if (setfsuid($<) != $rec->{fsuid});
    confess "cannot restore fsgid\n" if (setfsgid($<) != $rec->{fsgid});
    
    # restore group array to the real group array by
    # setting the effective group array $) to the real group array $(
    $) = $(;
}

sub setattr_time {
    my $psdev = shift;
    my $rec = shift;
    my $path = shift;
    my $latime = shift;
    my $lmtime = shift;
    my $lctime = shift;
    my $attr_flags = shift;
    my $slot_offset = shift;
    my $flags = shift;
    
    # right now no atime, but lento_setattr needs it
    my $local_mode = 0;
    my $local_valid = 0;
    my $local_flags = 0;
    
    die unless defined $flags; 
    die unless defined $attr_flags; 
    die unless defined $latime;
    die unless defined $lmtime;
    die unless defined $lctime;
    
    #   $local_valid = $rec->{valid} unless not defined $rec->{valid};
    $local_valid = $attr_flags;
    $local_mode = $rec->{mode} unless not defined $rec->{mode};
    $local_flags = $flags unless not defined $flags;
    $rec->{uid} = 0 unless defined $rec->{uid}; 
    $rec->{gid} = 0 unless defined $rec->{gid};
    $rec->{size} = 0 unless defined $rec->{size};
    
    $psdev->lento_setattr($path, $local_valid, $local_mode,
                          $rec->{uid}, $rec->{gid}, $rec->{size},
                          $latime, $lmtime, $lctime,
                          $attr_flags, $slot_offset, $rec->{recno},
                          $rec->{kml_offset} + $rec->{len}, $local_flags);
    return 0;
}

sub reint_setattr {
    my $rec = shift;
    my $mtpt = shift;
    my $psdev = shift;
    my $slot_offset = shift;
    my $flags = shift;
    # right now no atime, but lento_setattr needs it
    my $atime = 0;
    my $path = $mtpt . $rec->{path};
    
    reint_pre_secure($rec);
    my $localmode = 0;
    my $local_mtime = $rec->{mtime};
    my $local_ctime = $rec->{ctime};
    $local_mtime = $rec->{old_obj_mtime} unless $rec->{valid} & ATTR_MTIME;
    $local_ctime = $rec->{old_obj_ctime} unless $rec->{valid} & ATTR_CTIME;
    
    $psdev->lento_setattr($path, $rec->{valid} | ATTR_MTIME | ATTR_CTIME, 
                          $rec->{mode}, $rec->{uid}, $rec->{gid}, $rec->{size},
                          $atime, $local_mtime, $local_ctime,
                          $rec->{flags} | ATTR_CTIME,
                          $slot_offset, $rec->{recno},
                          $rec->{kml_offset} + $rec->{len}, $flags);
    # we should check if $rec->{valid} sets the flag of time 
    #   We don't set parentdir's time attrib here.
    #   my $parentdir = $path;
    #   $parentdir =~ s/(.+)\/.+$/$1/;
    #   setattr_time ($psdev, $rec, $parentdir, 
    #            # 0, $rec->{mtime}, $rec->{ctime},
    #            0, $rec->{ctime}, $rec->{ctime},
    #                ATTR_MTIME | ATTR_CTIME, $slot_offset, $flags);
    
    reint_post_secure($rec);
    return 0;
}

sub reint_create {
    my $rec = shift;
    my $mtpt = shift;
    my $psdev = shift;
    my $slot_offset = shift;
    my $flags = shift;
    my $path = $mtpt . $rec->{path}; 
    
    reint_pre_secure($rec);
    $psdev->lento_create($path, $rec->{mode},
                         $slot_offset, $rec->{recno}, 
                         $rec->{kml_offset} + $rec->{len}, $flags, $rec->{new_obj_ctime});
#     setattr_time ($psdev, $rec, $path, 
#                   0, $rec->{new_obj_mtime}, $rec->{new_obj_ctime},
#                   ATTR_MTIME | ATTR_CTIME, $slot_offset, $flags);
#     my $parentdir = $path;
#     $parentdir =~ s,(.+)[/][^/]+$,$1,;
#     setattr_time ($psdev, $rec, $parentdir, 
#                   0, $rec->{new_pdir_mtime}, $rec->{new_pdir_ctime},
#                   ATTR_MTIME | ATTR_CTIME, $slot_offset, $flags);
    reint_post_secure($rec);
    
    return 0;
}

sub reint_link () {
    my $rec = shift;
    my $mtpt = shift;
    my $psdev = shift;
    my $slot_offset = shift;
    my $flags = shift;
    my $oldname = $mtpt . $rec->{path}; 
    my $newname = $mtpt . $rec->{target};
    
    reint_pre_secure($rec);
    $psdev->lento_link($oldname, $newname, 
                       $slot_offset, $rec->{recno}, 
                       $rec->{kml_offset} + $rec->{len}, $flags, $rec->{new_pdir_ctime});
#     setattr_time ($psdev, $rec, $newname, 
#                   0, $rec->{new_obj_mtime}, $rec->{new_obj_ctime}, 
#                   ATTR_CTIME, $slot_offset, $flags);
#     my $parentdir = $newname;
#     $parentdir =~ s,(.+)[/][^/]+$,$1,;
#     setattr_time ($psdev, $rec, $parentdir, 
#                   0, $rec->{new_pdir_mtime}, $rec->{new_pdir_ctime}, 
#                   ATTR_MTIME | ATTR_CTIME, $slot_offset, $flags);
    reint_post_secure($rec);
    
    return 0;
}

sub reint_unlink () {
    my $rec = shift;
    my $mtpt = shift;
    my $psdev = shift;
    my $slot_offset = shift;
    my $flags = shift;
    my $path = $mtpt . $rec->{path} . '/' . $rec->{target}; 
    
    reint_pre_secure($rec);
    $psdev->lento_unlink($path, 
                         $slot_offset, $rec->{recno}, 
                         $rec->{kml_offset} + $rec->{len}, $flags, $rec->{new_pdir_ctime});
    
#    my $parentdir = $mtpt . $rec->{path};
#    setattr_time ($psdev, $rec, $parentdir,
#                  0, $rec->{new_pdir_mtime}, $rec->{new_pdir_ctime}, 
#                  ATTR_MTIME | ATTR_CTIME, $slot_offset, $flags);
    
    reint_post_secure($rec);
    return 0;
}

sub reint_symlink () {
    my $rec = shift;
    my $mtpt = shift;
    my $psdev = shift;
    my $slot_offset = shift;
    my $flags = shift;
    my $oldname = $rec->{target};
    my $newname = $mtpt . $rec->{path}; 
    
    reint_pre_secure($rec);
    $psdev->lento_symlink($oldname, $newname,
                          $slot_offset, $rec->{recno}, 
                          $rec->{kml_offset} + $rec->{len}, $flags, $rec->{new_pdir_ctime});
    
#     setattr_time ($psdev, $rec, $newname,
#                   0, $rec->{new_obj_mtime}, $rec->{new_obj_ctime},
#                   ATTR_MTIME | ATTR_CTIME, $slot_offset, $flags);
#     my $parentdir = $mtpt . $rec->{path};
#     setattr_time ($psdev, $rec, $parentdir,
#                   0, $rec->{new_pdir_mtime}, $rec->{new_pdir_ctime},
#                   ATTR_MTIME | ATTR_CTIME, $slot_offset, $flags);
    
    reint_post_secure($rec);
    return 0;
}

sub reint_mkdir {
    my $rec = shift;
    my $mtpt = shift;
    my $psdev = shift;
    my $slot_offset = shift;
    my $flags = shift;
    my $path = $mtpt . $rec->{path}; 
    
    reint_pre_secure($rec);
    DEBUG "XXXXX mkdir mode %o, ctime %d\n", $rec->{mode}, $rec->{new_obj_ctime};
    $psdev->lento_mkdir($path, $rec->{mode}, 
                        $slot_offset, $rec->{recno}, 
                        $rec->{kml_offset} + $rec->{len}, $flags, $rec->{new_obj_ctime});
    
#    setattr_time ($psdev, $rec, $path, 
#                  0, $rec->{new_obj_mtime}, $rec->{new_obj_ctime}, 
#                  ATTR_MTIME | ATTR_CTIME, $slot_offset, $flags, $rec->{new_obj_ctime});
#    my $parentdir = $path;
#    $parentdir =~ s,(.+)[/][^/]+$,$1,;
#    setattr_time ($psdev, $rec, $parentdir, 
#                  0, $rec->{new_pdir_mtime}, $rec->{new_pdir_ctime}, 
#                  ATTR_MTIME | ATTR_CTIME, $slot_offset, $flags);
    
    reint_post_secure($rec);
    
    return 0;
}


sub reint_rmdir () {
    my $rec = shift;
    my $mtpt = shift;
    my $psdev = shift;
    my $slot_offset = shift;
    my $flags = shift;
    my $path = $mtpt . $rec->{path} . '/' . $rec->{target}; 
    
    reint_pre_secure($rec);
    $psdev->lento_rmdir($path, 
                        $slot_offset, $rec->{recno}, 
                        $rec->{kml_offset} + $rec->{len}, $flags, $rec->{new_pdir_ctime});
    
#    my $parentdir = $mtpt . $rec->{path};
#    setattr_time ($psdev, $rec, $parentdir, 
#                  0, $rec->{new_pdir_mtime}, $rec->{new_pdir_ctime}, 
#                  ATTR_MTIME | ATTR_CTIME, $slot_offset, $flags);
    
    reint_post_secure($rec);
    return 0;
}

sub reint_mknod () {
    my $rec = shift;
    my $mtpt = shift;
    my $psdev = shift;
    my $slot_offset = shift;
    my $flags = shift;
    my $path = $mtpt . $rec->{path}; 
    
    reint_pre_secure($rec);
    $psdev->lento_mknod($path, $rec->{mode}, $rec->{major}, $rec->{minor},
                        $slot_offset, $rec->{recno}, 
                        $rec->{kml_offset} + $rec->{len}, $flags, $rec->{new_obj_ctime});
    
#    setattr_time ($psdev, $rec, $path, 
#                  0, $rec->{new_obj_mtime}, $rec->{new_obj_ctime},
#                  ATTR_MTIME | ATTR_CTIME, $slot_offset, $flags);
#    my $parentdir = $path;
#    $parentdir =~ s,(.+)[/][^/]+$,$1,;
#    setattr_time ($psdev, $rec, $parentdir, 
#                  0, $rec->{new_pdir_mtime}, $rec->{new_pdir_ctime},
#                  ATTR_MTIME | ATTR_CTIME, $slot_offset, $flags);
    
    reint_post_secure($rec);
    return 0;
}

sub reint_rename () {
    my $rec = shift;
    my $mtpt = shift;
    my $psdev = shift;
    my $slot_offset = shift;
    my $flags = shift;
    my $oldname = $mtpt . $rec->{path}; 
    my $newname = $mtpt . $rec->{target};
    
    reint_pre_secure($rec);
    $psdev->lento_rename($oldname, $newname,
                         $slot_offset, $rec->{recno}, 
                         $rec->{kml_offset} + $rec->{len}, $flags, $rec->{new_obj_ctime});
#     my $parentdir = $newname;
#     $parentdir =~ s,(.+)[/][^/]+$,$1,;
#     setattr_time ($psdev, $rec, $parentdir,
#                   0, $rec->{new_obj_mtime}, $rec->{new_obj_ctime},
#                   ATTR_MTIME | ATTR_CTIME, $slot_offset, $flags);
#     $parentdir = $oldname;
#     $parentdir =~ s,(.+)[/][^/]+$,$1,;
#     setattr_time ($psdev, $rec, $parentdir,
#                   0, $rec->{new_tgt_mtime}, $rec->{new_tgt_ctime},
#                   ATTR_MTIME | ATTR_CTIME, $slot_offset, $flags);
#     setattr_time ($psdev, $rec, $newname,
#                   0, $rec->{old_obj_mtime}, $rec->{old_obj_ctime},
#                   ATTR_CTIME, $slot_offset, $flags);
    reint_post_secure($rec);
    return 0;
}

sub reint_close_callback () { 
    my $closeinfo  = $poe_kernel->get_active_session()->get_heap()->{closeinfo};
    
    DEBUG("close for non-existing fd\n") 
        if !defined($closeinfo) or !defined($closeinfo->{fd});

    # This might be received after a "got_error" in which case we should simply
    # return
    return unless defined($closeinfo) and defined($closeinfo->{fd});
    
    DEBUG("Closing: fd $closeinfo->{fd}: recno $closeinfo->{rec}->{recno},
    offset: $closeinfo->{rec}->{kml_offset} + $closeinfo->{rec}->{len}, flags
    $closeinfo->{flags}); \n");

    reint_pre_secure($closeinfo->{rec});
    
    # Close the fd...
    $closeinfo->{psdev}->lento_close
        ($closeinfo->{fd}, $closeinfo->{slot_offset},
         $closeinfo->{rec}->{recno}, 
         $closeinfo->{rec}->{kml_offset} + $closeinfo->{rec}->{len}, 
         $closeinfo->{flags}) 
            if defined $closeinfo and defined $closeinfo->{fd};
    
    # ... and journal the operation
    #$closeinfo->{psdev}->cancel_lml($closeinfo,$closeinfo->{flags});
    
    setattr_time ($closeinfo->{psdev}, $closeinfo->{rec}, $closeinfo->{path}, 
                  0, $closeinfo->{rec}->{new_obj_mtime}, 
                  $closeinfo->{rec}->{new_obj_ctime}, ATTR_MTIME | ATTR_CTIME, 
                  $closeinfo->{slot_offset}, $closeinfo->{flags});
    # parent only change (atime) , so we ignore it !!!
    #                     Arthur Ma, 2001-03-19
    # my $parentdir = $closeinfo->{path};
    # $parentdir =~ s,(.+)[/][^/]+$,$1,;
    # setattr_time ($closeinfo->{psdev}, $closeinfo->{rec}, $parentdir, 
    #               0, $closeinfo->{rec}->{new_obj_mtime}, 
    #              $closeinfo->{rec}->{new_obj_ctime}, ATTR_MTIME | ATTR_CTIME, 
    #              $closeinfo->{slot_offset}, $closeinfo->{flags});
    reint_post_secure($closeinfo->{rec});
    
    delete $closeinfo->{rec};
    delete $poe_kernel->get_active_session()->get_heap()->{closeinfo};
}


sub reint_localclose {
    my $rec = shift;
    my $src_mtpt = shift;
    my $replay_mtpt = shift;
    my $psdev = shift;
    my $slot_offset = shift;
    my $flags = shift;
    
    DEBUG("record to reintegrate in reint_fakeclose:\n");
    DEBUG($rec->printme('brief'));
    
    # first open the file based on inode and generation number
    # this will really be done on the client...    
    my $remotefile = $src_mtpt . $rec->{path}; 
    my $localfile = $replay_mtpt . $rec->{path}; 
    my $fd = $psdev->lento_iopen($remotefile, $rec->{ino}, $rec->{generation},
                                 $rec->{flags});
    
    # if the fd is not valid, then it was because the inode or generation #s
    #    don't match, so throw away the kml record and return
    if ($fd < 0) {
        LOG("lento_iopen failed, fd: $fd\n");
        return 1;
    }
    
    # now if you have a valid filehandle, read it and write it to the 
    # replay location
    my $remotefh = IO::Handle->new_from_fd($fd,'r') or die "NO handle";
    
    # could copy the file or alternatively, open for append, truncate,
    # and then read each record and write it.
    # copy($remotefh,$localfh);
    # create replay location.  Open for append, seek to the beginning,
    # and truncate in case something else exists there.
    my $localfh = new IO::File;
    $localfh->open($localfile, "a") 
        or confess "can't open $localfile for writing\n";
    sysseek($localfh, 0, SEEK_SET);
    $localfh->truncate(0);
    while (<$remotefh>) {
        $localfh->print($_);
    }
    # copy($remotefh,$localfh);
    
    my $lfd = fileno($localfh);
    # close the file handles
    #    $psdev->lento_close($lfd, 
    #                        $slot_offset, $rec->{recno}, 
    #                        $rec->{kml_offset} + $rec->{len}, $flags);
    
    $remotefh->close();
    $localfh->close();
    
    reint_pre_secure($rec);
    reint_post_secure($rec);
    return 0;
}


sub reint_close {            # must do a backfetch here
    my $rec = shift;
    my $mtpt = shift;
    my $psdev = shift;
    my $session = $poe_kernel->get_active_session();
    my $slot_offset = shift;
    my $flags = shift;
    my $fsetname = shift;
    my $replicator = shift;
    
    my $fset=$replicator->{fileset};
    
    confess "No fileset corresponding to $fsetname" unless defined $fset;
    
    # close needs to do:
    #  get inode, iversion of file to backfetch from client
    #  send bulk-transfer request to client
    #  get descriptor from client
    #  setup to receive
    #  receive data
    #  write the data to the fd
    
    # $path is the full path name of the location of the file
    # that will be written from the backfetch of the remote file
    my $path = $mtpt . $rec->{path}; 
    
    my $heap = $session->get_heap();
    confess "no heap!" unless defined $heap;
    
    my $closeinfo = $heap->{closeinfo} = 
    {   path        => $mtpt.$rec->{path}, 
        psdev       => $psdev, 
        slot_offset => $slot_offset,
        flags       => $flags,
        rec         => $rec };

    my $fetchtype = $fset->{fetchtype} || "Rsync";
    DEBUG "fetching $path with fetchtype $fetchtype\n";
    
    # set up sink desc
    my $desc = "Lento::Bulk::$fetchtype"->new
        ( { sink_type => $Lento::Bulk::SINK_FILE,
            sink_object => $path, 
            role => $Lento::Bulk::SINK, 
            direction => $Lento::Bulk::SERVER_TO_CLIENT, 
            connection => $heap->{connection},
            callback_event => 'got_backfetch_file',
            callback_session => $session,
            close_callback => 'Lento::KML::ReintLento::reint_close_callback',
            debug => 1 } );
    # SMC source/sink act?
    
    DEBUG $desc->printme();
    
    my ($rc, $bulk_session) = $desc->setup();
    if ($rc) {
        $rc=EBADF unless $rc;

        # Check for deleted/unchanged files
        if (($rc == ENOENT) or ($rc == EEXIST)) {
            DEBUG "Skipping backfetch for file $path\n";
            # Resync offsets to accomodate missing record
            $replicator->sync_last_rcvd( $rec->{recno}, 
                                         $rec->{kml_offset} + $rec->{len});
            # fake the backfetch completion
            # $poe_kernel->post($session,'got_backfetch_file', 0);
            # If the above line is uncommented change return value to 1
            # That will cause the reint loop to break and get restarted from
            # got_backfetch_file -SHP

            # XXX: we should change attributes of the file/parent directory
            # here. However, parent appears to only change atime which we 
            # ignore. See comments in reint_close_callback -SHP
            return 0;
        }

        if ($rc == ENOTCONN) {
            DEBUG "Disconnected while trying to create sink desc\n";
            # Now break the reintegration loop since we 
            # are unable to continue with the backfetch.
            $poe_kernel->post($session,'got_backfetch_file', ENOTCONN);
            return 1;
        }

        $!=$rc;
        confess "Error setting up bulk transfer ($!)\n";
    }

    # XXX: Should we test for return values from this function?
    $desc->start();
    
    my $fetchinfo = 
    { 'path'       => $rec->{path},
      'ino'        => $rec->{ino},
      'generation' => $rec->{generation}};
    
    
    # Handle will not be defined if we are using rsync without a basis
    $closeinfo->{fd} = $desc->{handle}->fileno()
        if defined $desc->{handle};

    # XXX: we assume that the recorded LML will be at the last offset
    # This may not be true always, so modify the BACKFETCH_LML ioctl to
    # to return the lml_offset of the lml record created.
    $closeinfo->{lml_offset}=$fset->{lml}->last_offset();
    
    # Tell presto that a backfetch is about to start...
    #$::psdev->record_backfetch_lml($closeinfo);
    
    # send event to client to send data
    # what we know about the remote file:
    #    fullfilename (w/o cache mountpoint), inode, generation#
    DEBUG("SENDING backfetch request packet for %s, inode: %i, genno: %i\n",
          $rec->{path}, $rec->{ino}, $rec->{generation});
    $heap->{connection}->sendpacket
        ('REQ', [ 'BackFetch' , $fsetname, $fetchinfo, $fetchtype]);
    return 1;
}

sub reint_setextattr {
    my $rec = shift;
    my $mtpt = shift;
    my $psdev = shift;
    my $slot_offset = shift;
    my $flags = shift;
    # right now no atime, but lento_setattr needs it
    my $atime = 0;
    my $path = $mtpt . $rec->{path};
    
    reint_pre_secure($rec);
    
    $psdev->lento_setextattr($path, $rec->{attr_name}, $rec->{attr_value}, 
                             $rec->{attr_flags}, $rec->{mode},
                             $slot_offset, $rec->{recno},
                             $rec->{kml_offset} + $rec->{len}, $flags,
                             $rec->{new_obj_ctime});
    reint_post_secure($rec);
    return 0;
}

sub reint_delextattr {
    my $rec = shift;
    my $mtpt = shift;
    my $psdev = shift;
    my $slot_offset = shift;
    my $flags = shift;
    # right now no atime, but lento_setattr needs it
    my $atime = 0;
    my $path = $mtpt . $rec->{path};
    
    reint_pre_secure($rec);
    
    $psdev->lento_delextattr($path, $rec->{attr_name}, $rec->{attr_flags}, 
                             $rec->{mode}, $slot_offset, $rec->{recno},
                             $rec->{kml_offset} + $rec->{len}, $flags,
                             $rec->{new_obj_ctime});
    reint_post_secure($rec);
    return 0;
}

1;

__END__
    

=head1 NAME

ReintLento - Kernel Modification Log (KML) Reintegration for a local
InterMezzo file system.

=head1 SYNOPSIS

    use ReintLento;

# reintegrate a record for a specific operation

# all but the close reintegration operations have the same 
# call and parameters:
# $rec is a pointer to a hash containing an unpacked KML record, 
#    different fields of which are used for different operations
# $mtpt is the root mountpoint for the filesystem that will 
#    receive the reintegration.
# $psdev is the intermezzo pseudo device.
# $slot_offset is the offset into the expect file for the peer
#    who sent the KML record.
# $flags is currently the WLEXPECT and WLKML flags - used by 
#    presto to determine if the local expect and kml files should 
#    be written for the record.
ReintLento::reint_op($rec, $mtpt, $psdev, $slot_offset, $flags);

# the current operations are:
ReintLento::reint_setattr
    ReintLento::reint_create 
    ReintLento::reint_link 
    ReintLento::reint_unlink 
    ReintLento::reint_symlink
    ReintLento::reint_mkdir
    ReintLento::reint_rmdir
    ReintLento::reint_mknod
    ReintLento::reint_rename 
    ReintLento::reint_close 
    
    # the close record reintegration needs to also have the callback
    #    session object.
    # $rec is a pointer to a hash containing an unpacked KML record, 
    # $mtpt is the root mountpoint for the filesystem that will 
    #    receive the reintegration.
    # $psdev is the intermezzo pseudo device.
    # $session is the callback session.
    # $slot_offset is the offset into the expect file for the peer
    #    who sent the KML record.
    # $flags is currently the WLEXPECT and WLKML flags - used by 
    #    presto to determine if the local expect and kml files should 
    #    be written for the record.
    #
    # reint_close performs these steps:
    #    get inode, iversion of file to backfetch from client
    #    send bulk-transfer request to client
    #    get descriptor from client
    #    setup to receive
    #    receive data
    #    write the data to the fd
    ReintLento::reint_close($rec, $mtpt, $psdev, $session, $slot_offset, $flags);

=head1 DESCRIPTION

=item 

ReintLento reintegrations KML records into a local InterMezzo file system. 
    
    ReintLento is subordinate to Reint - only Reint should call ReintLento
    methods.

=head1 Uses

    use when running Lento locally to sync caches with another machine.
    use for the testing framework.
  
=head1 Bugs

  * Currently doing the backfetch by *name* and not by inode. Under 
    certain circumstances (involving rename reintegrations) the 
    correct file is not fetched from the server, leaving a zero
    size file. This should be resolved with the new iopen (open by inode).

  * The setfsuid/setfsgid are not done correctly for a close file
    at this time.  We need special handling here due to the blocking
    nature of the close and the asyncronicity of Lento.

  * Versions need to be checked for conflict resolution.
  

=cut

