
=pod

=head1 NAME

LML - Methods for manipulating the Local Modification Log file

=head1 SYNOPSIS

=over 4

=item B<new>: Build a new LML object.

=item B<start_offset>: Set the bytewise start offset is an arg is supplied.  
                Return the start offset in any case.

=item B<last_offset>: Set the bytewise start offset is an arg is supplied.  
                 Return the start offset in any case.

=item B<readrec>: Read a single record from a LML file into a buffer.

=back

=head1 DESCRIPTION

FIXME: Write a description.

=head1 BUGS

=head1 USAGE

=cut

package Lento::LML;
use Lento::LML::Rec;
use Lento::Debuggable;

use POSIX qw(:sys_stat_h :errno_h :unistd_h :fcntl_h);
use IO::File;
use File::stat;
use strict;
use Carp;

@Lento::LML::ISA=qw(Lento::Debuggable);
$Lento::LML::level=1;

######################################################################
#
# static method Lento::LML::new
#
# Create a new LML object with the inputs:
#   class name
#   path for the LML file
#   associated fileset (optional)
#
# Initialize the member data:
#   {path}         To the LML file
#   {handle}       File handle for the LML file
#   {fset}          The LML fileset, if it was input.
#   {start_offset} The initial offset in the LML file, set to zero.
#
#
######################################################################

sub new () {
    my $class = shift;
    my $path = shift;        # path of LML file
    my $fset = shift;        # associated fileset (when defined)
    my $self = {};
    
    $self->{path} = $path;
    $self->{handle} = new IO::File;
    $self->{handle}->open($path, &O_RDWR) or confess "Cannot open";
    $self->{fset} = $fset if defined $fset;
    $self->{start_offset} = 0;
        $self->{backfetch_hash}={};
    bless $self;
}

sub last_offset {
        my $self=shift;

        die "No lml\n" unless defined $self and defined $self->{path};

        my $st=stat($self->{path}) or die "No lml file $self->{path}\n";

    return $st->size;
}

sub populate_backfetch_hash {
    my $LML = shift;
    my $LML_offset = shift;
        my %localclosehash=();
        my $path;

        #die unless defined($LML->{fset});

    $LML_offset = 0 unless defined $LML_offset;

    while ($LML_offset < $LML->last_offset()) {
        my ($size, $offset, $buf) = $LML->readrec($LML_offset);
        last if $size == 0;

                # Keep this in place just in case we introduce chunks in LML
                #DEBUG "skipped chunk boundary at $LML_offset to $offset\n" 
        #       if ($offset > $LML_offset);
    
        my $rec = Lento::LML::Rec->Unpack($buf, 0, $offset);
    
        $LML_offset = $offset + $size;

                # The paths are relative to the fset
                if (defined $rec) {
                        if ($rec->{remote_ino}) {
                                # Only record remote backfetches
                                $LML->{backfetch_hash}->{$rec->{path}} = $rec
                        } else {
                                $localclosehash{$rec->{path}} = $rec;
                        }
                }
        }

        # now mark all backfetches which have local closes, since there is
        # evidence that these are suspect. However, we still need to retain them
        # to complete the backfetch operation.
        # XXX: actually we have a problem with all closes.
        foreach $path (keys %localclosehash) {
                if (exists ($LML->{backfetch_hash}->{$path})) {
                        $LML->{backfetch_hash}->{$path}->{dirty}=1;
                }
        }
}


sub printme { 
    my $LML = shift;
    my $LML_offset = shift;
    my $briefly = shift;
    my $result = "";
    my $count = 0;
    
    $briefly = "" unless defined $briefly;
    $LML_offset = 0 unless defined $LML_offset;
    
    while ($LML_offset < $LML->last_offset()) {
    my ($size, $offset, $buf) = $LML->readrec($LML_offset);
    last if $size == 0;

        # Keep this in place just in case we introduce chunks in LML
    #DEBUG "skipped chunk boundary at $LML_offset to $offset\n" 
    #    if ($offset > $LML_offset);
    
    my $rec = Lento::LML::Rec->Unpack($buf, 0, $offset);
    
    $LML_offset = $offset + $size;
    $count++;
    
    $result .= sprintf "Record $count size: $size, offset $LML_offset,  (total LML length %d)\n", $LML->last_offset();
    
    if (defined $rec) {
        $result .= $rec->printme($briefly);
        TACITUS($rec->printme("verybrief",$LML->{fset}));
    } else {
        $result .= "  NOOP\n";
        return "";
    }
    
    
    if ($briefly ne "brief") {
        $result .= "----------------------------------------------------------------\n";
    }
    }
    $result;
}



######################################################################
#
# method Lento::LML::start_offset
#
# With no args, return the start offset.
# With an integer arg, set the start offset.
#
######################################################################

sub start_offset {
    my $self = shift;
    my $arg = shift;
    
    if ($arg) {
    $self->{start_offset} = $arg;
    }
    
    return $self->{start_offset};
}


######################################################################
#
# method Lento::LML::readrec
#
# Read a single record from the LML into a buffer.  If called with
# optional position, first seek to that position then read the record.
# This routine also skips chunk boundaries if any are detected.
#
######################################################################

sub readrec () {
    #
    # Get the LML object and an optional position to seek to.
    #
    my $self = shift;
    my $pos = shift;
    my $buf = "";
    my $size=0; 
    
    sysseek($self->{handle}, $pos, SEEK_SET) if ( defined $pos );
    my $res = $self->{handle}->sysread($buf, 4);
    if ( $res != 4 ) {
        return (0, $pos, undef);
    }
    $size = unpack "I", $buf;
    
    sysseek($self->{handle}, $pos, SEEK_SET) if ( defined $pos );
    $res = $self->{handle}->sysread($buf, $size);
    DEBUG("Res=$res, size=$size, pos=$pos, length=%d\n", length($buf));
    confess "Could not read record in LML::readrec!\n"  if ( $res != $size );
    
    return ($size, $pos, $buf);
}

 
1;
