package Lento::List;
use strict;
use Carp;
use Lento::Debuggable;
use Tie::RefHash;

@Lento::List::ISA = qw(Lento::Debuggable);
$Lento::List::level = -1000;

=head1 NAME

List - a dll abstraction

=head1 SYNOPSIS

my $list = Lento::List->new();

$list->add($item, $desc);
$list->del($item);
$list->empty();
$list->iterate(\&fun, $parms);
$list->exists($elt);

=head1 DESCRIPTION

List is a class to give the functionality of double linked lists.
The List object has one attribute which is an anonymous hash, Tie''d with RefHash to allow references to be stored in the List.

The list goes away when no references to it are held.

=head1 Methods

=over 4

=item B<new>: Instantiate a new list

=item B<add>: Add all arguments to the list.

=item B<iterate>: take a ref to a function, and other params. Call the
function for each member of the list.

=item B<del>: Remove all arguments from the list

=over 4 B<exists>

Takes one argument returns true if the element is in the list

=back

=cut

sub new {
    my $class = shift;
    my $self = {};
    $self->{list} = {};
    tie %{$self->{list}}, 'Tie::RefHash';
    bless $self;

    return $self;
}

sub add ($$) {
    my $self = shift;
    my $elt = shift;
    my $mnemonic = shift;

    $mnemonic = "" unless defined $mnemonic;
    if ( defined $self->{list}->{$elt} ) {
	    DEBUG "Replacing an existing element $self->{list}->{$elt} in $self";
    }
    $self->{list}->{$elt} = $mnemonic;
}

sub empty ($) {
    my $self = shift;

    my $count = scalar keys %{$self->{list}};

    if ( $count ) {
       return 1; 
    } else {
       return 0;
    }
}

sub iterate {
    my $self = shift;
    my $reffun = shift;
    
    foreach my $elt (keys %{$self->{list}}) {
        if (!defined $self->{list}->{$elt}) {
            DEBUG "Removing undefined element $elt\n";
	        delete $self->{list}->{$elt};
            next;
        }
	    my ($class, $file, $line) = caller;
	    DEBUG "Iterator on %s (%s) class %s, file %s, line %d\n", $self->{list}->{$elt}, $elt, $class, $file, $line;
	    &$reffun($elt, @_);
    }
}

sub del {
    my $self = shift;
    
    foreach my $elt (@_) {
	    if ( ! defined $self->{list}->{$elt} ) {
	        DEBUG "Removing a non-existing element $elt from $self";
	    }
	    delete $self->{list}->{$elt};
    }
}

sub exists {
    my $self = shift;
    my $elt =shift;

    if ( defined ($self->{list}->{$elt} )) {
	    return 1;
    } else {
	    return undef;
    }
}

1;

__END__
