package Lento::Pinger;

use POE;
use POSIX qw(ETIMEDOUT);
use Carp;
use strict;
use Lento::Packet;
use Lento::Connection;
use Lento::EventSupport;
use Lento::Debuggable;


@Lento::Pinger::ISA = qw(Debuggable);
$Lento::Pinger::level = 1;

$Lento::Pinger::default_delay = 5;

sub new {
    my $class = shift;
    my $connector = shift;
    my $ping_delay = shift;
    my $self = {};
    bless $self;
    $self->{callback_session} = shift;

    # send got_connection to this session
    die unless defined $self->{callback_session};
    $ping_delay = $Lento::Pinger::default_delay unless $ping_delay;

    # start a Pinger for the given connector
    $self->{session}=new POE::Session
      ([$connector, $ping_delay, $self], 
       "Lento::EventSupport", ['_signal', '_child', '_stop', 'alias_set',
                              'alias_remove'],
       _start => sub {
            $_[HEAP]->{session_name} = $_[SESSION]->ID().
                                       ":Pinger for $_[ARG0]->{peer}";
            $_[HEAP]->{connector} = $_[ARG0];
            $_[HEAP]->{connection} = $_[HEAP]->{connector}->{connection};
            $_[HEAP]->{delay} = $_[ARG1];
            $_[HEAP]->{pinger} = $_[ARG2];
            $_[HEAP]->{session_level} = 0;

            Lento::Debuggable::set_option($_[SESSION],
                trace => 1, default => 1);

            $_[KERNEL]->alias_set("Pinger to ($_[ARG0]->{peer}) $_[SESSION]");
            $_[HEAP]->{last_life} = time();
            $_[KERNEL]->delay('check_ping', 2*$_[HEAP]->{delay});
            $_[HEAP]->{pinger}->{state} = 'starting';
            $_[KERNEL]->yield('do_connect');
       },
       # the following event handles a monitor loop
       check_ping => sub {
            my $connection = $_[HEAP]->{connection};

            if (  time() - $_[HEAP]->{last_life} > 
                  2*$Lento::Connection::MAX_SILENCE) {
                # if we are connected check the timeout
                if ($_[HEAP]->{connector}->{state} eq "connected") {
                    $_[KERNEL]->yield
                        ('got_error', 'check ping', ETIMEDOUT, "timed out");
                    $connection->set_error("check ping", ETIMEDOUT, "timed out");
                }
            }

            # reschedule ourselves
            $_[KERNEL]->delay('check_ping', 2*$_[HEAP]->{delay});
       },
       do_connect => sub {
            # This is an event handler to accomodate for the delay
            DEBUG "do_connect: State: $_[HEAP]->{pinger}->{state}\n";

            # Remove ourself from the error list, we are going back
            # on momentarily and don't want double ACT's.
            $_[HEAP]->{connection}->put();

            $_[HEAP]->{connector}->connectme();
       },
       got_connection => sub {
            DEBUG "got_connection: State: $_[HEAP]->{pinger}->{state}\n";
            $_[HEAP]->{last_life} = time();
            if ( $_[HEAP]->{pinger}->{state} ne 'connected' ) {
                $_[KERNEL]->post($_[HEAP]->{pinger}->{callback_session}, 
                    'got_connection');
                $_[HEAP]->{pinger}->{state} = 'connected';
                LOG("  >> Connected to $_[HEAP]->{connection}->{sysid}\n");
            }
            $_[HEAP]->{session_level} = LVL_PING;
            DEBUG("Sending packet\n");

            $_[HEAP]->{connection}->sendpacket('REQ', ['Ping', time()]);
       },
       REP => sub {
            my $packet = $_[ARG0];

            my $time = @{$packet->[PKT_PARMS]}[0];
            DEBUG "REP: State: $_[HEAP]->{pinger}->{state}\n";
            undef $packet;

            #print "Ping delta: " . (time() - $time) . "\n";

            $_[HEAP]->{last_life} = time();
            DEBUG("Calling DELAY with secs %d\n", $_[HEAP]->{delay});
            $_[KERNEL]->delay('do_connect', $_[HEAP]->{delay});
       },
       got_error => sub {
            DEBUG "got_error: State: $_[HEAP]->{pinger}->{state}\n";
            $_[HEAP]->{pinger}->{state} = 'disconnected';

            $_[HEAP]->{connection}->put() if defined $_[HEAP]->{connection};

            $_[HEAP]->{last_life} = time();

            $_[KERNEL]->delay('do_connect', $_[HEAP]->{delay});
            $_[HEAP]->{session_level} = LVL_NOTCONN;
       },
       _stop => sub {
            warn "Pinger should never stop unless Lento is closing";
            LOG "$_[HEAP]->{session_name} being stopped\n";
       }
      );

    return $self;
}

1;

__END__

=head1 NAME

Pinger - automatic connector and disconnection detector

=head1 DESCRIPTION 

This module describes a state machine that has three states:
disconnected, connected and starting. Starting state is only used when the
pinger is first getting initialized.

The transition from disconnected to connected follows the
got_connection event, the reverse a got_error event when in connected
mode. In disconnected mode, the pinger will try to do_connect every
B<delay> seconds.  The connector will return an error on failure, and
the do_connect event will be rescheduled.

In connected mode a ping message is sent out every delay seconds.  If
it fails to give a return message a got_error is generated in check
ping.

=over 4

=item * Try to connect periodically if the connection is not yet up. 
=item * Send out a ping packet periodically to keep the connection active
=item * Deliver a got_connection event upon success to the callback session

=back 

=head1 SYNOPSIS

use Lento::Pinger;
use Lento::Connector;
$connector = new Lento::Connector(...)
Lento::Pinger->new($connector, $ping_delay, $callback_session)

=head1 ACCESS METHODS

methods called by other modules

=over 4

=item I<new> (Method) 

Set up a periodic pinging session to poll liveness of the Peer.  Parameters:

=back 

=over 8

=item  a connector to use for connection setup, 

=item  a ping delay to determine frequency of pinging

=item  a callback session to notify upon success

=back

=head1 INTERNAL EVENT HANDLING

=over 4

=item B<_start> Sets up a B<check_ping> event after the designated delay.  Yields to  B<doconnect>. 

=item B<doconnect> Calls Connector::connectme

=item B<got_connection> (Sent by Connector).  Set state to
connected. Send Ping request over connection. Send B<got_connection>
event to callback session. Update last life.

=item B<check_ping> Schedule another B<check_ping>. If last life is
too old, generate an error on the Connection object.

=item Check that the connection or ping request is not quiet too
long. If so, we provoke an error and start again.

=back

=cut
