package Lento::Psdev;
use strict;
use File::stat;
use Carp;
use IO::File;
use POSIX qw(:fcntl_h :errno_h);
use Lento::Debuggable;
require "asm/ioctl.ph";

@Lento::Psdev::ISA = qw(Lento::Debuggable);
$Lento::Psdev::level = -1000;

# must pass in the path of the pseudo device
sub new {
    my $class = shift;
    my $self = {};
    my $stat;

    $self->{prestodev} = shift || die "No pathname passed";
    $stat = stat($self->{prestodev}) || 
    die "No such device $self->{prestodev}";
    $self->{minor} = ($stat->rdev) && 0xFF;
         
    bless $self, $class;
}

# open the device in this object
sub open {
    my $self = shift;
    my $whoami = shift;

    # guard against incomplete Perl installations, like mine
    die "Fcntl: O_RDWR not OK" unless O_RDWR == 2;

    $self->{handle} = new IO::File($self->{prestodev}, O_RDWR);
    if (!$self->{handle}) {
        die "Unable to open " . $self->{prestodev};
    }
    LOG ">>     Opened $self->{prestodev}, I'm pid $$\n";
    return 0;
}

#
# This is called if an EOF is ever received on this connection.  Since this is
# our link with the kernel, this is a Bad Thing(tm), and a fatal error.
#
sub cleanup {
    LOG("Aiee!  Lento received end-of-file on kernel link!\n");

    # If this function ever becomes non-fatal (!?), make sure to unregister()
    # the fd!

    die "Lento::Psdev: This is bad.  Email bugs\@inter-mezzo.org\n";
}


sub read {
    my $psdev = shift;
    my $buf;
    my $rc = $psdev->{handle}->sysread($buf, 4096);
    DEBUG "Lento::Psdev: Read $rc\n";
    return $buf;
}

eval 'sub PRESTO_SETPID () { &_IOC(1, ord(\'p\'),0x04, 4);}';
eval 'sub PRESTO_SETOPT () { &_IOC(1, ord(\'p\'),0x09, 4);}';
eval 'sub PRESTO_GETOPT () { &_IOC(1, ord(\'p\'),0x0a, 4);}';

sub setpid {
    my $psdev = shift;
    my $buf = pack "i", 0;

    my $rc = ioctl($psdev->{handle}, &PRESTO_SETPID(), $buf);
    if (!defined $rc) { 
     confess "PRESTO_SETPID ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
    1;
    } else {
        confess "PRESTO_SETPID: non-zero return $rc\n";
    }
}

eval 'sub PRESTO_SET_FSETROOT () { &_IOC(1, ord(\'p\'),0x07, 4);}';

sub setfsetroot {
    my $psdev = shift;
    my $path = shift;
    my $name = shift;
    my $id  = shift;
    my $flags = shift;

    $id = 0 unless defined $id;
    $flags = 0 unless defined $flags;

    my $buf = pack("pipiii", $path, length($path), $name, length($name),
           $id, $flags);
    
    my $rc = ioctl($psdev->{handle}, &PRESTO_SET_FSETROOT(), $buf);
    if (!defined $rc) { 
        # do not accept failures: they are protocol errors
        confess "PRESTO_SET_FSETROOT ioctl failed (os error = $!)\n"
          if ($! != EEXIST); 
        # EEXIST implies that the FSET root is already set, don't die but
        # return a failure
        return 0;
    } elsif ($rc eq "0 but true") {
        return 1;
    } else {
        confess "PRESTO_SET_FSETROOT: non-zero return $rc\n";
    }

    return 1;
}



eval 'sub PRESTO_VFS_SETATTR () { &_IOC(1, ord(\'p\'),0x10, 4);}';

sub lento_setattr {
    my $psdev = shift;
    my $path = shift;
    my $valid = shift;
    my $mode = shift;
    my $uid = shift;
    my $gid = shift;
    my $size = shift;
    my $atime = shift;
    my $mtime = shift;
    my $ctime  = shift;
    my $attr_flags  = shift;
    my $slot_offset  = shift;
    my $recno = shift;
    my $kml_offset  = shift;
    my $flags = shift;
    my $dummy = 0; # alignment for 64-bit packing

    DEBUG( "-----------> FLAGS %x\n", $flags);
    die unless defined $path;
    die unless defined $dummy;
    die unless defined $valid;
    die unless defined $mode;
    die unless defined $uid;
    die unless defined $gid;
    die unless defined $size;
    die unless defined $atime;
    die unless defined $ctime;
    die unless defined $mtime;
    die unless defined $attr_flags;
    die unless defined $slot_offset;
    die unless defined $recno;
    die unless defined $kml_offset;
    die unless defined $flags;

    my $buf = pack("p i i i i i LL LL LL LL i i i i i i i", $path, $dummy, $valid, $mode, $uid, $gid, $size, $dummy, $atime, $dummy, $mtime, $dummy, $ctime, $dummy, $attr_flags, $slot_offset, $recno, $kml_offset, 0, $flags, 0);
    
    my $rc = ioctl($psdev->{handle}, &PRESTO_VFS_SETATTR(), $buf);
    if (!defined $rc) { 
        if ($! != ENOENT) {
            confess "PRESTO_VFS_SETATTR ioctl failed (os error = $!)\n";
        } else {
            LOG "PRESTO_VFS_SETATTR ioctl failed (os error = $!)\n";
        }
    } elsif ($rc eq "0 but true") {
        DEBUG "PRESTO_VFS_SETATTR finished (success)\n";
    } else {
        confess "PRESTO_VFS_SETATTR: non-zero return $rc\n" 
            unless ($rc == ENOENT);
    }
}

eval 'sub PRESTO_VFS_CREATE () { &_IOC(1, ord(\'p\'),0x11, 4);}';

sub lento_create {
    my $psdev = shift;
    my $path = shift;
    my $mode = shift;
    my $slot_offset  = shift;
    my $recno = shift;
    my $kml_offset  = shift;
    my $flags = shift;
    my $updated_time = shift;

    my $buf = pack("piiiiiii", $path, $mode, $slot_offset, $recno, $kml_offset, 0, $flags, $updated_time);
    
    my $rc = ioctl($psdev->{handle}, &PRESTO_VFS_CREATE(), $buf);
    if (!defined $rc) { 
         confess "PRESTO_VFS_CREATE ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        DEBUG "PRESTO_VFS_CREATE finished (success)\n";
    } else {
        confess "PRESTO_VFS_CREATE: non-zero return $rc\n";
    }
}

eval 'sub PRESTO_VFS_LINK () { &_IOC(1, ord(\'p\'),0x12, 4);}';

sub lento_link {
    my $psdev = shift;
    my $oldname = shift;
    my $newname = shift;
    my $slot_offset  = shift;
    my $recno = shift;
    my $kml_offset  = shift;
    my $flags = shift;
    my $updated_time = shift;

    my $buf = pack("ppiiiiii", $oldname, $newname, $slot_offset, $recno, $kml_offset, 0, $flags, $updated_time);
    
    my $rc = ioctl($psdev->{handle}, &PRESTO_VFS_LINK(), $buf);
    if (!defined $rc) { 
        confess "PRESTO_VFS_LINK ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        DEBUG "PRESTO_VFS_LINK finished (success)\n";
    } else {
        confess "PRESTO_VFS_LINK: non-zero return $rc\n";
    }
}

eval 'sub PRESTO_VFS_UNLINK () { &_IOC(1, ord(\'p\'),0x13, 4);}';

sub lento_unlink {
    my $psdev = shift;
    my $path = shift;
    my $slot_offset  = shift;
    my $recno = shift;
    my $kml_offset  = shift;
    my $flags = shift;
    my $updated_time = shift;

    my $buf = pack("piiiiii", $path, $slot_offset, $recno, $kml_offset, 0, $flags, $updated_time);
    
    my $rc = ioctl($psdev->{handle}, &PRESTO_VFS_UNLINK(), $buf);
    if (!defined $rc) { 
     if ($! == ENOENT) {
            LOG "PRESTO_VFS_UNLINK ioctl failed (os error = $!) \n";
     } else {
           confess "PRESTO_VFS_UNLINK ioctl failed (os error = $!)";
     }
    } elsif ($rc eq "0 but true") {
        DEBUG "PRESTO_VFS_UNLINK finished (success)\n";
    } else {
        confess "PRESTO_VFS_UNLINK: non-zero return $rc\n";
    }
}

eval 'sub PRESTO_VFS_SYMLINK () { &_IOC(1, ord(\'p\'),0x14, 4);}';

sub lento_symlink {
    my $psdev = shift;
    my $oldname = shift;
    my $newname = shift;
    my $slot_offset  = shift;
    my $recno = shift;
    my $kml_offset  = shift;
    my $flags = shift;
    my $updated_time = shift;

    my $buf = pack("ppiiiiii", $oldname, $newname, $slot_offset, $recno, $kml_offset, 0, $flags, $updated_time);
    
    my $rc = ioctl($psdev->{handle}, &PRESTO_VFS_SYMLINK(), $buf);
    if (!defined $rc) { 
        confess "PRESTO_VFS_SYMLINK ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        DEBUG "PRESTO_VFS_SYMLINK finished (success)\n";
    } else {
        confess "PRESTO_VFS_SYMLINK: non-zero return $rc\n";
    }
}

eval 'sub PRESTO_VFS_MKDIR () { &_IOC(1, ord(\'p\'),0x15, 4);}';

sub lento_mkdir {
    my $psdev = shift;
    my $path = shift;
    my $mode = shift;
    my $slot_offset  = shift;
    my $recno = shift;
    my $kml_offset = shift;
    my $flags = shift;
    my $updated_time = shift;

    my $buf = pack("piiiiiii", $path, $mode, $slot_offset, $recno, $kml_offset, 0, $flags, $updated_time);

    my $rc = ioctl($psdev->{handle}, &PRESTO_VFS_MKDIR(), $buf);
    if (!defined $rc) { 
        confess "PRESTO_VFS_MKDIR ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        DEBUG "PRESTO_VFS_MKDIR finished (success)\n";
    } else {
        confess "PRESTO_VFS_MKDIR: non-zero return $rc\n";
    }
}

eval 'sub PRESTO_VFS_RMDIR () { &_IOC(1, ord(\'p\'),0x16, 4);}';

sub lento_rmdir {
    my $psdev = shift;
    my $path = shift;
    my $slot_offset  = shift;
    my $recno = shift;
    my $kml_offset  = shift;
    my $flags = shift;
    my $updated_time = shift;

    my $buf = pack("piiiiii", $path, $slot_offset, $recno, $kml_offset, 0, $flags, $updated_time);
    
    my $rc = ioctl($psdev->{handle}, &PRESTO_VFS_RMDIR(), $buf);
    if (!defined $rc) { 
        confess "PRESTO_VFS_RMDIR ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        DEBUG "PRESTO_VFS_RMDIR finished (success)\n";
    } else {
        confess "PRESTO_VFS_RMDIR: non-zero return $rc\n";
    }
}

eval 'sub PRESTO_VFS_MKNOD () { &_IOC(1, ord(\'p\'),0x17, 4);}';

sub lento_mknod {
    my $psdev = shift;
    my $path = shift;
    my $mode = shift;
    my $major = shift;
    my $minor = shift;
    my $slot_offset  = shift;
    my $recno = shift;
    my $kml_offset  = shift;
    my $flags = shift;
    my $updated_time = shift;

    my $buf = pack("piiiiiiiii", $path, $mode, $major, $minor, $slot_offset, $recno, $kml_offset, 0, $flags, $updated_time);
    
    my $rc = ioctl($psdev->{handle}, &PRESTO_VFS_MKNOD(), $buf);
    if (!defined $rc) { 
        confess "PRESTO_VFS_MKNOD ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        DEBUG "PRESTO_VFS_MKNOD finished (success)\n";
    } else {
        confess "PRESTO_VFS_MKNOD: non-zero return $rc\n";
    }
}

eval 'sub PRESTO_VFS_RENAME () { &_IOC(1, ord(\'p\'),0x18, 4);}';

sub lento_rename {
    my $psdev = shift;
    my $oldname = shift;
    my $newname = shift;
    my $slot_offset  = shift;
    my $recno = shift;
    my $kml_offset  = shift;
    my $flags = shift;
    my $updated_time = shift;

    my $buf = pack("ppiiiiii", $oldname, $newname, $slot_offset, $recno, $kml_offset, 0, $flags, $updated_time);
    
    my $rc = ioctl($psdev->{handle}, &PRESTO_VFS_RENAME(), $buf);
    if (!defined $rc) { 
        confess "PRESTO_VFS_RENAME ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        DEBUG "PRESTO_VFS_RENAME finished (success)\n";
    } else {
        confess "PRESTO_VFS_RENAME: non-zero return $rc\n";
    }
}

eval 'sub PRESTO_VFS_IOPEN () { &_IOC(1, ord(\'p\'),0x1b, 4);}';

sub lento_iopen {
    my $psdev = shift;
    my $path = shift;    # any file/path in the same cache, usually old filename
    my $ino  = shift;    # inode number for file, given in close KML record
    my $gen = shift;    # inode generation number, given in close KML record
    my $iflags = shift;    # same as unix open(2) flags, excluding O_CREAT

    my $dummy = 0;    # for 64-bit alignment
    my $fd = $dummy;    # return file descriptor value

    # pack data for ioctl into buffer.  Format:
    #   - path (string, p)
    #   - dummy (empty number for 64-bit alignment, i)
    #   - inode (number, L)
    #   - dummy (empty number for 64-bit alignment, L)
    #   - generation number (number, i)
    #   - iflags (number returned upon successful open of inode, i)
    #   - file descriptor (number, i)
    my $buf = pack("piLLiii", $path, $dummy, $ino, $dummy, $gen, $iflags, $fd);
    
    my $rc = ioctl($psdev->{handle}, &PRESTO_VFS_IOPEN(), $buf);

    # don't confess, just log so that we can deal with an error 
    #  that means, while we weren't able to find the correct inode
    #  we should just drop the kml entry and continue without a 
    #  fatal error.
    if (!defined $rc) { 
        LOG "PRESTO_VFS_IOPEN ioctl failed (os error = $!)\n";
    $fd = -1;
    } elsif ($rc eq "0 but true") {
        DEBUG "PRESTO_VFS_IOPEN finished (success)\n";
    $fd = unpack("x24 i", $buf);
    } else {
        LOG "PRESTO_VFS_IOPEN: non-zero return $rc\n";
    $fd = -1;
    }
    return $fd;
}

eval 'sub PRESTO_VFS_CLOSE () { &_IOC(1, ord(\'p\'),0x1a, 4);}';

sub lento_close {
    my $psdev = shift;
    my $fd = shift;
    my $slot_offset  = shift;
    my $recno = shift;
    my $kml_offset  = shift;
    my $flags = shift;

    # pack data for ioctl into buffer.  Format:
    #   - file descriptor (number, i)
    #   - slot_offset (number, i)
    #   - recno (number, i)
    #   - kml offset (number, i)
    #   - flags (number, i)

    DEBUG(" ** ** slotoffset: $slot_offset, kml_offset $kml_offset, recno $recno\n");     

    my $buf = pack("iiiiii", $fd, $slot_offset, $recno, $kml_offset, 0, $flags);
    
    my $rc = ioctl($psdev->{handle}, &PRESTO_VFS_CLOSE(), $buf);

    if (!defined $rc) { 
        confess "PRESTO_VFS_CLOSE ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        DEBUG "PRESTO_VFS_CLOSE finished (success)\n";
    } else {
        confess "PRESTO_VFS_CLOSE: non-zero return $rc\n";
    }
}

eval 'sub PRESTO_VFS_SETEXTATTR () { &_IOC(1, ord(\'p\'),0x1c, 4);}';
eval 'sub PRESTO_VFS_DELEXTATTR () { &_IOC(1, ord(\'p\'),0x1d, 4);}';

sub lento_setextattr {
    my $psdev = shift;
    my $path = shift;
    my $name = shift;
    my $value = shift;
    my $attrflags = shift;
    my $mode = shift;
    my $slot_offset  = shift;
    my $recno = shift;
    my $kml_offset  = shift;
    my $flags = shift;
    my $updated_time = shift;

    # pack data for ioctl into buffer.  Format:
    #   - path (pointer, p)
    #   - name (pointer, p)
    #   - namelen (number, i)
    #   - value (pointer, p)
    #   - valuelen (number, i)
    #   - attr flags (number, i)
    #   - mode (number, i)
    #   - slot_offset (number, i)
    #   - recno (number, i)
    #   - kml offset (number, i)
    #   - (dummy for 64 bit padding)
    #   - flags (number, i)
    #   - new ctime (number, i)


    DEBUG(" ** path $path, name $name, value $value length ".length($value).
          " ,attrflags $attrflags, mode $mode, slotoffset: $slot_offset,".
          " kml_offset $kml_offset, recno $recno, flags $flags,".
          " time $updated_time\n");     

    my $buf = pack("ppipiiiiiiiii", $path, $name, length($name), 
                   $value, length($value), $attrflags, $mode, $slot_offset, 
                   $recno, $kml_offset, 0, $flags, $updated_time);
    
    my $rc = ioctl($psdev->{handle}, &PRESTO_VFS_SETEXTATTR(), $buf);

    if (!defined $rc) { 
        confess "PRESTO_VFS_SETEXTATTR ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        DEBUG "PRESTO_VFS_SETEXTATTR finished (success)\n";
    } else {
        confess "PRESTO_VFS_SETEXTATTR: non-zero return $rc\n";
    }
}

sub lento_delextattr {
    my $psdev = shift;
    my $path = shift;
    my $name = shift;
    my $dummy;
    my $attrflags = shift;
    my $mode = shift;
    my $slot_offset  = shift;
    my $recno = shift;
    my $kml_offset  = shift;
    my $flags = shift;
    my $updated_time = shift;

    # pack data for ioctl into buffer.  Format:
    #   - path (pointer, p)
    #   - name (pointer, p)
    #   - namelen (number, i)
    #   - dummy value (pointer, p)
    #   - 0 valuelen (number, i)
    #   - attr flags (number, i)
    #   - mode (number, i)
    #   - slot_offset (number, i)
    #   - recno (number, i)
    #   - kml offset (number, i)
    #   - (dummy for 64 bit padding)
    #   - flags (number, i)
    #   - new ctime (number, i)


    DEBUG(" ** path $path, name $name, ".
          " attrflags $attrflags, mode $mode, slotoffset: $slot_offset,".
          " kml_offset $kml_offset, recno $recno, flags $flags,".
          " time $updated_time\n");     

    my $buf = pack("ppipiiiiiiiii", $path, $name, length($name), $dummy, 
                   0, $attrflags, $mode, $slot_offset, $recno, $kml_offset, 
                   0, $flags, $updated_time);
    
    my $rc = ioctl($psdev->{handle}, &PRESTO_VFS_DELEXTATTR(), $buf);

    if (!defined $rc) { 
        confess "PRESTO_VFS_DELEXTATTR ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        DEBUG "PRESTO_VFS_DELEXTATTR finished (success)\n";
    } else {
        confess "PRESTO_VFS_DELEXTATTR: non-zero return $rc\n";
    }
}


sub dogetopt { 
    my $psdev = shift;
    my $opt = shift;

    my $buf = pack "i", $opt;

    my $rc = ioctl($psdev->{handle}, &PRESTO_GETOPT(), $buf);
    if (!defined $rc) { 
        confess "PRESTO_GETOPT ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        DEBUG "PRESTO_GETOPT finished (success)\n";
    } else {
        confess "PRESTO_GETOPT: non-zero return $rc\n";
    }
    my ($junk,  $retval) = unpack("ii", $buf);
    return $retval;
}

eval 'sub PRESTO_CLEAR_FSETROOT () { &_IOC(1, ord(\'p\'),0x08, 4);}';

sub clearfsetroot {
    my $psdev = shift;
    my $path = shift;

    my $buf = pack "pi", $path, length($path);

    my $rc = ioctl($psdev->{handle}, &PRESTO_CLEAR_FSETROOT(), $buf);
    if (!defined $rc) { 
         confess "PRESTO_CLEAR_FSETROOT ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        # DEBUG "PRESTO_CLEAR_FSETROOT finished (success)\n";
        0;
    } else {
        confess "PRESTO_CLEAR_FSETROOT: non-zero return $rc\n";
    }

}


eval 'sub PRESTO_CLEAR_ALL_FSETROOTS () { &_IOC(1, ord(\'p\'),0x22, 4);}';

sub clearallfsetroots {
    my $psdev = shift;
    my $path = shift;

    my $buf = pack "pi", $path, length($path);

    my $rc = ioctl($psdev->{handle}, &PRESTO_CLEAR_ALL_FSETROOTS(), $buf);
    if (!defined $rc) { 
         confess "PRESTO_CLEAR_ALL_FSETROOTS ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        # DEBUG "PRESTO_CLEAR_ALL_FSETROOTS finished (success)\n";
        0;
    } else {
        confess "PRESTO_CLEAR_ALL_FSETROOTS: non-zero return $rc\n";
    }

}

eval 'sub PRESTO_GET_KMLSIZE () { &_IOC(1, ord(\'p\'),0x0b, 4);}';

# this must be fixed later for correct 64 bit operation
sub get_kml_size {
    my $psdev = shift;
    my $path = shift;
    my $size_low;
    my $size_high;
    my $buf = pack "iipi", 0, 0, $path, length($path);

    #DEBUG "KML size called for $::psdev->{prestodev} ";
    my $rc = ioctl($psdev->{handle}, &PRESTO_GET_KMLSIZE(), $buf);
    if (!defined $rc) { 
         confess "PRESTO_GET_KMLSIZE failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        ($size_low, $size_high) = unpack "ii", $buf;
        DEBUG "PRESTO_GET_KMLSIZE low: $size_low, high $size_high\n";
    } else {
        confess "PRESTO_GET_KMLSIZE: non-zero return $rc\n";
    }
    return $size_low;
}

eval 'sub PRESTO_GET_RECNO () { &_IOC(1, ord(\'p\'),0x0c, 4);}';

# this must be fixed later for correct 64 bit operation
sub get_last_recno {
    my $psdev = shift;
    my $path = shift;
    my $recno_low;
    my $recno_high;
    my $buf = pack "iipi", 0, 0, $path, length($path);

    DEBUG "last recno called for $::psdev->{prestodev} ";
    my $rc = ioctl($psdev->{handle}, &PRESTO_GET_RECNO(), $buf);
    if (!defined $rc) { 
         confess "PRESTO_GET_RECNO failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        ($recno_low, $recno_high) = unpack "ii", $buf;
        DEBUG "PRESTO_GET_RECNO low: $recno_low, high $recno_high\n";
    } else {
        confess "PRESTO_GET_RECNO: non-zero return $rc\n";
    }
    return $recno_low;
}

sub dosetopt {
    my $psdev = shift;
    my $opt = shift;
    my $newval = shift;

    my $buf = pack "ii", $opt, $newval;

    my $rc = ioctl($psdev->{handle}, &PRESTO_SETOPT(), $buf);
    if (!defined $rc) { 
         confess "PRESTO_SETOPT ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        1;
    } else {
        confess "PRESTO_SETOPT: non-zero return $rc\n";
    }
    my ($junk,  $retval) = unpack("ii", $buf);
    return $retval;
}

sub fix_path ($) {
    my $path = shift;
    if( $path eq "" ) { 
      return "/"; 
    }
    else { 
      return $path; 
    } 
}

eval 'sub PRESTO_RELEASE_PERMIT() { &_IOC(0x1, ord(\'p\'),0x21, 4);}';

sub revoke_permit ($) {
    my $psdev = shift;
    my $path = shift;
    my $cookie = shift;    # return value to wait for

    my $buf = pack "iip", $cookie, length($path), $path;

    DEBUG "Revoking permit for $path with cookie $cookie\n";

    my $rc = ioctl($psdev->{handle}, &PRESTO_RELEASE_PERMIT(), $buf);

    if (!defined $rc) { 
         confess "PRESTO_RELEASE_PERMIT ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        $cookie = unpack("i", $buf);
        return $cookie;
    } else {
        confess "PRESTO_RELEASE_PERMIT: non-zero return $rc\n";
    }
}

eval 'sub PRESTO_BACKFETCH_LML() { &_IOC(0x1, ord(\'p\'),0x23, 4);}';

# this must be fixed later for correct 64 bit operation
sub record_backfetch_lml {
    my $psdev=shift;
    my $inodeinfo=shift;
    my $path = $inodeinfo->{path};
    my $pathlen = length($path);
    my $remote_ino = $inodeinfo->{ino};    
    my $remote_generation = $inodeinfo->{generation};    
    my $remote_version = $inodeinfo->{version};    
    my $remote_mtime = $inodeinfo->{mtime};    
    my $remote_ctime = $inodeinfo->{ctime};    
    my $remote_size = $inodeinfo->{size};    

    DEBUG "LML args: $path, $pathlen, $remote_ino, $remote_generation, $remote_version, $remote_mtime, $remote_ctime, $remote_size\n";

    my $buf = pack "piiiiiiiiiii", $path, $pathlen, $remote_ino, 0, 
                $remote_generation, $remote_version, $remote_mtime, 0 ,
                $remote_ctime, 0, $remote_size, 0;

    #if quad words are supported
    #my $buf = pack "piqiiqqq", $path, $pathlen, $remote_ino, 
    #            $remote_generation, $remote_version, $remote_mtime, 
    #            $remote_ctime, $remote_size;

    DEBUG "Recording backfetch for path $path, remote ino $remote_ino\n";

    my $rc = ioctl($psdev->{handle}, &PRESTO_BACKFETCH_LML(), $buf);

    if (!defined $rc) { 
         confess "PRESTO_BACKFETCH_LML ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        return;
    } else {
        confess "PRESTO_BACKFETCH_LML: non-zero return $rc\n";
    }
}

eval 'sub PRESTO_REINT() { &_IOC(0x1, ord(\'p\'),0x24, 4);}';

sub reint_kml {
    my $psdev=shift;
    my $volname=shift;
    my $vollen = length ($volname);
    my $recbuf=shift;
    my $reclen= length($recbuf);

    my $buf = pack "pipi", $volname, $vollen, $recbuf, $reclen;

    my $rc = ioctl($psdev->{handle}, &PRESTO_REINT (), $buf);
    if (!defined $rc) { 
         confess "PRESTO_REINT ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        return;
    } else {
        confess "PRESTO_REINT: non-zero return $rc\n";
    }
}

eval 'sub PRESTO_CANCEL_LML() { &_IOC(0x1, ord(\'p\'),0x25, 4);}';

sub cancel_lml {
    my $psdev=shift;
    my $inodeinfo=shift;
    my $flags=shift;

    my $path = $inodeinfo->{path};
    my $lml_offset = $inodeinfo->{lml_offset};
    my $pathlen = length($path);
    my $remote_ino = $inodeinfo->{ino};    
    my $remote_generation = $inodeinfo->{generation};    
    my $remote_version = $inodeinfo->{version};    

    # the info structure
    my $slot_offset  = $inodeinfo->{slot_offset};
    my $recno = $inodeinfo->{recno};
    my $kml_offset  = $inodeinfo->{kml_offset};
    #my $flags = $inodeinfo->{flags};

    DEBUG "cancel lml args: $path, $lml_offset, $pathlen, $remote_ino, $remote_generation, $remote_version, $slot_offset, $recno, $kml_offset, $flags\n";

    my $buf = pack "piiiiiiiiiiii", $path, $lml_offset, 0, $pathlen, 
                $remote_ino, 0, $remote_generation, $remote_version, 
                $slot_offset, $recno, $kml_offset, 0, $flags;

    my $rc = ioctl($psdev->{handle}, &PRESTO_CANCEL_LML(), $buf);

    if (!defined $rc) { 
         confess "PRESTO_CANCEL_LML ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        return;
    } else {
        confess "PRESTO_CANCEL_LML: non-zero return $rc\n";
    }
}

eval 'sub PRESTO_RESET_FSET() { &_IOC(0x1, ord(\'p\'),0x26, 4);}';

sub reset_fset {
    my $psdev=shift;
    my $path = shift;
    my $pathlen = length($path);
    my $offset= shift;
     my $recno=shift;

    my $buf = pack "piiii", $path, $pathlen, $offset, 0 ,$recno;

    my $rc = ioctl($psdev->{handle}, &PRESTO_RESET_FSET (), $buf);
    if (!defined $rc) { 
         confess "PRESTO_RESET_FSET ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        return;
    } else {
        confess "PRESTO_RESET_FSET: non-zero return $rc\n";
    }
}

eval 'sub PRESTO_COMPLETE_CLOSES() { &_IOC(0x1, ord(\'p\'),0x27, 4);}';

sub complete_closes {
    my $psdev=shift;
    my $path = shift;
    my $pathlen = length($path);

    my $buf = pack "pi", $path, $pathlen;

    my $rc = ioctl($psdev->{handle}, &PRESTO_COMPLETE_CLOSES (), $buf);

    if (!defined $rc) { 
         confess "PRESTO_COMPLETE_CLOSES ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        return;
    } else {
        LOG("PRESTO_COMPLETE_CLOSES: non-zero return $rc\n");
    }
}

# data in dentry->d_fsdata and fset->fset_data
my $PRESTO_FSETROOT = 0x00000001;  # this dentry is root of fileset 
my $PRESTO_DATA     = 0x00000002;  # data cached
my $PRESTO_ATTR     = 0x00000004; # attributes cached

# flags in fset->fset_flags
my $FSET_CLIENT_RO = 0x00000001;
my $FSET_LENTO_RO = 0x00000002;
my $FSET_HASPERMIT = 0x00000004;  # we have a permit to WB
my $FSET_INSYNC= 0x0000008;  # this file set is in sync
my $FSET_PERMIT_WAITING= 0x00000010;  # this fileset has a revoke pending
my $FSET_STEAL_PERMIT=0x00000020;  # presto is allowed to steal a permit
my $FSET_JCLOSE_ON_WRITE=0x00000040;  # journal closes on writes

# flags in cache->cache_flags
my $PRESTO_CACHE_CLIENT_RO= 0x2;
my $PRESTO_CACHE_LENTO_RO=  0x4;

# constants to say mark this that or that...
my $MARK_DENTRY = 101;
my $MARK_FSET = 102;
my $MARK_CACHE = 103;

eval 'sub PRESTO_MARK () {&_IOC(0x1, ord(\'p\'),0x20, 4);}';

sub mark ($$$) {
    my $psdev = shift;
    my $mark_what = shift;
    my $path = shift;
    my $andflag = shift;
    my $orflag = shift;
    $path= fix_path($path);
    my $len = length($path) + 1;

    if (! -e $path) {
        confess "No such file $path\n";
    }

    if ($mark_what != $MARK_DENTRY &&
        $mark_what != $MARK_FSET && 
        $mark_what != $MARK_CACHE) {
        confess("wrong mark_what parameter ($mark_what)\n");
    }

    DEBUG(" MARK : what=$mark_what: a=%x : o=%x : p=$path\n",
     $andflag, $orflag); 
    my $msg = pack "iiiip", $mark_what, $andflag, $orflag, $len, $path;
    my $rc = ioctl($psdev->{handle}, &PRESTO_MARK(), $msg);

    if (!defined $rc) { 
         confess "PRESTO_MARK ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        my $flags = unpack "i", $msg;
        return (0, $flags);
    } else {
        confess "PRESTO_MARK: non-zero return $rc\n";
    }

}

sub mark_dentry ($$$) {
    my $psdev = shift;
    my $path = shift;
    my $andflag = shift;
    my $orflag = shift;

    return $psdev->mark($MARK_DENTRY, $path, $andflag, $orflag);
}

sub mark_fset ($$$) {
    my $psdev = shift;
    my $path = shift;
    my $andflag = shift;
    my $orflag = shift;

    return $psdev->mark($MARK_FSET, $path, $andflag, $orflag);
}

sub mark_cache ($$$) {
    my $psdev = shift;
    my $path = shift;
    my $andflag = shift;
    my $orflag = shift;

    return $psdev->mark($MARK_CACHE, $path, $andflag, $orflag);
}

# get dentry, fset and cache flags
sub get_flags($$) {
    my $psdev = shift;
    my $path = shift;

    $path= fix_path($path);
    my $len = length($path) + 1;

    if (! -e $path) {
        die "No such file $path\n";
    }

    # constants to say mark this that or that...
    my $MARK_GETFL = 104;

    my $msg = pack "iiiip", $MARK_GETFL, 0, 0, $len, $path;
    my $rc = ioctl($psdev->{handle}, &PRESTO_MARK(), $msg);

    if (!defined $rc) { 
         confess "PRESTO_GETFL ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        my ($dflags, $fflags, $cflags) = unpack "iii", $msg;
        DEBUG("GETFLAGS: p=$path - d: %x, f: %x, c: %x\n",
            $dflags, $fflags, $cflags); 
        return ($dflags, $fflags, $cflags);
    } else {
        confess "PRESTO_GETFL: non-zero return $rc\n";
    }
}

sub set_presto_read_only () {
    my $psdev = shift;
    my $path = shift;
    LOG("  >>  Calling set_presto_read_only : $path : ****************\n");
    $psdev->mark_cache($path, ~$PRESTO_CACHE_CLIENT_RO, 0);
} 

sub revoke_data ($$) {
    my $psdev = shift;
    my $path = shift;
    $path= fix_path($path);
    DEBUG("  >>  Psdev::revoke_data ($path)\n");
    $psdev->mark_dentry($path, ~$PRESTO_DATA, 0);
}

sub revoke_attr ($$) {
    my $psdev = shift;
    my $path = shift;
    $path= fix_path($path);
    DEBUG("  >>  Psdev::revoke_attr ($path)\n");
    $psdev->mark_dentry($path, ~$PRESTO_ATTR, 0);
}

sub setfsetinsync ($$) {
    my $psdev = shift;
    my $path = shift;
    $path= fix_path($path);
    LOG("  >>  Psdev::setfsetinsync ($path)\n");
    $psdev->mark_fset($path, ~0, $FSET_INSYNC);
}

sub setfsetallowsteal ($$) {
    my $psdev = shift;
    my $path = shift;
    $path= fix_path($path);
    LOG("  >>  Psdev::setfsetallowsteal ($path)\n");
    $psdev->mark_fset($path, ~0, $FSET_STEAL_PERMIT);
}

sub grant_permit ($) {
    my $psdev = shift;
    my $path = shift;
    $path= fix_path($path);
    DEBUG("  >>  Psdev::grant_permit ($path)\n");
    $psdev->mark_fset($path, ~0, $FSET_HASPERMIT);
}

sub enable_permit_stealing ($) {
    my $psdev = shift;
    my $path = shift;
    $path= fix_path($path);
    LOG("  >>  Psdev::enable_permit_stealing ($path)\n");
    $psdev->mark_fset($path, ~0, $FSET_STEAL_PERMIT);
}

sub disable_permit_stealing ($) {
    my $psdev = shift;
    my $path = shift;
    $path= fix_path($path);
    LOG("  >>  Psdev::disable_permit_stealing ($path)\n");
    $psdev->mark_fset($path, ~$FSET_STEAL_PERMIT, 0);
}

sub enable_jclose_on_write ($) {
    my $psdev = shift;
    my $path = shift;
    $path= fix_path($path);
    LOG("  >>  Psdev::enable_jclose_on_write ($path)\n");
    $psdev->mark_fset($path, 0, $FSET_JCLOSE_ON_WRITE);
}

sub disable_jclose_on_write ($) {
    my $psdev = shift;
    my $path = shift;
    $path= fix_path($path);
    LOG("  >>  Psdev::disable_jclose_on_write ($path)\n");
    $psdev->mark_fset($path, ~$FSET_JCLOSE_ON_WRITE, 0);
}

1;


#
#  Sysctl stuff: these are for those who don't have /proc
#

sub psdev_debug {
    my $psdev = shift;
    my $opt = shift;
    $psdev->dosetopt(1, $opt);
}
sub psdev_entry {
    my $psdev = shift;
    my $opt = shift;
    $psdev->dosetopt(2, $opt);
}
sub psdev_timeout {
    my $psdev = shift;
    my $opt = shift;
    $psdev->dosetopt(3, $opt);
}
sub psdev_hard {
    my $psdev = shift;
    my $opt = shift;
    $psdev->dosetopt(4, $opt);
}
sub psdev_no_filter {
    my $psdev = shift;
    my $opt = shift;
    $psdev->dosetopt(5, $opt);
}
sub psdev_no_journal {
    my $psdev = shift;
    my $opt = shift;
    $psdev->dosetopt(6, $opt);
}

eval 'sub PRESTO_GETMOUNT () { &_IOC(1, ord(\'p\'),0x03, 4);}';

sub readmount {
    my $psdev = shift;
    my $len = 8192;
    my $str = " "x$len;
    my $buf = pack("ip", $len, $str);
    
    my $rc = ioctl($psdev->{handle}, &PRESTO_GETMOUNT(), $buf);
    if (!defined $rc) { 
         confess "PRESTO_GETMOUNT ioctl failed (os error = $!)\n";
    } elsif ($rc eq "0 but true") {
        my $outlen = unpack("i", $buf);
        my $retval = substr($str, 0, $outlen);
        return $retval;
    } else {
        confess "PRESTO_GETMOUNT: non-zero return $rc\n";
    }
}          

sub write {
    my $psdev = shift;
    my $buf = shift;
    my $len = shift;
    DEBUG "Writing $len bytes of %d\n", length($buf);
    my $rc = $psdev->{handle}->syswrite($buf, $len);
    if (!defined $rc) {
#    die "syswrite error";
    } else {
        DEBUG "Wrote $rc\n";
    }

    if ($rc) {
        return $rc;
    } else {
        return -$psdev->{handle}->error();
    }
}

1;
