package Lento::Replicator;
use strict;
use Carp;
use POE;
use POSIX;
use DB_File;
use File::stat;
use Lento::EventSupport;
use Lento::List;
use Lento::Debuggable;

# Sometimes stats and opens might have to be done multiple times due
# to EGAIN failures
# This parameter controls how many
$Lento::Replicator::RETRY = 3;

my $SENDING = 0x1;
my $ONLINE = 0x2;
my $MOREPENDING = 0x4;
my $RECEIVING = 0x8;
my $SYNCING = 0x10;

$Lento::Replicator::state_string = { 0 => "NO STATE", 1 => "SENDING", 
    2 => "ONLINE", 4 => "MORE PENDING", 8 => "RECEIVING", 16 => "SYNCING" };

@Lento::Replicator::ISA = qw(Lento::Debuggable);
$Lento::Replicator::level = -1;

my $ReplicatorDB = {};

=pod

$ReplicatorDB_tied is a hash of ties to an anonymous hashes stored in DB files.
    The key in the anonymous hash is the binary string "sysid\0fsetname".  The
    values are a frozen copy of fields  slot_offset & state.
    
    An important function is the last_rcvd method which returns the
    last_rcvd record numbers and offsets, both the local and remote ones,
    from the last_rcvd file.

=cut
    
my %ReplicatorDB_tied = ();     # becomes a hash of the hashes for disk files
my %ReplicatorDB_handle= ();    # hash of handles to open databases

# XXX set $D to n or I depending on the host
# test the result of unpack("I" (pack "n", 1)) == 1
my $D ;

#
# XXX need
# - garbage collection from the DB when replicators are killed
#

sub new {
    my $class = shift;
    my $fset = shift;
    my $remoteid = shift || die "No peerid";
    
    my $fsetname = $fset->{name} || die "No file set name";
    my $cachemtpt = $fset->{cache_mtpt} || confess "No cache mountpoint";
    
    my $self = {};
    
    # this peer is either a client talking to "us the server"
    # or the server of the fileset, which "we the client" are talking to"
    $self->{sysid} = $remoteid;
    $self->{fsetname} = $fsetname;
    
    $self->{fileset}= $fset;
    
    $self->{journal_path}="$cachemtpt/.intermezzo/$fsetname/";
    
    $self->{last_rcvd_file} =  $self->{journal_path}."last_rcvd";
    
    $self->{state} = 0;         # XXX: for now, reset this each time.
    
    bless $self, $class;
    # if this guy is new assign a slot_offset to it
    my $data = $self->load_from_db();
    
    if (! defined  $data ) {
        my $st = stat $self->{last_rcvd_file} || 
            die "No last received file for $self->{fsetname}";
        # the size of an expect structure is 32 + 64 + 32 + 64 bits = 24 bytes
        $self->{slot_offset} = $st->size;
        
        # put stuff there to fill the slot
        # XXX I to be replaced with $D$D$D$D
        DEBUG "extending $self->{last_rcvd_file} from $self->{slot_offset} to $self->{slot_offset} + 24\n";
        my $buf = pack("IIIIII", 0, 0, 0, 0, 0, 0);
        sysopen FH, "$self->{last_rcvd_file}", O_RDWR | O_APPEND 
            or die "Cannot open $self->{last_rcvd_file} ($!)";
        die "write error ($!)" unless defined syswrite FH, $buf, 24;
        close FH;
        $self->save_to_db();
        $self->test_db();
        $ReplicatorDB_handle{$self->{fsetname}}->sync();
        $self->test_db();
    } else {
        DEBUG "slot_offset= $self->{slot_offset}";
        DEBUG "state= $self->{state}\n";
    }
    LOG(">>     New replicator: peer %s fset %s, slot offset %s\n", 
        $self->{sysid}, $self->{fsetname}, $self->{slot_offset});
    
    # and into the hashes
    $ReplicatorDB->{$self->{sysid}}->{$self->{fsetname}} = $self;
    
    $self->test_db();
    
    # next KML offset to send to peer (this system's numbering)
    $self->{next_to_send} = 0; 
    
    # XXX hack, should be established by ReplicatorStatus calls
    $self->online(1);
    
    # This is needed since the lento might have crashed in the middle
    # of a send_KML the last time around and this erroneous state will get
    # restored from the database
    $self->morepending(0);
    $self->clear_send_session();
    return $self;
}

sub printme {
    my $self = shift;
    my $str;
    $self->next_to_send();
    $str  = "REPLICATOR: sysid              $self->{sysid}\n";
    $str .= "REPLICATOR: fsetname           $self->{fsetname}\n";
    $str .= "REPLICATOR: next_to_send       $self->{next_to_send}\n";
    $str .= "REPLICATOR: last_local_offset  $self->{last_local_offset}\n";
    $str .= "REPLICATOR: last_local_recno   $self->{last_local_recno}\n";
    $str .= "REPLICATOR: last_remote_offset $self->{last_remote_offset}\n";
    $str .= "REPLICATOR: last_remote_recno  $self->{last_remote_recno}\n";
    $str .= "REPLICATOR: state              $Lento::Replicator::state_string{$self->{state}}\n" 
        if defined $self->{state} 
            and defined $Lento::Replicator::state_string{$self->{state}};
    
    return $str;
}


sub load_from_db () {
    my $self = shift;
    $self->initialize_tie() 
        unless defined $ReplicatorDB_tied{$self->{fsetname}};
    my $data = $ReplicatorDB_tied{$self->{fsetname}}->{$self->{sysid} . "\0" . $self->{fsetname}};
    
    DEBUG("READING $self->{sysid}, $self->{fsetname} from replicator db\n");
    if (defined $data) {
        $self->{slot_offset}
            = unpack "n", $data;
        DEBUG("---- DATA: $self->{slot_offset}\n");
    }
    
    DEBUG "kml size".$::psdev->get_kml_size($self->{fileset}->mtpt())."\n" 
        if defined $::fsetdb;
    
    return $data;
}

sub test_db () {
    my $self = shift;
    my $data = $ReplicatorDB_tied{$self->{fsetname}}->{$self->{sysid} . "\0" . $self->{fsetname}};
    confess "GRR: data still not defined\n" unless defined $data;
}


sub save_to_db () {
    my $self = shift;
    $self->initialize_tie() unless defined $ReplicatorDB_tied{$self->{fsetname}};
    die "No db\n" unless defined $ReplicatorDB_tied{$self->{fsetname}} and
        defined $ReplicatorDB_handle{$self->{fsetname}};
    die unless defined  $self->{slot_offset};
    my $data = pack("n", $self->{slot_offset});
    $ReplicatorDB_tied{$self->{fsetname}}->{
                       $self->{sysid} . "\0" . $self->{fsetname}} = $data;
}

sub initialize_tie {
    my $self=shift;
    my $replicator_db_file = $self->{journal_path} . "db/replicators.db";
    $ReplicatorDB_handle{$self->{fsetname}}= 
        tie (%{$ReplicatorDB_tied{$self->{fsetname}}}, 'DB_File', 
            $replicator_db_file, O_CREAT|O_RDWR, 0700, $DB_HASH) 
        or die "Cannot open replicator database - $replicator_db_file. ($!)\n";
}

sub sync_replicator_db {
    my $fsetname;
    foreach $fsetname (keys %ReplicatorDB_handle) {
        $ReplicatorDB_handle{$fsetname}->sync();
    }
}

sub last_rcvd {
    my $self = shift;
    my $buf;
    my $count=0;
    # needed for 64 bit packing
    my $dummy1;
    my $dummy2;
    
    # Sometimes under high loads the os might return an error (EAGAIN?)
    for ($count=0;$count < $Lento::Replicator::RETRY;$count++) {
        sysopen(FH, "$self->{last_rcvd_file}", O_RDONLY) && last;
    };
    
    confess "Cannot open $self->{last_rcvd_file} ($!)" 
        if ($count == $Lento::Replicator::RETRY);
    
    sysseek(FH, $self->{slot_offset}, SEEK_SET);
    sysread(FH, $buf, 24);
    close FH;
    ($self->{last_remote_recno}, $self->{last_remote_offset}, $dummy1, 
     $self->{last_local_recno}, $self->{last_local_offset}, $dummy2) = 
         unpack "IIIIII", $buf;
    
    $self->{last_remote_recno}=0 unless defined $self->{last_remote_recno};
    $self->{last_remote_offset}=0 unless defined $self->{last_remote_offset};
    $self->{last_local_recno}=0 unless defined $self->{last_local_recno};
    $self->{last_local_offset}=0 unless defined $self->{last_local_offset};

    DEBUG " dummy1=$dummy1, dummy2=$dummy2\n";
    
    DEBUG "---- slot $self->{slot_offset}, last local $self->{last_local_offset} last remote $self->{last_remote_offset}\n";
    $self->{last_local_offset} = 0 unless defined $self->{last_local_offset};
    return ($self->{last_remote_recno}, $self->{last_remote_offset}, 
            $self->{last_local_recno}, $self->{last_local_offset});
}    

sub set_remote_offset {
    my ($self, $last_remote_offset) = @_;

    my ($remote_recno,$remote_offset,$local_recno,$local_offset)=
        $self->last_rcvd();

    $self->write_last_rcvd($remote_recno,$last_remote_offset,
                           $local_recno,$local_offset);
}

sub sync_last_rcvd {
    my ($self,$last_remote_recno, $last_remote_offset) = @_;
    
    die "recno  not defined" unless defined $last_remote_recno;
    die "offset not defined" unless defined $last_remote_offset;
    
    my $last_local_recno=$::psdev->get_last_recno($self->{fileset}->mtpt());
    my $last_local_offset=$::psdev->get_kml_size($self->{fileset}->mtpt());
    
    DEBUG $self->printme();
    DEBUG "syncing last rcvd sequence with $last_remote_recno,".
        " $last_remote_offset, $last_local_recno, $last_local_offset";
    
    $self->write_last_rcvd($last_remote_recno, $last_remote_offset, 
                           $last_local_recno, $last_local_offset);
}

sub write_last_rcvd ($$$$) {
    my ($self, $last_remote_recno, $last_remote_offset, $last_local_recno,
        $last_local_offset)=@_;
    
    DEBUG("Writing last_rcvd for \n".$self->printme().
        "with $last_remote_recno, $last_remote_offset, $last_local_recno,
       $last_local_offset");
    
    $last_remote_recno=0  if !(defined $last_remote_recno);
    $last_remote_offset=0 if !(defined $last_remote_offset);
    $last_local_recno=0   if !(defined $last_local_recno);
    $last_local_offset=0  if !(defined $last_local_recno);
    
    # XXXX: 0s are for 64 bit packing
    my $buf= pack("IIIIII",$last_remote_recno, $last_remote_offset, 0,
                  $last_local_recno, $last_local_offset, 0);
    
    sysopen(FH, "$self->{last_rcvd_file}", O_RDWR) 
        or die "Cannot open $self->{last_rcvd_file}";
    sysseek(FH, $self->{slot_offset}, SEEK_SET);
    die "write error ($!)" unless defined syswrite FH, $buf, 24;
    close FH;
}


# class method
sub find {
    my $sysid = shift;
    my $fsetname = shift;
    
    if ( defined $ReplicatorDB->{$sysid}) {
        return $ReplicatorDB->{$sysid}->{$fsetname};
    }
    undef;
}

sub is_registered_replicator {
    my $sysid = shift;
    return ( defined $ReplicatorDB->{$sysid});
}

sub sysid {
    my $self = shift;
    return $self->{sysid};
}

#
# state management
#

sub send_session {
    my $replicator = shift;
    my $session = shift;
    my $name = undef;
    
   if (defined $session ) {
        $name = defined $session->get_heap()->{session_name} ? $session->get_heap()->{session_name} : $session;
    }

    $replicator->report_state("SENDING", $name);
    $replicator->{send_session} = $session if defined $session; 
    
    return $replicator->{send_session}; 
}

sub clear_send_session { 
    my $replicator = shift;
    my $session=shift;

    return unless defined $replicator->{send_session};
    return unless $session==$replicator->{send_session};
    
    $replicator->report_state("SENDING", 0);
    
    delete $replicator->{send_session};
} 

sub report_state {
    my $self = shift;
    my $state = shift;
    my $val = shift;
    
    my $msg = "Replicator $self->{sysid} $self->{fsetname}: $state ";
    if (defined $val) {
        $msg .= " SET $val";
    } else {
        $msg .= " QUERY";
    }
    my ($pkg, $file, $line) = caller 2;
    $msg .= " (CURRENT STATE $self->{state}) " . $pkg . "::" . "$file (line $line)\n";
    DEBUG($msg);
}


sub syncing {
    my $self = shift;
    my $val = shift;
    
    $self->report_state("SYNCING", $val);
    
    my $ret = Lento::EventSupport::state($self, $SYNCING, $val);
    return $ret;
}

sub online {
    my $self = shift;
    my $val = shift;
    
    #    $self->report_state("ONLINE", $val);
    
    my $ret = Lento::EventSupport::state($self, $ONLINE, $val);
    return $ret;
}

sub morepending {
    my $self = shift;
    my $val = shift;
    
    $self->report_state("MOREPENDING", $val);
    
    my $ret = Lento::EventSupport::state($self, $MOREPENDING, $val);
    return $ret;
}

sub receiving {
    my $self = shift;
    my $val = shift;
    
    $self->report_state("RECEIVING", $val);
    
    my $ret = Lento::EventSupport::state($self, $RECEIVING, $val);
    return $ret;
}


#
# next to send support
#

sub incr_next_to_send ($$) {
    my $self = shift;
    my $val = shift;
    
    die "No incr passed" unless defined $val;
    $self->{next_to_send} += $val;
    DEBUG "Setting next_to_send to $self->{next_to_send}\n";
    
    my $kmlsize=$::psdev->get_kml_size($self->{fileset}->mtpt());
    if ($self->{next_to_send} > $kmlsize) {
        DEBUG "next_to_send $self->{next_to_send} > kmlsize $kmlsize, ".
            "adjusting offsets\n";
        # be paranoid kml sizes may change!
        $kmlsize=$self->{next_to_send}=$::psdev->get_kml_size($self->{fileset}->mtpt());
    }
    
    DEBUG $self->printme();
    DEBUG "kml size $kmlsize\n";
}

sub last_remote_offset { 
    my $self = shift;
    
    $self->last_rcvd();
    return $self->{last_remote_offset};
}

sub next_to_send ($) {
    my $self = shift;
    my $next_expected= shift;
    
    $self->last_rcvd();
    
    confess "no last local" unless defined $self->{last_local_offset};
    confess "no next_to_send" unless defined $self->{next_to_send};
    
    if (defined $next_expected) {
        $self->{next_to_send}=$next_expected;
        DEBUG "next_to_send for $self->{sysid} set to $self->{next_to_send}\n";
        DEBUG "kml size".$::psdev->get_kml_size($self->{fileset}->mtpt())."\n";
    } else {
        if ($self->{last_local_offset} > $self->{next_to_send}) {
            $self->{next_to_send} = $self->{last_local_offset};
            DEBUG "next_to_send set for $self->{sysid} to ";
            DEBUG "$self->{next_to_send}";
            DEBUG "kml size ".$::psdev->get_kml_size($self->{fileset}->mtpt()).
                "\n";
        } 
    }
    
    my $kmlsize=$::psdev->get_kml_size($self->{fileset}->mtpt());
    if ($self->{next_to_send} > $kmlsize) {
        DEBUG "next_to_send $self->{next_to_send} > kmlsize $kmlsize, ".
            "adjusting offsets\n";
        # be paranoid kml sizes may change!
        $self->{next_to_send}=$::psdev->get_kml_size($self->{fileset}->mtpt());
    }
    
    
    $self->{next_to_send};
}


sub fix_offsets {
    my $replicator = shift;
    my $fsetinfo = shift;
    
    confess "No fsetinfo or replicator\n"
         unless (defined $fsetinfo and defined $replicator);
    
    DEBUG "Fixing offsets\n";
    
    # our status of the fileset - important at reconnect time
    my ($r_kmlsize, $r_remote_recno, $r_remote_offset, 
    $r_local_recno, $r_local_offset) = unpack("NNNNN", $fsetinfo);
    
    DEBUG("fsetinfo= $r_kmlsize, $r_remote_recno, $r_remote_offset,". 
      " $r_local_recno, $r_local_offset\n");
    
    my ($local_last_state, $local_kmlgrowth, $remote_kmlgrowth)=
        $replicator->compute_kmlgrowth
        ($r_kmlsize, $r_remote_recno, $r_remote_offset, 
         $r_local_recno, $r_local_offset);
    
    DEBUG "last_state ". $local_last_state. ", remote_kmlgrowth "
        .$remote_kmlgrowth. ", local_kmlgrowth "
        .$local_kmlgrowth. "\n";
    
    if ($local_kmlgrowth < 0) {
        # We lost records. 
        # XXX: we no longer die! we should now take care of this by 
        # sending in a fsetinfo in the reply
        LOG "kml records were lost for fileset $replicator->{fsetname}\n"; 
    }
    
    if ($remote_kmlgrowth < 0) {
        # Downstream lost records. reset our database and journal
        # last_rcvd and recompute kml growth
        # Note that we must reverse the order of parameters here
        # since remote's remote is local here!
        # XXX: check this code! Also should we add a check for remote
        # kml size of zero? The only case where lost records make
        # sense is when the remote was reformatted, but things may change with
        # snapshots.
        $replicator->write_last_rcvd
            ($r_local_recno,$r_local_offset,$r_remote_recno,$r_remote_offset);
        my ($last_state,$local_kmlgrowth,$remote_kmlgrowth) =
            $replicator->compute_kmlgrowth($r_kmlsize,
                       $r_remote_recno,$r_remote_offset,
                       $r_local_recno,$r_local_offset);
        LOG "Remote lost kml records for fileset $replicator->{fsetname}\n";
        LOG "Lost records will be resent from this host\n";
        DEBUG " Recomputed: last_state " .$local_last_state. 
            ", remote_kmlgrowth " .$remote_kmlgrowth. ", local_kmlgrowth "
            .$local_kmlgrowth. "\n";
    }
    
    # this is the confirmed next to send. Note that remote's remote is us.
    $replicator->next_to_send($r_remote_offset);
    
    1;
}

# Direction support, replaces fileset states for sending and receiving
# What was the last direction in which kml was being sent for this replicator?
sub test_direction($$$$) {
    my $self=shift;
    my $fset = $self->{fileset};
    
    #use remote's nomenclature, which means that remote's remote is local here
    my ($r_remote_recno, $r_remote_offset, 
        $r_local_recno, $r_local_offset) = @_;
    
    my ($remote_recno,$remote_offset,$local_recno,$local_offset)=
        $self->last_rcvd();
    
    
    DEBUG "\nremote's last_local_offset =".$r_local_offset. 
          "\n         last_local_recno  =".$r_local_recno.
          "\n         last_remote_offset=".$r_remote_offset . 
          "\n         last_remote_recno =".$r_remote_recno. 
          "\n".$self->printme();
    
    # make sure that the last sender state is consistent: At most one of the
    # two peers should have the greater record numbers than the other.
    # This is a severe problem
    confess "record sequence error; both sides are the last sender\n"
        if (($remote_recno > $r_local_recno) && 
            ($local_recno < $r_remote_recno));
    
    # I was receiving if my last_recvd is greater than peer's last_rcvd
    my $local_last_state=($remote_recno > $r_local_recno) 
                            ? "receiving" : "sending";
    
    DEBUG "\n Local last state is $local_last_state\n";
    
    return $local_last_state; 
}

sub repstatus {
    my $replicator=shift;
    my ($last_remote_recno,$last_remote_offset,$last_local_recno,
        $last_local_offset) = $replicator->last_rcvd();
    my $local_kmlsize=$::psdev->get_kml_size($replicator->{fileset}->{mtpt});
    my $status = pack("NNNNN", $local_kmlsize, $last_remote_recno, 
                $last_remote_offset, $last_local_recno, $last_local_offset);

    return $status;
}

# Disconnection support for ReplicatorStatus
# By what extent did kml grow for this fileset wrt to this replicator?
sub compute_kmlgrowth($$$$$) {
    my $self=shift;
    my $remote_kmlsize=shift;

    my $fset = $self->{fileset};
    
    DEBUG "remote_kmlsize=$remote_kmlsize\n".$self->printme();
    
    my $local_last_state=$self->test_direction(@_);
    
    if (!defined $local_last_state) {
        LOG("Direction error: local last state could not be determined\n");
        return undef;
    }
    
    #use remote's nomenclature
    # Note that remote's remote is us
    my ($r_remote_recno, $r_remote_offset, 
        $r_local_recno, $r_local_offset) = @_;
    
    my ($remote_recno,$remote_offset,$local_recno,$local_offset)=
        $self->last_rcvd();
    my $local_kmlsize=$::psdev->get_kml_size($fset->mtpt());
    
    #Now calculate both kml size increases. 
    # Note that these could be negative indicating lost records
    my $local_kmlgrowth=0;
    my $remote_kmlgrowth=0;
    
    if ($local_last_state eq "sending") {
        # Peer's last rcvd is authoritative
        $local_kmlgrowth=$local_kmlsize - $r_remote_offset;
        $remote_kmlgrowth=$remote_kmlsize - $r_local_offset;
    } else {
        # Our last rcvd is authoritative
        $local_kmlgrowth=$local_kmlsize - $local_offset;
        $remote_kmlgrowth=$remote_kmlsize - $remote_offset;
    }
    
    DEBUG "local kml growth=$local_kmlgrowth\n";
    DEBUG "remote kml growth=$remote_kmlgrowth\n";

    return ($local_last_state,$local_kmlgrowth,$remote_kmlgrowth);
}
                      
1;
