package Lento::ReqDispatcher;

use Carp;
use POE qw(Wheel::ReadWrite Driver::SysRW);
use Socket;
use Carp;
use Lento::ACT;
use Lento::Debuggable;
use Lento::Filter::Upcall;
use Lento::Packet;
use Lento::InControl;
use Lento::InterMezzo::ReqHandler;
use Lento::InterMezzo::UpcallHandler;
use Lento::Server;
use Lento::Fileset;

@Lento::ReqDispatcher::ISA = qw(Lento::Debuggable);
$Lento::ReqDispatcher::level = 1;

=head1 NAME

ReqDispatcher - dispatch upcall and network request

=head1 SYNOPSIS

$::reqdispatcher = Lento::ReqDispatcher->new();

=head1 DESCRIPTION

Dispatches a network or upcall request, usually to a new session.
Owns the upcall wheel and handles kernel requests entirely.


=head1 Events handled

Events handled by this class are:

=over 4

=item  B<UPC>: uses an upcall wheel to get data from the kernel

=item  B<REQ>: network requests from Connections

=back

The requests it gets need to be sent to

=over 8

=item - an already active ReqSession expecting data

=item - a new session

=back


If the PacketWheel tells us that the connection was closed, this class
will:

=over 8

=item  - destroy all sessions set up by this ServerConnection

=item  - ask the ServerSemantics to remove the client

=item - self destruct

=back


=cut

######################################################################
#
# ReqDispatcher::new
#
# Create a new ReqDispatcher with:
#    the name of the class
#    the presto pseudo-device
#    the ReqHandler class (defaults to Lento::InterMezzo::ReqHandler::)
#
# Initializes:
#
# {handler_class} and {psdev} from the input args.
#
# {session}: Builds a new session for this dispatcher.  It tells the 
# session that it can handle the signals _start, _signal, REQ, UPC, 
# got_error and flushed.
#
######################################################################

sub new {
    my ($class, $psdev, $handler_class) = @_;
    my $self = bless { } , $class;

    #
    # The handler class comes from the input or has a default.
    #
    if ($handler_class) {
        $handler_class .= "::";
    } else {
        $handler_class = "Lento::InterMezzo::ReqHandler::";
    }

    #
    # Save the handler class and pseudo device.
    #
    $self->{handler_class} = $handler_class;
    LOG ">>>>> Starting ReqDispatcher\n";
    LOG ">>    Handler class: $self->{handler_class}\n";
    $self->{psdev} = $psdev;

    #
    # Build a new session that holds this ReqDispatcher.
    #
    $self->{session}=
      new POE::Session( $self => ['_start', '_signal', 'REQ', 'UPC', 
                                  'got_error', 'flushed' ], 
                        'Lento::EventSupport',  ['alias_set', 'alias_remove', 
                                                 'AddUpcallWaiter', 'RemoveUpcallWaiter',
                                                 'init_upcall_waiting', , 'get_cookie', 
                         'AddCookieWaiter', 'RespondWithCookie',
                         'WakeCookieList', '_child' ],
                        [ $self]
                        
                    );
    #
    # Return yourself.
    #
    LOG "<<<<< Starting ReqDispatcher\n";
    $self;
}

######################################################################
#
# ReqDispatcher::_start
#
# Called by POE once to set things up in the HEAP and SESSION 
# data structures.
#
# Sets up:
#    $_[HEAP]->{session_name}
#    $heap
#
#
######################################################################

sub _start {
    my ($object, $heap, $session) = @_[ARG0, HEAP, SESSION];

    $_[HEAP]->{session_name} = $_[SESSION]->ID().":ReqDispatcher";

    Lento::Debuggable::set_option($_[SESSION], trace => 1, default => 1);

    $_[KERNEL]->alias_set("ReqDispatcher $_[SESSION]");

    DEBUG("BUILDING WHEEL\n");
    if (!$object->{psdev}) {
            return; 
    }

    #
    # start the read/write wheel for the intermezzo device
    #
    if ($object->{psdev}) {
        $heap->{'wheel'} = new POE::Wheel::ReadWrite
          ( Handle  => $object->{psdev}->{handle}, # on this socket handle
            Driver  => new POE::Driver::SysRW(), # use sysread and syswrite
            Filter  => new Lento::Filter::Upcall(), # parse streams as packets
            InputState   => 'UPC', # generate this upon receipt of input
            ErrorState   => 'got_error',  # and this event if an error occurs
            FlushedState => 'flushed'  # and this event when all output is sent
          );
    }

    #
    # Tell the POE kernel to start waiting for upcalls, and set the
    # alias to be ReqDispatcher.
    #
    $_[KERNEL]->call($_[SESSION], 'init_upcall_waiting' );
}

=pod

=over 4

=item B<_signal> (Event)

Let us know what signal we got!

=back

=cut

# event: We don't do anything with signals other than report it. 
sub _signal {
    my $signal = $_[ARG0];
    DEBUG "++> $_[HEAP]->{session_name} got signal $signal\n";
}


=pod

=over 4

=item B<REQ> (Event)

This event is sent by a Connection when a REQ packet has come in. It
instantiates a new sessions to handle the request.

=back

=cut

sub REQ {
  #
  # Get the packet out of the input args.
  #
  my ($object, $heap, $packet) = @_[OBJECT, HEAP, ARG0];

  #
  # Get the name of the method to be called on this side.
  #
  my $method = shift @{$packet->[PKT_PARMS]};

  #
  # If the method is a SysId, we need to grab the sysid out of 
  # the packet.
  #
  if ( $method eq "SysId" ) {
      DEBUG "Req ($method) from %s\n", @{$packet->[PKT_PARMS]}[0];
      $packet->[PKT_CONNECTION]->{sysid} = @{$packet->[PKT_PARMS]}[0];;
  } else {      
      DEBUG "Req ($method) from %s\n", $packet->[PKT_CONNECTION]->{sysid};
  }

  #
  # Get the full name of the handler method, including the class name.
  #
  $method = $object->{handler_class} . $method;

  #
  # Get the session this request should go to.
  #
  my $session = &{"$method"}($packet);

  #
  # Initiate a new 'act' from this packet for $session.
  #
  $packet->[PKT_STOKEN] =
    $packet->[PKT_CONNECTION]->{act_table}->newact($session);
}

# Event handler: Process an upcall from presto.
sub UPC {
  #
  # Get the upcall info, and pull the method out of it.
  #
  my ($object, $heap, $upcall) = @_[OBJECT, HEAP, ARG0];
  my $method = $upcall->op();
  
  #
  # Print debug info if the test flag is set.
  #
  if ($object->{test}) {
    DEBUG $upcall->printme();
  }

  #
  # Get the session to send this upcall to using the method in
  # the upcall.
  #
  # For some reason we need to turn off strict refs to do this.
  #
  no strict "refs";
  $upcall->{session} =  
    &{"Lento::InterMezzo::UpcallHandler::$method"}($upcall);
  use strict "refs";
  DEBUG("In ReqDispatcher::UPC with seq no %d\n", $upcall->{uniq});
  
  #
  # Check if there are sessions waiting for this upcall.
  # XXX Bug 3 If there is nobody waiting for this upcall do we just 
  # drop this UPC on the ground?
  #
  if ( $_[HEAP]->{num_uc_waiting_sessions} > 0 ) {
    #
    # Get the seq number from the upcall and use that to wake up
    # the appropriate handler(?)
    #
    my $seq_no= $upcall->{uniq};
    if ( ! defined $seq_no ) {
      confess "ReqDispatcher::UPC did not define {uniq}!\n";
    }
    DEBUG("Waking up sessions\n");
    Lento::EventSupport::WakeUpcallList( $seq_no ); 
  }
  DEBUG("Out of ReqDispatcher::UPC with seq no %d\n", $upcall->{uniq});
}

# Method: ReqDispatcher::connectme
# Connect to the server for a given fileset with a given level of connection?
#
# Inputs:
#    session
#    fileset: From this we get the server for this fileset.
#    level: The level of the connection.  By setting the level for 
#           the connection we can specify crude priorities for messages
#           since messages at a higher(?) level won't go across this link?
sub connectme {
  #
  # Get the session, fileset and level from the input args.
  #
  my ($session, $fset, $level) = @_[SESSION, ARG0, ARG1];
  
  # XXXX BUG 4
  # This is supposed to make a decision which session
  # will finish the upcall request.  For now just plough on:
  
  # make sure we have a connection to the server(s)
  DEBUG "In connectme for $fset with level $level\n";
  if (!$fset or !$fset->server()) {
    $poe_kernel->yield('got_error',"connectme",ENOENT,"no server or fileset");
    return;
  }

  #
  # Make the connection.
  #
  $fset->server()->connectme( $level );
}


#######################################################################
#
# ReqDispatcher::got_error
#
# Handle I/O errors.  If we get an I/O error we print an error 
# message and delete the wheel, which will delete this session.
# 
#######################################################################

sub got_error {
  my ($object, $heap, $operation, $errnum, $errstr) = 
    @_[OBJECT, HEAP, ARG0, ARG1, ARG2];

  DEBUG("$object detected $operation error $errnum: $errstr\n");

  # The SocketFactory wheel is the only thing keeping this session
  # alive.  Deleting it causes its destructor to be called.  The
  # session has nothing further to do after the socketfactory is
  # destroyed, so the kernel stops it.

  delete $heap->{'wheel'};
}

######################################################################
#
# ReqDispatcher::flushed
#
# When all output is flushed, check to see if the client requested a
# quit.  If they did, then delete the wheel, which will delete this 
# session.
#
######################################################################

sub flushed {
  my $heap = $_[HEAP];

  if ($heap->{'protocol state'} eq 'quitting') {
    delete $heap->{'wheel'};
  }
}

#######################################################################
# EOF
#######################################################################
1;
