package Lento::Server;
use strict;
use POE;
use POSIX qw(:errno_h);
use XML::Simple;
use Lento::Connector;
use Lento::Pinger;
use Lento::Connection;
use Lento::Debuggable;
use Lento::EventSupport;
use Lento::List;
use Lento::Packet;
use Carp;

@Lento::Server::ISA = qw(Lento::Debuggable);
$Lento::Server::level = 1;


=head1 NAME

Server - objects for a server database and server management session.

=head1 SYNOPSIS

Describe a server and manage connections to it:

Lento::Server->setup_server_db($local)

=head1 DESCRIPTION

A Server object contains the name and port of the server to contact.
For each server a connection and connector object is maintained that
try to connect to the server.  A hash table maintains which filesets
can be found on a particular server.

The state machine for a server reconnection manager is as follows:

=over 4

=item * connecting: initial state and state after a got_error is handled. 

=item * binding: from receiving a got_connection event, where the
SysId REQ packet is sent to receiving the reply.

=item * syncing: during the initial reintegration

=item * online:

=cut


my %SERVERDB;

my $local;  # to use stuff without pingers etc

sub setup_server_db  {
    my $class = shift;
    my $localp = shift;  
    my $self = {};

    $local = $localp;
    $self->{srv_by_name} = {};

    my $database= $::configdir . "/serverdb";
    LOG(">>>>> Opening $database\n");

    my $serverhash = eval { XMLin($database, forcearray => 1)};
    die "Can't parse server database $database\n" if ($@);

    my $servers = $serverhash->{server};

    foreach  my $sysid ( keys %{$servers} ) {
        my $entry = $servers->{$sysid}; 
#        DEBUG("server= $sysid, " .
#              (defined ($entry->{port}) ? $entry->{port} : '') . ', ' .
#              (defined ($entry->{ipaddr}) ? $entry->{ipaddr} : '') . ", " . 
#             (defined ($entry->{bindaddr})?$entry->{binaddr} : '') . "\n");
        my $server = 
            new Lento::Server($sysid,
                              $entry->{port}, 
                              $entry->{ipaddr},
                              $entry->{bindaddr}
                              );
        $SERVERDB{$sysid} = $server; 
    }

    die "No servers found in serverdb\n" unless scalar %SERVERDB;

    LOG ">>>>> SERVERDB -- \n";
    foreach my $sysid ( keys %SERVERDB ) {
    my $server = $SERVERDB{$sysid};
    LOG ">>    SERVER $sysid, ip $server->{ipaddr}, port $server->{port}\n";
    }
    LOG "<<<<< SERVERDB -- \n";
}

sub find {
    my $sysid = shift || die "No sysid passed";

    return $SERVERDB{$sysid};
}


# XXX We should do a DNS lookup on the server if no IPADDR is given. 
# otherwise long hangs can occur for non-existent servers.
# perldoc Net::DNS has an example.
# how to do this asynchronously

sub new { 
    my $class = shift || die; 
    my $self = {}; bless $self,
    $class;

    $self->{name} = shift;
    $self->{port} = shift;
    $self->{ipaddr} = shift;
    $self->{bindaddr} = shift;
    $self->{fsets} = {};
    $self->{sessions} = new Lento::List;
    $self->{replist} = new Lento::List;
    $self->{port} = 369 unless $self->{port};
    if ($::noclient)  {
        return $self;
    }

    $self->{connection_obj} = new Lento::Connection($self->{name});

    LOG(">>    Server::new : $self->{name} ($::mysysid)(%d)\n", (!$local));
    if ( $self->{name} ne $::mysysid && ! $local) {
        $self->server_reconnect_mgr();
    }

    return $self;
}

sub server_reconnect_mgr {
    my $self = shift;

    $self->{manager} = new POE::Session
      ([$self], 
       "Lento::EventSupport", ['_signal', '_child', '_stop', 'alias_set',
                   'alias_remove', 'get_cookie'],
       _start => sub {
            my $server = $_[ARG0];
            $_[HEAP]->{server_obj} = $server;
            $_[HEAP]->{session_level} = 0;
            $_[HEAP]->{session_name} = $_[SESSION]->ID().
                            ": Reconnect Manager to " . $server->{name};
            Lento::Debuggable::set_option($_[SESSION],
                     trace => 1, default => 1);
            $_[KERNEL]->alias_set($_[SESSION]);
            DEBUG "Setting state to connecting\n";
            $server->{state} = 'connecting';
            $server->{connector} = new Lento::Connector 
                ($server->{connection_obj}, $server->{ipaddr},
                    $server->{port}, $::mysysip);
            LOG(">>>>  Reconnect manager to $server->{name} starting pinger\n");
            $server->{pinger} = new Lento::Pinger($server->{connector}, undef,
                                $_[SESSION]);
       },
       got_connection => sub {
            my $server = $_[HEAP]->{server_obj};
            $_[HEAP]->{session_level} = LVL_BIND;
            $_[HEAP]->{server_obj}->{connection_obj}->level(LVL_BIND);
            $_[HEAP]->{server_obj}->{connection_obj}->get();
      
            TACITUS("CONNECTING\n");

            if ($server->{state} eq 'connecting') {
                DEBUG "setting state to binding\n";
                $server->{state} = 'binding';
                LOG("   >> Connected to %s\n", $server->{name});
                LOG("   >> Binding with SysId %s\n", $::mysysid);
                
                $server->{connection_obj}->register_in_monitor_list
                        ($_[SESSION]); 
                $server->{connection_obj}->sendpacket
                        ('REQ', ['SysId', $::mysysid]);
            }
       },
       REP => sub {
            my $server = $_[HEAP]->{server_obj};
            my $packet = $_[ARG0];
            my $result = shift @{$packet->[PKT_PARMS]};

            $server->{connection_obj}->remove_from_monitor_list($_[SESSION]); 

            TACITUS( "BOUND\n");

            DEBUG " >>   Bind result $result\n";
            DEBUG "setting state to syncing\n";
            $server->{state} = 'syncing';
            $server->{connection_obj}->level(LVL_RECOVER);
            $_[KERNEL]->yield('_release_permits');
       },
       _release_permits => sub {
            my $server = $_[HEAP]->{server_obj};
            $_[HEAP]->{revokes_waiting}= $server->revoke_permits('_got_cookie');
            DEBUG " >>   $_[HEAP]->{revokes_waiting} revokes_waiting\n";
            if ( $_[HEAP]->{revokes_waiting} == 0 ) {
                $_[HEAP]->{revokes_waiting}= 1;
                $_[KERNEL]->yield('_got_cookie');
            }
       },
       _got_cookie => sub {
            $_[HEAP]->{revokes_waiting}--;
            if ($_[HEAP]->{revokes_waiting} == 0) {
                DEBUG("  >>  All permits revoked\n");
                $_[KERNEL]->yield('_synchronize_fsets');
            } else {
                DEBUG("  >>  Permit revoked: num remaining %d\n",
                    $_[HEAP]->{revokes_waiting} );
            }
       },
       _synchronize_fsets => sub {
            my $server = $_[HEAP]->{server_obj};
            $_[HEAP]->{syncs_waiting}= $server->sync_fsets('_sync_done');
            DEBUG " >>   $_[HEAP]->{syncs_waiting} syncs_waiting\n";
            if ( $_[HEAP]->{syncs_waiting} == 0 ) {
                $_[HEAP]->{syncs_waiting}= 1;
                $_[KERNEL]->yield('_sync_done');
            }
       },
       _sync_done => sub {
            $_[HEAP]->{syncs_waiting}--;
       
            if ($_[HEAP]->{syncs_waiting} == 0) {
                LOG("  >>  All filesets synchronised\n");
                $_[KERNEL]->yield('notify_sessions');
            } else {
                DEBUG("  >>  Fileset synced: num remaining %d\n",
                    $_[HEAP]->{syncs_waiting} );
            }
       },
       got_error => sub {
            my $operation=$_[ARG0];
            my $errnum=$_[ARG1];
            my $errstr=$_[ARG2];
    
            my $server = $_[HEAP]->{server_obj};
            DEBUG "got_error: setting state to connecting\n";
            $server->{state} = 'connecting';

            TACITUS("DISCONNECTED\n"); 

            if ( $server->{sessions} ) {
                $server->{sessions}->iterate
                ( sub { 
                           $poe_kernel->post(@_); 
                     }, 'got_error', 'server_mgr', 
                        ENOTCONN, 'failed to connect');
                $server->{sessions} = new Lento::List;
            }
       },
       notify_sessions => sub {
            my $server = $_[HEAP]->{server_obj};
            # fsets synced; recovery complete.
            DEBUG("  >>  Connection level set to FULL\n");
            $server->{connection_obj}->level(LVL_FULL);

            TACITUS("CONNECTED");

            # The unless is required since a got_error event might occur
            # just before notify_sessions. If this is not done, the system
            # will never do a GetPermit/GetML when the peer comes back online.
            # Another possible fix is to have Fileset::sync() loop when it
            # receives a got_error
            DEBUG "setting state to online\n";
            $server->{state} = 'online' 
                unless ($server->{state} eq 'connecting');

            $server->{sessions}->iterate
                ( sub {
                        $poe_kernel->post(@_);
                        $server->{connection_obj}->get($_[0]);
                       }, 'got_connection');
            # we are no longer responsible for these sessions
            $server->{sessions} = new Lento::List;
       }
      );
}

sub revoke_permits {
    my $server = shift;
    my $completion = shift || die;
    my $cookie_count= 0;
    my $session=$poe_kernel->get_active_session();

    DEBUG "In revoke_permits\n";
    
    $server->{replist}->iterate
        (
         sub {
             my $replicator = shift;
             my $fset = $::fsetdb->find_by_name($replicator->{fsetname});
             my $cookie=$poe_kernel->call($session,'get_cookie');
             
             DEBUG "revoking permit for  ". $replicator->{fsetname} . "\n";
             
             # the following lines are for partial caching
             # XXX: Restore when partial caches are in place. We need to fix
             # the OpenDir and OpenFile upcalls before we do this.
             #$::psdev->revoke_attr($fset->{mtpt});
             #$::psdev->revoke_data($fset->{mtpt});
             
             $fset->{permitholder}= "";
             my $kernel_cookie = $::psdev->revoke_permit
                 ($fset->{mtpt}, $cookie);
             if ( $kernel_cookie ) {
                 $cookie_count++;
                 Lento::EventSupport::WaitForCookie
                     ($kernel_cookie, $completion);
             } 
         }
        );
    return $cookie_count;
}

# XXX
# rewrite this routine such that 
# 1) we can wait for a permit on a cookie
#    it is safe to assume that this permit will be released quickly
# 2) we don't start reintegration on more than N filesets 
#    at the same time
#

sub sync_fsets {
    my $server = shift;
    my $rep_count= 0;
    my $completion = shift || die;

    DEBUG "In sync_fsets\n";
    
    $server->{replist}->iterate
      (
       sub {
        my $replicator = shift;
        my $connection = shift;
        my $fset = $::fsetdb->find_by_name($replicator->{fsetname});

        DEBUG "syncing ". $replicator->{fsetname} . "\n";
       
        my $sess= $fset->reconnect($replicator, $connection);
        if ( $sess ) {
            $rep_count++;
            Lento::EventSupport::WaitForSession($sess, $completion);
        } else {
            LOG("  >>  fset->sync no session ($replicator->{fsetname})\n");
        }

        undef $sess;
       }, $server->{connection_obj}
      );

    return $rep_count;
}

sub name ($) {
    my $self = shift;

    return $self->{name};
}

# client RPC routines get their connection object through this method
# they must register with the connection for errors to avoid
# lost error events bubbling up from the network layer
sub connectme {
    my $self = shift;
    my $level = shift;

    confess if (defined $level) && ($level eq '_got_permit');

    DEBUG "my state is ". $self->{state}."\n";

    if ( (defined $level) && ($level == LVL_RECOVER) ) {
        DEBUG( "  >>  Level is RECOVER\n");
        $self->{connection_obj}->get();
        $poe_kernel->yield('got_connection');    
    } elsif ( $self->{state} eq 'online' ) {
        $self->{connection_obj}->get();
        $poe_kernel->yield('got_connection');
    } elsif ( $self->{state} eq 'connecting' ) {
        # we are not yet connected so return error.
        $poe_kernel->yield('got_error', "connectme", ENOTCONN, "not connected");
    } elsif (($self->{state} eq 'syncing') || ($self->{state} eq 'binding')) {
        my $session= $poe_kernel->get_active_session();
        $self->{sessions}->add($session, $session->get_heap()->{session_name});
    } else { 
        confess "Bad state in server!\n";
    }
    return;
}

sub pushfset ($$) {
    my $self = shift;
    my $fset = shift;

    return $self->{fsets}->{$fset->name} = $fset;
}

sub printme ($) {
    my $self = shift;
    my $result;
    my $fsetname;
    $result = "Server \"$self->{name}\", state $self->{state} holds filesets: ";
    foreach $fsetname ( keys %{$self->{fsets}} ) { 
       $result .= $self->{fsets}->{$fsetname}->name . " ";
    };
    $result .= "\n";

    return $result;
}


sub printdb($) {
    my $self = shift;
    my $result;
    foreach my $name (keys %{$self->{srv_by_name}}) {
       $result .= $self->{srv_by_name}->{$name}->printme();
    }

    return $result;
}

1;

__END__

