/*
 *
 *
 *  Copyright (C) 2000 Stelias Computing, Inc.
 *  Copyright (C) 2000 Red Hat, Inc.
 *  Copyright (C) 2000 Tacitus Systems
 *  Copyright (C) 2000 Peter J. Braam
 *
 */


#include <stdarg.h>

#include <asm/bitops.h>
#include <asm/uaccess.h>
#include <asm/system.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/ext2_fs.h>
#include <linux/malloc.h>
#include <linux/vmalloc.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/blkdev.h>
#include <linux/init.h>
#define __NO_VERSION__
#include <linux/module.h>

#include <linux/intermezzo_fs.h>
#include <linux/intermezzo_upcall.h>
#include <linux/intermezzo_psdev.h>
/*
 * these are initialized in super.c
 */
extern int presto_permission(struct inode *inode, int mask);
int presto_ilookup_uid = 0;
extern int presto_opendir_upcall(int minor, struct dentry *de, int async);

extern int presto_prep(struct dentry *, struct presto_cache **,
                       struct presto_file_set **);

static inline void presto_setup_info(struct presto_cache *cache, 
                                    struct lento_vfs_context *info)
{
        memset(info, 0, sizeof(*info));
        info->flags = LENTO_FL_IGNORE_TIME;
        if (!ISLENTO(cache->cache_psdev->uc_minor))
                info->flags |= LENTO_FL_KML;
        
}

static int dentry2id(struct dentry *dentry, ino_t *id, unsigned int *generation)
{
        char *tmpname;
        char *next;
        int error = 0;

        ENTRY;
        if (dentry->d_name.len > EXT2_NAME_LEN) {
                EXIT;
                return -ENAMETOOLONG;
        }

        /* prefix is 7 characters: '...ino:' */
        if ( dentry->d_name.len < 7 ||
             memcmp(dentry->d_name.name, PRESTO_ILOOKUP_MAGIC, 7) != 0 ) {
                EXIT;
                return 1;
        }

        PRESTO_ALLOC(tmpname, char *, dentry->d_name.len - 7 + 1);
        if ( !tmpname ) {
                EXIT;
                return -ENOMEM;
        }

        memcpy(tmpname, dentry->d_name.name + 7, dentry->d_name.len - 7);
        *(tmpname + dentry->d_name.len) = '\0';

        /* name is of the form <inode number>:<generation> */
        *id = simple_strtoul(tmpname, &next, 0);
        if ( *next == PRESTO_ILOOKUP_SEP ) {
                *generation = simple_strtoul(next + 1, 0, 0);
                CDEBUG(D_INODE, "INO to find = %s\n", tmpname);
                CDEBUG(D_INODE, "Id = %lx (%lu), generation %x (%d)\n",
                       *id, *id, *generation, *generation);
        } else
                error = 1;

        PRESTO_FREE(tmpname, dentry->d_name.len - 7 + 1);
        EXIT;
        return error;
}


inline int presto_can_ilookup(void)
{
        return (current->euid == presto_ilookup_uid ||
                capable(CAP_DAC_READ_SEARCH));
}

struct dentry *presto_ilookup(struct inode *dir, struct dentry *dentry,
                            ino_t ino, unsigned int generation)
{
        struct inode *inode;
        int error;

        ENTRY;

        /* if we can't ilookup, forbid anything with this name to
         * avoid any security issues/name clashes.
         */
        if ( !presto_can_ilookup() ) {
                CDEBUG(D_CACHE, "ilookup denied: euid %u, ilookup_uid %u\n",
                       current->euid, presto_ilookup_uid);
                EXIT;
                return ERR_PTR(-EPERM);
        }
        inode = iget(dir->i_sb, ino);
        if (!inode || !inode->i_nlink || is_bad_inode(inode)) {
                CDEBUG(D_PIOCTL, "fatal: invalid inode %ld (%s).\n",
                       ino, inode ? inode->i_nlink ? "bad inode" :
                       "no links" : "NULL");
                inode->i_sb = NULL;
                clear_inode(inode);
                error = -ENOENT;
                EXIT;
                goto cleanup_iput;
        }

        /* We need to make sure we have the right inode (by checking the
         * generation) so we don't write into the wrong file (old inode was
         * deleted and then a new one was created with the same number).
         */
        if (inode->i_generation != generation) {
                CDEBUG(D_PIOCTL, "fatal: bad generation %u (want %u)\n",
                       inode->i_generation, generation);
                error = -ENOENT;
                EXIT;
                goto cleanup_iput;
        }

        d_instantiate(dentry, inode);
        EXIT;
        return NULL;

cleanup_iput:
        iput(inode);
        return ERR_PTR(error);
}


struct dentry *presto_lookup(struct inode * dir, struct dentry *dentry)
{
        int rc = 0;
        struct dentry *de;
        struct presto_cache *cache;
        int minor;
        ino_t ino;
        unsigned int generation;

        ENTRY;
        cache = presto_get_cache(dir);
        if ( !cache ) {
                printk("PRESTO: BAD, BAD: cannot find cache\n");
                make_bad_inode(dir);
                rc = -EBADF;
                goto exit;
        }
        minor = cache->cache_psdev->uc_minor;

        CDEBUG(D_CACHE, "dir ino: %ld, name: %*s\n",
               dir->i_ino, dentry->d_name.len, dentry->d_name.name);
        if ( ISLENTO(minor) )
                CDEBUG(D_CACHE, "We are lento\n");

        rc = dentry2id(dentry, &ino, &generation);
        CDEBUG(D_CACHE, "dentry2id returned %d\n", rc);
        if ( rc < 0 ) {
                EXIT;
                goto exit;
        }

        if ( rc == 0 ) {
                de = presto_ilookup(dir, dentry, ino, generation);
        } else {
                struct inode_operations *iops = filter_c2cdiops(cache->cache_filter);
                rc = 0;
                /* recursively do a cache lookup in dir */
                de = iops->lookup(dir, dentry);
        }
        /* XXX this needs some work to handle returning de if we get it */
        CDEBUG(D_CACHE, "\n");
        filter_setup_dentry_ops(cache->cache_filter, 
                                dentry->d_op, &presto_dentry_ops);
        dentry->d_op = filter_c2udops(cache->cache_filter);
        CDEBUG(D_CACHE, "\n");
        if ( IS_ERR(de) ) {
                rc = PTR_ERR(de);
                CDEBUG(D_CACHE, "dentry lookup error %d\n", rc);
        }
        CDEBUG(D_CACHE, "\n");

        /* get all entries and attrs for a new directory */
        if (!ISLENTO(minor) && dentry->d_inode &&
            !rc && !presto_chk(dentry, PRESTO_DATA) &&
            S_ISDIR(dentry->d_inode->i_mode)) {
                rc = presto_opendir_upcall(minor, dentry, SYNCHRONOUS);
                if ( rc == EISFSETROOT )
                        presto_set(dentry, PRESTO_FSETROOT);
                /* XXX what do we do with other errors ? */
                CDEBUG(D_CACHE, "setting ATTR & DATA for %ld\n",
                               dentry->d_inode->i_ino);
                presto_set(dentry, PRESTO_DATA | PRESTO_ATTR);
        } else if ( !rc || rc == -ENOENT ) {
                presto_set(dentry, PRESTO_ATTR);
        }
        EXIT;
exit:
        return ERR_PTR(rc);
}


/*
 *  in 2.3 the following inline functions are in dcache_func.h
 */


static inline void unlock_dir(struct dentry *dir)
{
        fs_up(&dir->d_inode->i_sem);
        dput(dir);
}

/*
 * Locking the parent is needed to:
 *  - serialize directory operations
 *  - make sure the parent doesn't change from
 *    under us in the middle of an operation.
 *
 * NOTE! Right now we'd rather use a "struct inode"
 * for this, but as I expect things to move toward
 * using dentries instead for most things it is
 * probably better to start with the conceptually
 * better interface of relying on a path of dentries.
 */
static inline struct dentry *lock_parent(struct dentry *dentry)
{
        struct dentry *dir = dget(dentry->d_parent);

        fs_down(&dir->d_inode->i_sem);
        return dir;
}

/*
 * Whee.. Deadlock country. Happily there is only one VFS
 * operation that does this..
 */
static inline void double_lock(struct dentry *d1, struct dentry *d2)
{
        struct semaphore *s1 = &d1->d_inode->i_sem;
        struct semaphore *s2 = &d2->d_inode->i_sem;

        if (s1 != s2) {
                if ((unsigned long) s1 < (unsigned long) s2) {
                        struct semaphore *tmp = s2;
                        s2 = s1; s1 = tmp;
                }
                fs_down(s1);
        }
        fs_down(s2);
}

static inline void double_unlock(struct dentry *d1, struct dentry *d2)
{
        struct semaphore *s1 = &d1->d_inode->i_sem;
        struct semaphore *s2 = &d2->d_inode->i_sem;

        fs_up(s1);
        if (s1 != s2)
                fs_up(s2);
        dput(d1);
        dput(d2);
}


/*
 *  Now the meat: the fs operations that require journaling
 *
 *
 *  XXX: some of these need modifications for hierarchical filesets
 */

int presto_prep(struct dentry *dentry, struct presto_cache **cache,
                struct presto_file_set **fset)
{
        *fset = presto_fset(dentry);
        if ( !*fset ) {
                CDEBUG(D_INODE, "No file set for dentry at %p\n", dentry);
                return -EROFS;
        }

        *cache = (*fset)->fset_cache;
        if ( !*cache ) {
                printk("PRESTO: BAD, BAD: cannot find cache\n");
                return -EBADF;
        }

        CDEBUG(D_PIOCTL, "---> cache flags %x, fset flags %x\n",
              (*cache)->cache_flags, (*fset)->fset_flags);
        if( presto_is_read_only(*fset) ) {
                printk("PRESTO: cannot modify read-only fileset, minor %d.\n",
                       (*cache)->cache_psdev->uc_minor);
                return -EROFS;
        }
        return 0;
}


static int presto_create(struct inode * dir, struct dentry * dentry, int mode)
{
        int error;
        struct presto_cache *cache;
        struct presto_file_set *fset;
        struct dentry *parent = dentry->d_parent;
        struct lento_vfs_context info;

        ENTRY;
        error = presto_prep(dentry, &cache, &fset);
        if ( error ) {
                EXIT;
                return error;
        }

        unlock_dir(parent);
        /* Does blocking and non-blocking behavious need to be 
           checked for.  Without blocking (return 1), the permit
           was acquired without reintegration
        */
        if ( presto_get_permit(dir) < 0 ) {
                EXIT;
                lock_parent(dentry);
                return -EROFS;
        }

        parent = lock_parent(dentry);
        presto_setup_info(cache, &info);
        error = presto_do_create(fset, parent, dentry, mode, &info);
        presto_put_permit(dir);

        EXIT;
        return error;
}

static int presto_link(struct dentry *old_dentry, struct inode *dir,
                struct dentry *new_dentry)
{
        int error;
        struct presto_cache *cache;
        struct presto_file_set *fset;
        struct dentry *parent = new_dentry->d_parent;
        struct lento_vfs_context info;

        ENTRY;
        error = presto_prep(old_dentry, &cache, &fset);
        if ( error ) {
                EXIT;
                return error;
        }

        unlock_dir(parent);
        if ( presto_get_permit(old_dentry->d_inode) < 0 ) {
                EXIT;
                lock_parent(new_dentry);
                return -EROFS;
        }

        if ( presto_get_permit(dir) < 0 ) {
                presto_put_permit(old_dentry->d_inode);
                EXIT;
                lock_parent(new_dentry);
                return -EROFS;
        }
        parent = lock_parent(new_dentry);
        presto_setup_info(cache, &info);
        error = presto_do_link(fset, old_dentry, parent, new_dentry, &info);
        presto_put_permit(dir);
        presto_put_permit(old_dentry->d_inode);

        EXIT;
        return error;
}

static int presto_mkdir(struct inode * dir, struct dentry * dentry, int mode)
{
        int error;
        struct presto_file_set *fset;
        struct presto_cache *cache;
        struct dentry *parent = dentry->d_parent;
        struct lento_vfs_context info;

        ENTRY;

        error = presto_prep(dentry, &cache, &fset);
        if ( error  ) {
                EXIT;
                return error;
        }

        /* do VFS style checking, until we have lookup contexts */
        unlock_dir(parent);
        if ( presto_get_permit(dir) < 0 ) {
                EXIT;
                lock_parent(dentry);
                return -EROFS;
        }
        parent = lock_parent(dentry);

        presto_setup_info(cache, &info);
        error = presto_do_mkdir(fset, parent, dentry, mode, &info);
        presto_put_permit(dir);

        EXIT;
        return error;
}


static int presto_symlink(struct inode *dir, struct dentry *dentry,
                   const char *name)
{
        int error;
        struct presto_cache *cache;
        struct presto_file_set *fset;
        struct dentry *parent = dentry->d_parent;
        struct lento_vfs_context info;

        ENTRY;
        error = presto_prep(dentry, &cache, &fset);
        if ( error ) {
                EXIT;
                return error;
        }

        unlock_dir(parent);
        if ( presto_get_permit(dir) < 0 ) {
                EXIT;
                lock_parent(dentry);
                return -EROFS;
        }
        parent = lock_parent(dentry);

        presto_setup_info(cache, &info);
        error = presto_do_symlink(fset, parent, dentry, name, &info);
        presto_put_permit(dir);

        EXIT;
        return error;
}

int presto_unlink(struct inode *dir, struct dentry *dentry)
{
        int error;
        struct presto_cache *cache;
        struct presto_file_set *fset;
        struct dentry *parent = dentry->d_parent;
        struct lento_vfs_context info;

        ENTRY;
        error = presto_prep(dentry, &cache, &fset);
        if ( error  ) {
                EXIT;
                return error;
        }

        unlock_dir(parent);
        if ( presto_get_permit(dir) < 0 ) {
                EXIT;
                lock_parent(dentry);
                return -EROFS;
        }
        parent = lock_parent(dentry);

        presto_setup_info(cache, &info);
        error = presto_do_unlink(fset, parent, dentry, &info);
        presto_put_permit(dir);

        EXIT;
        return error;
}

static int presto_rmdir(struct inode *dir, struct dentry *dentry)
{
        int error;
        struct presto_cache *cache;
        struct presto_file_set *fset;
        struct dentry *parent = dentry->d_parent;
        struct lento_vfs_context info;

        ENTRY;
        CDEBUG(D_FILE, "prepping presto\n");
        error = presto_prep(dentry, &cache, &fset);
        if ( error ) {
                EXIT;
                return error;
        }

        CDEBUG(D_FILE, "unlocking\n");
        /* We need to dget() before the dput in double_unlock, to ensure we
         * still have dentry references.  double_lock doesn't do dget for us.
         */
        dget(dentry);
        dget(dentry->d_parent);
        double_unlock(parent, dentry);

        CDEBUG(D_FILE, "getting permit\n");
        if ( presto_get_permit(parent->d_inode) < 0 ) {
                EXIT;
                double_lock(parent, dentry);
                return -EROFS;
        }
        CDEBUG(D_FILE, "locking\n");
        double_lock(parent, dentry);

        presto_setup_info(cache, &info);
        error = presto_do_rmdir(fset, parent, dentry, &info);
        presto_put_permit(parent->d_inode);

        EXIT;
        return error;
}

static int presto_mknod(struct inode * dir, struct dentry * dentry, int mode, int rdev)
{
        int error;
        struct presto_cache *cache;
        struct presto_file_set *fset;
        struct dentry *parent = dentry->d_parent;
        struct lento_vfs_context info;

        ENTRY;
        error = presto_prep(dentry, &cache, &fset);
        if ( error  ) {
                EXIT;
                return error;
        }

        unlock_dir(parent);
        if ( presto_get_permit(dir) < 0 ) {
                EXIT;
                lock_parent(dentry);
                return -EROFS;
        }
        parent = lock_parent(dentry);

        presto_setup_info(cache, &info);
        error = presto_do_mknod(fset, parent, dentry, mode, rdev, &info);
        presto_put_permit(dir);

        EXIT;
        return error;
}

int presto_rename(struct inode *old_dir, struct dentry *old_dentry,
                  struct inode *new_dir, struct dentry *new_dentry)
{
        int error;
        struct presto_cache *cache;
        struct presto_file_set *fset;
        struct dentry *old_parent = old_dentry->d_parent;
        struct dentry *new_parent = new_dentry->d_parent;
        struct lento_vfs_context info;

        ENTRY;
        error = presto_prep(old_dentry, &cache, &fset);
        if ( error ) {
                EXIT;
                return error;
        }

        /* We need to do dget before the dput in double_unlock, to ensure we
         * still have dentry references.  double_lock doesn't do dget for us.
         */
        dget(new_parent);
        dget(old_parent);
        double_unlock(new_parent, old_parent);
        if ( presto_get_permit(old_dir) < 0 ) {
                EXIT;
                double_lock(new_parent, old_parent);
                return -EROFS;
        }
        if ( presto_get_permit(new_dir) < 0 ) {
                presto_put_permit(old_dir);
                EXIT;
                double_lock(new_parent, old_parent);
                return -EROFS;
        }
        double_lock(new_parent, old_parent);

        presto_setup_info(cache, &info);
        error = presto_do_rename(fset, old_parent, old_dentry, new_parent,
                                 new_dentry, &info);

        presto_put_permit(new_dir);
        presto_put_permit(old_dir);

        EXIT;
        return error;
}

/* basically this allows the ilookup processes access to all files for
 * reading, while not making ilookup totally insecure.  This could all
 * go away if we could set the CAP_DAC_READ_SEARCH capability for the client.
 */
/* If posix acls are available, the underlying cache fs will export the
 * appropriate permission function. Thus we do not worry here about ACLs
 * or EAs. -SHP
 */
int presto_permission(struct inode *inode, int mask)
{
        unsigned short mode = inode->i_mode;
        struct presto_cache *cache;
        int rc;

        ENTRY;
        if ( presto_can_ilookup() && !(mask & S_IWOTH)) {
                CDEBUG(D_CACHE, "ilookup on %ld OK\n", inode->i_ino);
                EXIT;
                return 0;
        }

        cache = presto_get_cache(inode);

        if ( cache ) {
                /* we only override the file/dir permission operations */
                struct inode_operations *fiops = filter_c2cfiops(cache->cache_filter);
                struct inode_operations *diops = filter_c2cdiops(cache->cache_filter);

                if ( S_ISREG(mode) && fiops && fiops->permission ) {
                        EXIT;
                        return fiops->permission(inode, mask);
                }
                if ( S_ISDIR(mode) && diops && diops->permission ) {
                        EXIT;
                        return diops->permission(inode, mask);
                }
        }

        /* The cache filesystem doesn't have its own permission function,
         * but we don't want to duplicate the VFS code here.  In order
         * to avoid looping from permission calling this function again,
         * we temporarily override the permission operation while we call
         * the VFS permission function.
         */
        inode->i_op->permission = NULL;
        rc = permission(inode, mask);
        inode->i_op->permission = &presto_permission;

        EXIT;
        return rc;
}

int presto_opendir_upcall(int minor, struct dentry *de, int async)
{
        int rc;
        char *path, *buffer;
        int pathlen;

        PRESTO_ALLOC(buffer, char *, PAGE_SIZE);
        if ( !buffer ) {
                printk("PRESTO: out of memory!\n");
                return ENOMEM;
        }
        path = d_path(de, buffer, PAGE_SIZE);
        pathlen = MYPATHLEN(buffer, path);
        CDEBUG(D_INODE, "path: %*s, len %d\n", pathlen, path, pathlen);
        rc = lento_opendir(minor, pathlen, path, async);
        PRESTO_FREE(buffer, PAGE_SIZE);
        return rc;
}

static int presto_dir_open(struct inode *inode, struct file *file)
{
        int rc = 0;
        struct dentry *de = file->f_dentry;
        struct file_operations *fops;
        struct presto_cache *cache;
        int minor;

        ENTRY;

        cache = presto_get_cache(inode);
        if ( !cache ) {
                printk("PRESTO: BAD, BAD: cannot find cache\n");
                make_bad_inode(inode);
                EXIT;
                return -EBADF;
        }
        minor = cache->cache_psdev->uc_minor;

        CDEBUG(D_CACHE, "minor %d, DATA_OK: %d, ino: %ld\n",
               minor, presto_chk(de, PRESTO_DATA), inode->i_ino);

        if ( ISLENTO(minor) )
                goto cache;

        if ( !presto_chk(de, PRESTO_DATA) ) {
                CDEBUG(D_CACHE, "doing lento_opendir\n");
                rc = presto_opendir_upcall(minor, file->f_dentry, SYNCHRONOUS);
        }

        if ( rc ) {
                printk("presto_dir_open: DATA_OK: %d, ino: %ld, error %d\n",
                       presto_chk(de, PRESTO_DATA), inode->i_ino, rc);
                EXIT;
                return rc ;
        }

 cache:
        fops = filter_c2cdfops(cache->cache_filter);
        if ( fops->open ) {
                rc = fops->open(inode, file);
        }
        presto_set(de, PRESTO_DATA | PRESTO_ATTR);
        CDEBUG(D_CACHE, "returns %d, data %d, attr %d\n", rc,
               presto_chk(de, PRESTO_DATA), presto_chk(de, PRESTO_ATTR));

        EXIT;
        return 0;
}

static struct file_operations presto_dir_fops = {
        NULL,                   /* lseek - default */
        NULL,                   /* read */
        NULL,                   /* write - bad */
        NULL,                   /* readdir */
        NULL,                   /* poll - default */
        NULL,                   /* ioctl */
        NULL,                   /* mmap */
        presto_dir_open,        /* open */
        NULL,                   /* flush */
        NULL,                   /* no special release code */
        NULL,                   /* fsync */
        NULL,                   /* fasync */
        NULL,                   /* check_media_change */
        NULL                    /* revalidate */
};

struct inode_operations presto_dir_iops = {
        &presto_dir_fops,       /* no file-operations */
        presto_create,          /* create */
        presto_lookup,          /* lookup */
        presto_link,            /* link */
        presto_unlink,          /* unlink */
        presto_symlink,         /* symlink */
        presto_mkdir,           /* mkdir */
        presto_rmdir,           /* rmdir */
        presto_mknod,           /* mknod */
        presto_rename,          /* rename */
        NULL,                   /* readlink */
        NULL,                   /* follow_link */
        NULL,                   /* readpage */
        NULL,                   /* writepage */
        NULL,                   /* bmap */
        NULL,                   /* truncate */
        presto_permission,      /* permission */
        NULL,                   /* smap */
        NULL,                   /* updatepage */
        NULL,                   /* revalidate */
        NULL,                   /* prepare_write */
        NULL,                   /* sync_page */
#ifdef CONFIG_FS_EXT_ATTR
        NULL,                   /* get_ext_attr */
        presto_set_ext_attr,    /* set_ext_attr */
#endif
};
