/*
 * Super block/filesystem wide operations
 *
 * Copryright (C) 1996 Peter J. Braam <braam@maths.ox.ac.uk> and
 * Michael Callahan <callahan@maths.ox.ac.uk>
 *
 * Rewritten for Linux 2.1.  Peter Braam <braam@cs.cmu.edu>
 * Copyright (C) Carnegie Mellon University
 */

#define __NO_VERSION__
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/errno.h>
#include <linux/locks.h>
#include <linux/unistd.h>

#include <asm/system.h>
#include <asm/uaccess.h>

#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/errno.h>
#include <linux/locks.h>
#include <linux/string.h>
#include <asm/uaccess.h>
#include <linux/malloc.h>
#include <linux/vmalloc.h>
#include <asm/segment.h>

#include <linux/intermezzo_fs.h>
#include <linux/intermezzo_upcall.h>
#include <linux/intermezzo_psdev.h>

extern int presto_remount(struct super_block *, int *, char *);

int presto_excluded_gid = PRESTO_EXCL_GID;

extern int presto_prep(struct dentry *, struct presto_cache **,
                              struct presto_file_set **);
extern void presto_free_cache(struct presto_cache *);

void presto_read_inode(struct inode *inode)
{
        struct presto_cache *cache;

        cache = presto_get_cache(inode);
        if ( !cache ) {
                printk("PRESTO: BAD, BAD: cannot find cache\n");
                make_bad_inode(inode);
                return ;
        }

        filter_c2csops(cache->cache_filter)->read_inode(inode);

        CDEBUG(D_INODE, "presto_read_inode: ino %ld\n", inode->i_ino);

        if (S_ISREG(inode->i_mode)) {
                if ( !filter_c2cfiops(cache->cache_filter) ) {
                       filter_setup_file_ops(cache->cache_filter, 
                                              inode->i_op, &
                                              presto_file_iops);
                }
                if (inode->i_gid == presto_excluded_gid )
                        inode->i_op = filter_c2cfiops(cache->cache_filter);
                else
                        inode->i_op = filter_c2ufiops(cache->cache_filter);
                CDEBUG(D_INODE, "set file methods for %ld to %p\n",
                       inode->i_ino, inode->i_op);
        } else if (S_ISDIR(inode->i_mode)) {
                if (inode->i_gid == presto_excluded_gid )
                        inode->i_op = filter_c2cdiops(cache->cache_filter);
                else
                        inode->i_op = filter_c2udiops(cache->cache_filter);
                CDEBUG(D_INODE, "set dir methods for %ld to %p\n",
                       inode->i_ino, inode->i_op);
        } else if (S_ISLNK(inode->i_mode)) {
                if ( !filter_c2csiops(cache->cache_filter)) {
                        filter_setup_symlink_ops(cache->cache_filter, 
                                                 inode->i_op,
                                                 &presto_sym_iops);
                }
                if (inode->i_gid == presto_excluded_gid )
                        inode->i_op = filter_c2csiops(cache->cache_filter);
                else
                        inode->i_op = filter_c2usiops(cache->cache_filter);
                CDEBUG(D_INODE, "set link methods for %ld to %p\n",
                       inode->i_ino, inode->i_op);
        }

        /* XXX handle special inodes here or not? */

}

int presto_notify_change(struct dentry *de, struct iattr *iattr)
{
        int error;
        struct presto_cache *cache;
        struct presto_file_set *fset;
        struct lento_vfs_context info = { 0, 0, 0 };
        int sem_locked;

        ENTRY;
        error = presto_prep(de, &cache, &fset);
        if ( error ) {
                EXIT;
                return error;
        }

        CDEBUG(D_INODE, "valid %#x, mode %#o, uid %u, gid %u, size %lu, "
               "atime %lu mtime %lu ctime %lu flags %d\n",
               iattr->ia_valid, iattr->ia_mode, iattr->ia_uid, iattr->ia_gid,
               iattr->ia_size, iattr->ia_atime, iattr->ia_mtime,
               iattr->ia_ctime, iattr->ia_attr_flags);
        
        if (!ISLENTO(cache->cache_psdev->uc_minor)) {
                /* if this is a truncation, the process holds the i_sem,
                   but lento may need that for reintegration */ 
                sem_locked = down_trylock(&de->d_inode->i_sem); 
                CDEBUG(D_INODE, "sem_locked %d minor %d\n", sem_locked, 
                       cache->cache_psdev->uc_minor);
                /* 
                 *  it's locked now - so undo that
                 *  distinguish between the case we locked it or the fs did 
                 */ 
                if (sem_locked) { 
                        fs_up(&de->d_inode->i_sem); 
                } else { 
                        up(&de->d_inode->i_sem); 
                }
                if ( presto_get_permit(de->d_inode) < 0 ) {
                        EXIT;
                        if (sem_locked) {
                                CDEBUG(D_INODE, "relocking semaphore, %ld\n",
                                       de->d_inode->i_ino);
                                fs_down(&de->d_inode->i_sem);
                        }
                        return -EROFS;
                }
                if (sem_locked)  {
                        CDEBUG(D_INODE, "relocking semaphore, %ld, minor %d\n",
                               de->d_inode->i_ino,
                               cache->cache_psdev->uc_minor);
                        fs_down(&de->d_inode->i_sem);
                }
                info.flags = LENTO_FL_KML;
        }
        error = presto_do_setattr(fset, de, iattr, &info);
        
        if (!ISLENTO(cache->cache_psdev->uc_minor)) 
                presto_put_permit(de->d_inode);

        EXIT;
        return error;
}

void presto_put_super(struct super_block *sb)
{
        struct presto_cache *cache;
        struct upc_comm *psdev;
        struct super_operations *sops;
        struct list_head *lh;

        cache = presto_find_cache(sb->s_dev);
        if (!cache)
                goto exit;

        sops = filter_c2csops(cache->cache_filter);
        if (sops->put_super)
                sops->put_super(sb);

        /* free any remaining async upcalls when the filesystem is unmounted */
        psdev = cache->cache_psdev;
        if (!psdev) {
                EXIT;
                goto exit;
        }
        lh = psdev->uc_pending.next;
        while ( lh != &psdev->uc_pending) {
                struct upc_req *req;
                req = list_entry(lh, struct upc_req, rq_chain);

                /* assignment must be here: we are about to free &lh */
                lh = lh->next;
                if ( ! (req->rq_flags & REQ_ASYNC) ) 
                        continue;
                list_del(&(req->rq_chain));
                PRESTO_FREE(req->rq_data, req->rq_bufsize);
                PRESTO_FREE(req, sizeof(struct upc_req));
        }


exit:
        list_del(&cache->cache_chain); 
        list_del(&cache->cache_psdev_chain);
        presto_free_cache(cache);

        CDEBUG(D_MALLOC, "after umount: kmem %ld, vmem %ld\n",
               presto_kmemory, presto_vmemory);
        MOD_DEC_USE_COUNT;
        return ;
}


/* Like psdev.c (upc_comms), the initializer is just to prevent
   presto_sym_iops from appearing as a COMMON symbol (and therefore
   interfering with other modules that use the same variable name. */
struct inode_operations presto_sym_iops = {NULL}; /* NULL for now */

struct super_operations presto_super_ops = {
        read_inode:     presto_read_inode,
        put_super:      presto_put_super,
        remount_fs:     presto_remount,
    notify_change: presto_notify_change
};
