/*
 * vfs.c
 *
 * This file implements kernel downcalls from lento.
 *
 * Author: Rob Simmonds <simmonds@stelias.com>
 *         Andreas Dilger <adilger@stelias.com>
 * Copyright (C) 2000 Stelias Computing Inc
 * Copyright (C) 2000 Red Hat Inc.
 *
 * This code is based on code from namei.c in the linux file system;
 * see copyright notice below.
 *
 * Extended Attribute Support 
 * Copyright (C) 2001, Tacitus Systems, Inc.
 * Shirish H. Phatak
 */


/** namei.c copyright **/

/*
 *  linux/fs/namei.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */

/*
 * Some corrections by tytso.
 */

/* [Feb 1997 T. Schoebel-Theuer] Complete rewrite of the pathname
 * lookup logic.
 */

/** end of namei.c copyright **/

#include <linux/mm.h>
#include <linux/proc_fs.h>
#include <linux/smp_lock.h>
#include <linux/quotaops.h>

#include <asm/uaccess.h>
#include <asm/unaligned.h>
#include <asm/semaphore.h>
#include <asm/pgtable.h>

#include <asm/namei.h>

#include <linux/file.h>
#include <linux/fs.h>
#include <linux/blk.h>

#include <linux/intermezzo_fs.h>
#include <linux/intermezzo_kml.h>
#include <linux/intermezzo_upcall.h>
#include <linux/intermezzo_psdev.h>

#ifdef CONFIG_FS_EXT_ATTR
#include <linux/ext_attr.h>

#ifdef CONFIG_FS_POSIX_ACL
#include <linux/posix_acl.h>
#endif
#endif


extern struct inode_operations presto_sym_iops;

/*
 * It's inline, so penalty for filesystems that don't use sticky bit is
 * minimal.
 */
static inline int check_sticky(struct inode *dir, struct inode *inode)
{
        if (!(dir->i_mode & S_ISVTX))
                return 0;
        if (inode->i_uid == current->fsuid)
                return 0;
        if (dir->i_uid == current->fsuid)
                return 0;
        return !capable(CAP_FOWNER);
}

/*
 *      Check whether we can remove a link victim from directory dir, check
 *  whether the type of victim is right.
 *  1. We can't do it if dir is read-only (done in permission())
 *  2. We should have write and exec permissions on dir
 *  3. We can't remove anything from append-only dir
 *  4. We can't do anything with immutable dir (done in permission())
 *  5. If the sticky bit on dir is set we should either
 *      a. be owner of dir, or
 *      b. be owner of victim, or
 *      c. have CAP_FOWNER capability
 *  6. If the victim is append-only or immutable we can't do antyhing with
 *     links pointing to it.
 *  7. If we were asked to remove a directory and victim isn't one - ENOTDIR.
 *  8. If we were asked to remove a non-directory and victim isn't one - EISDIR.
 *  9. We can't remove a root or mountpoint.
 */
static inline int may_delete(struct inode *dir,struct dentry *victim, int isdir)
{
        int error;
        if (!victim->d_inode || victim->d_parent->d_inode != dir)
                return -ENOENT;
        error = permission(dir,MAY_WRITE | MAY_EXEC);
        if (error)
                return error;
        if (IS_APPEND(dir))
                return -EPERM;
        if (check_sticky(dir, victim->d_inode)||IS_APPEND(victim->d_inode)||
            IS_IMMUTABLE(victim->d_inode))
                return -EPERM;
        if (isdir) {
                if (!S_ISDIR(victim->d_inode->i_mode))
                        return -ENOTDIR;
                if (IS_ROOT(victim))
                        return -EBUSY;
                if (victim->d_mounts != victim->d_covers)
                        return -EBUSY;
        } else if (S_ISDIR(victim->d_inode->i_mode))
                return -EISDIR;
        return 0;
}

/* Check whether we can create an object with dentry child in directory dir.
 *  1. We can't do it if child already exists (open has special treatment for
 *     this case, but since we are inlined it's OK)
 *  2. We can't do it if dir is read-only (done in permission())
 *  3. We should have write and exec permissions on dir
 *  4. We can't do it if dir is immutable (done in permission())
 */
static inline int may_create(struct inode *dir, struct dentry *child)
{
        if (child->d_inode)
                return -EEXIST;
        return permission(dir,MAY_WRITE | MAY_EXEC);
}

static inline struct dentry *get_parent(struct dentry *dentry)
{
        return dget(dentry->d_parent);
}

static inline void unlock_dir(struct dentry *dir)
{
        fs_up(&dir->d_inode->i_sem);
        dput(dir);
}

/*
 * We need to do a check-parent every time
 * after we have locked the parent - to verify
 * that the parent is still our parent and
 * that we are still hashed onto it..
 *
 * This is requied in case two processes race
 * on removing (or moving) the same entry: the
 * parent lock will serialize them, but the
 * other process will be too late..
 */
#define check_parent(dir, dentry) \
        ((dir) == (dentry)->d_parent && !list_empty(&dentry->d_hash))

/*
 * Locking the parent is needed to:
 *  - serialize directory operations
 *  - make sure the parent doesn't change from
 *    under us in the middle of an operation.
 *
 * NOTE! Right now we'd rather use a "struct inode"
 * for this, but as I expect things to move toward
 * using dentries instead for most things it is
 * probably better to start with the conceptually
 * better interface of relying on a path of dentries.
 */
static inline struct dentry *lock_parent(struct dentry *dentry)
{
        struct dentry *dir = dget(dentry->d_parent);

        fs_down(&dir->d_inode->i_sem);
        return dir;
}

/*
 * Whee.. Deadlock country. Happily there are only two VFS
 * operations that do this..
 */
static inline void double_lock(struct dentry *d1, struct dentry *d2)
{
        struct semaphore *s1 = &d1->d_inode->i_sem;
        struct semaphore *s2 = &d2->d_inode->i_sem;

        if (s1 != s2) {
                if ((unsigned long) s1 < (unsigned long) s2) {
                        struct semaphore *tmp = s2;
                        s2 = s1; s1 = tmp;
                }
                fs_down(s1);
        }
        fs_down(s2);
}

static inline void double_unlock(struct dentry *d1, struct dentry *d2)
{
        struct semaphore *s1 = &d1->d_inode->i_sem;
        struct semaphore *s2 = &d2->d_inode->i_sem;

        fs_up(s1);
        if (s1 != s2)
                fs_up(s2);
        dput(d1);
        dput(d2);
}


#ifdef PRESTO_DEBUG
/* The loop_discard_io() function is available via a kernel patch to the
 * loop block device.  It "works" by accepting writes, but throwing them
 * away, rather than trying to write them to disk.  The old method worked
 * by setting the underlying device read-only, but that has the problem
 * that dirty buffers are kept in memory, and ext3 didn't like that at all.
 */
#ifdef CONFIG_LOOP_DISCARD
#define BLKDEV_FAIL(dev,fail) loop_discard_io(dev,fail)
#else
#define BLKDEV_FAIL(dev,fail) set_device_ro(dev, 1)
#endif

/* If a breakpoint has been set via /proc/sys/intermezzo/intermezzoX/errorval,
 * that is the same as "value", the underlying device will "fail" now.
 */
inline void presto_debug_fail_blkdev(struct presto_file_set *fset,
                                     unsigned long value)
{
        int minor = fset->fset_cache->cache_psdev->uc_minor;
        int errorval = upc_comms[minor].uc_errorval;
        kdev_t dev = fset->fset_mtpt->d_inode->i_dev;

        if (errorval && errorval == (long)value && !is_read_only(dev)) {
                CDEBUG(D_SUPER, "setting device %s read only\n", kdevname(dev));
                BLKDEV_FAIL(dev, 1);
                upc_comms[minor].uc_errorval = -dev;
        }
}
#else
#define presto_debug_fail_blkdev(dev,value) do {} while (0)
#endif


static inline int presto_do_kml(struct lento_vfs_context *info, struct inode* inode)
{
        if ( ! (info->flags & LENTO_FL_KML) ) 
                return 0;
        if ( inode->i_gid == presto_excluded_gid ) 
                return 0;
        return 1;
}

static inline int presto_do_expect(struct lento_vfs_context *info, struct inode *inode)
{
        if ( ! (info->flags & LENTO_FL_EXPECT) ) 
                return 0;
        if ( inode->i_gid == presto_excluded_gid ) 
                return 0;
        return 1;
}

int presto_settime(struct presto_file_set *fset, 
                   struct dentry *dentry, 
                   struct lento_vfs_context *ctx, 
                   int valid)
{
        int error; 
        struct inode *inode = dentry->d_inode;
        struct super_operations *sops;
        struct iattr iattr;

        ENTRY;
        if (ctx->flags &  LENTO_FL_IGNORE_TIME ) { 
                EXIT;
                return 0;
        }
        iattr.ia_ctime = ctx->updated_time;
        iattr.ia_mtime = ctx->updated_time;
        iattr.ia_valid = valid;

        error = -EROFS;
        if (IS_RDONLY(inode)) {
                EXIT;
                return -EROFS;
        }

        if (IS_IMMUTABLE(inode) || IS_APPEND(inode)) {
                EXIT;
                return -EPERM;
        }

        error = -EPERM;
        sops = filter_c2csops(fset->fset_cache->cache_filter); 
        if (!sops &&
            !sops->notify_change) {
                EXIT;
                return error;
        }

        error = sops->notify_change(dentry, &iattr);
        EXIT;
        return error;
}

int presto_do_setattr(struct presto_file_set *fset, struct dentry *dentry,
                      struct iattr *iattr, struct lento_vfs_context *info)
{
        struct rec_info rec;
        struct inode *inode = dentry->d_inode;
        struct super_operations *sops;
        int error;
        struct presto_version old_ver;
        void *handle;

        ENTRY;
        error = -EROFS;
        if (IS_RDONLY(inode)) {
                EXIT;
                return -EROFS;
        }

        if (IS_IMMUTABLE(inode) || IS_APPEND(inode)) {
                EXIT;
                return -EPERM;
        }

        presto_getversion(&old_ver, dentry->d_inode);
        error = -EPERM;
        sops = filter_c2csops(fset->fset_cache->cache_filter); 
        if (!sops &&
            !sops->notify_change) {
                EXIT;
                return error;
        }

	error = presto_reserve_space(fset->fset_cache, PRESTO_REQHIGH); 
	if (error) {
		EXIT;
		return error;
	}
	
	if  (iattr->ia_valid & ATTR_SIZE) { 
		handle = presto_trans_start(fset, dentry->d_inode, PRESTO_OP_TRUNC);
	} else {
		handle = presto_trans_start(fset, dentry->d_inode, PRESTO_OP_SETATTR);
	}
        if ( IS_ERR(handle) ) {
                printk("presto_do_setattr: no space for transaction\n");
		presto_release_space(fset->fset_cache, PRESTO_REQHIGH); 
                return -ENOSPC;
        }
        error = sops->notify_change(dentry, iattr);
        if (error) {
                EXIT;
                goto exit;
        }

        presto_debug_fail_blkdev(fset, PRESTO_OP_SETATTR | 0x10);
        if ( presto_do_kml(info, dentry->d_inode) )
                error = presto_journal_setattr(&rec, fset, dentry, &old_ver, iattr);

        presto_debug_fail_blkdev(fset, PRESTO_OP_SETATTR | 0x20);
        if ( presto_do_expect(info, dentry->d_inode) )
                error = presto_write_last_rcvd(&rec, fset, info);

	if ( !error && (iattr->ia_valid & ATTR_SIZE) ) { 
		vmtruncate(inode, inode->i_size);
		if (inode->i_op && inode->i_op->truncate)
			inode->i_op->truncate(inode);
	}
	
        presto_debug_fail_blkdev(fset, PRESTO_OP_SETATTR | 0x30);
        EXIT;
exit:
	presto_release_space(fset->fset_cache, PRESTO_REQHIGH); 
        presto_trans_commit(fset, handle);
        return error;
}

int lento_setattr(const char *name, struct iattr *iattr,
                  struct lento_vfs_context *info)
{
        struct dentry *dentry;
        struct presto_file_set *fset;
        int error;
#ifdef  CONFIG_FS_POSIX_ACL
        int (*set_posix_acl)(struct inode *, int type, posix_acl_t *)=NULL;
#endif

        ENTRY;
        CDEBUG(D_PIOCTL,"name %s, valid %#x, mode %#o, uid %d, gid %d, size %ld\n",
               name, iattr->ia_valid, iattr->ia_mode, iattr->ia_uid,
               iattr->ia_gid, (long) iattr->ia_size);
        CDEBUG(D_PIOCTL, "atime %#lx, mtime %#lx, ctime %#lx, attr_flags %#x\n",
               iattr->ia_atime, iattr->ia_mtime, iattr->ia_ctime,
               iattr->ia_attr_flags);
        CDEBUG(D_PIOCTL, "offset %d, recno %d, flags %#x\n",
               info->slot_offset, info->recno, info->flags);
        lock_kernel();
        dentry = lnamei(name);
        error = PTR_ERR(dentry);
        if (IS_ERR(dentry)) {
                EXIT;
                goto exit;
        }

        fset = presto_fset(dentry);
        error = -EINVAL;
        if ( !fset ) {
                printk("No fileset!\n");
                EXIT;
                goto exit_lock;
        }

        /* NOTE: this prevents us from changing the filetype on setattr,
         *       as we normally only want to change permission bits.
         *       If this is not correct, then we need to fix the perl code
         *       to always send the file type OR'ed with the permission.
         */
        if (iattr->ia_valid & ATTR_MODE) {
                int set_mode = iattr->ia_mode;
                iattr->ia_mode = (iattr->ia_mode & S_IALLUGO) |
                                 (dentry->d_inode->i_mode & ~S_IALLUGO);
                CDEBUG(D_PIOCTL, "chmod: orig %#o, set %#o, result %#o\n",
                       dentry->d_inode->i_mode, set_mode, iattr->ia_mode);
#ifdef CONFIG_FS_POSIX_ACL
                /* ACl code interacts badly with setattr 
                 * since it tries to modify the ACL using 
                 * set_ext_attr which recurses back into presto.  
                 * This only happens if ATTR_MODE is set.
                 * Here we are doing a "forced" mode set 
                 * (initiated by lento), so we disable the 
                 * set_posix_acl operation which 
                 * prevents such recursion.  -SHP
                 *
                 * This will probably still be required when native
                 * acl journalling is in place.
                 */
                set_posix_acl=dentry->d_inode->i_op->set_posix_acl;
                dentry->d_inode->i_op->set_posix_acl=NULL;
#endif
        }
        error = presto_do_setattr(fset, dentry, iattr, info);
#ifdef CONFIG_FS_POSIX_ACL
        /* restore the inode_operations if we changed them*/
        if (iattr->ia_valid & ATTR_MODE) 
                dentry->d_inode->i_op->set_posix_acl=set_posix_acl;
#endif

        EXIT;
exit_lock:
        dput(dentry);
exit:
        unlock_kernel();
        return error;
}



int presto_do_statfs (struct presto_file_set *fset, 
                   struct statfs * buf, int bufsiz)
{
        struct super_operations *sops;
        struct super_block *sb;
        int result;
        ENTRY;

        if ( !fset ) {
                EXIT;
                return -EINVAL;
        }
        if ( !fset->fset_cache ) {
                EXIT;
                return -EINVAL;
        }
        if ( !fset->fset_cache->cache_filter ) {
                EXIT;
                return -EINVAL;
        }

        sops = filter_c2csops(fset->fset_cache->cache_filter);
        if ( ! sops ) {
                EXIT;
                return -EINVAL;
        }
        if ( ! fset->fset_cache->cache_mtde ) {
                EXIT;
                return -EINVAL;
        }

        if ( ! fset->fset_cache->cache_mtde->d_inode ) {
                EXIT;
                return -EINVAL;
        }

        if ( ! fset->fset_cache->cache_mtde->d_inode->i_sb ) {
                EXIT;
                return -EINVAL;
        }
        sb = fset->fset_cache->cache_mtde->d_inode->i_sb;

        if (sops->statfs) {
                mm_segment_t old_fs = get_fs();

                return 0; 
                set_fs(get_ds());
                result = sops->statfs(sb, buf, bufsiz);
                set_fs(old_fs);
        } else {
                result = -EINVAL;
        }

        return result;
}



int presto_do_create(struct presto_file_set *fset, struct dentry *dir,
                     struct dentry *dentry, int mode,
                     struct lento_vfs_context *info)
{
        struct rec_info rec;
        int error;
        struct presto_version tgt_dir_ver, new_file_ver;
        struct inode_operations *iops;
        void *handle;

        ENTRY;
        /*
         * EEXIST is kind of a strange error code to
         * return, but basically if the dentry was moved
         * or unlinked while we locked the parent, we
         * do know that it _did_ exist before, and as
         * such it makes perfect sense.. In contrast,
         * ENOENT doesn't make sense for create.
         */
        error = -EEXIST;
        if (!check_parent(dir, dentry)) {
                EXIT;
                return error;
        }

#if 0   
	/* XXX the VFS needs to give us the flag to the open method! */
        /*
         * Somebody might have created the file while we
         * waited for the directory lock.. So we have to
         * re-do the existence test.
         */
        if (dentry->d_inode) {
                error = 0;
                if (flag & O_EXCL) {
                        EXIT;
                        return -EEXIST;
                }
        }
#endif

        error = may_create(dir->d_inode, dentry);
        if (error) { 
                /* there is a tricky case here: the file was created
                   during reintegration and we would get -EEXIST here
                */
                if (error == -EEXIST && !ISLENTO(fset->fset_cache->cache_psdev->uc_minor) ) {
                        EXIT;
                        return 0;
                }
                EXIT;
                return error;
        }

        error = -EPERM;
        iops = filter_c2cdiops(fset->fset_cache->cache_filter);
        if (!iops->create) {
                EXIT;
                return error;
        }
	error = presto_reserve_space(fset->fset_cache, PRESTO_REQHIGH); 
	if (error) {
		EXIT;
		return error;
	}

        presto_getversion(&tgt_dir_ver, dir->d_inode);
        handle = presto_trans_start(fset, dir->d_inode, PRESTO_OP_CREATE);
        if ( IS_ERR(handle) ) {
		presto_release_space(fset->fset_cache, PRESTO_REQHIGH); 
                printk("presto_do_create: no space for transaction\n");
                return -ENOSPC;
        }
        DQUOT_INIT(dir->d_inode);
        error = iops->create(dir->d_inode, dentry, mode);
        if (error) {
                EXIT;
                goto exit;
        }

        if (dentry->d_inode && 
            dentry->d_inode->i_gid != presto_excluded_gid) {
                struct presto_cache *cache = fset->fset_cache;
                /* was this already done? */
                if ( !filter_c2cfiops(cache->cache_filter) )
                        filter_setup_file_ops(cache->cache_filter, 
                                              dentry->d_inode->i_op,
                                              &presto_file_iops);

                /* make the new inode ours */
                dentry->d_inode->i_op = 
                        filter_c2ufiops(cache->cache_filter);
                filter_setup_dentry_ops(cache->cache_filter, 
                                        dentry->d_op, 
                                        &presto_dentry_ops);
                dentry->d_op = filter_c2udops(cache->cache_filter);

                /* if Lento creates this file, we won't have data */
                if ( ISLENTO(cache->cache_psdev->uc_minor) ) {
                        presto_set(dentry, PRESTO_ATTR);
                } else {
                        presto_set(dentry, PRESTO_ATTR | PRESTO_DATA);
                }
        }

        error = presto_settime(fset, dir, info, ATTR_CTIME | ATTR_MTIME);
        if (error) { 
                EXIT;
                goto exit;
        }
        error = presto_settime(fset, dentry, info, ATTR_CTIME | ATTR_MTIME);
        if (error) { 
                EXIT;
                goto exit;
        }

        presto_debug_fail_blkdev(fset, PRESTO_OP_CREATE | 0x10);
        presto_getversion(&new_file_ver, dentry->d_inode);
        /* if acl support is enabled, i_mode gets modified by 
         * posix_acl_init_cred. We need to journal this change. So instead
         * of passing mode to presto_journal_create, we pass i_mode. -SHP
         */
        if ( presto_do_kml(info, dentry->d_inode) )
                error = presto_journal_create(&rec, fset, dentry, &tgt_dir_ver,                                              &new_file_ver, 
                                             dentry->d_inode->i_mode);

        presto_debug_fail_blkdev(fset, PRESTO_OP_CREATE | 0x20);
        if ( presto_do_expect(info, dentry->d_inode) )
                error = presto_write_last_rcvd(&rec, fset, info);

        presto_debug_fail_blkdev(fset, PRESTO_OP_CREATE | 0x30);
        EXIT;
exit:
	presto_release_space(fset->fset_cache, PRESTO_REQHIGH); 
        presto_trans_commit(fset, handle);
        return error;
}

int lento_create(const char *name, int mode, struct lento_vfs_context *info)
{
        int error;
        char * pathname;
        struct dentry *dir;
        struct dentry *dentry;
        struct presto_file_set *fset;

        ENTRY;
        lock_kernel();
        pathname = getname(name);
        error = PTR_ERR(pathname);
        if (IS_ERR(pathname)) {
                EXIT;
                goto exit;
        }

        dentry = lookup_dentry(pathname, NULL, LOOKUP_SLASHOK);
        error = PTR_ERR(dentry);
        if (IS_ERR(dentry)) {
                EXIT;
                goto exit_put;
        }

        dir = lock_parent(dentry);

        fset = presto_fset(dentry);
        error = -EINVAL;
        if ( !fset ) {
                printk("No fileset!\n");
                EXIT;
                goto exit_lock;
        }
        error = presto_do_create(fset, dir, dentry, (mode&S_IALLUGO)|S_IFREG,
                                 info);

        EXIT;
exit_lock:
        unlock_dir(dir);
        dput(dentry);
exit_put:
        putname(pathname);
exit:
        unlock_kernel();
        return error;
}


int presto_do_link(struct presto_file_set *fset, struct dentry *old_dentry,
                   struct dentry *dir, struct dentry *new_dentry,
                   struct lento_vfs_context *info)
{
        struct rec_info rec;
        int error;
        struct inode *inode;
        struct inode_operations *iops;
        struct presto_version tgt_dir_ver, new_link_ver;
        void *handle;

        ENTRY;
        error = -ENOENT;
        if (!check_parent(dir, new_dentry)) {
                EXIT;
                return error;
        }

        error = -ENOENT;
        inode = old_dentry->d_inode;
        if (!inode) {
                EXIT;
                return error;
        }

        error = may_create(dir->d_inode, new_dentry);
        if (error) {
                EXIT;
                return error;
        }

        error = -EXDEV;
        if (dir->d_inode->i_dev != inode->i_dev) {
                EXIT;
                return error;
        }

        /*
         * A link to an append-only or immutable file cannot be created.
         */
        error = -EPERM;
        if (IS_APPEND(inode) || IS_IMMUTABLE(inode)) {
                EXIT;
                return error;
        }

        error = -EPERM;
        iops = filter_c2cdiops(fset->fset_cache->cache_filter);
        if (!iops->link) {
                EXIT;
                return error;
        }

	error = presto_reserve_space(fset->fset_cache, PRESTO_REQHIGH); 
	if (error) {
		EXIT;
		return error;
	}
        presto_getversion(&tgt_dir_ver, dir->d_inode);
        handle = presto_trans_start(fset, dir->d_inode, PRESTO_OP_LINK);
        if ( IS_ERR(handle) ) {
		presto_release_space(fset->fset_cache, PRESTO_REQHIGH); 
                printk("presto_do_link: no space for transaction\n");
                return -ENOSPC;
        }

        DQUOT_INIT(dir->d_inode);
        error = iops->link(old_dentry, dir->d_inode, new_dentry);
        if (error) {
                EXIT;
                goto exit;
        }

        error = presto_settime(fset, dir, info, ATTR_CTIME | ATTR_MTIME);
        if (error) { 
                EXIT;
                goto exit;
        }
        error = presto_settime(fset, new_dentry, info, ATTR_CTIME);
        if (error) { 
                EXIT;
                goto exit;
        }
        presto_debug_fail_blkdev(fset, PRESTO_OP_LINK | 0x10);
        presto_getversion(&new_link_ver, new_dentry->d_inode);
        if ( presto_do_kml(info, old_dentry->d_inode) )
                error = presto_journal_link(&rec, fset, old_dentry, new_dentry,
                                            &tgt_dir_ver, &new_link_ver);

        presto_debug_fail_blkdev(fset, PRESTO_OP_LINK | 0x20);
        if ( presto_do_expect(info, old_dentry->d_inode) )
                error = presto_write_last_rcvd(&rec, fset, info);

        presto_debug_fail_blkdev(fset, PRESTO_OP_LINK | 0x30);
        EXIT;
exit:
	presto_release_space(fset->fset_cache, PRESTO_REQHIGH); 
        presto_trans_commit(fset, handle);

        return error;
}

int lento_link(const char *oldname, const char *newname,
               struct lento_vfs_context *info)
{
        int error;
        char *from;
        char *to;
        struct dentry *old_dentry, *new_dentry;
        struct dentry *dir;
        struct presto_file_set *fset;

        ENTRY;
        lock_kernel();
        from = getname(oldname);
        error = PTR_ERR(from);
        if (IS_ERR(from)) {
                EXIT;
                goto exit;
        }

        to = getname(newname);
        error = PTR_ERR(to);
        if (IS_ERR(to)) {
                EXIT;
                goto exit_from;
        }

        /*
         * Hardlinks are often used in delicate situations.  We avoid
         * security-related surprises by not following symlinks on the
         * newname.  --KAB
         *
         * We don't follow them on the oldname either to be compatible
         * with linux 2.0, and to avoid hard-linking to directories
         * and other special files.  --ADM
         */
        old_dentry = lookup_dentry(oldname, NULL, 0);
        error = PTR_ERR(old_dentry);
        if (IS_ERR(old_dentry)) {
                EXIT;
                goto exit_to;
        }

        new_dentry = lookup_dentry(newname, NULL, 0);
        error = PTR_ERR(new_dentry);
        if (IS_ERR(new_dentry)) {
                EXIT;
                goto exit_old;
        }

        dir = lock_parent(new_dentry);

        fset = presto_fset(new_dentry);
        error = -EINVAL;
        if ( !fset ) {
                printk("No fileset!\n");
                EXIT;
                goto exit_lock;
        }

        error = presto_do_link(fset, old_dentry, dir, new_dentry, info);
        EXIT;
exit_lock:
        unlock_dir(dir);
        dput(new_dentry);
exit_old:
        dput(old_dentry);
exit_to:
        putname(to);
exit_from:
        putname(from);
exit:
        unlock_kernel();
        return error;
}


int presto_do_unlink(struct presto_file_set *fset, struct dentry *dir,
                     struct dentry *dentry, struct lento_vfs_context *info)
{
        struct rec_info rec;
        int error;
        int linkno;
        struct inode_operations *iops;
        struct presto_version tgt_dir_ver, old_file_ver;
        void *handle;
        int do_kml, do_expect;

        ENTRY;
        error = -ENOENT;
        if (!check_parent(dir, dentry)) {
                EXIT;
                return error;
        }

        error = may_delete(dir->d_inode, dentry, 0);
        if (error) {
                EXIT;
                return error;
        }

        error = -EPERM;
        iops = filter_c2cdiops(fset->fset_cache->cache_filter);
        if (!iops->unlink) {
                EXIT;
                return error;
        }

	error = presto_reserve_space(fset->fset_cache, PRESTO_REQLOW); 
	if (error) {
		EXIT;
		return error;
	}

        presto_getversion(&tgt_dir_ver, dir->d_inode);
        presto_getversion(&old_file_ver, dentry->d_inode);
        handle = presto_trans_start(fset, dir->d_inode, PRESTO_OP_UNLINK);
        if ( IS_ERR(handle) ) {
		presto_release_space(fset->fset_cache, PRESTO_REQLOW); 
                printk("ERROR: presto_do_unlink: no space for transaction. Tell Peter.\n");
                return -ENOSPC;
        }
        DQUOT_INIT(dir->d_inode);
        do_kml = presto_do_kml(info, dir->d_inode);
        do_expect = presto_do_expect(info, dir->d_inode);

        linkno = dentry->d_inode->i_nlink;
        if (linkno > 1) {
                dget(dentry);
        }

        error = iops->unlink(dir->d_inode, dentry);
        if (error) {
                EXIT;
                goto exit;
        }

        if (linkno > 1) { 
                error = presto_settime(fset, dentry, info, ATTR_CTIME);
                dput(dentry); 
                if (error) { 
                        EXIT;
                        goto exit;
                }
        }

        error = presto_settime(fset, dir, info, ATTR_CTIME | ATTR_MTIME);
        if (error) { 
                EXIT;
                goto exit;
        }


        presto_debug_fail_blkdev(fset, PRESTO_OP_UNLINK | 0x10);
        if ( do_kml )
                error = presto_journal_unlink(&rec, fset, dir, &tgt_dir_ver,
                                              &old_file_ver,
                                              dentry->d_name.len,
                                              dentry->d_name.name);

        presto_debug_fail_blkdev(fset, PRESTO_OP_UNLINK | 0x20);
        if ( do_expect )
                error = presto_write_last_rcvd(&rec, fset, info);

        presto_debug_fail_blkdev(fset, PRESTO_OP_UNLINK | 0x30);
        EXIT;
exit:
	presto_release_space(fset->fset_cache, PRESTO_REQLOW); 

        presto_trans_commit(fset, handle);
        return error;
}

int lento_unlink(const char *name, struct lento_vfs_context *info)
{
        int error;
        char *tmp;
        struct dentry *dir;
        struct dentry *dentry;
        struct presto_file_set *fset;

        ENTRY;
        lock_kernel();
        tmp = getname(name);
        error = PTR_ERR(tmp);
        if (IS_ERR(tmp)) {
                EXIT;
                goto exit;
        }

        dentry = lookup_dentry(tmp, NULL, 0);
        error = PTR_ERR(dentry);
        if (IS_ERR(dentry)) {
                EXIT;
                goto exit_put;
        }

        dir = lock_parent(dentry);
        fset = presto_fset(dentry);
        error = -EINVAL;
        if ( !fset ) {
                printk("No fileset!\n");
                EXIT;
                goto exit_lock;
        }
        error = presto_do_unlink(fset, dir, dentry, info);

        EXIT;
exit_lock:
        unlock_dir(dir);
        dput(dentry);
exit_put:
        putname(tmp);
exit:
        unlock_kernel();
        return error;
}


int presto_do_symlink(struct presto_file_set *fset, struct dentry *dir,
                      struct dentry *dentry, const char *oldname,
                      struct lento_vfs_context *info)
{
        struct rec_info rec;
        int error;
        struct presto_version tgt_dir_ver, new_link_ver;
        struct inode_operations *iops;
        void *handle;

        ENTRY;
        error = -ENOENT;
        if (!check_parent(dir, dentry)) {
                EXIT;
                return error;
        }

        error = may_create(dir->d_inode, dentry);
        if (error) {
                EXIT;
                return error;
        }

        error = -EPERM;
        iops = filter_c2cdiops(fset->fset_cache->cache_filter);
        if (!iops->symlink) {
                EXIT;
                return error;
        }
	/* record + max path len + space to free */ 
	error = presto_reserve_space(fset->fset_cache, PRESTO_REQHIGH + 4096); 
	if (error) {
		EXIT;
		return error;
	}
        presto_getversion(&tgt_dir_ver, dir->d_inode);
        handle = presto_trans_start(fset, dir->d_inode, PRESTO_OP_SYMLINK);
        if ( IS_ERR(handle) ) {
		presto_release_space(fset->fset_cache, PRESTO_REQHIGH + 4096); 
                printk("ERROR: presto_do_symlink: no space for transaction. Tell Peter.\n"); 
                EXIT;
                return -ENOSPC;
        }
        DQUOT_INIT(dir->d_inode);
        error = iops->symlink(dir->d_inode, dentry, oldname);
        if (error) {
                EXIT;
                goto exit;
        }

        if (dentry->d_inode &&
            dentry->d_inode->i_gid != presto_excluded_gid) {
                struct presto_cache *cache = fset->fset_cache;
                /* was this already done? */
                if ( !filter_c2csiops(cache->cache_filter) )
                        filter_setup_symlink_ops(cache->cache_filter, 
                                                 dentry->d_inode->i_op,
                                                 &presto_sym_iops);

                /* make the new inode ours */
                dentry->d_inode->i_op = filter_c2usiops(cache->cache_filter);
                filter_setup_dentry_ops(cache->cache_filter, dentry->d_op, 
                                        &presto_dentry_ops);
                dentry->d_op = filter_c2udops(cache->cache_filter);

                /* XXX ? Cache state ? if Lento creates a symlink */
                if ( ISLENTO(cache->cache_psdev->uc_minor) ) {
                        presto_set(dentry, PRESTO_ATTR);
                } else {
                        presto_set(dentry, PRESTO_ATTR | PRESTO_DATA);
                }
        }
        error = presto_settime(fset, dir, info, ATTR_CTIME | ATTR_MTIME);
        if (error) { 
                EXIT;
                goto exit;
        }
        error = presto_settime(fset, dentry, info, ATTR_CTIME | ATTR_MTIME);
        if (error) { 
                EXIT;
                goto exit;
        }


        presto_debug_fail_blkdev(fset, PRESTO_OP_SYMLINK | 0x10);
        presto_getversion(&new_link_ver, dentry->d_inode);
        if ( presto_do_kml(info, dentry->d_inode) )
                error = presto_journal_symlink(&rec, fset, dentry, oldname,
                                               &tgt_dir_ver, &new_link_ver);

        presto_debug_fail_blkdev(fset, PRESTO_OP_SYMLINK | 0x20);
        if ( presto_do_expect(info, dentry->d_inode) )
                error = presto_write_last_rcvd(&rec, fset, info);

        presto_debug_fail_blkdev(fset, PRESTO_OP_SYMLINK | 0x30);
        EXIT;
exit:
	presto_release_space(fset->fset_cache, PRESTO_REQHIGH + 4096); 
        presto_trans_commit(fset, handle);
        return error;
}

int lento_symlink(const char *oldname, const char *newname,
                  struct lento_vfs_context *info)
{
        int error;
        char *from;
        char *to;
        struct dentry *dir;
        struct dentry *dentry;
        struct presto_file_set *fset;

        ENTRY;
        lock_kernel();
        from = getname(oldname);
        error = PTR_ERR(from);
        if (IS_ERR(from)) {
                EXIT;
                goto exit;
        }

        to = getname(newname);
        error = PTR_ERR(to);
        if (IS_ERR(to)) {
                EXIT;
                goto exit_from;
        }

        dentry = lookup_dentry(to, NULL, 0);

        error = PTR_ERR(dentry);
        if (IS_ERR(dentry)) {
                EXIT;
                goto exit_to;
        }

        dir = lock_parent(dentry);

        fset = presto_fset(dentry);
        error = -EINVAL;
        if ( !fset ) {
                printk("No fileset!\n");
                EXIT;
                goto exit_lock;
        }

        error = presto_do_symlink(fset, dir, dentry, oldname, info);

        EXIT;
exit_lock:
        unlock_dir(dir);
        dput(dentry);
exit_to:
        putname(to);
exit_from:
        putname(from);
exit:
        unlock_kernel();
        return error;
}


int presto_do_mkdir(struct presto_file_set *fset, struct dentry *dir,
                    struct dentry *dentry, int mode,
                    struct lento_vfs_context *info)
{
        struct rec_info rec;
        int error;
        struct presto_version tgt_dir_ver, new_dir_ver;
        void *handle;

        ENTRY;
        /*
         * EEXIST is kind of a strange error code to
         * return, but basically if the dentry was moved
         * or unlinked while we locked the parent, we
         * do know that it _did_ exist before, and as
         * such it makes perfect sense.. In contrast,
         * ENOENT doesn't make sense for mkdir.
         */
        error = -EEXIST;
        if (!check_parent(dir, dentry)) {
                EXIT;
                return error;
        }

        error = may_create(dir->d_inode, dentry);
        if (error) {
                EXIT;
                return error;
        }

        error = -EPERM;
        if (!filter_c2cdiops(fset->fset_cache->cache_filter)->mkdir) {
                EXIT;
                return error;
        }
	/* one journal record + directory block + room for removals*/ 
	error = presto_reserve_space(fset->fset_cache, PRESTO_REQHIGH + 4096); 
	if (error) { 
                EXIT;
                return error;
        }

        presto_getversion(&tgt_dir_ver, dir->d_inode);
        handle = presto_trans_start(fset, dir->d_inode, PRESTO_OP_MKDIR);
        if ( IS_ERR(handle) ) {
		presto_release_space(fset->fset_cache, PRESTO_REQHIGH + 4096); 
                printk("presto_do_mkdir: no space for transaction\n");
                EXIT;
                return -ENOSPC;
        }
        DQUOT_INIT(dir->d_inode);

        error = filter_c2cdiops(fset->fset_cache->cache_filter)->mkdir(dir->d_inode, dentry, mode);
        if (error) {
                EXIT;
                goto exit;
        }

        if ( dentry->d_inode && !error && 
             dentry->d_inode->i_gid != presto_excluded_gid) {
                struct presto_cache *cache = fset->fset_cache;
                /* make it ours */
                dentry->d_inode->i_op = filter_c2udiops(cache->cache_filter);
                filter_setup_dentry_ops(cache->cache_filter, 
                                        dentry->d_op, 
                                        &presto_dentry_ops);
                dentry->d_op = filter_c2udops(cache->cache_filter);
                /* if Lento does this, we won't have data */
                if ( ISLENTO(cache->cache_psdev->uc_minor) ) {
                        presto_set(dentry, PRESTO_ATTR);
                } else {
                        presto_set(dentry, PRESTO_ATTR | PRESTO_DATA);
                }
        }
        error = presto_settime(fset, dir, info, ATTR_CTIME | ATTR_MTIME);
        if (error) { 
                EXIT;
                goto exit;
        }
        error = presto_settime(fset, dentry, info, ATTR_CTIME | ATTR_MTIME);
        if (error) { 
                EXIT;
                goto exit;
        }

        presto_debug_fail_blkdev(fset, PRESTO_OP_MKDIR | 0x10);
        presto_getversion(&new_dir_ver, dentry->d_inode);
        /* POSIX ACL code modifies mode. Hence we need to pull the actual
         * mode from the inode rather than the mode passed to us.
         * Also see comments in presto_do_create. -SHP
         */
        if ( presto_do_kml(info, dentry->d_inode) ) {
                error = presto_journal_mkdir(&rec, fset, dentry, &tgt_dir_ver,
                                             &new_dir_ver, 
                                             dentry->d_inode->i_mode);
        }

        presto_debug_fail_blkdev(fset, PRESTO_OP_MKDIR | 0x20);
        if ( presto_do_expect(info, dentry->d_inode) ) {

                error = presto_write_last_rcvd(&rec, fset, info);
        }
        presto_debug_fail_blkdev(fset, PRESTO_OP_MKDIR | 0x30);
        EXIT;
exit:
	presto_release_space(fset->fset_cache, PRESTO_REQHIGH + 4096); 
        presto_trans_commit(fset, handle);
        return error;
}

/*
 * Look out: this function may change a normal dentry
 * into a directory dentry (different size)..
 */
int lento_mkdir(const char *name, int mode, struct lento_vfs_context *info)
{
        int error;
        char *pathname;
        struct dentry *dir;
        struct dentry *dentry;
        struct presto_file_set *fset;

        ENTRY;
        CDEBUG(D_PIOCTL, "name: %s, mode %o, offset %d, recno %d, flags %x\n",
               name, mode, info->slot_offset, info->recno, info->flags);
        lock_kernel();
        pathname = getname(name);
        error = PTR_ERR(pathname);
        if (IS_ERR(pathname)) {
                EXIT;
                goto exit;
        }

        dentry = lookup_dentry(pathname, NULL, LOOKUP_SLASHOK);
        error = PTR_ERR(dentry);
        if (IS_ERR(dentry)) {
                EXIT;
                goto exit_put;
        }

        dir = lock_parent(dentry);

        fset = presto_fset(dentry);
        error = -EINVAL;
        if ( !fset ) {
                printk("No fileset!\n");
                EXIT;
                goto exit_lock;
        }

	
        error = presto_do_mkdir(fset, dir, dentry, mode & S_IALLUGO, info);
        EXIT;
exit_lock:
        unlock_dir(dir);
        dput(dentry);
exit_put:
        putname(pathname);
exit:
        unlock_kernel();
        return error;
}


int presto_do_rmdir(struct presto_file_set *fset, struct dentry *dir,
                    struct dentry *dentry, struct lento_vfs_context *info)
{
        struct rec_info rec;
        int error;
        struct presto_version tgt_dir_ver, old_dir_ver;
        struct inode_operations *iops;
        void *handle;
        int do_kml, do_expect;
	int size;

        ENTRY;
        error = -EPERM;
        iops = filter_c2cdiops(fset->fset_cache->cache_filter);
        if (!iops->rmdir) {
                EXIT;
                return error;
        }

	size = PRESTO_REQHIGH - dentry->d_inode->i_size; 
	error = presto_reserve_space(fset->fset_cache, size); 
	if (error) { 
		EXIT;
		return error;
	}

        presto_getversion(&tgt_dir_ver, dir->d_inode);
        presto_getversion(&old_dir_ver, dentry->d_inode);
        handle = presto_trans_start(fset, dir->d_inode, PRESTO_OP_RMDIR);
        if ( IS_ERR(handle) ) {
		presto_release_space(fset->fset_cache, size); 
                printk("ERROR: presto_do_rmdir: no space for transaction. Tell Peter.\n");
                return -ENOSPC;
        }

        DQUOT_INIT(dir->d_inode);
        do_kml = presto_do_kml(info, dir->d_inode);
        do_expect = presto_do_expect(info, dir->d_inode);

        error = iops->rmdir(dir->d_inode, dentry);
        if (error) {
                EXIT;
                goto exit;
        }
        error = presto_settime(fset, dir, info, ATTR_CTIME | ATTR_MTIME);
        if (error) { 
                EXIT;
                goto exit;
        }

        presto_debug_fail_blkdev(fset, PRESTO_OP_RMDIR | 0x10);
        if ( do_kml )
                error = presto_journal_rmdir(&rec, fset, dir, &tgt_dir_ver,
                                             &old_dir_ver,
                                             dentry->d_name.len,
                                             dentry->d_name.name);

        presto_debug_fail_blkdev(fset, PRESTO_OP_RMDIR | 0x20);
        if ( do_expect ) 
                error = presto_write_last_rcvd(&rec, fset, info);

        presto_debug_fail_blkdev(fset, PRESTO_OP_RMDIR | 0x30);
        EXIT;
exit:
	presto_release_space(fset->fset_cache, size); 
        presto_trans_commit(fset, handle);
        return error;
}

int lento_rmdir(const char *name, struct lento_vfs_context *info)
{
        int error;
        char * tmp;
        struct dentry *dir;
        struct dentry *dentry;
        struct presto_file_set *fset;

        ENTRY;
        lock_kernel();
        tmp = getname(name);
        error = PTR_ERR(tmp);
        if (IS_ERR(tmp)) {
                EXIT;
                goto exit;
        }

        dentry = lookup_dentry(tmp, NULL, 0);
        error = PTR_ERR(dentry);
        if (IS_ERR(dentry)) {
                EXIT;
                goto exit_put;
        }

        error = -ENOENT;
        if (!dentry->d_inode) {
                EXIT;
                goto exit_dput;
        }

        dir = get_parent(dentry);

        /*
         * The dentry->d_count stuff confuses d_delete() enough to
         * not kill the inode from under us while it is locked. This
         * wouldn't be needed, except the dentry semaphore is really
         * in the inode, not in the dentry..
         */
        dentry->d_count++;
        double_lock(dir, dentry);

        fset = presto_fset(dentry);
        error = -EINVAL;
        if ( !fset ) {
                printk("No fileset!\n");
                EXIT;
                goto exit_lock;
        }

        error = -ENOENT;
        if (!check_parent(dir, dentry)) {
                EXIT;
                goto exit_lock;
        }
        error = may_delete(dir->d_inode, dentry, 1);
        if (error) {
                EXIT;
                goto exit_lock;
        }

        /*
         * We try to drop the dentry early: we should have
         * a usage count of 2 if we're the only user of this
         * dentry, and if that is true (possibly after pruning
         * the dcache), then we drop the dentry now.
         *
         * A low-level filesystem can, if it choses, legally
         * do a
         *
         *      if (!list_empty(&dentry->d_hash))
         *              return -EBUSY;
         *
         * if it cannot handle the case of removing a directory
         * that is still in use by something else..
         */
        switch (dentry->d_count) {
        default:
                CDEBUG(D_PIOCTL, "shrinking dcache\n");
                shrink_dcache_parent(dentry);
                CDEBUG(D_PIOCTL, "shrunk dcache\n");
                if (dentry->d_count != 2)
                        break;
        case 2:
                CDEBUG(D_PIOCTL, "dropping dentry\n");
                d_drop(dentry);
        }

        error = presto_do_rmdir(fset, dir, dentry, info);

        EXIT;
exit_lock:
        double_unlock(dentry, dir);
exit_dput:
        dput(dentry);
exit_put:
        putname(tmp);
exit:
        unlock_kernel();
        CDEBUG(D_PIOCTL, "real exit\n");
        return error;
}


int presto_do_mknod(struct presto_file_set *fset, struct dentry *dir,
                    struct dentry *dentry, int mode, dev_t dev,
                    struct lento_vfs_context *info)
{
        struct rec_info rec;
        int error;
        struct inode_operations *iops;
        struct presto_version tgt_dir_ver, new_node_ver;
        void *handle;

        ENTRY;
        error = -EEXIST;
        if (!check_parent(dir, dentry)) {
                EXIT;
                return error;
        }

        error = may_create(dir->d_inode, dentry);
        if (error) {
                EXIT;
                return error;
        }

        error = -EPERM;
        iops = filter_c2cdiops(fset->fset_cache->cache_filter);
        if (!iops->mknod) {
                EXIT;
                return error;
        }
	/* one KML entry */ 
	error = presto_reserve_space(fset->fset_cache, PRESTO_REQHIGH); 
	if (error) {
		EXIT;
		return error;
	}

        presto_getversion(&tgt_dir_ver, dir->d_inode);
        handle = presto_trans_start(fset, dir->d_inode, PRESTO_OP_MKNOD);
        if ( IS_ERR(handle) ) {
		presto_release_space(fset->fset_cache, PRESTO_REQHIGH); 
                printk("presto_do_mknod: no space for transaction\n");
                return -ENOSPC;
        }
        DQUOT_INIT(dir->d_inode);
        error = iops->mknod(dir->d_inode, dentry, mode, dev);
        if (error) {
                EXIT;
                goto exit;
        }

        if ( dentry->d_inode &&
             dentry->d_inode->i_gid != presto_excluded_gid) {
                struct presto_cache *cache = fset->fset_cache;
                /* make it ours */
                dentry->d_inode->i_op = filter_c2udiops(cache->cache_filter);
                filter_setup_dentry_ops(cache->cache_filter, dentry->d_op, 
                                        &presto_dentry_ops);
                dentry->d_op = filter_c2udops(cache->cache_filter);

                /* if Lento does this, we won't have data */
                if ( ISLENTO(cache->cache_psdev->uc_minor) ) {
                        presto_set(dentry, PRESTO_ATTR);
                } else {
                        presto_set(dentry, PRESTO_ATTR | PRESTO_DATA);
                }
        }

        error = presto_settime(fset, dir, info, ATTR_MTIME);
        if (error) { 
                EXIT;
                goto exit;
        }
        error = presto_settime(fset, dentry, info, ATTR_CTIME | ATTR_MTIME);
        if (error) { 
                EXIT;
                goto exit;
        }

        presto_debug_fail_blkdev(fset, PRESTO_OP_MKNOD | 0x10);
        presto_getversion(&new_node_ver, dentry->d_inode);
        /* journal real mode rather than mode passed to us, since posix acls
         * can modify the actual mode. -SHP
         */
        if ( presto_do_kml(info, dentry->d_inode) )
                error = presto_journal_mknod(&rec, fset, dentry, &tgt_dir_ver,
                                             &new_node_ver, 
                                             dentry->d_inode->i_mode,
                                             MAJOR(dev), MINOR(dev) );

        presto_debug_fail_blkdev(fset, PRESTO_OP_MKNOD | 0x20);
        if ( presto_do_expect(info, dentry->d_inode) ) 
                error = presto_write_last_rcvd(&rec, fset, info);

        presto_debug_fail_blkdev(fset, PRESTO_OP_MKNOD | 0x30);
        EXIT;
exit:
	presto_release_space(fset->fset_cache, PRESTO_REQHIGH); 
        presto_trans_commit(fset, handle);
        return error;
}

int lento_mknod(const char *filename, int mode, dev_t dev,
                struct lento_vfs_context *info)
{
        struct dentry *dentry;
        struct dentry *dir;
        int error;
        char *tmp;
        struct presto_file_set *fset;

        ENTRY;
        lock_kernel();
        error = -EPERM;
        if (S_ISDIR(mode) || (!S_ISFIFO(mode) && !capable(CAP_SYS_ADMIN))) {
                EXIT;
                goto exit;
        }
        error = -EINVAL;
        switch (mode & S_IFMT) {
        case 0:
                mode |= S_IFREG;
                break;
        case S_IFREG: case S_IFCHR: case S_IFBLK: case S_IFIFO: case S_IFSOCK:
                break;
        default:
                EXIT;
                goto exit;
        }

        tmp = getname(filename);
        error = PTR_ERR(tmp);
        if (IS_ERR(tmp)) {
                EXIT;
                goto exit;
        }

        dentry = lookup_dentry(tmp, NULL, 0);
        error = PTR_ERR(dentry);
        if (IS_ERR(dentry)) {
                EXIT;
                goto exit_put;
        }

        dir = lock_parent(dentry);
        fset = presto_fset(dentry);
        error = -EINVAL;
        if ( !fset ) {
                printk("No fileset!\n");
                EXIT;
                goto exit_lock;
        }

        error = presto_do_mknod(fset, dir, dentry, mode, dev, info);

        EXIT;
exit_lock:
        unlock_dir(dir);
        dput(dentry);
exit_put:
        putname(tmp);
exit:
        unlock_kernel();
        return error;
}

int presto_do_rename(struct presto_file_set *fset,
                     struct dentry *old_parent, struct dentry *old_dentry,
                     struct dentry *new_parent, struct dentry *new_dentry,
                     struct lento_vfs_context *info)
{
        struct rec_info rec;
        int error;
        struct inode_operations *iops;
        struct presto_version src_dir_ver, tgt_dir_ver;
        void *handle;
        int new_inode_unlink = 0;
        struct inode *old_dir = old_parent->d_inode;
        struct inode *new_dir = new_parent->d_inode;

        ENTRY;
        presto_getversion(&src_dir_ver, old_dir);
        presto_getversion(&tgt_dir_ver, new_dir);

        error = -EPERM;
        iops = filter_c2cdiops(fset->fset_cache->cache_filter);
        if (!iops || !iops->rename) {
                EXIT;
                return error;
        }

	error = presto_reserve_space(fset->fset_cache, PRESTO_REQHIGH); 
	if (error) {
		EXIT;
		return error;
	}

        handle = presto_trans_start(fset, old_dir, PRESTO_OP_RENAME);
        if ( IS_ERR(handle) ) {
		presto_release_space(fset->fset_cache, PRESTO_REQHIGH); 

                printk("presto_do_rename: no space for transaction\n");
                return -ENOSPC;
        }

        if (new_dentry->d_inode && new_dentry->d_inode->i_nlink > 1) { 
                dget(new_dentry); 
                new_inode_unlink = 1;
        }

        error = iops->rename(old_dir, old_dentry, new_dir, new_dentry);

        if (error) {
                EXIT;
                dput(new_dentry);
                goto exit;
        }


        if (new_inode_unlink) { 
                error = presto_settime(fset, old_dentry, info, ATTR_CTIME);
                dput(old_dentry); 
                if (error) { 
                        EXIT;
                        goto exit;
                }
        }
        error = presto_settime(fset, old_parent, info, ATTR_CTIME | ATTR_MTIME);
        if (error) { 
                EXIT;
                goto exit;
        }
        error = presto_settime(fset, new_parent, info, ATTR_CTIME | ATTR_MTIME);
        if (error) { 
                EXIT;
                goto exit;
        }


        /* XXX make a distinction between cross file set
         * and intra file set renames here
         */
        presto_debug_fail_blkdev(fset, PRESTO_OP_RENAME | 0x10);
        if ( presto_do_kml(info, old_dir) )
                error = presto_journal_rename(&rec, fset, old_dentry, new_dentry,
                                              &src_dir_ver, &tgt_dir_ver);

        presto_debug_fail_blkdev(fset, PRESTO_OP_RENAME | 0x20);

        if ( presto_do_expect(info, new_dir) ) 
                error = presto_write_last_rcvd(&rec, fset, info);

        presto_debug_fail_blkdev(fset, PRESTO_OP_RENAME | 0x30);
        EXIT;
exit:
	presto_release_space(fset->fset_cache, PRESTO_REQHIGH); 
        presto_trans_commit(fset, handle);
        return error;
}

static
int presto_rename_dir(struct presto_file_set *fset, struct dentry *old_parent, 
                      struct dentry *old_dentry, struct dentry *new_parent, 
                      struct dentry *new_dentry, struct lento_vfs_context *info)
{
        int error;
        int need_rehash = 0;
        struct inode *old_dir = old_parent->d_inode;
        struct inode *new_dir = new_parent->d_inode;
        ENTRY;
        if (old_dentry->d_inode == new_dentry->d_inode) {
                EXIT;
                return 0;
        }

        error = may_delete(old_dir, old_dentry, 1);
        if (error) {
                EXIT;
                return error;
        }

        if (new_dir->i_dev != old_dir->i_dev) {
                EXIT;
                return -EXDEV;
        }

        if (!new_dentry->d_inode)
                error = may_create(new_dir, new_dentry);
        else
                error = may_delete(new_dir, new_dentry, 1);
        if (error) {
                EXIT;
                return error;
        }

        /*
         * If we are going to change the parent - check write permissions,
         * we'll need to flip '..'.
         */
        if (new_dir != old_dir)
                error = permission(old_dentry->d_inode, MAY_WRITE);
        if (error) {
                EXIT;
                return error;
        }

        DQUOT_INIT(old_dir);
        DQUOT_INIT(new_dir);
        fs_down(&old_dir->i_sb->s_vfs_rename_sem);
        error = -EINVAL;
        if (is_subdir(new_dentry, old_dentry)) {
                EXIT;
                goto out_unlock;
        }
        if (new_dentry->d_inode) {
                error = -EBUSY;
                if (d_invalidate(new_dentry) < 0) {
                        EXIT;
                        goto out_unlock;
                }
                need_rehash = 1;
        }
        error = presto_do_rename(fset, old_parent, old_dentry,
                                 new_parent, new_dentry, info);
        if (need_rehash)
                d_rehash(new_dentry);
        if (!error)
                d_move(old_dentry, new_dentry);
        EXIT;
out_unlock:
        fs_up(&old_dir->i_sb->s_vfs_rename_sem);
        return error;
}

static
int presto_rename_other(struct presto_file_set *fset, struct dentry *old_parent, 
                        struct dentry *old_dentry,  struct dentry *new_parent, 
                        struct dentry *new_dentry, struct lento_vfs_context *info)
{
        int error;
        struct inode *old_dir = old_parent->d_inode;
        struct inode *new_dir = new_parent->d_inode;

        ENTRY;
        if (old_dentry->d_inode == new_dentry->d_inode) {
                EXIT;
                return 0;
        }

        error = may_delete(old_dir, old_dentry, 0);
        if (error) {
                EXIT;
                return error;
        }

        if (new_dir->i_dev != old_dir->i_dev) {
                EXIT;
                return -EXDEV;
        }
        if (!new_dentry->d_inode)
                error = may_create(new_dir, new_dentry);
        else
                error = may_delete(new_dir, new_dentry, 0);
        if (error) {
                EXIT;
                return error;
        }

        DQUOT_INIT(old_dir);
        DQUOT_INIT(new_dir);
        error = presto_do_rename(fset, old_parent, old_dentry,
                                 new_parent, new_dentry, info);
        if (error) {
                EXIT;
                return error;
        }
        /* The following d_move() should become unconditional */
        if (!(old_dir->i_sb->s_flags & MS_ODD_RENAME))
                d_move(old_dentry, new_dentry);
        EXIT;
        return 0;
}


int lento_rename(const char *oldname, const char *newname,
                 struct lento_vfs_context *info)
{
        int error;
        char *from;
        char *to;
        struct dentry *old_dir, *new_dir;
        struct dentry *old_dentry, *new_dentry;
        struct presto_file_set *fset;

        ENTRY;
        lock_kernel();
        from = getname(oldname);
        error = PTR_ERR(from);
        if (IS_ERR(from)) {
                EXIT;
                goto exit;
        }

        to = getname(newname);
        error = PTR_ERR(to);
        if (IS_ERR(to)) {
                EXIT;
                goto exit_from;
        }

        old_dentry = lookup_dentry(from, NULL, 0);

        error = PTR_ERR(old_dentry);
        if (IS_ERR(old_dentry)) {
                EXIT;
                goto exit_to;
        }

        error = -ENOENT;
        if (!old_dentry->d_inode) {
                EXIT;
                goto exit_old;
        }

        {
                unsigned int flags = 0;
                if (S_ISDIR(old_dentry->d_inode->i_mode))
                        flags = LOOKUP_SLASHOK;
                new_dentry = lookup_dentry(newname, NULL, flags);
        }

        error = PTR_ERR(new_dentry);
        if (IS_ERR(new_dentry)) {
                EXIT;
                goto exit_old;
        }

        new_dir = get_parent(new_dentry);
        old_dir = get_parent(old_dentry);

        double_lock(new_dir, old_dir);

        fset = presto_fset(new_dentry);
        error = -EINVAL;
        if ( !fset ) {
                printk("No fileset!\n");
                EXIT;
                goto exit_lock;
        }

        if (!check_parent(old_dir, old_dentry) ||
            !check_parent(new_dir, new_dentry)) {
                EXIT;
                error = -ENOENT;
                goto exit_lock;
        }

        if (S_ISDIR(old_dentry->d_inode->i_mode))
                error = presto_rename_dir(fset, old_dir, old_dentry,
                                          new_dir, new_dentry, info);
        else
                error = presto_rename_other(fset, old_dir, old_dentry,
                                            new_dir, new_dentry, info);

        EXIT;
exit_lock:
        double_unlock(new_dir, old_dir);
        dput(new_dentry);
exit_old:
        dput(old_dentry);
exit_to:
        putname(to);
exit_from:
        putname(from);
exit:
        unlock_kernel();
        return error;
}

struct dentry *presto_iopen(struct dentry *dentry,
                            ino_t ino, unsigned int generation)
{
        struct presto_file_set *fset;
        char name[48];
        int error;

        ENTRY;
        /* see if we already have the dentry we want */
        if (dentry->d_inode && dentry->d_inode->i_ino == ino &&
            dentry->d_inode->i_generation == generation) {
                EXIT;
                return dentry;
        }

        /* Make sure we have a cache beneath us.  We should always find at
         * least one dentry inside the cache (if it exists), otherwise not
         * even the cache root exists, or we passed in a bad name.
         */
        fset = presto_fset(dentry);
        error = -EINVAL;
        if (!fset) {
                printk("No fileset for %*s!\n",
                       dentry->d_name.len, dentry->d_name.name);
                EXIT;
                dput(dentry);
                return ERR_PTR(error);
        }
        dput(dentry);

        sprintf(name, "%s%#lx%c%#x",
                PRESTO_ILOOKUP_MAGIC, ino, PRESTO_ILOOKUP_SEP, generation);
        CDEBUG(D_PIOCTL, "opening %ld by number (as %s)\n", ino, name);
        return lookup_dentry(name, fset->fset_mtpt, 0);
}

static struct file *presto_filp_dopen(struct dentry *dentry, int flags)
{
        struct file *f;
        struct inode *inode;
        int flag, error;

        ENTRY;
        error = -ENFILE;
        f = get_empty_filp();
        if (!f) {
                CDEBUG(D_PIOCTL, "error getting file pointer\n");
                EXIT;
                goto out;
        }
        f->f_flags = flag = flags;
        f->f_mode = (flag+1) & O_ACCMODE;
        inode = dentry->d_inode;
        if (f->f_mode & FMODE_WRITE) {
                error = get_write_access(inode);
                if (error) {
                        CDEBUG(D_PIOCTL, "error getting write access\n");
                        EXIT;
                        goto cleanup_file;
                }
        }

        f->f_dentry = dentry;
        f->f_pos = 0;
        f->f_reada = 0;
        f->f_op = NULL;
        if (inode->i_op)
                /* XXX should we set to presto ops, or leave at cache ops? */
                f->f_op = inode->i_op->default_file_ops;
        if (f->f_op && f->f_op->open) {
                error = f->f_op->open(inode, f);
                if (error) {
                        CDEBUG(D_PIOCTL, "error calling cache 'open'\n");
                        EXIT;
                        goto cleanup_all;
                }
        }
        f->f_flags &= ~(O_CREAT | O_EXCL | O_NOCTTY | O_TRUNC);

        return f;

cleanup_all:
        if (f->f_mode & FMODE_WRITE)
                put_write_access(inode);
cleanup_file:
        put_filp(f);
out:
        return ERR_PTR(error);
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,2,18))
/* duplicated from fs/open.c */
inline void put_unused_fd(unsigned int fd)
{
        FD_CLR(fd, current->files->open_fds);
        if (fd < current->files->next_fd)
                current->files->next_fd = fd;
}
#endif

/* Open an inode by number.  We pass in the cache root name (or a subdirectory
 * from the cache that is guaranteed to exist) to be able to access the cache.
 */
int lento_iopen(const char *name, ino_t ino, unsigned int generation,
                int flags)
{
        char * tmp;
        struct dentry *dentry;
        int fd;

        ENTRY;
        CDEBUG(D_PIOCTL,
               "open %s:inode %#lx (%ld), generation %x (%d), flags %d \n",
               name, ino, ino, generation, generation, flags);
        /* We don't allow creation of files by number only, as it would
         * lead to a dangling files not in any directory.  We could also
         * just turn off the flag and ignore it.
         */
        if (flags & O_CREAT) {
                printk(KERN_WARNING __FUNCTION__
                       ": create file by inode number (%ld) not allowed\n",ino);
                EXIT;
                return -EACCES;
        }

        tmp = getname(name);
        if (IS_ERR(tmp)) {
                EXIT;
                return PTR_ERR(tmp);
        }

        lock_kernel();
again:  /* look the named file or a parent directory so we can get the cache */
        CDEBUG(D_PIOCTL, "looking up %s\n", tmp);
        dentry = lookup_dentry(tmp, NULL, 0);
        fd = PTR_ERR(dentry);
        if (fd == -ENOENT) /* this is OK for us right here */
                dentry = NULL;
        if (IS_ERR(dentry)) {
                EXIT;
                goto exit;
        }
        /* we didn't find the named file, so see if a parent exists */
        if (!dentry) {
                char *slash;

                slash = strrchr(tmp, '/');
                if (slash && slash != tmp) {
                        *slash = '\0';
                        goto again;
                }
                /* we should never get here... */
                CDEBUG(D_PIOCTL, "no more path components to try??!!\n");
                fd = -ENOENT;
                goto exit;
        }
        CDEBUG(D_PIOCTL, "returned dentry %p\n", dentry);

        dentry = presto_iopen(dentry, ino, generation);
        fd = PTR_ERR(dentry);
        if (IS_ERR(dentry)) {
                EXIT;
                goto exit;
        }

        /* XXX start of code that might be replaced by something like:
         * if (flags & (O_WRONLY | O_RDWR)) {
         *      error = get_write_access(dentry->d_inode);
         *      if (error) {
         *              EXIT;
         *              goto cleanup_dput;
         *      }
         * }
         * fd = open_dentry(dentry, flags);
         *
         * including the presto_filp_dopen() function (check dget counts!)
         */
        fd = get_unused_fd();
        if (fd < 0) {
                EXIT;
                goto cleanup_dput;
        }

        {
                int error;
                struct file * f = presto_filp_dopen(dentry, flags);
                error = PTR_ERR(f);
                if (IS_ERR(f)) {
                        put_unused_fd(fd);
                        fd = error;
                        EXIT;
                        goto cleanup_dput;
                }
                fd_install(fd, f);
        }
        /* end of code that might be replaced by open_dentry */

        EXIT;
exit:
        unlock_kernel();
        putname(tmp);
        return fd;

cleanup_dput:
        dput(dentry);
        goto exit;
}

int lento_close(unsigned int fd, struct lento_vfs_context *info)
{
        struct rec_info rec;
        int error;
        struct file * filp;
        struct dentry *dentry;
        int do_kml, do_expect;

        ENTRY;
        lock_kernel();

        error = -EBADF;
        filp = fcheck(fd);

        if (filp  ) {
                struct files_struct * files = current->files;
                dentry = filp->f_dentry;
                dget(dentry);
                do_kml = presto_do_kml(info, dentry->d_inode);
                do_expect = presto_do_expect(info, dentry->d_inode);
                files->fd[fd] = NULL;
                put_unused_fd(fd);
                FD_CLR(fd, files->close_on_exec);
                error = filp_close(filp, files);
        } else {
                EXIT;
                return error;
        }

        if (error) {
                EXIT;
                goto exit;
        }

        if ( do_kml ) { 
                struct presto_file_set *fset;
                struct presto_version new_file_ver;
                struct file file;
                void *handle;

                fset = presto_fset(dentry);
                error = -EINVAL;
                if (!fset) {
                        printk("No fileset for %*s!\n",
                               dentry->d_name.len, dentry->d_name.name);
                        EXIT;
                        goto exit;
                }
                handle = presto_trans_start(fset, dentry->d_inode, 
                                       PRESTO_OP_RELEASE); 
                if ( IS_ERR(handle) ) {
                        printk("lento_close: no space for transaction\n");
                        return -ENOSPC;
                }
                presto_getversion(&new_file_ver, dentry->d_inode);
                file.f_dentry = dentry;
                file.private_data = NULL;
                error = presto_journal_close(&rec, fset, &file,
                                             &new_file_ver);
                if ( error ) {
                        presto_trans_commit(fset, handle); 
                        printk("presto: close error %d!\n", error);
                        EXIT;
                        goto exit;
                }
                if ( do_expect ) 
                        error = presto_write_last_rcvd(&rec, fset, info);
                presto_trans_commit(fset, handle); 
        }

        EXIT;
exit:
        dput(dentry);
        unlock_kernel();
        return error;
}


#ifdef CONFIG_FS_EXT_ATTR

#ifdef CONFIG_FS_POSIX_ACL
/* Posix ACL code changes i_mode without using a notify_change (or
 * a mark_inode_dirty!). We need to duplicate this at the reintegrator
 * which is done by this function. This function also takes care of 
 * resetting the cached posix acls in this inode. If we don't reset these
 * VFS continues using the old acl information, which by now may be out of
 * date.
 */
int presto_setmode(struct presto_file_set *fset, struct dentry *dentry,
                   mode_t mode)
{
        struct inode *inode = dentry->d_inode;

        ENTRY;
        /* The extended attributes for this inode were modified. 
         * At this point we can not be sure if any of the ACL 
         * information for this inode was updated. So we will 
         * force VFS to reread the acls. Note that we do this 
         * only when called from the SETEXTATTR ioctl, which is why we
         * do this while setting the mode of the file. Also note
         * that mark_inode_dirty is not be needed for i_*acl only
         * to force i_mode info to disk, and should be removed once
         * we use notify_change to update the mode.
         * XXX: is mode setting really needed? Just setting acl's should
         * be enough! VFS should change the i_mode as needed? SHP
         */
        if (inode->i_acl && 
            inode->i_acl != POSIX_ACL_NOT_CACHED) 
            posix_acl_release(inode->i_acl);
        if (inode->i_default_acl && 
            inode->i_default_acl != POSIX_ACL_NOT_CACHED) 
            posix_acl_release(inode->i_default_acl);
        inode->i_acl = POSIX_ACL_NOT_CACHED;
        inode->i_default_acl = POSIX_ACL_NOT_CACHED;
        inode->i_mode = mode;
        /* inode should already be dirty...but just in case */
        mark_inode_dirty(inode);
        return 0;

#if 0
        /* XXX: The following code is the preferred way to set mode, 
         * however, I need to carefully go through possible recursion
         * paths back into presto. See comments in presto_do_setattr.
         */
        {    
        int error=0; 
        struct super_operations *sops;
        struct iattr iattr;

        iattr.ia_mode = mode;
        iattr.ia_valid = ATTR_MODE|ATTR_FORCE;

        error = -EPERM;
        sops = filter_c2csops(fset->fset_cache->cache_filter); 
        if (!sops &&
            !sops->notify_change) {
                EXIT;
                return error;
        }

        error = sops->notify_change(dentry, &iattr);

        EXIT;
        return error;
        }
#endif
}
#endif

/* setextattr Interface to cache filesystem */
int presto_do_set_ext_attr(struct presto_file_set *fset, 
                           struct dentry *dentry, 
                           const char *name, void *buffer,
                           size_t buffer_len, int flags, mode_t *mode,
                           struct lento_vfs_context *info) 
{
        struct rec_info rec;
        struct inode *inode = dentry->d_inode;
        struct inode_operations *iops;
        int error;
        struct presto_version ver;
        void *handle;
        char temp[PRESTO_EXT_ATTR_NAME_MAX+1];

        ENTRY;
        error = -EROFS;
        if (IS_RDONLY(inode)) {
                EXIT;
                return -EROFS;
        }

        if (IS_IMMUTABLE(inode) || IS_APPEND(inode)) {
                EXIT;
                return -EPERM;
        }

        presto_getversion(&ver, inode);
        error = -EPERM;
        /* We need to invoke different filters based on whether
         * this dentry is a regular file, directory or symlink.
         */
        switch (inode->i_mode & S_IFMT) {
                case S_IFLNK: /* symlink */
                    iops = filter_c2csiops(fset->fset_cache->cache_filter); 
                    break;
                case S_IFDIR: /* directory */
                    iops = filter_c2cdiops(fset->fset_cache->cache_filter); 
                    break;
                case S_IFREG:
                default: /* everything else including regular files */
                    iops = filter_c2cfiops(fset->fset_cache->cache_filter); 
        }

        if (!iops && !iops->set_ext_attr) {
                EXIT;
                return error;
        }

        error = presto_reserve_space(fset->fset_cache, PRESTO_REQHIGH); 
        if (error) {
                EXIT;
                return error;
        }

        
        handle = presto_trans_start(fset,dentry->d_inode,PRESTO_OP_SETEXTATTR);
        if ( IS_ERR(handle) ) {
                printk("presto_do_set_ext_attr: no space for transaction\n");
                presto_release_space(fset->fset_cache, PRESTO_REQHIGH); 
                return -ENOSPC;
        }

        /* We first "truncate" name to the maximum allowable in presto */
        /* This simulates the strncpy_from_use code in fs/ext_attr.c */
        strncpy(temp,name,sizeof(temp));

        /* Pass down to cache*/
        error = iops->set_ext_attr(inode,temp,buffer,buffer_len,flags);
        if (error) {
                EXIT;
                goto exit;
        }

#ifdef CONFIG_FS_POSIX_ACL
        /* Reset mode if specified*/
        /* XXX: when we do native acl support, move this code out! */
        if (mode != NULL) {
                error = presto_setmode(fset, dentry, *mode);
                if (error) { 
                    EXIT;
                    goto exit;
                }
        }
#endif

        /* Reset ctime. Only inode change time (ctime) is affected */
        error = presto_settime(fset, dentry, info, ATTR_CTIME);
        if (error) { 
                EXIT;
                goto exit;
        }

        if (flags & EXT_ATTR_FLAG_USER) {
                printk(" USER flag passed to presto_do_set_ext_attr!\n");
                *(int *)0 = 1;
        }

        /* We are here, so set_ext_attr succeeded. We no longer need to keep
         * track of EXT_ATTR_FLAG_{EXISTS,CREATE}, instead, we will force
         * the attribute value during log replay. -SHP
         */
        flags &= ~(EXT_ATTR_FLAG_EXISTS | EXT_ATTR_FLAG_CREATE);

        presto_debug_fail_blkdev(fset, PRESTO_OP_SETEXTATTR | 0x10);
        if ( presto_do_kml(info, dentry->d_inode) )
                error = presto_journal_set_ext_attr
                        (&rec, fset, dentry, &ver, name, buffer, 
                         buffer_len, flags);

        presto_debug_fail_blkdev(fset, PRESTO_OP_SETEXTATTR | 0x20);
        if ( presto_do_expect(info, dentry->d_inode) )
                error = presto_write_last_rcvd(&rec, fset, info);

        presto_debug_fail_blkdev(fset, PRESTO_OP_SETEXTATTR | 0x30);
        EXIT;
exit:
        presto_release_space(fset->fset_cache, PRESTO_REQHIGH); 
        presto_trans_commit(fset, handle);

        return error;
}
#endif
